/* global define */
define('WS.Data/Adapter/Abstract', [
   'WS.Data/Entity/Abstract',
   'WS.Data/Adapter/IAdapter',
   'WS.Data/Entity/SerializableMixin'
], function (
   AbstractEntity,
   IAdapter,
   SerializableMixin
) {
   'use strict';

   /**
    * Абстрактный адаптер для данных.
    * Это абстрактный класс, не предназначенный для создания самостоятельных экземпляров.
    * @class WS.Data/Adapter/Abstract
    * @extends WS.Data/Entity/Abstract
    * @implements WS.Data/Adapter/IAdapter
    * @mixes WS.Data/Entity/SerializableMixin
    * @public
    * @author Мальцев Алексей
    */

   var Abstract = AbstractEntity.extend([IAdapter, SerializableMixin], /** @lends WS.Data/Adapter/Abstract.prototype */{
      '[WS.Data/Adapter/Abstract]': true,

      /**
       * @member {String} Разделитель для обозначения пути в данных
       */
      _pathSeparator: '.',

      getProperty: function (data, property) {
         property = property || '';
         var parts = property.split(this._pathSeparator),
            result;
         for (var i = 0; i < parts.length; i++) {
            result = i ?
               (result ? result[parts[i]] : undefined) :
               (data ? data[parts[i]] : undefined);
         }
         return result;
      },

      setProperty: function (data, property, value) {
         if (!data || !(data instanceof Object)) {
            return;
         }
         property = property || '';
         var parts = property.split(this._pathSeparator),
            current = data;
         for (var i = 0, max = parts.length - 1; i <= max; i++) {
            if (i === max) {
               current[parts[i]] = value;
            } else {
               if (current[parts[i]] === undefined) {
                  current[parts[i]] = {};
               }
               current = current[parts[i]];
            }
         }
      },

      serialize: function (data) {
         return serializer.serialize(data);
      }
   });

   var serializer = (function() {
      var serialize = function(data) {
            if (data instanceof Array) {
               return serializeArray(data);
            } else if (data && typeof data === 'object') {
               return serializeObject(data);
            } else {
               return data;
            }
         },
         serializeArray = function (arr) {
            return arr.map(function(item) {
               return serialize(item);
            });
         },
         serializeObject = function (obj) {
            if (typeof obj.getRawData === 'function') {
               //Instance of WS.Data/Entity/Record || WS.Data/Collection/RecordSet || WS.Data/Source/DataSet
               return obj.getRawData(true);
            } else if (obj instanceof Date) {
               return obj.toSQL(Date.SQL_SERIALIZE_MODE_AUTO);
            } else {
               //Check if 'obj' is a scalar value wrapper
               if (obj.valueOf) {
                  obj = obj.valueOf();
               }
               if (obj && typeof obj === 'object') {
                  return serializePlainObject(obj);
               }
               return obj;
            }
         },
         serializePlainObject = function(obj) {
            var result = {},
               proto = Object.getPrototypeOf(obj),
               keys,
               key,
               i;
            if (proto !== null && proto !== Object.prototype) {
               throw new TypeError('Unsupported object type. Only plain objects can be serialized.');
            }
            keys = Object.keys(obj);
            for (i = 0; i < keys.length; i++) {
               key = keys[i];
               result[key] = serialize(obj[key]);
            }
            return result;
         };

      return {
         serialize: serialize
      };
   })();

   return Abstract;
});
