/* global define */
define('WS.Data/Adapter/SbisRecord', [
   'WS.Data/Entity/Abstract',
   'WS.Data/Adapter/IRecord',
   'WS.Data/Entity/ICloneable',
   'WS.Data/Adapter/SbisFormatMixin'
], function (
   Abstract,
   IRecord,
   ICloneable,
   SbisFormatMixin
) {
   'use strict';

   /**
    * Адаптер для записи таблицы данных в формате СБиС.
    * Работает с данными, представленными в виде объекта ({_type: 'record', d: [], s: []}), где
    * <ul>
    *    <li>d - значения полей записи;</li>
    *    <li>s - описание полей записи.</li>
    * </ul>
    *
    * Создадим адаптер для записи:
    * <pre>
    *    var adapter = new SbisRecord({
    *       _type: 'record',
    *       d: [1, 'Test'],
    *       s: [
    *          {n: 'id', t: 'Число целое'},
    *          {n: 'title', t: 'Строка'}
    *       ]
    *    });
    *    adapter.get('title');//'Test'
    * </pre>
    * @class WS.Data/Adapter/SbisRecord
    * @extends WS.Data/Entity/Abstract
    * @implements WS.Data/Adapter/IRecord
    * @implements WS.Data/Entity/ICloneable
    * @mixes WS.Data/Adapter/SbisFormatMixin
    * @public
    * @author Мальцев Алексей
    */
   var SbisRecord = Abstract.extend([IRecord, ICloneable, SbisFormatMixin], /** @lends WS.Data/Adapter/SbisRecord.prototype */{
      '[WS.Data/Adapter/SbisRecord]': true,

      _type: 'record',

      /**
       * @member {String} Разделитель значений при сериализации сложных типов
       */
      _castSeparator: ',',

      /**
       * Конструктор
       * @param {*} data Сырые данные
       */
      constructor: function (data) {
         SbisRecord.superclass.constructor.call(this, data);
         SbisFormatMixin.constructor.call(this, data);
      },

      //region WS.Data/Entity/ICloneable

      clone: function (shallow) {
         return new SbisRecord(shallow ? this.getData() : this._cloneData());
      },

      //endregion WS.Data/Entity/ICloneable

      //region WS.Data/Adapter/SbisFormatMixin

      _buildD: function(at, value) {
         this._data.d.splice(at, 0, value);
      },

      _removeD: function(at) {
         this._data.d.splice(at, 1);
      },

      //endregion WS.Data/Adapter/SbisFormatMixin

      //region Public methods

      has: function (name) {
         return this._has(name);
      },

      get: function (name) {
         var index = this._getFieldIndex(name);
         return index >= 0 ?
            this._cast(this._data.s[index], this._data.d[index]) :
            undefined;
      },

      set: function (name, value) {
         var index = this._getFieldIndex(name);
         if (index < 0) {
            throw new ReferenceError(this._moduleName + '::set(): field "' + name + '" is not defined');
         }
         this._data.d[index] = this._uncast(this._data.s[index], value);
      },

      clear: function () {
         this._touchData();
         SbisFormatMixin.clear.call(this);
         this._data.s.length = 0;
      },

      //endregion Public methods

      //region Protected methods

      _cast: function (format, value) {
         switch (format && format.t) {
            case 'Идентификатор':
               return value instanceof Array ?
                  (value[0] === null ? value[0] : value.join(this._castSeparator, value)) :
                  value;
         }
         return value;
      },

      _uncast: function (format, value) {
         switch (format && format.t) {
            case 'Идентификатор':
               if (value instanceof Array) {
                  return value;
               }
               return typeof value === 'string' ?
                  value.split(this._castSeparator) :
                  [value];
         }
         return value;
      }

      //endregion Protected methods
   });

   return SbisRecord;
});
