/* global define */
define('WS.Data/Adapter/SbisTable', [
   'WS.Data/Entity/Abstract',
   'WS.Data/Format/FieldsFactory',
   'WS.Data/Adapter/ITable',
   'WS.Data/Adapter/IMetaData',
   'WS.Data/Entity/ICloneable',
   'WS.Data/Adapter/SbisRecord',
   'WS.Data/Adapter/SbisFormatMixin',
   'Core/core-merge'
], function (
   Abstract,
   FieldsFactory,
   ITable,
   IMetaData,
   ICloneable,
   SbisRecord,
   SbisFormatMixin,
   coreMerge
) {
   'use strict';

   /**
    * Адаптер для таблицы данных в формате СБиС.
    * Работает с данными, представленными в виде объекта ({_type: 'recordset', d: [], s: []}), где
    * <ul>
    *    <li>d - значения полей для каждой записи;</li>
    *    <li>s - описание полей записи.</li>
    * </ul>
    *
    * Создадим адаптер для таблицы:
    * <pre>
    *    var adapter = new SbisTable({
    *       _type: 'recordset',
    *       d: [
    *          [1, 'Test 1'],
    *          [2, 'Test 2']
    *       ],
    *       s: [
    *          {n: 'id', t: 'Число целое'},
    *          {n: 'title', t: 'Строка'}
    *       ]
    *    });
    *    adapter.at(0);//{d: [1, 'Test 1'], s: [{n: 'id', t: 'Число целое'}, {n: 'title', t: 'Строка'}]}
    * </pre>
    * @class WS.Data/Adapter/SbisTable
    * @extends WS.Data/Entity/Abstract
    * @implements WS.Data/Adapter/ITable
    * @implements WS.Data/Adapter/IMetaData
    * @implements WS.Data/Entity/ICloneable
    * @mixes WS.Data/Adapter/SbisFormatMixin
    * @public
    * @author Мальцев Алексей
    */
   var SbisTable = Abstract.extend([ITable, IMetaData, ICloneable, SbisFormatMixin], /** @lends WS.Data/Adapter/SbisTable.prototype */{
      '[WS.Data/Adapter/SbisTable]': true,

      _type: 'recordset',

      /**
       * Конструктор
       * @param {*} data Сырые данные
       */
      constructor: function (data) {
         SbisTable.superclass.constructor.call(this, data);
         SbisFormatMixin.constructor.call(this, data);
      },

      //region WS.Data/Adapter/IMetaData

      getMetaDataDescriptor: function () {
         var result = [],
            data = this.getData();

         if (!(data instanceof Object)) {
            return result;
         }

         if (data.hasOwnProperty('r')) {
            result.push(FieldsFactory.create({
               name: 'results',
               type: 'record'
            }));
         }

         if (data.hasOwnProperty('p')) {
            result.push(FieldsFactory.create({
               name: 'path',
               type: 'recordset'
            }));
         }

         if (data.hasOwnProperty('n')) {
            var type = 'integer';
            switch (typeof data.n) {
               case 'boolean':
                  type = 'boolean';
                  break;
               case 'object':
                  if (data.n) {
                     type = 'object';
                  }
                  break;
            }

            result.push(FieldsFactory.create({
               name: 'total',
               type: type
            }));

            result.push(FieldsFactory.create({
               name: 'more',
               type: type
            }));
         }

         if (data.hasOwnProperty('m')) {
            var meta = new SbisRecord(data.m);
            meta.getFields().forEach(function(name) {
               result.push(meta.getFormat(name));
            });
         }

         return result;
      },

      getMetaData: function (name) {
         var alias = this._getMetaDataAlias(name),
            data = this.getData();

         if (alias) {
            return data && data instanceof Object ? data[alias] : undefined;
         }

         var meta = new SbisRecord(data.m);
         return meta.get(name);
      },

      setMetaData: function (name, value) {
         var alias = this._getMetaDataAlias(name),
            data = this.getData();

         if (alias) {
            if (data && data instanceof Object) {
               data[alias] = value;
            }
            return;
         }

         var meta = new SbisRecord(data.m);
         meta.set(name, value);
      },

      _getMetaDataAlias: function(name) {
         switch (name) {
            case 'results':
               return 'r';
            case 'path':
               return 'p';
            case 'more':
            case 'total':
               return 'n';
         }
      },

      //endregion WS.Data/Adapter/IMetaData

      //region WS.Data/Entity/ICloneable

      clone: function (shallow) {
         return new SbisTable(shallow ? this.getData() : this._cloneData());
      },

      //endregion WS.Data/Entity/ICloneable

      //region WS.Data/Adapter/SbisFormatMixin

      _buildD: function(at, value) {
         this._data.d.forEach(function(item) {
            item.splice(at, 0, value);
         });
      },

      _removeD: function(at) {
         this._data.d.forEach(function(item) {
            item.splice(at, 1);
         });
      },

      //endregion WS.Data/Adapter/SbisFormatMixin

      //region Public methods

      getCount: function () {
         return this._isValidData() ? this._data.d.length : 0;
      },

      add: function (record, at) {
         this._touchData();
         record = this._normalizeData(record, SbisRecord.prototype._type);

         if (this._data.d.length === 0 && record.s) {
            this._data.s = record.s;
         }

         this._checkFormat(record, '::add()');
         record.s = this._data.s;

         if (at === undefined) {
            this._data.d.push(record.d);
         } else {
            this._checkRowIndex(at, true);
            this._data.d.splice(at, 0, record.d);
         }
      },

      at: function (index) {
         return this._isValidData() && this._data.d[index] ? {
            d: this._data.d[index],
            s: this._data.s
         } : undefined;
      },

      remove: function (at) {
         this._touchData();
         this._checkRowIndex(at);
         this._data.d.splice(at, 1);
      },

      replace: function (record, at) {
         this._touchData();
         this._checkRowIndex(at);
         if (!this._data.s.length && record.s.length) {
            this._data.s = record.s;
         }
         this._checkFormat(record, '::replace()');
         this._data.d[at] = record.d;
      },

      move: function(source, target) {
         this._touchData();
         if (target === source) {
            return;
         }
         var removed = this._data.d.splice(source, 1);
         target === -1 ? this._data.d.unshift(removed.shift()) : this._data.d.splice(target, 0, removed.shift());
      },

      merge: function(acceptor, donor){
         this._touchData();
         this._checkRowIndex(acceptor);
         this._checkRowIndex(donor);
         coreMerge(
            this._data.d[acceptor],
            this._data.d[donor]
         );
         this.remove(donor);
      },

      copy: function(index){
         this._touchData();
         this._checkRowIndex(index);
         var source = this._data.d[index],
            clone = [];
         coreMerge(clone, source, {clone: true});
         this._data.d.splice(1 + index, 0, clone);
         return clone;
      },

      //endregion Public methods

      //region Protected methods

      _checkRowIndex: function(index, addMode) {
         var max = this._data.d.length + (addMode ? 0 : -1);
         if (!(index >= 0 && index <= max)) {
            throw new RangeError(this._moduleName + ': row index ' + index + ' is out of bounds.');
         }
      }

      //endregion Protected methods

   });

   return SbisTable;
});
