/* global define */
define('WS.Data/Collection/ObservableList', [
   'WS.Data/Collection/List',
   'WS.Data/Mediator/IReceiver',
   'WS.Data/Collection/IBind',
   'WS.Data/Collection/EventRaisingMixin',
   'WS.Data/Di'
], function (
   List,
   IMediatorReceiver,
   IBindCollection,
   EventRaisingMixin,
   Di
) {
   'use strict';

   /**
    * Список, в котором можно отслеживать изменения.
    * <pre>
    *    define(['WS.Data/Collection/ObservableList', 'WS.Data/Collection/IBind'], function(ObservableList, IBindCollection) {
    *       var list = new ObservableList({
    *          items: [1, 2, 3]
    *       });
    *
    *       list.subscribe('onCollectionChange', function(eventObject, action, newItems, newItemsIndex, oldItems, oldItemsIndex) {
    *          if (action == IBindCollection.ACTION_REMOVE){
    *             console.log(oldItems);//[1]
    *             console.log(oldItemsIndex);//0
    *          }
    *       });
    *
    *       list.removeAt(0);
    *    });
    * </pre>
    * @class WS.Data/Collection/ObservableList
    * @extends WS.Data/Collection/List
    * @implements WS.Data/Collection/IBind
    * @implements WS.Data/Mediator/IReceiver
    * @mixes WS.Data/Collection/EventRaisingMixin
    * @public
    * @author Мальцев Алексей
    */

   var ObservableList = List.extend([IBindCollection, IMediatorReceiver, EventRaisingMixin], /** @lends WS.Data/Collection/ObservableList.prototype */{
      '[WS.Data/Collection/ObservableList]': true,

      /**
       * @member {Number} Количество измененившихся элементов, при превышении которого генерируется одно событие onCollectionChange с ACTION_RESET, вместо нескольких c другими action
       */
      _resetChangesCount: 100,

      constructor: function ObservableList(options) {
         ObservableList.superclass.constructor.call(this, options);
         EventRaisingMixin.constructor.call(this, options);

         this._publish('onCollectionChange', 'onCollectionItemChange');
      },
      
      //region WS.Data/Collection/List

      assign: function (items) {
         var oldItems = this._itemsSlice(),
            eventsWasRaised = this._eventRaising;

         this._eventRaising = false;
         ObservableList.superclass.assign.call(this, items);
         this._eventRaising = eventsWasRaised;

         if (oldItems.length > 0 || this._$items.length > 0) {
            this._notifyCollectionChange(
               IBindCollection.ACTION_RESET,
               this._itemsSlice(),
               0,
               oldItems,
               0
            );
         }
      },

      append: function (items) {
         var eventsWasRaised = this._eventRaising;

         this._eventRaising = false;
         var count = this.getCount();
         ObservableList.superclass.append.call(this, items);
         this._eventRaising = eventsWasRaised;

         this._notifyCollectionChange(
            IBindCollection.ACTION_ADD,
            this._itemsSlice(count),
            count,
            [],
            0
         );
      },

      prepend: function (items) {
         var eventsWasRaised = this._eventRaising;

         this._eventRaising = false;
         var length = this.getCount();
         ObservableList.superclass.prepend.call(this, items);
         this._eventRaising = eventsWasRaised;

         this._notifyCollectionChange(
            IBindCollection.ACTION_ADD,
            this._itemsSlice(0, this.getCount() - length),
            0,
            [],
            0
         );
      },

      clear: function () {
         var oldItems = this._$items.slice(),
            eventsWasRaised = this._eventRaising;

         this._eventRaising = false;
         ObservableList.superclass.clear.call(this);
         this._eventRaising = eventsWasRaised;

         this._notifyCollectionChange(
            IBindCollection.ACTION_RESET,
            this._itemsSlice(),
            0,
            oldItems,
            0
         );
      },

      add: function (item, at) {
         ObservableList.superclass.add.call(this, item, at);
         at = this._isValidIndex(at) ? at : this.getCount() - 1;
         this._notifyCollectionChange(
            IBindCollection.ACTION_ADD,
            this._itemsSlice(at, at + 1),
            at,
            [],
            0
         );
      },

      removeAt: function (index) {
         var item = ObservableList.superclass.removeAt.call(this, index);
         this._notifyCollectionChange(
            IBindCollection.ACTION_REMOVE,
            [],
            0,
            [item],
            index
         );
         return item;
      },

      replace: function (item, at) {
         var oldItem = this._$items[at];
         ObservableList.superclass.replace.call(this, item, at);

         //Replace with itself has no effect
         if (oldItem !== item) {
            this._notifyCollectionChange(
               IBindCollection.ACTION_REPLACE,
               this._itemsSlice(at, at + 1),
               at,
               [oldItem],
               at
            );
         }
      },

      move: function (from, to) {
         var item = this._$items[from];
         ObservableList.superclass.move.call(this, from, to);

         if (from !== to) {
            this._notifyCollectionChange(
               IBindCollection.ACTION_MOVE,
               [item],
               to,
               [item],
               from
            );
         }
      },

      //endregion WS.Data/Collection/List

      //region WS.Data/Mediator/IReceiver

      relationChanged: function (which, route) {
         var index = this.getIndex(which.target);
         if (index > -1) {
            this._reindex(
               '',
               index,
               1
            );
         }

         var name = route[0];
         if (!name) {
            this._notifyItemChange(
               which.target,
               which.data || {}
            );
         }
      },

      //endregion WS.Data/Mediator/IReceiver

      //region WS.Data/Collection/EventRaisingMixin

      setEventRaising: function(enabled, analyze) {
         EventRaisingMixin.setEventRaising.call(this, enabled, analyze);

         //если стрелять событиями до синхронизации то проекция не всегда сможет найти стрельнувший item или найдет не тот
         if (enabled && analyze && this._silentChangedItems) {
            if (this._silentChangedItems.length >= Math.min(this._resetChangesCount, this._$items.length)) {
               //Если изменилось критическое число элементов, то генерируем reset
               this._notifyCollectionChange(
                  IBindCollection.ACTION_RESET,
                  this._itemsSlice(),
                  0,
                  [],
                  0
               );
            } else {
               //Собираем изменившиеся элементы в пачки
               EventRaisingMixin._extractPacksByList(
                  this,
                  this._silentChangedItems,
                  function(pack, index) {
                     this._notifyCollectionChange(
                        IBindCollection.ACTION_CHANGE,
                        pack,
                        index,
                        pack,
                        index
                     );
                  }.bind(this)
               );
            }
         }
         delete this._silentChangedItems;
      },

      //endregion WS.Data/Collection/EventRaisingMixin

      //region Protected methods

      /**
       * Генерирует событие об изменении элемента
       * @param {*} item Элемент
       * @param {Object} properties Изменившиеся свойства
       */
      _notifyItemChange: function (item, properties) {
         if (this._isNeedNotifyCollectionItemChange()) {
            var index = this.getIndex(item);
            this._notify(
               'onCollectionItemChange',
               this._$items[index],
               index,
               properties
            );
         }

         if (
            (this.hasEventHandlers('onCollectionItemChange') || this.hasEventHandlers('onCollectionChange')) &&
            !this._eventRaising
         ) {
            if (!this._silentChangedItems) {
               this._silentChangedItems = [];
            }
            this._silentChangedItems.push(item);
         }
      },

      /**
       * Извлекает элементы, входящие в указанный отрезок
       * @param {Number} [begin] Индекс, по которому начинать извлечение.
       * @param {Number} [end] Индекс, по которому заканчивать извлечение.
       * @return {Array}
       * @protected
       */
      _itemsSlice: function () {
         return ArraySlice.apply(this._$items, arguments);
      },

      /**
       * Возвращает признак, что нужно генерировать события об изменениях элементов коллекции
       * @return {Boolean}
       * @protected
       */
      _isNeedNotifyCollectionItemChange: function () {
         return this._eventRaising && this.hasEventHandlers('onCollectionItemChange');
      }

      //endregion Protected methods
   }),

   ArraySlice = Array.prototype.slice;

   Di.register('collection.observable-list', ObservableList);

   return ObservableList;
});
