/* global define */
define('WS.Data/Display/Collection', [
   'WS.Data/Collection/IEnumerable',
   'WS.Data/Collection/IList',
   'WS.Data/Collection/EventRaisingMixin',
   'WS.Data/Entity/SerializableMixin',
   'WS.Data/Display/IBindCollection',
   'WS.Data/Display/CollectionEnumerator',
   'WS.Data/Display/Display',
   'WS.Data/Display/CollectionItem',
   'WS.Data/Display/GroupItem',
   'WS.Data/Display/ItemsStrategy/Direct',
   'WS.Data/Display/ItemsStrategy/Group',
   'WS.Data/Functor/Compute',
   'WS.Data/Shim/Set',
   'WS.Data/Di',
   'WS.Data/Utils'
], function (
   IEnumerable,
   IList,
   EventRaisingMixin,
   SerializableMixin,
   IBindCollectionDisplay,
   CollectionEnumerator,
   Display,
   CollectionItem,
   GroupItem,
   DirectItemsStrategy,
   GroupItemsStrategy,
   ComputeFunctor,
   Set,
   Di,
   Utils
) {
   'use strict';

   /**
    * Проекция коллекции - предоставляет методы навигации, фильтрации и сортировки, не меняя при этом оригинальную коллекцию.
    * @class WS.Data/Display/Collection
    * @extends WS.Data/Display/Display
    * @implements WS.Data/Collection/IEnumerable
    * @implements WS.Data/Collection/IList
    * @implements WS.Data/Display/IBindCollection
    * @mixes WS.Data/Entity/SerializableMixin
    * @mixes WS.Data/Collection/EventRaisingMixin
    * @ignoreMethods notifyItemChange
    * @public
    * @author Мальцев Алексей
    */

   var Collection = Display.extend([IEnumerable, IList, IBindCollectionDisplay, SerializableMixin, EventRaisingMixin], /** @lends WS.Data/Display/Collection.prototype */{
      '[WS.Data/Display/Collection]': true,

      /**
       * @typedef {Object} UserSortItem
       * @property {WS.Data/Display/CollectionItem} item Элемент проекции
       * @property {*} collectionItem Элемент коллекции
       * @property {Number} index Индекс элемента проекции
       * @property {Number} collectionIndex Индекс элемента коллекции
       */

      /**
       * @event onBeforeCollectionChange Перед началом изменений коллекции
       */

      /**
       * @event onAfterCollectionChange После окончания изменений коллекции
       */

      _sessionItemContentsGetter: 'getContents',
      _localize: true,

      /**
       * @cfg {WS.Data/Collection/IEnumerable} Оригинальная коллекция
       * @name WS.Data/Display/Collection#collection
       * @see getCollection
       */
      _$collection: null,

      /**
       * @cfg {Array.<Function(*, Number, WS.Data/Display/CollectionItem, Number): Boolean>|Function(*, Number, WS.Data/Display/CollectionItem, Number): Boolean} Пользовательские методы фильтрации элементов проекциию. Аргументы: элемент коллекции, позиция в коллекции, элемент проекции, позиция в проекции. Должен вернуть Boolean - признак, что элемент удовлетворяет условиям фильтрации.
       * @name WS.Data/Display/Collection#filter
       * @example
       * Отберем персонажей женского пола:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list,
       *             filter: function(collectionItem) {
       *                return collectionItem.gender === 'F';
       *             }
       *          });
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().name);
       *       });
       *       //output: 'Turanga Leela', 'Amy Wong'
       *    });
       * </pre>
       * @see getFilter
       * @see setFilter
       * @see addFilter
       * @see removeFilter
       */
      _$filter: null,

      /**
       * @cfg {Function(*, Number, WS.Data/Display/CollectionItem): String|null} Метод группировки элементов проекции. Аргументы: элемент коллекции, позиция в коллекции, элемент проекции. Должен вернуть идентификатор группы.
       * @name WS.Data/Display/Collection#group
       * @example
       * Сгруппируем персонажей по полу:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List',
       *       'WS.Data/Display/Collection',
       *       'WS.Data/Display/GroupItem'
       *    ], function(List, CollectionDisplay, GroupItem) {
       *       var list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list
       *             group: function(collectionItem, index, item) {
       *                return collectionItem.gender;
       *             }
       *          });
       *
       *       display.each(function(item, index) {
       *          if (item instanceof GroupItem) {
       *             console.log('[' + item.getContents() + ']';
       *          } else {
       *             console.log(item.getContents().name);
       *          }
       *       });
       *       //output: '[M]', 'Philip J. Fry', 'Professor Farnsworth', '[F]', 'Turanga Leela', 'Amy Wong', '[R]', 'Bender Bending Rodriguez'
       *    });
       * </pre>
       * @see getGroup
       * @see setGroup
       */
      _$group: null,

      /**
       * @cfg {Array.<Function(UserSortItem, UserSortItem): Number>|Function(UserSortItem, UserSortItem): Number} Пользовательские методы сортировки элементов. Аргументы: 2 объекта типа {@link UserSortItem}, должен вернуть -1|0|1 (см. Array.prototype.sort())
       * @name WS.Data/Display/Collection#sort
       * @example
       * Отсортируем коллекцию по возрастанию значения поля title:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var display = new CollectionDisplay({
       *          collection: new List({
       *             items: [
       *                {id: 1, title: 'foo'},
       *                {id: 2, title: 'bar'}
       *             ]
       *          }),
       *          sort: function(a, b) {
       *             return a.collectionItem.title - b.collectionItem.title;
       *          }
       *       });
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().title;
       *       });
       *       //output: 'bar', 'foo'
       *    });
       * </pre>
       * @see getSort
       * @see setSort
       * @see addSort
       * @see removeSort
       */
      _$sort: null,

      /**
       * @cfg {String} Название свойства элемента коллекции, содержащего его уникальный идентификатор.
       * @name WS.Data/Display/Collection#idProperty
       */
      _$idProperty: '',

      /**
       * @cfg {Boolean} Обеспечивать уникальность элементов (элементы с повторяющимися идентфикаторами будут игнорироваться). Работает только если задано {@link idProperty}.
       * @name WS.Data/Display/Collection#unique
       */
      _$unique: false,

      /**
       * @cfg {Array.<String>} Названия свойств элемента коллекции, от которых зависят фильтрация, сортировка, группировка.
       * @name WS.Data/Display/Collection#importantItemProperties
       * @remark
       * Изменение любого из указанных свойств элемента коллекции приведет к перерасчету фильтрации, сортировки и группировки.
       */
      _$importantItemProperties: null,

      /**
       * @member {String|Function} Тип элемента проекции
       */
      _itemModule: 'display.collection-item',

      /**
       * @member {WS.Data/Display/ItemsStrategy/Abstract} Стратегия получения элементов проекции
       */
      _itemsStrategy: null,

      /**
       * @member {WS.Data/Display/ItemsStrategy/Group} Стратегия группировки элементов проекции
       */
      _itemsGroupStrategy: null,

      /**
       * @member {Boolean} Коллекция синхронизирована с проекцией (все события, приходящие от нее, соответсвуют ее состоянию)
       */
      _sourceCollectionSynchronized: true,

      /**
       * @member {Array.<Function>} Обработчики событий коллекции, отложенные до момента синхронизации
       */
      _sourceCollectionDelayedCallbacks: null,

      /**
       * @member {Array.<Boolean>} Результат применения фильтра: индекс элемента проекции -> прошел фильтр
       */
      _filterMap: null,

      /**
       * @member {Array.<Number>} Результат применения группировки: индекс элемента проекции -> индекс группы
       */
      _groupMap: null,

      /**
       * @member {Array.<Object>} Доступные способы сортировки
       */
      _sorters: null,

      /**
       * @member {Array.<Number>} Результат применения сортировки: индекс после сортировки -> индекс элемента проекции
       */
      _sortMap: null,

      /**
       * @member {WS.Data/Display/CollectionEnumerator} Служебный энумератор проекции - для организации курсора
       */
      _cursorEnumerator: null,

      /**
       * @member {WS.Data/Display/CollectionEnumerator} Служебный энумератор проекции - для поиска по свойствам и поиска следующего или предыдущего элемента относительно заданного
       */
      _utilityEnumerator: null,

      /**
       * @member {Function} Обработчик события об изменении коллекции
       */
      _onSourceCollectionChange: null,

      /**
       * @member {Function} Обработчик события об изменении элемента коллекции
       */
      _onSourceCollectionItemChange: null,

      /**
       * @member {Function} Обработчик события об изменении генерации событий коллекции
       */
      _onSourceCollectionEventRaisingChange: null,

      constructor: function Collection(options) {
         this._$importantItemProperties = [];
         this._$filter = [];
         this._$sort = [];
         this._filterMap = [];
         this._sortMap = [];
         this._groupMap = [];

         Collection.superclass.constructor.call(this, options);
         EventRaisingMixin.constructor.call(this, options);

         if (!this._$collection) {
            throw new Error(this._moduleName + ': source collection is empty');
         }
         if (this._$collection instanceof Array) {
            this._$collection = Di.create('collection.list', {items: this._$collection});
         }
         if (!this._$collection['[WS.Data/Collection/IEnumerable]']) {
            throw new TypeError(this._moduleName + ': source collection should implement WS.Data/Collection/IEnumerable');
         }

         if (typeof this._$sort === 'function') {
            this._$sort = [this._$sort];
         }
         this._$sort = this._$sort instanceof Array ? this._$sort.filter(function(item) {
            return typeof item === 'function';
         }) : [];

         if (typeof this._$filter === 'function') {
            this._$filter = [this._$filter];
         }
         this._$filter = this._$filter instanceof Array ? this._$filter.filter(function(item) {
            return typeof item === 'function';
         }) : [];

         if (this._$idProperty) {
            this._setImportantProperty(this._$idProperty);
         }

         this._publish('onCurrentChange', 'onCollectionChange', 'onBeforeCollectionChange', 'onAfterCollectionChange');

         this._initSorters();
         this._switchSortersByUserSort();
         this._switchSorterInner('group', this._isGrouped());

         this._reBuild();
         this._bindHandlers();
         if (this._$collection['[WS.Data/Collection/IBind]']) {
            //implements WS.Data/Collection/IBind
            this._$collection.subscribe('onCollectionChange', this._onSourceCollectionChange);
            this._$collection.subscribe('onCollectionItemChange', this._onSourceCollectionItemChange);
         }
         if (this._$collection['[WS.Data/Collection/EventRaisingMixin]']) {
            //mixes WS.Data/Collection/EventRaisingMixin
            this._$collection.subscribe('onEventRaisingChange', this._onSourceCollectionEventRaisingChange);
         }
      },

      destroy: function () {
         if (!this._$collection.isDestroyed()) {
            if (this._$collection['[WS.Data/Collection/IBind]']) {
               //implements WS.Data/Collection/IBind
               this._$collection.unsubscribe('onCollectionChange', this._onSourceCollectionChange);
               this._$collection.unsubscribe('onCollectionItemChange', this._onSourceCollectionItemChange);
            }
            if (this._$collection['[WS.Data/Collection/EventRaisingMixin]']) {
               //mixes WS.Data/Collection/EventRaisingMixin
               this._$collection.unsubscribe('onEventRaisingChange', this._onSourceCollectionEventRaisingChange);
            }
         }

         this._unbindHandlers();
         this._itemsStrategy = null;
         this._itemsGroupStrategy = null;
         this._filterMap = [];
         this._sortMap = [];
         this._groupMap = [];
         this._cursorEnumerator = null;
         this._utilityEnumerator = null;

         Collection.superclass.destroy.call(this);
      },

      //region mutable

      /**
       * Возвращает элемент проекции с указанным идентификатором экземпляра.
       * @param {String} instanceId Идентификатор экземпляра.
       * @return {WS.Data/Display/CollectionItem}
       * @state mutable
       */
      getByInstanceId: function(instanceId) {
         return this.at(
            this._getUtilityEnumerator().getIndexByValue('instanceId', instanceId)
         );
      },

      /**
       * Возвращает индекс элемента проекции с указанным идентификатором экземпляра.
       * @param {String} instanceId Идентификатор экземпляра.
       * @return {Number}
       * @state mutable
       */
      getIndexByInstanceId: function (instanceId) {
         return this._getUtilityEnumerator().getIndexByValue('instanceId', instanceId);
      },

      //endregion mutable

      //region WS.Data/Collection/IEnumerable

      /**
       * Возвращает энумератор для перебора элементов проекции
       * @return {WS.Data/Display/CollectionEnumerator}
       */
      getEnumerator: function () {
         return this._getEnumerator();
      },

      /**
       * Перебирает все элементы проекции, начиная с первого.
       * @param {Function(WS.Data/Display/CollectionItem, Number)} callback Ф-я обратного вызова для каждого элемента коллекции (аргументами придут элемент коллекции и его порядковый номер)
       * @param {Object} [context] Контекст вызова callback
       * @example
       * Сгруппируем персонажей по полу:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List',
       *       'WS.Data/Display/Collection',
       *       'WS.Data/Display/GroupItem'
       *    ], function(List, CollectionDisplay, GroupItem) {
       *       var list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list
       *          });
       *
       *       display.setGroup(function(collectionItem, index, item) {
       *          return collectionItem.gender;
       *       });
       *
       *       display.each(function(item, index) {
       *          if (item instanceof GroupItem) {
       *             console.log('[' + item.getContents() + ']');
       *          } else {
       *             console.log(item.getContents().name);
       *          }
       *       });
       *       //output: '[M]', 'Philip J. Fry', 'Professor Farnsworth', '[F]', 'Turanga Leela', 'Amy Wong', '[R]', 'Bender Bending Rodriguez'
       *    });
       * </pre>
       */
      each: function (callback, context) {
         var enumerator = this.getEnumerator(),
            index;
         while (enumerator.moveNext()) {
            index = enumerator.getCurrentIndex();
            callback.call(
               context,
               enumerator.getCurrent(),
               index
            );
         }
      },

      //endregion WS.Data/Collection/IEnumerable

      //region WS.Data/Collection/IList

      assign: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      append: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      prepend: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      clear: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      add: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      at: function (index) {
         return this._getUtilityEnumerator().at(index);
      },

      remove: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      removeAt: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      replace: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      move: function () {
         throw new Error(MESSAGE_READ_ONLY);
      },

      getIndex: function (item) {
         if (!(item instanceof CollectionItem)) {
            return -1;
         }
         return this.getIndexByInstanceId(item.getInstanceId());
      },

      /**
       * Возвращает количество элементов проекции.
       * @param {Boolean} [skipGroups=false] Не считать группы
       * @return {Number}
       */
      getCount: function (skipGroups) {
         var count = 0;
         if (skipGroups && this._isGrouped()) {
            this.each(function(item) {
               if (!(item instanceof GroupItem)) {
                  count++;
               }
            });
         } else {
            count = this._getUtilityEnumerator().getCount();
         }
         return count;
      },

      //endregion WS.Data/Collection/IList

      //region WS.Data/Entity/SerializableMixin

      _getSerializableState: function(state) {
         state = SerializableMixin._getSerializableState.call(this, state);

         state.items = this._getItems();

         return state;
      },

      _setSerializableState: function(state) {
         var fromSerializableMixin = SerializableMixin._setSerializableState(state);
         return function() {
            fromSerializableMixin.call(this);

            var items = state.items;
            this._getItemsStrategy().setItems(items);

            for (var i = 0, count = items.length; i < count; i++) {
               items[i].setOwner(this);
            }

            this._reIndex();
            this._reAnalize();
         };
      },

      //endregion WS.Data/Entity/SerializableMixin

      //region Public methods

      //region Access

      /**
       * Возвращает оригинальную коллекцию
       * @return {WS.Data/Collection/IEnumerable}
       * @see collection
       */
      getCollection: function () {
         return this._$collection;
      },

      /**
       * Возвращает число элементов оригинальной коллекции
       * @return {Number}
       * @see collection
       */
      getCollectionCount: function () {
         var collection = this.getCollection();
         if (collection['[WS.Data/Collection/IList]']) {
            //implements WS.Data/Collection/IList
            return collection.getCount();
         }

         var enumerator = collection.getEnumerator(),
            count = 0;
         enumerator.reset();
         while (enumerator.moveNext()) {
            count++;
         }
         return count;
      },

      /**
       * Возвращает элементы проекции (без учета сортировки, фильтрации и группировки)
       * @return {Array.<WS.Data/Display/CollectionItem>}
       */
      getItems: function () {
         return this._getItems().slice();
      },

      //endregion Access

      //region Navigation

      /**
       * Возвращает текущий элемент
       * @return {WS.Data/Display/CollectionItem}
       */
      getCurrent: function () {
         return this._getCursorEnumerator().getCurrent();
      },

      /**
       * Устанавливает текущий элемент
       * @param {WS.Data/Display/CollectionItem} item Новый текущий элемент
       * @param {Boolean} [silent=false] Не генерировать событие onCurrentChange
       */
      setCurrent: function (item, silent) {
         var oldCurrent = this.getCurrent();
         if (oldCurrent !== item) {
            var enumerator = this._getCursorEnumerator(),
               oldPosition = this.getCurrentPosition();
            enumerator.setCurrent(item);

            if (!silent) {
               this._notifyCurrentChange(
                  this.getCurrent(),
                  oldCurrent,
                  enumerator.getPosition(),
                  oldPosition
               );
            }
         }
      },

      /**
       * Возвращает позицию текущего элемента
       * @return {Number}
       */
      getCurrentPosition: function () {
         return this._getCursorEnumerator().getPosition();
      },

      /**
       * Устанавливает позицию текущего элемента
       * @param {Number} position Позиция текущего элемента. Значение -1 указывает, что текущий элемент не выбран.
       * @param {Boolean} [silent=false] Не генерировать событие onCurrentChange
       */
      setCurrentPosition: function (position, silent) {
         var oldPosition = this.getCurrentPosition();
         if (position !== oldPosition) {
            var oldCurrent = this.getCurrent();
            this._getCursorEnumerator().setPosition(position);
            if (!silent) {
               this._notifyCurrentChange(
                  this.getCurrent(),
                  oldCurrent,
                  position,
                  oldPosition
               );
            }
         }
      },

      /**
       * Возвращает первый элемент
       * @return {WS.Data/Display/CollectionItem}
       */
      getFirst: function () {
         var enumerator = this._getUtilityEnumerator(),
            item;

         enumerator.setPosition(0);
         item = enumerator.getCurrent();

         if (item instanceof GroupItem) {
            return this._getNearbyItem(
               enumerator,
               item,
               true,
               true
            );
         }

         return item;
      },

      /**
       * Возвращает последний элемент
       * @return {WS.Data/Display/CollectionItem}
       */
      getLast: function () {
         var enumerator = this._getUtilityEnumerator(),
            lastIndex = enumerator.getCount() - 1,
            item;

         enumerator.setPosition(lastIndex);
         item = enumerator.getCurrent();

         if (item instanceof GroupItem) {
            return this._getNearbyItem(
               enumerator,
               undefined,
               false,
               true
            );
         }

         return item;
      },

      /**
       * Возвращает следующий элемент относительно item
       * @param {WS.Data/Display/CollectionItem} item элемент проекции
       * @return {WS.Data/Display/CollectionItem}
       */
      getNext: function (item) {
         return this._getNearbyItem(
            this._getUtilityEnumerator(),
            item,
            true,
            true
         );
      },

      /**
       * Возвращает предыдущий элемент относительно item
       * @param {WS.Data/Display/CollectionItem} index элемент проекции
       * @return {WS.Data/Display/CollectionItem}
       */
      getPrevious: function (item, skipGroups) {
         return this._getNearbyItem(
            this._getUtilityEnumerator(),
            item,
            false,
            true
         );
      },

      /**
       * Устанавливает текущим следующий элемент
       * @return {Boolean} Есть ли следующий элемент
       */
      moveToNext: function () {
         var oldCurrent = this.getCurrent(),
             oldCurrentPosition = this.getCurrentPosition(),
             hasNext = this._getCursorEnumerator().moveNext();
         if (hasNext) {
            this._notifyCurrentChange(
               this.getCurrent(),
               oldCurrent,
               this.getCurrentPosition(),
               oldCurrentPosition
            );
         }
         return hasNext;
      },

      /**
       * Устанавливает текущим предыдущий элемент
       * @return {Boolean} Есть ли предыдущий элемент
       */
      moveToPrevious: function () {
         var oldCurrent = this.getCurrent(),
             oldCurrentPosition = this.getCurrentPosition(),
             hasPrevious = this._getCursorEnumerator().movePrevious();
         if (hasPrevious) {
            this._notifyCurrentChange(
               this.getCurrent(),
               oldCurrent,
               this.getCurrentPosition(),
               oldCurrentPosition
            );
         }
         return hasPrevious;
      },

      /**
       * Устанавливает текущим первый элемент
       * @return {Boolean} Есть ли первый элемент
       */
      moveToFirst: function () {
         if (this.getCurrentPosition() === 0) {
            return false;
         }
         this.setCurrentPosition(0);
         return this._getCursorEnumerator().getPosition() === 0;
      },

      /**
       * Устанавливает текущим последний элемент
       * @return {Boolean} Есть ли последний элемент
       */
      moveToLast: function () {
         var position = this.getCount() - 1;
         if (this.getCurrentPosition() === position) {
            return false;
         }
         this.setCurrentPosition(position);
         return this.getCurrentPosition() === position;
      },

      /**
       * Возвращает индекс элемента в коллекции по его индексу в проекции
       * @param {Number} index Индекс элемента в проекции
       * @return {Number} Индекс элемента в коллекции
       */
      getSourceIndexByIndex: function (index) {
         var sourceIndex = this._getUtilityEnumerator().getSourceByInternal(index);
         sourceIndex = sourceIndex === undefined || sourceIndex === null ? -1 : sourceIndex;
         return this._getSourceIndex(sourceIndex);
      },

      /**
       * Возвращает индекс элемента проекции в коллекции
       * @param {WS.Data/Display/CollectionItem} item Элемент проекции
       * @return {Number} Индекс элемента проекции в коллекции
       */
      getSourceIndexByItem: function (item) {
         var index = this.getIndex(item);
         return index === -1 ? -1 : this.getSourceIndexByIndex(index);
      },

      /**
       * Возвращает индекс элемента в проекции по индексу в коллекции
       * @param {Number} index Индекс элемента в коллекции
       * @return {Number} Индекс элемента в проекции
       */
      getIndexBySourceIndex: function (index) {
         index = this._getItemIndex(index);
         var itemIndex = this._getUtilityEnumerator().getInternalBySource(index);
         return itemIndex === undefined || itemIndex === null ? -1 : itemIndex;
      },

      /**
       * Возвращает позицию элемента коллекции в проекции.
       * @param {*} item Элемент коллекции
       * @return {Number} Позиция элемента в проекции или -1, если не входит в проекцию
       */
      getIndexBySourceItem: function (item) {
         var collection = this.getCollection(),
            sourceIndex = -1;

         if (collection && collection['[WS.Data/Collection/IList]']) {
            //implements WS.Data/Collection/IList
            sourceIndex = collection.getIndex(item);
         } else {
            collection.each(function(value, index) {
               if (sourceIndex === -1 && value === item) {
                  sourceIndex = index;
               }
            }, this, this._localize);
         }
         return sourceIndex === -1 ? -1 : this.getIndexBySourceIndex(sourceIndex);
      },

      /**
       * Возвращает элемент проекции по индексу коллекции.
       * @param {Number} index Индекс элемента в коллекции
       * @return {WS.Data/Display/CollectionItem} Элемент проекции или undefined, если index не входит в проекцию
       */
      getItemBySourceIndex: function (index) {
         index = this.getIndexBySourceIndex(index);
         return index === -1 ? undefined : this.at(index);
      },

      /**
       * Возвращает элемент проекции для элемента коллекции.
       * @param {*} item Элемент коллекции
       * @return {WS.Data/Display/CollectionItem} Элемент проекции или undefined, если item не входит в проекцию
       */
      getItemBySourceItem: function (item) {
         var index = this.getIndexBySourceItem(item);
         return index === -1 ? undefined : this.at(index);
      },

      //endregion Navigation

      //region Changing

      /**
       * Возвращает пользовательские методы фильтрации элементов проекции
       * @return {Array.<Function(*, Number, WS.Data/Display/CollectionItem, Number): Boolean>}
       * @see filter
       * @see setFilter
       * @see addFilter
       * @see removeFilter
       */
      getFilter: function () {
         return this._$filter.slice();
      },

      /**
       * Устанавливает пользовательские методы фильтрации элементов проекции. Вызов метода без аргументов приведет к удалению всех пользовательских фильтров.
       * @param {...Function(*, Number, WS.Data/Display/CollectionItem, Number): Boolean} [filter] Методы фильтрации элементов: аргументами приходят элемент коллекции, позиция в коллекции, элемент проекции, позиция в проекции. Должен вернуть Boolean - признак, что элемент удовлетворяет условиям фильтрации.
       * @see filter
       * @see getFilter
       * @see addFilter
       * @see removeFilter
       * @example
       * Отберем персонажей женского пола:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list
       *          });
       *
       *       display.setFilter(function(collectionItem, index, item) {
       *          return collectionItem.gender === 'F';
       *       });
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().name);
       *       });
       *       //output: 'Turanga Leela', 'Amy Wong'
       *    });
       * </pre>
       */
      setFilter: function () {
         var filters;
         if (arguments[0] instanceof Array) {
            filters = arguments[0];
         } else {
            filters = Array.prototype.slice.call(arguments);
         }

         if (this._$filter.length === filters.length) {
            var changed = false;
            for (var i = 0; i < filters.length; i++) {
               if (this._$filter[i] !== filters[i]) {
                  changed = true;
                  break;
               }
            }

            if (!changed) {
               return;
            }
         }

         this._$filter = filters.filter(function(item) {
            return typeof item === 'function';
         });

         var session = this._startUpdateSession();
         this._reFilter();
         this._removeEmptyGroups();
         this._finishUpdateSession(session);
      },

      /**
       * Добавляет пользовательский метод фильтрации элементов проекции, если такой еще не был задан.
       * @param {Function(*, Number, WS.Data/Display/CollectionItem, Number): Boolean} filter Метод фильтрации элементов: аргументами приходят элемент коллекции, позиция в коллекции, элемент проекции, позиция в проекции. Должен вернуть Boolean - признак, что элемент удовлетворяет условиям фильтрации.
       * @param {Number} [at] Порядковый номер метода (если не передан, добавляется в конец)
       * @see filter
       * @see getFilter
       * @see setFilter
       * @see removeFilter
       * @example
       * Отберем персонажей женского пола:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list
       *          });
       *
       *       display.addFilter(function(collectionItem, index, item) {
       *          return collectionItem.gender === 'F';
       *       });
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().name);
       *       });
       *       //output: 'Turanga Leela', 'Amy Wong'
       *    });
       * </pre>
       */
      addFilter: function (filter, at) {
         if (this._$filter.indexOf(filter) > -1) {
            return;
         }
         if (at === undefined) {
            this._$filter.push(filter);
         } else {
            this._$filter.splice(at, 0, filter);
         }

         var session = this._startUpdateSession();
         this._reFilter();
         this._removeEmptyGroups();
         this._finishUpdateSession(session);
      },

      /**
       * Удаляет пользовательский метод фильтрации элементов проекции.
       * @param {Function(*, Number, WS.Data/Display/CollectionItem, Number): Boolean} filter Метод фильтрации элементов: аргументами приходят элемент коллекции, позиция в коллекции, элемент проекции, позиция в проекции. Должен вернуть Boolean - признак, что элемент удовлетворяет условиям фильтрации.
       * @return {Boolean} Был ли установлен такой метод фильтрации
       * @see filter
       * @see getFilter
       * @see setFilter
       * @see addFilter
       * @example
       * Уберем фильтрацию персонажей по полу:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var filter = function(collectionItem, index, item) {
       *             return collectionItem.gender === 'F';
       *          }),
       *          list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list,
       *             filter: filter
       *          });
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().name);
       *       });
       *       //output: 'Turanga Leela', 'Amy Wong'
       *
       *       display.removeFilter(filter);
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().name);
       *       });
       *       //output: 'Philip J. Fry', 'Turanga Leela', 'Professor Farnsworth', 'Amy Wong', 'Bender Bending Rodriguez'
       *    });
       * </pre>
       */
      removeFilter: function (filter) {
         var at = this._$filter.indexOf(filter);
         if (at === -1) {
            return false;
         }

         this._$filter.splice(at, 1);

         var session = this._startUpdateSession();
         this._reFilter();
         this._removeEmptyGroups();
         this._finishUpdateSession(session);

         return true;
      },

      /**
       * Возвращает метод группировки элементов проекции
       * @return {Function}
       * @see group
       * @see setGroup
       */
      getGroup: function () {
         return this._$group;
      },

      /**
       * Устанавливает метод группировки элементов проекции. Для сброса ранее установленной группировки следует вызвать этот метод без параметров.
       * @param {Function(*, Number, WS.Data/Display/CollectionItem): String|null} group Метод группировки элементов: аргументами приходят элемент коллекции, его позиция, элемент проекции. Должен вернуть String|Number - группу, в которую входит элемент.
       * @see group
       * @see getGroup
       */
      setGroup: function (group) {
         if (this._$group === group) {
            return;
         }
         
         var session = this._startUpdateSession();
         this._$group = group;
         this._switchSorterInner('group', this._isGrouped());
         this._resetGroups();
         if (this._isGrouped()) {
            this._reGroup();
         }
         this._reSort();
         if (this._isFiltered()) {
            this._reFilter();
         }
         this._removeEmptyGroups();
         this._finishUpdateSession(session);
      },

      /**
       * Возвращает элементы группы. Учитывается сортировка и фильтрация.
       * @param {String} groupId Идентификатор группы, элементы которой требуется получить
       * @return {Array.<WS.Data/Display/CollectionItem>}
       * @example
       * Получим персонажей мужского пола:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list
       *          });
       *
       *       display.setGroup(function(collectionItem, index, item) {
       *          return collectionItem.gender;
       *       });
       *
       *       var males = display.getGroupItems('M'),
       *          male,
       *          i;
       *       for (i = 0; i < males.length; i++) {
       *          male = males[i].getContents();
       *          console.log(male.name);
       *       }
       *       //output: 'Philip J. Fry', 'Professor Farnsworth'
       *    });
       * </pre>
       */
      getGroupItems: function (groupId) {
         var filterMap = this._filterMap,
            sortMap = this._sortMap,
            groupMap = this._groupMap,
            groupIndex = this._itemsGroupStrategy.groups.reduce(function(prev, curr, index) {
               return prev === -1 ? (curr.getContents() === groupId ? index : -1 ) : prev;
            }, -1),
            groupItems = [],
            items = this._getItems(),
            index,
            sortedIndex;

         if (groupIndex > -1) {
            for (index = 0; index < groupMap.length; index++) {
               if (groupMap[index] === groupIndex && filterMap[index]) {
                  sortedIndex = sortMap.indexOf(index);
                  if (sortedIndex > -1 && !(items[index] instanceof GroupItem)) {
                     groupItems.push(sortedIndex);
                  }
               }
            }
            groupItems.sort(function (a, b) {
               return a - b;
            });
         }

         return groupItems.map(function(sortedIndex) {
            return items[sortMap[sortedIndex]];
         });
      },

      /**
       * Возвращает идентификтор группы по индексу элемента в проекции
       * @param {Number} index Индекс элемента в проекции
       * @return {String|Number}
       * @example
       * Сгруппируем персонажей по полу:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var list = new List({
       *             items: [
       *                {name: 'Philip J. Fry', gender: 'M'},
       *                {name: 'Turanga Leela', gender: 'F'},
       *                {name: 'Professor Farnsworth', gender: 'M'},
       *                {name: 'Amy Wong', gender: 'F'},
       *                {name: 'Bender Bending Rodriguez', gender: 'R'}
       *             ]
       *          }),
       *          display = new CollectionDisplay({
       *             collection: list
       *          });
       *
       *       display.setGroup(function(collectionItem, index, item) {
       *          return collectionItem.gender;
       *       });
       *
       *       var enumerator = display.getEnumerator(),
       *          index = 0,
       *          item,
       *          group,
       *          contents;
       *       while (enumerator.moveNext()) {
       *          item = enumerator.getCurrent();
       *          group = display.getGroupByIndex(index);
       *          contents = item.getContents();
       *          console.log(group + ': ' + contents.name);
       *          index++;
       *       }
       *       //output: 'M: Philip J. Fry', 'M: Professor Farnsworth', 'F: Turanga Leela', 'F: Amy Wong', 'R: Bender Bending Rodriguez'
       *    });
       * </pre>
       */
      getGroupByIndex: function (index) {
         var groupIndex = this._getGroupIndexByIndex(index);
         return this._getGroupIdByIndex(groupIndex);
      },

      /**
       * Возвращает пользовательские методы сортировки элементов проекции
       * @return {Array.<Function>}
       * @see sort
       * @see setSort
       * @see addSort
       */
      getSort: function () {
         return this._$sort.slice();
      },

      /**
       * Устанавливает пользовательские методы сортировки элементов проекции. Вызов метода без аргументов приведет к удалению всех пользовательских сортировок.
       * @param {...Function(UserSortItem, UserSortItem): Number} [sort] Методы сортировки элементов: аргументами приходят 2 объекта типа {@link UserSortItem}, должен вернуть -1|0|1 (см. Array.prototype.sort())
       * @see sort
       * @see getSort
       * @see addSort
       * @see removeSort
       * @example
       * Отсортируем коллекцию по возрастанию значения поля title:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var display = new CollectionDisplay({
       *          collection: new List({
       *             items: [
       *                {id: 1, title: 'foo'},
       *                {id: 2, title: 'bar'}
       *             ]
       *          })
       *       });
       *
       *       display.setSort(function(a, b) {
       *          return a.collectionItem.title > b.collectionItem.title;
       *       });
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().title;
       *       });
       *       //output: 'bar', 'foo'
       *    });
       * </pre>
       * Отсортируем коллекцию сначала по title, а потом - по id:
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var display = new CollectionDisplay({
       *          collection: new List({
       *             items: [
       *                {id: 4, title: 'foo'},
       *                {id: 3, title: 'bar'},
       *                {id: 2, title: 'foo'}
       *             ]
       *          })
       *       });
       *
       *       display.setSort(function(a, b) {
       *          return a.collectionItem.title > b.collectionItem.title;
       *       }, function(a, b) {
       *          return a.collectionItem.id - b.collectionItem.id;
       *       });
       *
       *       display.each(function(item) {
       *          console.log(item.getContents().id;
       *       });
       *       //output: 3, 2, 4
       *    });
       * </pre>
       */
      setSort: function () {
         var sorts;
         if (arguments[0] instanceof Array) {
            sorts = arguments[0];
         } else {
            sorts = Array.prototype.slice.call(arguments);
         }

         if (this._$sort.length === sorts.length) {
            var changed = false;
            for (var i = 0; i < sorts.length; i++) {
               if (this._$sort[i] !== sorts[i]) {
                  changed = true;
                  break;
               }
            }

            if (!changed) {
               return;
            }
         }

         this._$sort = sorts.filter(function(item) {
            return typeof item === 'function';
         });

         var session = this._startUpdateSession();
         this._switchSortersByUserSort();
         this._reSort();
         if (this._isFiltered()) {
            this._reFilter();
            this._removeEmptyGroups();
         }
         this._finishUpdateSession(session);
      },

      /**
       * Добавляет пользовательский метод сортировки элементов проекции, если такой еще не был задан.
       * @param {Function(UserSortItem, UserSortItem): Number} [sort] Метод сортировки элементов: аргументами приходят 2 объекта типа {@link UserSortItem}, должен вернуть -1|0|1 (см. Array.prototype.sort())
       * @param {Number} [at] Порядковый номер метода (если не передан, добавляется в конец)
       * @see sort
       * @see getSort
       * @see setSort
       * @see removeSort
       * @example
       * Отсортируем коллекцию по возрастанию значения поля id
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var display = new CollectionDisplay({
       *          collection: new List({
       *             items: [
       *                {id: 1, title: 'foo'},
       *                {id: 2, title: 'bar'}
       *             ]
       *          })
       *       });
       *
       *       display.addSort(function(a, b) {
       *          return a.collectionItem.id - b.collectionItem.id
       *       });
       *    });
       * </pre>
       */
      addSort: function (sort, at) {
         if (this._$sort.indexOf(sort) > -1) {
            return;
         }
         if (at === undefined) {
            this._$sort.push(sort);
         } else {
            this._$sort.splice(at, 0, sort);
         }

         var session = this._startUpdateSession();
         this._switchSortersByUserSort();
         this._reSort();
         if (this._isFiltered()) {
            this._reFilter();
            this._removeEmptyGroups();
         }
         this._finishUpdateSession(session);
      },

      /**
       * Удаляет пользовательский метод сортировки элементов проекции.
       * @param {Function(UserSortItem, UserSortItem): Number} [sort] Метод сортировки элементов: аргументами приходят 2 объекта типа {@link UserSortItem}, должен вернуть -1|0|1 (см. Array.prototype.sort())
       * @return {Boolean} Был ли установлен такой метод сортировки
       * @see sort
       * @see getSort
       * @see setSort
       * @see addSort
       * @example
       * Отсортируем коллекцию по возрастанию значения поля id
       * <pre>
       *    require([
       *       'WS.Data/Collection/List'
       *       'WS.Data/Display/Collection'
       *    ], function(List, CollectionDisplay) {
       *       var handler = function(a, b) {
       *             return a.item.id - b.item.id
       *          },
       *          display = new CollectionDisplay({
       *             collection: new List({
       *                items: [
       *                   {id: 1, title: 'foo'},
       *                   {id: 2, title: 'bar'}
       *                ]
       *             }),
       *             sort: handler
       *          });
       *
       *       //...
       *       display.removeSort(handler);
       *    });
       * </pre>
       */
      removeSort: function (sort) {
         var at = this._$sort.indexOf(sort);
         if (at === -1) {
            return false;
         }

         this._$sort.splice(at, 1);

         var session = this._startUpdateSession();
         this._switchSortersByUserSort();
         this._reSort();
         if (this._isFiltered()) {
            this._reFilter();
            this._removeEmptyGroups();
         }
         this._finishUpdateSession(session);

         return true;
      },

      /**
       * Включает-выключает способ сортировки
       * @param {String} name Название способа сортировки
       * @param {Boolean} enabled Включен или выключен
       */
      switchSorter: function (name, enabled) {
         if (this._switchSorterInner(name, enabled)) {
            this._reSort();
         }
      },

      /**
       * Возвращает признак, включен ли способ сортировки
       * @param {String} name Название способа сортировки
       * @return {Boolean|undefined}
       */
      isSorterEnabled: function (name) {
         for (var i = 0; i < this._sorters.length; i++) {
            if (this._sorters[i].name === name) {
               return this._sorters[i].enabled;
            }
         }
      },

      /**
       * Возвращает Название свойства элемента коллекции, содержащего его уникальный идентификатор.
       * @return {String}
       */
      getIdProperty: function () {
         return this._$idProperty;
      },

      /**
       * Возвращает признак обеспечивания уникальности элементов
       * @return {Boolean}
       */
      isUnique: function () {
         return this._$unique;
      },

      /**
       * Возвращает признак обеспечивания уникальности элементов
       * @param {String} unique Обеспечивать уникальность элементов
       */
      setUnique: function (unique) {
         if (this._$unique === unique) {
            return;
         }
         this._$unique = unique;

         var session = this._startUpdateSession();
         this._reSort();
         this._finishUpdateSession(session);
      },

      /**
       * Уведомляет подписчиков об изменении элемента коллекции
       * @param {WS.Data/Display/CollectionItem} item Элемент проекции
       * @param {Object} [properties] Изменившиеся свойства
       */
      notifyItemChange: function (item, properties) {
         var isFiltered = this._isFiltered(),
            isGrouped = this._isGrouped();
         
         if (isFiltered || isGrouped) {
            var session;
            session = this._startUpdateSession();
            if (isGrouped) {
               this._reGroup();
               this._reSort();
            }
            if (isFiltered) {
               this._reFilter();
            }
            this._removeEmptyGroups();
            this._finishUpdateSession(session);
         }

         if (!this.isEventRaising()) {
            return;
         }

         var index = this.getIndex(item),
            items = [item];
         items.properties = properties;

         this._notifyBeforeCollectionChange();
         this._notifyCollectionChange(
            IBindCollectionDisplay.ACTION_CHANGE,
            items,
            index,
            items,
            index
         );
         this._notifyAfterCollectionChange();
      },

      //endregion Changing

      //endregion Public methods

      //region WS.Data/Collection/EventRaisingMixin

      _analizeUpdateSession: function (session) {
         if (session) {
            this._notifyBeforeCollectionChange();
         }

         EventRaisingMixin._analizeUpdateSession.call(this, session);

         if (session) {
            this._notifyAfterCollectionChange();
         }
      },

      _notifyCollectionChange: function (action, newItems, newItemsIndex, oldItems, oldItemsIndex, session) {
         if (!this._isNeedNotifyCollectionChange()) {
            return;
         }
         if (
            !session ||
            action === IBindCollectionDisplay.ACTION_RESET ||
            !this._isGrouped()
         ) {
            this._notifyLater(
               'onCollectionChange',
               action,
               newItems,
               newItemsIndex,
               oldItems,
               oldItemsIndex
            );
            return;
         }

         //Split by groups and notify
         var notify = function(start, finish) {
               if (start < finish) {
                  this._notifyLater(
                     'onCollectionChange',
                     action,
                     newItems.slice(start, finish),
                     newItems.length ? newItemsIndex + start : 0,
                     oldItems.slice(start, finish),
                     oldItems.length ? oldItemsIndex + start : 0
                  );
               }
            }.bind(this),
            isRemove = action === IBindCollectionDisplay.ACTION_REMOVE,
            max = isRemove ? oldItems.length : newItems.length,
            notifyIndex = 0,
            item,
            i;

         for (i = 0; i < max; i++) {
            item = isRemove ? oldItems[i] : newItems[i];
            if (item instanceof GroupItem) {
               notify(notifyIndex, i);
               notifyIndex = i;
            }
            if (i === max - 1) {
               notify(notifyIndex, i + 1);
            }
         }
      },

      //endregion WS.Data/Collection/EventRaisingMixin

      //region Multiselectable

      /**
       * Возвращает массив выбранных элементов
       * @remark Метод возвращает выбранные элементы не зависимо от фильтра проекции.
       * @return {Array.<WS.Data/Display/CollectionItem>}
       */
      getSelectedItems: function() {
         var items = this._getItems(),
            result = [];
         for (var i = items.length - 1; i >= 0; i--) {
            if (items[i].isSelected()) {
               result.push(items[i]);
            }
         }
         return result;
      },

      /**
       * Устанавливает признак, что элемент выбран, переданным элементам.
       * @remark Метод зависит от фильтра проекции.
       * @param {Array} items Массив элементов коллекции
       * @param {Boolean} selected Элемент выбран.
       * @example
       * <pre>
       *     var list = new List({...}),
       *        display = new CollectionDisplay({
       *           collection: list
       *        });
       *    display.setSelectedItems([list.at(0), list.at(1)], true) //установит признак двум элементам;
       * </pre>
       */
      setSelectedItems: function(items, selected) {
         var sourceItems = [];
         for (var i = 0, count = items.length; i < count; i++) {
            sourceItems.push(
               this.getItemBySourceItem(items[i])
            );
         }
         this._setSelectedItems(sourceItems, selected);
      },

      /**
       * Устанавливает признак, что элемент выбран, всем элементам проекции.
       * @remark Метод устанавливает  признак всем элементам не зависимо от фильтра проекции.
       * @param {Boolean} selected Элемент выбран.
       * @return {Array}
       */
      setSelectedItemsAll: function(selected) {
         this._setSelectedItems(this._getItems(), selected);
      },

      /**
       * Инвертирует признак, что элемент выбран, у всех элементов проекции
       * @remark Метод инвертирует выделение у всех элементов не зависимо от фильтра проекции.
       */
      invertSelectedItemsAll: function() {
         var items = this._getItems(),
            i;
         for (i = items.length - 1; i >= 0; i--) {
            items[i].setSelected(!items[i].isSelected(), true);
         }
         this._notifyBeforeCollectionChange();
         this._notifyCollectionChange(
            IBindCollectionDisplay.ACTION_RESET,
            items,
            0,
            items,
            0
         );
         this._notifyAfterCollectionChange();
      },

      /**
       * Устанавливает признак, переданным, элементам проекции.
       * @param {Array} selecItems массив элементов проекции
       * @param {Boolean} selected Элемент выбран.
       */
      _setSelectedItems: function(selecItems, selected) {
         var items = [];
         selected = !!selected;
         for (var i = selecItems.length-1; i >= 0; i--){
            if (selecItems[i].isSelected() !== selected) {
               selecItems[i].setSelected(selected, true);
               items.push(selecItems[i]);
            }
         }
         if (items.length > 0) {
            var index = this.getIndex(items[0]);
            this._notifyBeforeCollectionChange();
            this._notifyCollectionChange(
               IBindCollectionDisplay.ACTION_REPLACE,
               items,
               index,
               items,
               index
            );
            this._notifyAfterCollectionChange();
         }
      },

      //endregion Multiselectable

      //region Protected methods

      //region Access

      /**
       * Добавляет свойство в importantItemProperties, если его еще там нет
       * @param {String} name Название свойства
       * @protected
       */
      _setImportantProperty: function(name) {
         var index = this._$importantItemProperties.indexOf(name);
         if (index === -1) {
            this._$importantItemProperties.push(name);
         }
      },

      /**
       * Удаляет свойство из importantItemProperties, если оно там есть
       * @param {String} name Название свойства
       * @protected
       */
      _unsetImportantProperty: function(name) {
         var index = this._$importantItemProperties.indexOf(name);
         if (index !== -1) {
            this._$importantItemProperties.splice(index, 1);
         }
      },

      /**
       * Настраивает контекст обработчиков
       * @protected
       */
      _bindHandlers: function() {
         this._onSourceCollectionChange = onSourceCollectionChange.bind(this);
         this._onSourceCollectionItemChange = onSourceCollectionItemChange.bind(this);
         this._onSourceCollectionEventRaisingChange = onSourceCollectionEventRaisingChange.bind(this);
      },

      _unbindHandlers: function() {
         this._onSourceCollectionChange = null;
         this._onSourceCollectionItemChange = null;
         this._onSourceCollectionEventRaisingChange = null;
      },

      //endregion Access

      //region Navigation

      /**
       * Возвращает элементы проекции
       * @return {Array.<WS.Data/Display/CollectionItem>}
       * @protected
       */
      _getItems: function () {
         return this._getItemsStrategy().getItems();
      },

      /**
       * Возвращает cтратегию получения элементов проекции
       * @return {WS.Data/Display/ItemsStrategy/Abstract}
       * @protected
       */
      _getItemsStrategy: function () {
         if (this._itemsStrategy) {
            return this._itemsStrategy;
         }

         this._itemsStrategy = this._createItemsStrategy();
         this._wrapItemsStrategy();

         return this._itemsStrategy;
      },

      /**
       * Создает cтратегию получения элементов проекции
       * @return {WS.Data/Display/ItemsStrategy/Abstract}
       * @protected
       */
      _createItemsStrategy: function () {
         return new DirectItemsStrategy(this, {
            itemModule: this._itemModule,
            localize: this._localize
         });
      },

      /**
       * Создает декораторы cтратегии получения элементов проекции
       * @protected
       */
      _wrapItemsStrategy: function() {
         if (this._itemsStrategy) {
            if (!this._itemsGroupStrategy) {
               this._itemsGroupStrategy = new GroupItemsStrategy(this._itemsStrategy);
            }
            if (!(this._itemsStrategy instanceof GroupItemsStrategy)) {
               this._itemsStrategy = this._itemsGroupStrategy;
            }
         }
      },

      /**
       * Возвращает энумератор
       * @param {Boolean} unlink Отвязать от состояния проекции
       * @return {WS.Data/Display/CollectionEnumerator}
       * @protected
       */
      _getEnumerator: function (unlink) {
         return this._buildEnumerator(
            unlink ? this._getItems().slice() : this._getItems.bind(this),
            unlink ? this._filterMap.slice() : this._filterMap,
            unlink ? this._sortMap.slice() : this._sortMap
         );
      },

      /**
       * Конструирует энумератор по входным данным
       * @param {Array.<WS.Data/Display/CollectionItem>} items Элементы проекции
       * @param {Array.<Boolean>} filterMap Фильтр: индекс в коллекции -> прошел фильтр
       * @param {Array.<WS.Data/Display/CollectionItem>} sortMap Сортировка: индекс в проекции -> индекс в коллекции
       * @return {WS.Data/Display/CollectionEnumerator}
       * @protected
       */
      _buildEnumerator: function (items, filterMap, sortMap) {
         return new CollectionEnumerator({
            items: items,
            filterMap: filterMap,
            sortMap: sortMap
         });
      },

      /**
       * Возвращает служебный энумератор для организации курсора
       * @return {WS.Data/Display/CollectionEnumerator}
       * @protected
       */
      _getCursorEnumerator: function () {
         return this._cursorEnumerator || (this._cursorEnumerator = this._getEnumerator());
      },

      /**
       * Возвращает служебный энумератор для для поиска по свойствам и поиска следующего или предыдущего элемента относительно заданного
       * @return {WS.Data/Display/CollectionEnumerator}
       * @protected
       */
      _getUtilityEnumerator: function() {
         return this._utilityEnumerator || (this._utilityEnumerator = this._getEnumerator());
      },

      /**
       * Возвращает соседний элемент проекции
       * @param {WS.Data/Collection/IEnumerator} enumerator Энумератор элементов
       * @param {WS.Data/Display/CollectionItem} item Элемент проекции относительно которого искать
       * @param {Boolean} isNext Следующий или предыдущий элемент
       * @param {Boolean} [skipGroups=false] Пропускать группы
       * @return {WS.Data/Display/CollectionItem}
       * @protected
       */
      _getNearbyItem: function (enumerator, item, isNext, skipGroups) {
         var method = isNext ? 'moveNext' : 'movePrevious',
            nearbyItem;

         enumerator.setCurrent(item);
         while (enumerator[method]()) {
            nearbyItem = enumerator.getCurrent();
            if (skipGroups && nearbyItem instanceof GroupItem) {
               nearbyItem = undefined;
               continue;
            }
            break;
         }

         return nearbyItem;
      },

      /**
       * Возвращает индекс элемента проекции по индексу в коллекции
       * @param {Number} index Индекс в коллекции
       * @return {Number}
       * @protected
       */
      _getItemIndex: function (index) {
         return this._getItemsStrategy().getItemIndex(index);
      },

      /**
       * Возвращает индекс в коллекци по индексу в проекции
       * @param {Number} index Индекс в проекции
       * @return {Number}
       * @protected
       */
      _getSourceIndex: function (index) {
         return this._getItemsStrategy().getCollectionIndex(index);
      },

      //endregion Navigation

      //region Calculation

      /**
       * Перерасчитывает все данные заново
       * @param {Boolean} [reset=false] Сбросить все созданные элементы
       * @protected
       */
      _reBuild: function (reset) {
         this._groupMap.length = 0;
         this._filterMap.length = 0;
         this._reIndex();

         var itemsStrategy = this._getItemsStrategy(),
            count,
            index;

         if (reset) {
            itemsStrategy.reset();
         }

         count = itemsStrategy.getCount();
         for (index = 0; index < count; index++) {
            this._filterMap.push(true);
         }

         if (this._isGrouped()) {
            this._reGroup();
         }
         this._reSort();
         if (this._isFiltered()) {
            this._reFilter();
         }
         this._removeEmptyGroups();
      },

      /**
       * Производит фильтрацию и сортировку и анализ изменений для набора элементов проекции
       * @param {Number} [start=0] Начальный индекс
       * @param {Number} [count] Кол-во элементов (по умолчанию - все)
       * @protected
       */
      _reAnalize: function (start, count) {
         start = start || 0;
         count = count || this._getItemsStrategy().getCount() - start;

         var session = this._startUpdateSession();
         if (this._isGrouped()) {
            start = this._reGroup(start, count);
         }
         this._reSort();
         if (this._isFilteredByIndex()) {
            this._reFilter();
         } else {
            this._reFilter(start, count);
         }
         this._removeEmptyGroups();
         this._finishUpdateSession(session);
      },

      /**
       * Вызывает переиндексацию
       * @protected
       */
      _reIndex: function () {
         this._getCursorEnumerator().reIndex();
         this._getUtilityEnumerator().reIndex();
      },

      //endregion Calculation

      //region Changing

      /**
       * Производит фильтрацию для набора элементов проекции
       * @param {Number} [start=0] Начальный индекс
       * @param {Number} [count] Кол-во элементов (по умолчанию - все)
       * @protected
       */
      _reFilter: function (start, count) {
         start = start || 0;
         count = count || this._getItemsStrategy().getCount() - start;

         var isMatch = function(item, index, position, hasMembers) {
               var result = true,
                  filter,
                  filterIndex;
               for (filterIndex = 0; filterIndex < filtersLength; filterIndex++) {
                  filter = filters[filterIndex];
                  result = filter(
                     item.getContents(),
                     index,
                     item,
                     position,
                     hasMembers
                  );
                  if (!result) {
                     break;
                  }
               }
               return result;
            },
            applyMatch = function(match, index) {
               var oldMatch = filterMap[index];
               if (match === oldMatch) {
                  return false;
               }
               if (match) {
                  filterMap[index] = match;
                  return true;
               } else if (oldMatch !== undefined) {
                  filterMap[index] = match;
                  return true;
               }
               return false;
            },
            filters = this._$filter,
            filtersLength = filters.length,
            items = this._getItems(),
            sortMap = this._sortMap,
            sortMapLength = sortMap.length,
            filterMap = this._filterMap,
            processedIndices = new Set(),
            finish = start + count,
            changed = false,
            item,
            position,
            index,
            prevGroup,
            prevGroupIndex = -1,
            prevGroupPosition = -1,
            prevGroupHasMembers = false,
            match;

         //Lookup every item in _sortMap order
         for (position = 0; position < sortMapLength; position++) {
            index = sortMap[position];
            //Check item index in range
            if (index === undefined || index < start || index >= finish) {
               continue;
            }

            processedIndices.add(index);
            item = items[index];
            match = true;
            if (item instanceof GroupItem) {
               //A new group begin, check match for previous
               if (prevGroup) {
                  match = isMatch(prevGroup, prevGroupIndex, prevGroupPosition, prevGroupHasMembers);
                  changed = applyMatch(match, prevGroupIndex) || changed;
               }
               //Remember current group as previous
               prevGroup = item;
               prevGroupIndex = index;
               prevGroupPosition = position;
               prevGroupHasMembers = false;
            } else {
               //Check item match
               match = isMatch(item, index, position);
               changed = applyMatch(match, index) || changed;
               if (match) {
                  prevGroupHasMembers = true;
               }
            }
         }

         for (index = start; index < finish; index++) {
            if (!processedIndices.has(index)) {
               filterMap[index] = undefined;
            }
         }

         //Check last group match
         if (prevGroup) {
            match = isMatch(prevGroup, prevGroupIndex, prevGroupPosition, prevGroupHasMembers);
            changed = applyMatch(match, prevGroupIndex) || changed;
         }

         if (changed) {
            this._reIndex();
         }
      },

      /**
       * Производит сортировку элементов
       * @protected
       */
      _reSort: function () {
         this._sortMap.length = 0;
         Array.prototype.push.apply(this._sortMap, this._buildSortMap());

         this._reIndex();
      },

      /**
       * Инициализирует сортировщики
       * @protected
       */
      _initSorters: function () {
         this._sorters = [{
            name: 'natural',
            enabled: true,
            method: Collection.sorters.natural,
            options: function() {
               return {
                  idProperty: this._$idProperty,
                  unique: this._$unique
               };
            }
         }, {
            name: 'user',
            enabled: false,
            method: Collection.sorters.user,
            options: function() {
               return {
                  userSort: this._$sort
               };
            }
         }, {
            name: 'group',
            enabled: false,
            method: Collection.sorters.group,
            options: function() {
               return {
                  groupMap: this._groupMap
               };
            }
         }];

         this._getItemsStrategy().addSorters(this._sorters);
      },

      /**
       * Переключает сортировики, зависящие от наличия пользовательской сортировки
       * @return {Boolean} Сортировки были переключены
       * @protected
       */
      _switchSortersByUserSort: function () {
         var isUserSorted = this._isUserSorted();
         return this._switchSorterInner('user', isUserSorted);
      },

      /**
       * Включает-выключает способ сортировки
       * @param {String} name Название способа сортировки
       * @param {Boolean} enabled Включен или выключен
       * @return {Boolean} Способ сортировки был переключен
       * @protected
       */
      _switchSorterInner: function (name, enabled) {
         var applyImportantProps = function(enabled) {
               var item,
                  properties,
                  i,
                  j;

               for (i = 0; i < this._$sort.length; i++) {
                  item = this._$sort[i];
                  if (ComputeFunctor.isFunctor(item)) {
                     properties = item.properties;
                     for (j = 0; j < properties.length; j++) {
                        if (enabled) {
                           this._setImportantProperty(properties[j]);
                        } else {
                           this._unsetImportantProperty(properties[j]);
                        }
                     }
                  }
               }

            }.bind(this),
            hasAffect = false,
            sorter,
            i;

         for (i = 0; i < this._sorters.length; i++) {
            sorter = this._sorters[i];
            if (sorter.name === name && sorter.enabled !== enabled) {
               sorter.enabled = enabled;
               hasAffect = true;
               if (name === 'user') {
                  applyImportantProps(enabled);
               }
            }
         }

         return hasAffect;
      },

      /**
       * Производит построение sortMap
       * @return {Array.<Number>}
       * @protected
       */
      _buildSortMap: function () {
         var i,
            sorter,
            items = this._getItems(),
            result = [];

         for (i = 0; i < this._sorters.length; i++) {
            sorter = this._sorters[i];
            if (sorter.enabled) {
               result = sorter.method.call(
                  null,
                  items,
                  result,
                  (typeof sorter.options === 'function' ? sorter.options.call(this) : sorter.options) || {}
               );
            }
         }

         return result;
      },

      /**
       * Производит группировку для набора элементов проекции
       * @param {Number} [start=0] Начальный индекс
       * @param {Number} [count] Кол-во элементов (по умолчанию - все)
       * @return {Number} Новый начальный индекс (если перед start были вставлены новые группы)
       * @protected
       */
      _reGroup: function (start, count) {
         start = start || 0;
         count = count || this._getItemsStrategy().getCount() - start;

         var shiftUp = function(fromIndex) {
               return function(index) {
                  return index >= fromIndex ? 1 + index : index;
               };
            },
            shiftOffset = 0,
            strategy = this._itemsGroupStrategy,
            items = this._getItems().slice(),
            finish = start + count,
            filterMap = this._filterMap,
            sortMap = this._sortMap,
            groupMap = this._groupMap,
            groupMethod = this._$group,
            group,
            groups,//Group index -> group ID
            groupId,
            groupIndex,
            item,
            index,
            prevIndex,
            hasChanges;

         groups = strategy.groups.map(function(item) {
            return item.getContents();
         });
         prevIndex = groupMap.slice(0, start).reduce(function(prev, curr, index) {
            return items[index] instanceof GroupItem ? prev : Math.max(prev, curr);
         }, -1);

         //Check for the new groups
         for (index = start; index < finish; index++) {
            item = items[index];
            if (item instanceof GroupItem) {
               groupId = item.getContents();
               groupIndex = groups.indexOf(groupId);
               groupMap[index] = groupIndex;
            } else {
               groupId = groupMethod ? groupMethod(item.getContents(), index, item) : undefined;
               groupIndex = groups.indexOf(groupId);
               if (groupIndex === -1) {
                  //Add a new group
                  groupIndex = 1 + prevIndex;

                  group = new GroupItem({
                     owner: this,
                     contents: groupId
                  });
                  strategy.addGroup(group, groupIndex);

                  //Increase group index for groups after groupIndex
                  groupMap = groupMap.map(shiftUp(groupIndex));

                  items.splice(groupIndex, 0, group);
                  groups.splice(groupIndex, 0, groupId);
                  filterMap.splice(groupIndex, 0, true);
                  sortMap.splice(groupIndex, 0, groupIndex);
                  groupMap.splice(groupIndex, 0, groupIndex);

                  if (index >= groupIndex) {
                     index++;
                     shiftOffset++;
                  }
                  finish++;
               }
               if (groupMap[index] !== groupIndex) {
                  hasChanges = true;
               }
               groupMap[index] = groupIndex;
            }
            prevIndex = Math.max(prevIndex, groupIndex);
         }

         this._groupMap = groupMap;

         if (hasChanges) {
            this._reIndex();
         }

         return start + shiftOffset;
      },

      /**
       * Удаляет пустые группы путем исключения их из _sortMap
       * @protected
       */
      _removeEmptyGroups: function() {
         //Presence of empty groups is ruled by filter
         if (this._isFiltered()) {
            return;
         }

         var items = this._getItems().slice(),
            filterMap = this._filterMap,
            sortMap = this._sortMap,
            groupMethod = this._$group,
            toRemove = [],
            item,
            current,
            currentIndex,
            hasItems;

         currentIndex = -1;
         hasItems = false;
         if (groupMethod) {
            //Lookup for every group in _sortMap order
            sortMap.forEach(function(index, position) {
               item = items[index];
               if (item instanceof GroupItem) {
                  //Has previous group
                  if (currentIndex > -1) {
                     //Remove previous group that doesn't have any members
                     if (!hasItems) {
                        toRemove.unshift(currentIndex);
                     }
                     //Remove previous group if it repeat
                     if (current.getContents() === item.getContents()) {
                        toRemove.unshift(position);
                     }
                  }
                  current = item;
                  currentIndex = position;
                  hasItems = false;
               } else if (filterMap[index]) {
                  //Previous group has some members
                  hasItems = true;
               }
            });

            //Remove last group that doesn't have any members
            if (currentIndex > -1 && !hasItems) {
               toRemove.unshift(currentIndex);
            }
         } else {
            //Remove every group if grouping no longer used
            sortMap.forEach(function(index, position) {
               if (item instanceof GroupItem) {
                  toRemove.unshift(position);
               }
            });
         }

         if (toRemove.length) {
            toRemove.forEach(function(position) {
               sortMap.splice(position, 1);
            });
            this._reIndex();
         }
      },

      /**
       * Удаляет все данные о группах
       * @protected
       */
      _resetGroups: function() {
         this._groupMap.length = 0;
         var strategy = this._itemsGroupStrategy;
         strategy.groups.slice().forEach(function(group) {
            strategy.removeGroup(group);
         });
      },

      /**
       * Проверяет, что используется фильтрация
       * @return {Boolean}
       * @protected
       */
      _isFiltered: function () {
         return this._$filter.length > 0;
      },

      /**
       * Проверяет, что хотя бы в один из методов фильтрации использует аргумент index
       * @return {Boolean}
       * @protected
       */
      _isFilteredByIndex: function () {
         return this._$filter.some(function(filter) {
            return this._isFilterUseIndex(filter);
         }.bind(this));
      },

      /**
       * Проверяет, что метод фильтрации использует аргумент index
       * @param {Function} filter
       * @return {Boolean}
       * @protected
       */
      _isFilterUseIndex: function (filter) {
         return filter.length > 3;
      },

      /**
       * Проверяет, что используется пользовательская сортировка
       * @return {Boolean}
       * @protected
       */
      _isUserSorted: function () {
         return this._$sort.length > 0;
      },

      /**
       * Проверяет, что используется группировка
       * @return {Boolean}
       * @protected
       */
      _isGrouped: function () {
         return !!this._$group;
      },

      /**
       * Возвращает идентификатор группы по ее индексу
       * @param {Number} groupIndex Индекс группы
       * @return {String}
       * @protected
       */
      _getGroupIdByIndex: function (groupIndex) {
         var group = this._itemsGroupStrategy.groups[groupIndex];
         return group && group.getContents();
      },

      /**
       * Возвращает индекс группы элемента
       * @param {Number} index Индекс элемента в проекции
       * @return {Number}
       * @protected
       */
      _getGroupIndexByIndex: function (index) {
         var collectionIndex = this._getUtilityEnumerator().getSourceByInternal(index),
            groupIndex = this._groupMap[collectionIndex];

         return groupIndex === undefined ? -1 : groupIndex;
      },

      /**
       * Дробавляет набор элементов коллекции в проекцию
       * @param {Number} start Начальный индекс (в коллекци)
       * @param {Array} items Элементы
       * @protected
       */
      _addItems: function (start, items) {
         var isFiltered = this._isFiltered(),
            splice = Array.prototype.splice,
            strategy = this._getItemsStrategy(),
            filterMap = [],
            sortMap = [],
            groupMap = [];

         items.forEach(function(item, index) {
            filterMap.push(!isFiltered);
            sortMap.push(start + index);
            groupMap.push(undefined);
         });

         strategy.splice(start, 0, items);
         splice.apply(this._filterMap, [start, 0].concat(filterMap));
         splice.apply(this._sortMap, [start, 0].concat(sortMap));
         if (this._isGrouped()) {
            splice.apply(this._groupMap, [start, 0].concat(groupMap));
         }
      },

      /**
       * Удаляет набор элементов проекции
       * @param {Number} start Начальный индекс
       * @param {Number} [count] Кол-во элементов (по умолчанию - все)
       * @return {Array.<WS.Data/Display/CollectionItem>} Удаленные элементы
       * @protected
       */
      _removeItems: function (start, count) {
         var strategy = this._getItemsStrategy(),
            result;

         start = start || 0;
         count = count === undefined ? strategy.getCount() - start : count;

         result = strategy.splice(start, count);
         this._filterMap.splice(start, count);
         this._removeFromSortMap(start, count);
         if (this._isGrouped()) {
            this._groupMap.splice(start, count);
         }

         return result;
      },

      /**
       * Заменяет набор элементов проекции
       * @param {Number} start Начальный индекс
       * @param {Array} newItems Новые элементы
       * @return {Array.<WS.Data/Display/CollectionItem>} Замененные элементы
       * @protected
       */
      _replaceItems: function (start, newItems) {
         var strategy = this._getItemsStrategy();
         return strategy.splice(start, newItems.length, newItems);
      },

      /**
       * Перемещает набор элементов проекции
       * @param {Number} newIndex Старый индекс
       * @param {Number} oldIndex Новый индекс
       * @param {Array} items Перемещаемые элементы
       * @protected
       */
      _moveItems: function (newIndex, oldIndex, items) {
         var length = items.length,
            strategy = this._getItemsStrategy(),
            currentItems;

         currentItems = strategy.splice(oldIndex, length);
         strategy.splice(newIndex, 0, currentItems);
      },

      /**
       * Удаляет из индекса сортировки срез элементов
       * @param {Number} start Начальный индекс (в коллекци)
       * @param {Number} count Кол-во элементов
       * @return {Array.<Number>}
       * @protected
       */
      _removeFromSortMap: function (start, count) {
         start = start || 0;
         count = count || 0;
         var finish = start + count,
            index,
            sortIndex,
            toRemove = [],
            removed = {};

         //Collect indices to remove
         for (index = start; index < finish; index++) {
            sortIndex = this._sortMap.indexOf(index);
            if (sortIndex > -1) {
               toRemove.push(sortIndex);
               removed[sortIndex] = this._sortMap[sortIndex];
            }
         }

         //Remove collected indices from _sortMap
         toRemove.sort(function (a, b) {
            return a - b;
         });
         for (index = toRemove.length - 1; index >= 0; index--) {
            this._sortMap.splice(toRemove[index], 1);
         }

         //Shift _sortMap values by count from start index
         for (index = 0; index < this._sortMap.length; index++) {
            if (this._sortMap[index] >= start) {
               this._sortMap[index] -= count;
            }
         }

         this._reIndex();

         return removed;
      },

      /**
       * Возвращает набор контрольных свойств элемента проекции для анализа его состояния
       * @param {WS.Data/Display/CollectionItem} item Элемент проекции
       * @return {Object}
       * @protected
       */
      _getItemState: function(item) {
         return {
            item: item,
            selected: item.isSelected()
         };
      },

      /**
       * Возвращает состояния элементов
       * @param {Array.<WS.Data/Display/CollectionItem>} items Элементы проекции
       * @return {Array.<Object>}
       * @protected
       */
      _getItemsState: function(items) {
         return items.map(this._getItemState);
      },

      /**
       * Возвращает разницу между двумя состояниями элементов (результатами работы метода _getItemsState)
       * @param {Array.<Object>} before Состояния до изменений
       * @param {Array.<Object>} after Состояния после изменений
       * @return {Array.<Object>} Отличающиеся состояния
       * @protected
       */
      _getItemsDiff: function(before, after) {
         var itemThen;
         return after.filter(function(itemNow, index) {
            itemThen = before[index];
            return Object.keys(itemNow).some(function(prop) {
               return itemNow[prop] !== itemThen[prop];
            });
         })
         .map(function(element) {
            return element.item;
         });
      },

      /**
       * Генерирует события об изменении элементов проекции при изменении их состояния
       * @param {Object} session Сессия изменений
       * @param {Array.<WS.Data/Display/CollectionItem>} items Измененные элементы
       * @param {Array} state Состояние элементов до изменений
       * @param {Function} beforeCheck Функция обратного вызова перед проверкой изменений состояния
       * @protected
       */
      _checkItemsDiff: function (session, items, state, beforeCheck) {
         var diff = state ? this._getItemsDiff(
            state,
            this._getItemsState(items)
         ) : [];

         if (beforeCheck) {
            beforeCheck(diff, items);
         }

         //Notify changes by the diff
         if (diff.length) {
            this._notifyBeforeCollectionChange();
            EventRaisingMixin._extractPacksByList(this, diff, function(items, index) {
               this._notifyCollectionChange(
                  IBindCollectionDisplay.ACTION_CHANGE,
                  items,
                  index,
                  items,
                  index,
                  session
               );
            }.bind(this));
            this._notifyAfterCollectionChange();
         }
      },

      /**
       * Генерирует событие об изменении текущего элемента проекции коллекции
       * @param {WS.Data/Display/CollectionItem} newCurrent Новый текущий элемент
       * @param {WS.Data/Display/CollectionItem} oldCurrent Старый текущий элемент
       * @param {Number} newPosition Новая позиция
       * @param {Number} oldPosition Старая позиция
       * @protected
       */
      _notifyCurrentChange: function (newCurrent, oldCurrent, newPosition, oldPosition) {
         if (!this.isEventRaising()) {
            return;
         }

         this._removeFromQueue('onCurrentChange');
         this._notify(
            'onCurrentChange',
            newCurrent,
            oldCurrent,
            newPosition,
            oldPosition
         );
      },

      /**
       * Нотифицирует событие change для измененных элементов
       * @param {Array} changed Измененные элементы исходной коллекции.
       * @param {Number} index Индекс исходной коллекции, в котором находятся элементы.
       * @protected
       */
      _notifyCollectionItemsChange: function (changed, index, session) {
         var items = this._getItems(),
            last = index + changed.length,
            changedItems = [],
            i;

         //Extract display items contains changed
         for (i = index; i < last; i++) {
            changedItems.push(items[this._getItemIndex(i)]);
         }

         this._notifyBeforeCollectionChange();
         EventRaisingMixin._extractPacksByList(
            this,
            changedItems,
            function(pack, index) {
               this._notifyCollectionChange(
                  IBindCollectionDisplay.ACTION_CHANGE,
                  pack,
                  index,
                  pack,
                  index,
                  session
               );
            }.bind(this)
         );
         this._notifyAfterCollectionChange();
      },

      /**
       * Генерирует событие об изменении элемента проекции
       * @param {Core/EventObject} event Дескриптор события.
       * @param {*} item Измененный элемент коллекции.
       * @param {Number} index Индекс измененного элемента.
       * @param {Object} [properties] Изменившиеся свойства
       * @protected
       */
      _notifySourceCollectionItemChange: function (event, item, index, properties) {
         var internalItems = this._getItems(),
            internalIndex = this._getItemIndex(index),
            internalItem = internalItems[internalIndex],
            enumerator = this._getUtilityEnumerator(),
            indexBefore = enumerator.getInternalBySource(internalIndex),
            isEventRaising = this.isEventRaising(),
            session = this._startUpdateSession(),
            state;

         //Only changes of important properties can run analysis
         for (var key in properties) {
            if (properties.hasOwnProperty(key)) {
               if (this._$importantItemProperties.indexOf(key) > -1) {
                  if (isEventRaising) {
                     //Fix the state before analysis
                     state = this._getItemsState(internalItems);
                  }
                  this._reAnalize(
                     internalIndex,
                     1
                  );
                  break;
               }
            }
         }

         //Return here if events are disabled
         if (!isEventRaising) {
            return;
         }

         this._finishUpdateSession(session, false);

         //Check changes by state
         this._checkItemsDiff(session, internalItems, state, function(diff) {
            //Some hard logic related to the character of item change.
            var indexAfter = enumerator.getInternalBySource(internalIndex),
               isMoved = indexBefore !== indexAfter,
               internalItemIndex = diff.indexOf(internalItem);
            if (isMoved) {
               //Item change the position
               if (internalItemIndex > -1 && indexBefore > indexAfter) {
                  //Changed item is presented in the diff and moved up.
                  //It will be presented as a move event with that item in _finishUpdateSession.
                  //We should not notify about item change with the diff.
                  diff.splice(internalItemIndex, 1);
               } else if (internalItemIndex === -1 && indexBefore < indexAfter) {
                  //Changed item isn't presented in the diff and moved down.
                  //It won't be presented as a move event with that item in _finishUpdateSession (items after will move up).
                  //We should notify about item change with the diff.
                  diff.push(internalItem);
               }
            } else if (!isMoved && internalItemIndex === -1) {
               //Changed item don't change the position and not presented in the diff.
               //We should notify about item change with the diff.
               diff.push(internalItem);
            }
         });
      },

      /**
       * Генерирует событие о начале изменений коллекции
       * @protected
       */
      _notifyBeforeCollectionChange: function () {
         if (!this.isEventRaising()) {
            return;
         }
         this._notifyLater('onBeforeCollectionChange');
      },

      /**
       * Генерирует событие об окончании изменений коллекции
       * @protected
       */
      _notifyAfterCollectionChange: function () {
         if (!this.isEventRaising()) {
            return;
         }
         this._notify('onAfterCollectionChange');
      }

      //endregion Changing

      //endregion Protected methods

   });

   var MESSAGE_READ_ONLY = 'The Display is read only. You should modify the source collection instead.',

      /**
       * Преобразует проекцию в массив из ее элементов
       * @param {WS.Data/Display/Display} display Проекция.
       * @return {Array.<WS.Data/Display/CollectionItem>}
       */
      toArray = function (display) {
         var result = [];
         display.each(function(item) {
            result.push(item);
         });
         return result;
      },

      /**
       * Обрабатывает событие об изменении коллекции
       * @param {Core/EventObject} event Дескриптор события.
       * @param {String} action Действие, приведшее к изменению.
       * @param {Array.<*>} newItems Новые элементы коллекции.
       * @param {Number} newItemsIndex Индекс, в котором появились новые элементы.
       * @param {Array.<*>} oldItems Удаленные элементы коллекции.
       * @param {Number} oldItemsIndex Индекс, в котором удалены элементы.
       */
      onSourceCollectionChange = function (event, action, newItems, newItemsIndex, oldItems, oldItemsIndex) {
         var projectionNewItemsIndex = this._getItemIndex(newItemsIndex),
            projectionOldItemsIndex = this._getItemIndex(oldItemsIndex),
            session;

         switch (action) {
            case IBindCollectionDisplay.ACTION_RESET:
               var projectionOldItems = toArray(this),
                  projectionNewItems;
               this._reBuild(true);
               projectionNewItems = toArray(this);
               this._notifyBeforeCollectionChange();
               this._notifyCollectionChange(
                  action,
                  projectionNewItems,
                  0,
                  projectionOldItems,
                  0
               );
               this._notifyAfterCollectionChange();
               return;

            case IBindCollectionDisplay.ACTION_CHANGE:
               session = this._startUpdateSession();
               if (this._isGrouped()) {
                  //FIXME: newItems.length - FIXME[OrderMatch]
                  projectionNewItemsIndex = this._reGroup(projectionNewItemsIndex, newItems.length);
               }
               this._reSort();
               if (this._isFiltered()) {
                  //FIXME: newItems.length - FIXME[OrderMatch]
                  this._reFilter(projectionNewItemsIndex, newItems.length);
               }
               if (this._isGrouped()) {
                  this._removeEmptyGroups();
               }
               this._finishUpdateSession(session, false);
               this._notifyCollectionItemsChange(newItems, newItemsIndex, session);
               return;
         }

         session = this._startUpdateSession();

         switch (action) {
            case IBindCollectionDisplay.ACTION_ADD:
               this._addItems(projectionNewItemsIndex, newItems);
               if (this._isGrouped()) {
                  //FIXME: newItems.length - FIXME[OrderMatch]
                  projectionNewItemsIndex = this._reGroup(projectionNewItemsIndex, newItems.length);
               }
               this._reSort();
               if (this._isFiltered()) {
                  if (this._isFilteredByIndex()) {
                     this._reFilter();
                  } else {
                     //FIXME: newItems.length - FIXME[OrderMatch]
                     this._reFilter(projectionNewItemsIndex, newItems.length);
                  }
               }
               if (this._isGrouped()) {
                  this._removeEmptyGroups();
               }
               break;

            case IBindCollectionDisplay.ACTION_REMOVE:
               //FIXME: oldItems.length - FIXME[OrderMatch]
               this._removeItems(projectionOldItemsIndex, oldItems.length);
               this._reSort();
               if (this._isFiltered() && this._isFilteredByIndex()) {
                  this._reFilter();
               }
               if (this._isGrouped()) {
                  this._removeEmptyGroups();
               }
               break;

            case IBindCollectionDisplay.ACTION_REPLACE:
               //FIXME: newItems - FIXME[OrderMatch]
               this._replaceItems(projectionNewItemsIndex, newItems);
               if (this._isGrouped()) {
                  //FIXME: newItems.length - FIXME[OrderMatch]
                  projectionNewItemsIndex = this._reGroup(projectionNewItemsIndex, newItems.length);
               }
               this._reSort();
               if (this._isFiltered()) {
                  //FIXME: newItems.length - FIXME[OrderMatch]
                  this._reFilter(projectionNewItemsIndex, newItems.length);
               }
               if (this._isGrouped()) {
                  this._removeEmptyGroups();
               }
               break;

            case IBindCollectionDisplay.ACTION_MOVE:
               //FIXME: newItems - FIXME[OrderMatch]
               this._moveItems(projectionNewItemsIndex, projectionOldItemsIndex, newItems);
               this._reSort();
               if (this._isFiltered()) {
                  this._reFilter();
               }
               if (this._isGrouped()) {
                  this._removeEmptyGroups();
               }
               break;
         }

         this._finishUpdateSession(session);
      },

      /**
       * Обрабатывает событие об изменении элемента коллекции
       * @param {Core/EventObject} event Дескриптор события.
       * @param {*} item Измененный элемент коллекции.
       * @param {Number} index Индекс измененного элемента.
       * @param {Object} [properties] Изменившиеся свойства
       */
      onSourceCollectionItemChange = function (event, item, index, properties) {
         if (!this.isEventRaising()) {
            return;
         }

         if (this._sourceCollectionSynchronized) {
            this._notifySourceCollectionItemChange(event, item, index, properties);
         } else {
            this._sourceCollectionDelayedCallbacks = this._sourceCollectionDelayedCallbacks || [];
            this._sourceCollectionDelayedCallbacks.push([this._notifySourceCollectionItemChange, arguments]);
         }
      },

      /**
       * Обрабатывает событие об изменении элемента коллекции
       * @param {Core/EventObject} event Дескриптор события.
       * @param {Boolean} enabled Включена или выключена генерация событий
       * @param {Boolean} analyze Включен или выключен анализ изменений
       */
      onSourceCollectionEventRaisingChange = function (event, enabled, analyze) {
         //Если без выключили без анализа изменений, то при следующем включении генерации надо актуализировать состояние
         if (!analyze && enabled) {
            this._reBuild(true);
         }

         this._sourceCollectionSynchronized = enabled;

         //Call delayed handlers if get back to synchronize
         var callbacks = this._sourceCollectionDelayedCallbacks,
            callback;
         if (this._sourceCollectionSynchronized && callbacks) {
            while (callbacks.length > 0) {
               callback = callbacks[0];
               callback[0].apply(this, callback[1]);
               callbacks.shift();
            }
         }
      };

   /**
    * @member {Object.<Function>} Алгоритмы сортировки элементов проекции.
    * @static
    */
   Collection.sorters = {
      /**
       * Создает индекс сортировки в том же порядке, что и коллекция
       * @param {Array.<WS.Data/Display/CollectionItem>} items Элементы проекции.
       * @param {Array.<Number>} current Текущий индекс сортировки
       * @param {Object} options Опции
       * @return {Array.<Number>}
       */
      natural: function (items, current, options) {
         var unique = options.unique,
            idProperty = options.idProperty,
            result = items.map(function(item, index) {
               return index;
            });

         if (unique && idProperty) {
            var processed = new Set(),
               item,
               itemId;

            result = result.filter(function(index) {
               item = items[index];
               if (item instanceof GroupItem) {
                  return true;
               }

               itemId = Utils.getItemPropertyValue(
                  item.getContents(),
                  idProperty
               );

               if (processed.has(itemId)) {
                  return false;
               }

               processed.add(itemId);
               return true;
            });
         }

         return result;
      },

      /**
       * Создает индекс сортировки в порядке, определенном набором пользовательских обработчиков
       * @param {Array.<WS.Data/Display/CollectionItem>} items Элементы проекции.
       * @param {Array.<Number>} current Текущий индекс сортировки
       * @param {Object} options Пользовательские обработчики для Array.prototype.sort
       */
      user: function(items, current, options) {
         var map = [],
            sorted = [],
            handlers = options.userSort,
            index,
            item,
            i,
            count;

         //Make utilitary array
         for (i = 0, count = current.length; i < count; i++) {
            index = current[i];
            item = items[index];
            if (item instanceof GroupItem) {
               map.push(index);
            } else {
               sorted.push({
                  item: item,
                  collectionItem: item.getContents(),
                  index: index,
                  collectionIndex: index
               });
            }
         }

         //Sort utilitary array
         for (i = handlers.length - 1; i >= 0; i--) {
            sorted.sort(handlers[i]);
         }

         //Create map from utilitary array
         for (index = 0, count = sorted.length; index < count; index++) {
            map.push(sorted[index].collectionIndex);
         }

         return map;
      },

      /**
       * Создает индекс сортировки в порядке группировки
       * @param {Array.<WS.Data/Display/CollectionItem>} items Элементы проекции.
       * @param {Array.<Number>} current Текущий индекс сортировки
       * @param {Object} options Опции
       * @return {Array.<Number>}
       */
      group: function (items, current, options) {
         var groupMap = options.groupMap,
            groupOrder = [],
            groupPosition = 0,
            groupsCount = 0,
            item,
            index,
            position,
            groupId;

         //Fill groupOrder by first member
         for (position = 0; position < current.length; position++) {
            index = current[position];
            item = items[index];
            groupId = groupMap[index];
            if (item instanceof GroupItem) {
               groupsCount++;
               continue;
            }
            if (groupOrder[groupId] === undefined) {
               groupOrder[groupId] = groupPosition++;
            }
         }

         //Fill empty array indices in groupOrder as last group
         //FIXME: there is the reason to move empty groups to the end, it can be an unexpected behavior
         for (groupId = 0; groupId < groupsCount; groupId++) {
            if (groupOrder[groupId] === undefined) {
               groupOrder[groupId] = groupPosition++;
            }
         }

         //Sort by group membership use groupOrder
         return current.map(function(index, positon) {
            return [index, positon];
         }).sort(function (a, b) {
            return groupOrder[groupMap[a[0]]] - groupOrder[groupMap[b[0]]] || a[1] - b[1];
         }).map(function(item) {
            return item[0];
         });
      }
   };

   //Deprecated methods
   Collection.prototype.getByHash = Collection.prototype.getByInstanceId;
   Collection.prototype.getIndexByHash = Collection.prototype.getIndexByInstanceId;

   Di.register('display.collection', Collection);

   return Collection;
});
