/* global define */
define('WS.Data/Display/ItemsStrategy/AdjacencyList', [
   'WS.Data/Display/ItemsStrategy/Direct',
   'WS.Data/Display/GroupItem',
   'WS.Data/Shim/Set',
   'WS.Data/Shim/Map',
   'WS.Data/Utils',
   'WS.Data/Di'
], function (
   Direct,
   GroupItem,
   Set,
   Map,
   Utils,
   Di
) {
   'use strict';

   /**
    * Стратегия получения элементов проекции по списку смежных вершин
    * @class WS.Data/Display/ItemsStrategy/AdjacencyList
    * @extends WS.Data/Display/ItemsStrategy/Direct
    * @author Мальцев Алексей
    */

   var AdjacencyList = Direct.extend(/** @lends WS.Data/Display/ItemsStrategy/AdjacencyList.prototype */{
      '[WS.Data/Display/ItemsStrategy/AdjacencyList]': true,

      /**
       * @typedef {Object} Options
       * @property {String|Function} itemModule Алиас зависимости или конструктора элементов проекции
       * @property {String} idProperty Имя свойства, хранящего первичный ключ
       * @property {String} parentProperty Имя свойства, хранящего первичный ключ родителя
       * @property {String} nodeProperty Имя свойства, хранящего признак "узел/лист"
       * @property {String} loadedProperty Имя свойства, хранящего признак "загружен"
       * @property {WS.Data/Display/TreeItem} root Корень
       */

      constructor: function AdjacencyList(display, options) {
         options = options || {};

         AdjacencyList.superclass.constructor.call(this, display, options);

         if (!options.idProperty) {
            Utils.logger.info(this._moduleName +'::constructor(): option "idProperty" is not defined - only root elements will be fetched');
         }
      },

      //region Public methods

      addSorters: function (sorters) {
         AdjacencyList.superclass.addSorters.call(this, sorters);

         sorters.push({
            name: 'tree',
            enabled: true,
            method: AdjacencyList.sortItems,
            options: this._options
         });
      },

      convertToItem: function (source, options) {
         var loadedProperty = this._options.loadedProperty,
            invertLoaded = false,
            loaded;
         if (typeof loadedProperty === 'string' && loadedProperty[0] === '!') {
            loadedProperty = loadedProperty.substr(1);
            invertLoaded = true;
         }
         loaded = Utils.getItemPropertyValue(source, loadedProperty);

         options = options || {};
         options.contents = source;
         options.owner = this._display;
         options.loaded = invertLoaded ? !loaded : loaded;
         if (!('node' in options)) {
            options.node = Utils.getItemPropertyValue(source, this._options.nodeProperty);
         }

         return Di.resolve(this.getItemModule(), options);
      }

      //endregion Public methods

      //region Protected methods

      //endregion Protected methods
   });

   /**
    * Создает индекс сортировки в порядке списка смежных вершин - от корневой вершины вглубь до конечных вершин
    * @param {Array.<WS.Data/Display/CollectionItem>} items Элементы проекции.
    * @param {Array.<Number>} current Текущий индекс сортировки
    * @param {Object} options Опции
    * @return {Array.<Number>}
    * @static
    */
   AdjacencyList.sortItems = function (items, current, options) {
      var push = Array.prototype.push,
         root = typeof options.root === 'function' ? options.root() : options.root,
         idProperty = options.idProperty,
         parentProperty = options.parentProperty,
         treeIndex = new Map(),//Map<*, Array<Number>>: parentId => [childIndex, childIndex, ...]
         treeGroups = new Map(),//Map<Number, Number>: itemIndex => groupIndex
         parentsProcessing = new Set(),//Set<*>
         lastGroupId,
         result,
         /**
          * Build treeIndex and treeGroups
          */
         buildIndex = function() {
            var index,
               count,
               item,
               itemIndex,
               group = -1,
               children,
               parentId;

            for (index = 0, count = current.length; index < count; index++) {
               itemIndex = current[index];
               item = items[itemIndex];

               //Remember current group
               if (item instanceof GroupItem) {
                  group = itemIndex;
                  continue;
               }

               //Save relation with first group member
               if (group > -1) {
                  treeGroups.set(itemIndex, group);
                  group = -1;
               }

               //Don't add root to the index. Otherwise it can cause a recursion.
               if (item === root) {
                  continue;
               }

               //TODO: work with parentId === Object|Array
               parentId = Utils.getItemPropertyValue(
                  item.getContents(),
                  parentProperty
               );
               if (typeof parentId === 'number') {
                  parentId = String(parentId);
               }

               if (treeIndex.has(parentId)) {
                  children = treeIndex.get(parentId);
               } else {
                  children = [];
               }
               children.push(itemIndex);
               treeIndex.set(parentId, children);
            }
         },
         /**
          * Lookup tree from given parent
          */
         getChildren = function(parent) {
            var result = [],
               parentContents = parent.getContents(),
               parentId,
               parentIsRoot = parent.isRoot(),
               parentGroupId,
               child,
               children,
               childrenCount,
               childIndex,
               i;

            if (parentIsRoot && parentContents instanceof Object) {
               parentContents = parentContents.valueOf();
            }

            parentId = parentContents instanceof Object ? Utils.getItemPropertyValue(
               parentContents,
               idProperty
            ) : parentContents;
            if (typeof parentId === 'number') {
               parentId = String(parentId);
            }

            children = treeIndex.has(parentId) ? treeIndex.get(parentId) : [];
            childrenCount = children.length;

            //FIXME: backward compatibility with controls logic: 1st level items may don\'t have parentProperty
            if (childrenCount === 0 && parentId === null && parentIsRoot) {
               children = treeIndex.has(undefined) ? treeIndex.get(undefined) : [];
               childrenCount = children.length;
            }

            //Check if that parentId is already behind
            if (parentsProcessing.has(parentId)) {
               Utils.logger.error(AdjacencyList.prototype._moduleName + ': recursive traversal detected: parent with id "' + parentId + '" is already in progress.');
            } else {
               for (i = 0; i < childrenCount; i++) {
                  childIndex = children[i];
                  child = items[childIndex];

                  if (child) {
                     child.setParent(parent, true);
                  }

                  //A new group beginning
                  if (treeGroups.has(childIndex)) {
                     lastGroupId = treeGroups.get(childIndex);
                     result.push(lastGroupId);
                  }

                  result.push(childIndex);
                  if (child && idProperty && parentProperty) {
                     parentGroupId = lastGroupId;
                     parentsProcessing.add(parentId);
                     //Lookup for children
                     push.apply(
                        result,
                        getChildren(child)
                     );
                     parentsProcessing.delete(parentId);

                     //If any child has join another group - revert parent's group
                     if (parentGroupId !== lastGroupId && i < childrenCount - 1) {
                        lastGroupId = parentGroupId;
                        result.push(lastGroupId);
                     }
                  }
               }
            }

            return result;
         };

      buildIndex();
      result = getChildren(root);
      //Save root if it's the first element
      if (items[0] === root) {
         result.unshift(0);
      }

      treeIndex.clear();
      treeGroups.clear();

      return result;
   };

   return AdjacencyList;
});
