define('WS.Data/Display/Ladder', [
   'WS.Data/Collection/IBind',
   'WS.Data/Display/Collection',
   'WS.Data/Entity/Abstract',
   'WS.Data/Entity/SerializableMixin',
   'WS.Data/Shim/Map',
   'WS.Data/Utils'
], function (
   IBindCollection,
   Collection,
   Abstract,
   SerializableMixin,
   Map,
   Utils
) {
   'use strict';

   /**
    * Лесенка - позволяет отслеживать повторяющиеся значения в колонках таблицы.
    * @class WS.Data/Display/Ladder
    * @extends WS.Data/Entity/Abstract
    * @mixes WS.Data/Entity/SerializableMixin
    * @public
    * @author Мальцев Алексей Александрович
    */
   var Ladder = Abstract.extend([SerializableMixin], /** @lends WS.Data/Display/Ladder.prototype */{
      '[WS.Data/Display/Ladder]': true,

      /**
       * @member {WS.Data/Display/Collection} Проекция, по которой строится лесенка
       */
      _collection: null,

      /**
       * @member {Number} Позиция в коллекции, с которой начинает строиться лесенка
       */
      _offset: 0,

      /**
       * @member {Object.<Function>|null} Конвертеры значений
       */
      _converters: null,

      /**
       * @member {Array.<String>} Названия колонок, входящих в лесенку
       */
      _columnNames: null,

      /**
       * @member {Map.<String, Map.<String, Boolean>>} Лесенка по ключу элементов для каждого поля
       */
      _column2primaryId: null,

      /**
       * @member {Function} Обработчик события изменения проекции
       */
      _onCollectionChangeHandler: null,

      /**
       * Конструктор лесенки.
       * @param {WS.Data/Display/Collection} collection Проекция, по которой строится лесенка.
       */
      constructor: function Ladder(collection) {
         Ladder.superclass.constructor.call(this);

         this._columnNames = [];
         this._onCollectionChangeHandler = this._onCollectionChange.bind(this);
         if (collection) {
            this.setCollection(collection);
         } else {
            this.reset();
         }
      },

      destroy: function() {
         this.setCollection(null);
         Ladder.superclass.destroy.call(this);
      },

      //region WS.Data/Entity/SerializableMixin

      _getSerializableState: function(state) {
         state = SerializableMixin._getSerializableState.call(this, state);

         if (this._collection) {
            state.$options = this._collection;
         } else {
            delete state.$options;
         }
         if (this._offset) {
            state._offset = this._offset;
         }
         if (this._columnNames.length) {
            state._columnNames = this._columnNames;
         }
         //FIXME: what about _converters?

         return state;
      },

      _setSerializableState: function(state) {
         var fromSerializableMixin = SerializableMixin._setSerializableState(state);
         return function() {
            fromSerializableMixin.call(this);

            if (state._offset) {
               this._offset = state._offset;
            }

            if (state._columnNames) {
               this._columnNames = state._columnNames;

               //Restore _column2primaryId on wake up
               if (this._collection) {
                  this._checkRange(0, this._collection.getCount());
               }
            }
         };
      },

      //endregion WS.Data/Entity/SerializableMixin

      //region Public methods

      /**
       * Возвращает проекцию коллекции, по которой строится лесенка.
       * @return {WS.Data/Display/Collection|null}
       */
      getCollection: function() {
         return this._collection;
      },

      /**
       * Устанавливает проекцию коллекции, по которой строится лесенка.
       * @param {WS.Data/Display/Collection|null} collection Проекция, по которой строится лесенка.
       */
      setCollection: function(collection) {
         if (collection !== null && !(collection instanceof Collection)) {
            throw new TypeError('Argument "collection" should be an instance of WS.Data/Display/Collection');
         }

         //Reset for  the new collection
         var reset = collection !== this._collection;

         //For the same collection just move event handler to the end (unsubscribe and then subscribe)
         if (this._collection && !this._collection.isDestroyed()) {
            this._collection.unsubscribe('onCollectionChange', this._onCollectionChangeHandler);
         }
         if (collection && !collection.isDestroyed()) {
            collection.subscribe('onCollectionChange', this._onCollectionChangeHandler);
         }

         this._collection = collection;

         if (reset) {
            this._columnNames = [];
            this.reset();
         }
      },

      /**
       * Устанавливает позицию в коллекции, с которой начинает строиться лесенка
       * @param {Number} offset Позиция.
       */
      setOffset: function(offset) {
         offset = parseInt(offset, 10);

         var prev = this._offset;
         this._offset = offset;
         var result = this._checkRange(this._offset, 2);
         this._notifyPrimaryChanges(result);

         if (Math.abs(prev - offset) > 1) {
            result = this._checkRange(prev, 1);
            this._notifyPrimaryChanges(result);
         }
      },

      reset: function() {
         this._column2primaryId = new Map();
      },

      /**
       * Устанавливает конвертер значения поля
       * @param {String} columnName Название поля
       * @param {Function(*): String} converter Конвертер значения поля
       * @return {*}
       */
      setConverter: function(columnName, converter) {
         this._converters = this._converters || {};
         this._converters[columnName] = converter;
      },

      /**
       * Возвращает значение поля с учетом лесенки
       * @param {*} item Элемент коллекции, для котрой построена проекция
       * @param {String} columnName Название поля
       * @return {String}
       */
      get: function(item, columnName) {
         return this.isPrimary(item, columnName) ?
            Utils.getItemPropertyValue(item, columnName) :
            '';
      },

      /**
       * Возвращает признак, что значение является основным (отображается)
       * @param {*} item Элемент коллекции, для котрой построена проекция
       * @param {String} columnName Название поля
       * @return {Boolean}
       */
      isPrimary: function(item, columnName) {
         if (!this._collection) {
            return true;
         }

         this._applyColumn(columnName);

         var id = this._getItemId(item),
            data = this._getColumnData(columnName),
            idx;
         if (!data.has(id)) {
            idx = this._collection.getIndexBySourceItem(item);
            this._checkRange(idx, 1);
         }

         return data.has(id) ? !!data.get(id) : true;
      },

      /**
       * Проверяет, что колонка входит в лесенку
       * @param {String} columnName Название поля
       * @return {Boolean}
       */
      isLadderColumn: function(columnName) {
         return this._column2primaryId.has(columnName);
      },

      //endregion Public methods

      //region Protected methods

      _applyColumn: function(columnName) {
         var columnNames = this._columnNames;
         if (columnNames.indexOf(columnName) > -1) {
            return;
         }
         columnNames.push(columnName);
      },

      _getColumnData: function(columnName) {
         var map = this._column2primaryId;
         if (map.has(columnName)) {
            return map.get(columnName);
         }

         var data = new Map();
         map.set(columnName, data);
         return data;
      },

      _onCollectionChange: function(event, action, newItems, newItemsIndex, oldItems, oldItemsIndex) {
         var push = Array.prototype.push,
            result = [],
            removeData = function(items) {
               if (items.length) {
                  var columnNames = this._columnNames,
                     columnIndex,
                     columnData,
                     delta,
                     itemId;
                  for (columnIndex = 0; columnIndex < columnNames.length; columnIndex++) {
                     columnData = this._getColumnData(columnNames[columnIndex]);
                     for (delta = 0; delta < items.length; ++delta) {
                        itemId = this._getCollectionItemId(items[delta]);
                        columnData.delete(itemId);
                     }
                  }
               }

            }.bind(this),
            checkItems = function(items, count, startIdx) {
               return count > 0 ? this._checkRange(startIdx - 1, items.length + 2) : [];
            }.bind(this);

         switch (action) {
            case IBindCollection.ACTION_ADD:
               push.apply(result, checkItems(newItems, newItems.length, newItemsIndex));
               break;
            case IBindCollection.ACTION_REMOVE:
               removeData(oldItems);
               push.apply(result, checkItems([], oldItems.length, oldItemsIndex));
               break;
            case IBindCollection.ACTION_CHANGE:
               result = this._checkRange(newItemsIndex - 1, 3);
               break;
            case IBindCollection.ACTION_MOVE:
               //если запись перемещают на верх то индекс сдвинется
               var startIndex = (newItemsIndex - oldItemsIndex) > 0 ? oldItemsIndex : oldItemsIndex+newItems.length;
               push.apply(result, checkItems([], oldItems.length, startIndex));
               push.apply(result, checkItems(newItems, newItems.length, newItemsIndex));
               break;
            default:
               removeData(oldItems);
               push.apply(result, checkItems(newItems, newItems.length, newItemsIndex));
         }

         this._notifyPrimaryChanges(result);
      },

      _notifyPrimaryChanges: function(changesArray) {
         var collection = this._collection,
            idx,
            columnName,
            item,
            optimized;

         optimized = changesArray.reduce(function(prev, curr) {
            if (curr !== null) {
               idx = curr[0];
               columnName = curr[1];
               item = collection.at(idx);

               prev[idx] = prev[idx] || {};
               prev[idx][columnName] = this.get(item.getContents(), columnName);
            }

            return prev;
         }.bind(this), {});

         for (idx in optimized) {
            if (optimized.hasOwnProperty(idx)) {
               collection.notifyItemChange(
                  collection.at(idx),
                  {contents: optimized[idx]}
               );
            }
         }
      },

      _checkRange: function(startIdx, length) {
         var result = [],
            columnNames = this._columnNames,
            columnIndex,
            finishIdx,
            idx,
            adjusted;

         finishIdx = Math.min(startIdx + length, this._collection.getCount());
         startIdx = Math.max(0, startIdx);

         for (columnIndex = 0; columnIndex < columnNames.length; columnIndex++) {
            for (idx = startIdx; idx < finishIdx; ++idx) {
               adjusted = this._adjustPrimary(idx, this._collection.at(idx), columnNames[columnIndex]);
               if (adjusted !== null) {
                  result.push(adjusted);
               }
            }
         }

         return result;
      },

      _adjustPrimary: function(idx, item, columnName) {
         if (!item) {
            return null;
         }

         var id = this._getCollectionItemId(item),
            data,
            nowIsPrimary,
            thenIsPrimary;

         if (id !== undefined) {
            data = this._getColumnData(columnName);
            nowIsPrimary = this._isPrimaryIndex(idx, columnName);
            thenIsPrimary = data.get(id);
            if (nowIsPrimary !== thenIsPrimary) {
               data.set(id, nowIsPrimary);
               return [idx, columnName];
            }
         }

         return null;
      },

      _isPrimaryIndex: function(idx, columnName) {
         if (idx === 0 || idx === this._offset) {
            return true;
         }

         var prev = this._collection.at(idx - 1).getContents(),
            curr = this._collection.at(idx).getContents(),
            prevVal = Utils.getItemPropertyValue(prev, columnName),
            currVal = Utils.getItemPropertyValue(curr, columnName);

         if (this._converters && this._converters.hasOwnProperty(columnName)) {
            prevVal = this._converters[columnName](prevVal, prev);
            currVal = this._converters[columnName](currVal, curr);
         }

         if ((prevVal instanceof Object) && (currVal instanceof Object)) {
            prevVal = prevVal.valueOf();
            currVal = currVal.valueOf();
         }

         return prevVal !== currVal;
      },

      _getCollectionItemId: function(collectionItem) {
         return this._getItemId(collectionItem.getContents());
      },

      _getItemId: function(item) {
         if (item.getInstanceId instanceof Function) {
            return item.getInstanceId();
         } else if (item.getId instanceof Function) {
            return item.getId();
         } else if (item.get instanceof Function) {
            return item.get('id');
         }
         return item && item.id;
      }
      
      //endregion Protected methods
   });

   return Ladder;
});
