/* global define */
define('WS.Data/Display/Tree', [
   'WS.Data/Display/Collection',
   'WS.Data/Display/GroupItem',
   'WS.Data/Display/CollectionItem',
   'WS.Data/Display/TreeItem',
   'WS.Data/Display/TreeChildren',
   'WS.Data/Display/ItemsStrategy/AdjacencyList',
   'WS.Data/Display/ItemsStrategy/MaterializedPath',
   'WS.Data/Display/ItemsStrategy/Root',
   'WS.Data/Di'
], function (
   Collection,
   GroupItem,
   CollectionItem,
   TreeItem,
   TreeChildren,
   AdjacencyListStrategy,
   MaterializedPathStrategy,
   RootStrategy,
   Di
) {
   'use strict';

   /**
    * Проекция в виде дерева - предоставляет методы навигации, фильтрации и сортировки, не меняя при этом оригинальную коллекцию.
    * Дерево может строиться по алгоритму {@link https://en.wikipedia.org/wiki/Adjacency_list Adjacency List} или {@link https://docs.mongodb.com/v3.2/tutorial/model-tree-structures-with-materialized-paths/ Materialized Path}. Выбор алгоритма выполняется в зависимости от настроек.
    * @class WS.Data/Display/Tree
    * @extends WS.Data/Display/Collection
    * @public
    * @author Мальцев Алексей
    */

   var Tree = Collection.extend(/** @lends WS.Data/Display/Tree.prototype */{
      '[WS.Data/Display/Tree]': true,

      _itemModule: 'display.tree-item',

      /**
       * @cfg {String} Название свойства, содержащего идентификатор родительского узла. Дерево в этом случае строится по алгоритму Adjacency List (список смежных вершин). Также требуется задать {@link idProperty}
       * @name WS.Data/Display/Tree#parentProperty
       */
      _$parentProperty: '',

      /**
       * @cfg {String} Название свойства, содержащего признак узла. Требуется для различения узлов и листьев.
       * @name WS.Data/Display/Tree#nodeProperty
       */
      _$nodeProperty: '',

      /**
       * @cfg {String} Название свойства, содержащего дочерние элементы узла. Дерево в этом случае строится по алгоритму Materialized Path (материализованный путь).
       * @remark Если задано, то опция {@link parentProperty} не используется.
       * @name WS.Data/Display/Tree#childrenProperty
       */
      _$childrenProperty: '',

      /**
       * @cfg {String} Название свойства, содержащего признак загруженности узла
       * @name WS.Data/Display/Tree#loadedProperty
       * @example
       * Зададим поле "Раздел$" отвечающим за признак загруженности:
       * <pre>
       *    new Tree({
       *       parentProperty: 'Раздел',
       *       loadedProperty: 'Раздел$'
       *    })
       * </pre>
       *
       */
      _$loadedProperty: '',

      /**
       * @cfg {WS.Data/Display/TreeItem|*} Корневой узел или его содержимое
       * @name WS.Data/Display/Tree#root
       */
      _$root: undefined,

      /**
       * @cfg {Boolean} Включать корневой узел в список элементов
       * @name WS.Data/Display/Tree#rootEnumerable
       * @example
       * Получим корень как первый элемент проекции:
       * <pre>
       *    var tree = new Tree({
       *       root: {id: null, title: 'Root'},
       *       rootEnumerable: true
       *    });
       *    tree.at(0).getContents().title;//'Root'
       * </pre>
       *
       */
      _$rootEnumerable: false,

      /**
       * @member {WS.Data/Display/TreeItem} Корневой элемент дерева
       */
      _root: null,

      /**
       * @member {Object.<Array.<WS.Data/Display/TreeItem>>} Соответствие узлов и их потомков
       */
      _childrenMap: {},

      constructor: function Tree(options) {
         this._childrenMap = {};
         Tree.superclass.constructor.call(this, options);

         if (this._$parentProperty) {
            this._setImportantProperty(this._$parentProperty);
         }
         if (this._$childrenProperty) {
            this._setImportantProperty(this._$childrenProperty);
         }
      },

      destroy: function () {
         this._childrenMap = {};

         Tree.superclass.destroy.call(this);
      },

      //region WS.Data/Collection/IEnumerable

      //endregion WS.Data/Collection/IEnumerable

      //region WS.Data/Entity/SerializableMixin

      _getSerializableState: function(state) {
         state = Tree.superclass._getSerializableState.call(this, state);

         state._root = this._root;

         return state;
      },

      _setSerializableState: function(state) {
         var fromSuper = Tree.superclass._setSerializableState(state);
         return function() {
            this._root = state._root;
            fromSuper.call(this);
         };
      },

      //region WS.Data/Display/Collection

      /**
       * Устанавливает текущим следующий элемент родительского узла.
       * @return {Boolean} Есть ли следующий элемент в родительском узле
       */
      moveToNext: function () {
         return this._moveTo(true);
      },

      /**
       * Устанавливает текущим предыдущий элемент родительского узла
       * @return {Boolean} Есть ли предыдущий элемент в родительском узле
       */
      moveToPrevious: function () {
         return this._moveTo(false);
      },

      //endregion WS.Data/Display/Collection

      //region Public methods

      /**
       * Возвращает название свойства, содержащего идентификатор родительского узла
       * @return {String}
       */
      getParentProperty: function () {
         return this._$parentProperty;
      },

      /**
       * Устанавливает название свойства, содержащего идентификатор родительского узла
       * @param {String} name
       */
      setParentProperty: function (name) {
         this._unsetImportantProperty(this._$parentProperty);
         this._$parentProperty = name;
         this._getItemsStrategy().getOptions().parentProperty = name;
         this._setImportantProperty(name);
         this._reIndex();
         this._reAnalize();
      },

      /**
       * Возвращает название свойства, содержащего признак узла
       * @return {String}
       */
      getNodeProperty: function () {
         return this._$nodeProperty;
      },

      /**
       * Возвращает название свойства, содержащего дочерние элементы узла
       * @return {String}
       */
      getChildrenProperty: function () {
         return this._$childrenProperty;
      },

      /**
       * Возвращает название свойства, содержащего признак загруженности узла
       * @return {String}
       */
      getLoadedProperty: function () {
         return this._$loadedProperty;
      },

      /**
       * Возвращает корневой узел дерева
       * @return {WS.Data/Display/TreeItem}
       */
      getRoot: function () {
         if (this._root === null) {
            this._root = this._$root;
            if (!(this._root instanceof TreeItem)) {
               this._root = new TreeItem({
                  contents: this._root,
                  owner: this,
                  node: true,
                  expanded: true,
                  loaded: true
               });
            }
         }

         return this._root;
      },

      /**
       * Устанавливает корневой узел дерева
       * @param {WS.Data/Display/TreeItem|*} root Корневой узел или его содержимое
       */
      setRoot: function (root) {
         this._$root = root;
         this._root = null;
         this._reIndex();
         this._reAnalize();
      },

      /**
       * Возвращает признак, что корневой узел включен в список элементов
       * @return {Boolean}
       */
      isRootEnumerable: function () {
         return this._$rootEnumerable;
      },

      /**
       * Устанавливает признак, что корневой узел включен в список элементов
       * @param {Boolean} enumerable Корневой узел включен в список элементов
       */
      setRootEnumerable: function (enumerable) {
         if (this._$rootEnumerable === enumerable) {
            return;
         }
         this._$rootEnumerable = enumerable;

         var session = this._startUpdateSession();
         if (enumerable) {
            this._wrapRootStrategy();
         } else {
            this._unwrapRootStrategy();
         }
         this._reSort();
         this._reFilter();
         this._finishUpdateSession(session);
      },

      /**
       * Возвращает коллекцию потомков элемента коллекции
       * @param {WS.Data/Display/TreeItem} parent Родительский узел
       * @param {Boolean} [withFilter=true] Учитывать {@link setFilter фильтр}
       * @return {WS.Data/Display/TreeChildren}
       */
      getChildren: function (parent, withFilter) {
         return new TreeChildren({
            owner: parent,
            items: this._getChildrenArray(parent, withFilter)
         });
      },

      /**
       * Устанавливает текущим родителя текущего элемента
       * @return {Boolean} Есть ли родитель
       */
      moveToAbove: function () {
         var current = this.getCurrent();
         if (!current) {
            return false;
         }
         var parent = current.getParent();
         if (!parent || parent.isRoot()) {
            return false;
         }

         this.setCurrent(parent);
         return true;
      },

      /**
       * Устанавливает текущим первого непосредственного потомка текущего элемента
       * @return {Boolean} Есть ли первый потомок
       */
      moveToBelow: function () {
         var current = this.getCurrent();
         if (!current || !current.isNode()) {
            return false;
         }
         var children = this._getChildrenArray(current);
         if (children.length === 0) {
            return false;
         }

         this.setCurrent(children[0]);
         return true;
      },

      //endregion Public methods

      //region Protected methods

      _createItemsStrategy: function () {
         if (this._$childrenProperty) {
            return new MaterializedPathStrategy(this, {
               itemModule: this._itemModule,
               childrenProperty: this._$childrenProperty,
               nodeProperty: this._$nodeProperty,
               loadedProperty: this._$loadedProperty,
               root: this.getRoot.bind(this)
            });
         } else {
            return new AdjacencyListStrategy(this, {
               itemModule: this._itemModule,
               idProperty: this._$idProperty,
               parentProperty: this._$parentProperty,
               nodeProperty: this._$nodeProperty,
               loadedProperty: this._$loadedProperty,
               root: this.getRoot.bind(this)
            });
         }
      },

      _wrapItemsStrategy: function() {
         Tree.superclass._wrapItemsStrategy.call(this);
         this._wrapRootStrategy();
      },

      _wrapRootStrategy: function() {
         if (this._itemsStrategy && this._$rootEnumerable && !(this._itemsStrategy instanceof RootStrategy)) {
            this._itemsStrategy = new RootStrategy(this._itemsStrategy,  this.getRoot.bind(this));
         }
      },

      _unwrapRootStrategy: function() {
         if (this._itemsStrategy && !this._$rootEnumerable && (this._itemsStrategy instanceof RootStrategy)) {
            this._itemsStrategy = this._itemsStrategy.source;
         }
      },

      _reIndex: function() {
         Tree.superclass._reIndex.call(this);
         this._childrenMap = {};
      },

      _bindHandlers: function() {
         Tree.superclass._bindHandlers.call(this);

         this._onSourceCollectionChange = onSourceCollectionChange.bind({
            instance: this,
            prev: this._onSourceCollectionChange
         });

         this._onSourceCollectionItemChange = onSourceCollectionItemChange.bind({
            instance: this,
            prev: this._onSourceCollectionItemChange
         });
      },

      _replaceItems: function (start, newItems) {
         var replaced = Tree.superclass._replaceItems.call(this, start, newItems),
            strategy = this._getItemsStrategy();

         replaced.forEach(function(item, index) {
            strategy.at(start + index).setExpanded(item.isExpanded(), true);
         });

         return replaced;
      },

      _getItemState: function(item) {
         var state = Tree.superclass._getItemState.call(this, item);

         if (item instanceof TreeItem) {
            state.parent = item.getParent();
            state.childrenCount = item.getOwner()._getChildrenArray(item, false).length;
            state.level = item.getLevel();
            state.node = item.isNode();
            state.expanded = item.isExpanded();
         }

         return state;
      },

      /**
       * Проверяет валидность элемента проекции
       * @param {*} item Элемент проекции
       * @protected
       */
      _checkItem: function (item) {
         if (!item || !(item instanceof CollectionItem)) {
            throw new Error(this._moduleName + '::_checkItem(): item should be in instance of WS.Data/Display/CollectionItem');
         }
      },

      /**
       * Возвращает массив детей для указанного родителя
       * @param {WS.Data/Display/TreeItem} parent Родительский узел
       * @param {Boolean} [withFilter=true] Учитывать {@link setFilter фильтр}
       * @return {Array.<WS.Data/Display/TreeItem>}
       * @protected
       */
      _getChildrenArray: function (parent, withFilter) {
         this._checkItem(parent);

         withFilter = withFilter === undefined ? true : !!withFilter;
         var iid = parent.getInstanceId(),
            key = iid + '|' + withFilter;

         if (!(key in this._childrenMap)) {
            var children = [],
               enumerator;

            if (withFilter) {
               enumerator = this.getEnumerator();
            } else {
               enumerator = this._buildEnumerator(
                  this._getItems.bind(this),
                  this._filterMap.map(function() {
                     return true;
                  }),
                  this._sortMap
               );
            }

            enumerator.setCurrent(parent);
            if (enumerator.getCurrent() === parent || parent.isRoot()) {
               var item;
               while (enumerator.moveNext()) {
                  item = enumerator.getCurrent();
                  if (!(item instanceof TreeItem)) {
                     continue;
                  }
                  if (item.getParent() === parent) {
                     children.push(item);
                  } else if (item.getLevel() === parent.getLevel()) {
                     break;
                  }
               }
            }

            this._childrenMap[key] = children;
         }

         return this._childrenMap[key];
      },

      _getNearbyItem: function (enumerator, item, isNext, skipGroups) {
         var method = isNext ? 'moveNext' : 'movePrevious',
            parent = item && item.getParent && item.getParent() || this.getRoot(),
            hasItem = true,
            initial = enumerator.getCurrent(),
            sameParent = false,
            current,
            nearbyItem;

         enumerator.setCurrent(item);
         //TODO: отлеживать по level, что вышли "выше"
         while (hasItem && !sameParent) {
            hasItem = enumerator[method]();
            nearbyItem = enumerator.getCurrent();

            if (skipGroups && nearbyItem instanceof GroupItem) {
               nearbyItem = undefined;
               continue;
            }

            sameParent = nearbyItem ? nearbyItem.getParent() === parent : false;
            current = (hasItem && sameParent) ? nearbyItem : undefined;
         }

         enumerator.setCurrent(initial);

         return current;
      },

      _moveTo: function (isNext) {
         var enumerator = this._getCursorEnumerator(),
            initial = this.getCurrent(),
            item = this._getNearbyItem(enumerator, initial, isNext, true),
            hasMove = !!item;
         if (hasMove) {
            this.setCurrent(item);
         } else {
            enumerator.setCurrent(initial);
         }
         return hasMove;
      },

      _notifyItemsParent: function(treeItem, oldParent, properties) {
         if (properties.hasOwnProperty(this.getParentProperty())) {
            this._notifyItemsParentByItem(treeItem.getParent());
            this._notifyItemsParentByItem(oldParent);
         }
      },

      _notifyItemsParentByItem: function(treeItem) {
         while (treeItem !== this.getRoot()) {
            this.notifyItemChange(treeItem, {children: []});
            treeItem = treeItem.getParent();
         }
      }
      //endregion Protected methods

   });

   var /**
       * Обрабатывает событие об изменении коллекции
       * @param {Core/EventObject} event Дескриптор события.
       * @param {String} action Действие, приведшее к изменению.
       * @param {*[]} newItems Новые элементы коллекции.
       * @param {Number} newItemsIndex Индекс, в котором появились новые элементы.
       * @param {*[]} oldItems Удаленные элементы коллекции.
       * @param {Number} oldItemsIndex Индекс, в котором удалены элементы.
       */
      onSourceCollectionChange = function (event, action, newItems, newItemsIndex, oldItems, oldItemsIndex) {
         //Fix state of all nodes
         var nodes = this.instance._getItems().filter(function(item) {
               return item.isNode && item.isNode();
            }),
            state = this.instance._getItemsState(nodes),
            session = this.instance._startUpdateSession();

         this.instance._reIndex();
         this.prev(event, action, newItems, newItemsIndex, oldItems, oldItemsIndex);

         //Check state of all nodes. They can change children count (include hidden by filter).
         this.instance._finishUpdateSession(session, false);
         this.instance._checkItemsDiff(session, nodes, state);
      },

      /**
       * Обрабатывает событие об изменении элемента коллекции
       * @param {Core/EventObject} event Дескриптор события.
       * @param {*} item Измененный элемент коллекции.
       * @param {Number} index Индекс измененного элемента.
       * @param {Object} properties Объект содержащий измененные свойства элемента
       */
      onSourceCollectionItemChange = function (event, item, index, properties) {
         this.instance._reIndex();
         this.prev(event, item, index, properties);
      };

   Di.register('display.tree', Tree);

   return Tree;
});
