/* global define */
define('WS.Data/Entity/ManyToManyMixin', [
   'WS.Data/Mediator/ManyToMany'
], function (
   ManyToManyMediator
) {
   'use strict';

   /**
    * Миксин, позволяющий строить отношения "многие ко многим"
    * @mixin WS.Data/Entity/ManyToManyMixin
    * @public
    * @author Мальцев Алексей
    */

   var ManyToManyMixin = /**@lends WS.Data/Entity/ManyToManyMixin.prototype  */{
      '[WS.Data/Entity/ManyToManyMixin]': true,

      /**
       * @member {WS.Data/Mediator/ManyToMany} Медиатор, отвечающий за связи между сущностями
       */
      _mediator: null,

      //region Public methods

      destroy: function() {
         var mediator = this._getMediator(),
            slaves = [],
            slave,
            i,
            count;

         mediator.hasMany(this, function(slave) {
            slaves.push(slave);
         });

         mediator.clear(this);

         for (i = 0, count = slaves.length; i < count; i++) {
            slave = slaves[i];
            if (slave.destroy) {
               slave.destroy();
            }
         }

         this._setMediator(null);
      },

      //endregion Public methods

      //region Protected methods

      /**
       * Добавляет отношение с другой сущностью
       * @param {WS.Data/Mediator/IReceiver} child Другая сущность
       * @param {String} [name] Название отношения
       * @protected
       */
      _addChild: function(child, name) {
         if (child instanceof Object) {
            var mediator = this._getMediator();
            mediator.addRelationship(this, child, name);


            if (child['[WS.Data/Entity/ManyToManyMixin]'] && //mixes WS.Data/Entity/ManyToManyMixin
               !child._hasSameMediator(mediator)
            ) {
               if (!child._hasMediator()) {
                  child._setMediator(this._createMediator());
               }
               child._getMediator().addRelationship(this, child, name);
            }
         }
      },

      /**
       * Удаляет отношение с другой сущностью
       * @param {WS.Data/Mediator/IReceiver} child Другая сущность
       * @protected
       */
      _removeChild: function(child) {
         if (child instanceof Object) {
            var mediator = this._getMediator();
            mediator.removeRelationship(this, child);

            if (child['[WS.Data/Entity/ManyToManyMixin]'] && //mixes WS.Data/Entity/ManyToManyMixin
               child._hasMediator() &&
               !child._hasSameMediator(mediator)
            ) {
               child._getMediator().removeRelationship(this, child);
            }
         }
      },

      /**
       * Уведомляет дочерние сущности об изменении родительской
       * @param {*} [data] Данные об изменениях
       * @protected
       */
      _parentChanged: function(data) {
         var which = {
            target: this,
            data: data
         };
         this._getMediator().hasMany(this, function(slave, name) {
            if (slave['[WS.Data/Mediator/IReceiver]']) {
               //implements WS.Data/Mediator/IReceiver
               slave.relationChanged(which, [name]);
            }
         }.bind(this));
      },

      /**
       * Рекурсивно уведомляет родительские сущности об изменении дочерней
       * @param {*} [data] Данные об изменениях
       * @protected
       */
      _childChanged: function (data) {
         var notifyParent = function(mediator, child, route) {
            mediator.belongsTo(child, function(parent, name) {
               var childRoute = route.slice();
               var which = {
                  target: child,
                  data: data
               };
               childRoute.unshift(name);
               if (parent['[WS.Data/Mediator/IReceiver]']) {
                  //implements WS.Data/Mediator/IReceiver
                  parent.relationChanged(which, childRoute);
               }

               notifyParent(parent._getMediator(), parent, childRoute);
            });
         };

         notifyParent(this._getMediator(), this, []);
      },

      /**
       * Возвращает признак наличия посредника
       * @return {Boolean}
       * @protected
       */
      _hasMediator: function() {
         return !!this._mediator;
      },

      /**
       * Возвращает признак наличия одинакового посредника
       * @param {WS.Data/Mediator/ManyToMany} mediator
       * @return {Boolean}
       * @protected
       */
      _hasSameMediator: function(mediator) {
         return this._mediator === mediator;
      },

      /**
       * Создает посредника для установления отношений с детьми
       * @return {WS.Data/Mediator/ManyToMany}
       * @protected
       */
      _createMediator: function() {
         return new ManyToManyMediator();
      },

      /**
       * Возвращает посредника для установления отношений с детьми
       * @return {WS.Data/Mediator/ManyToMany}
       * @protected
       */
      _getMediator: function() {
         return this._mediator || (this._mediator = this._createMediator());
      },

      /**
       * Устанавливает посредника для установления отношений с детьми
       * @param {WS.Data/Mediator/ManyToMany|null} mediator
       * @protected
       */
      _setMediator: function(mediator) {
         this._mediator = mediator;
      }

      //endregion Protected methods
   };

   return ManyToManyMixin;
});
