/* global define */
define('WS.Data/Entity/Model', [
   'WS.Data/Entity/Record',
   'WS.Data/Entity/IInstantiable',
   'WS.Data/Entity/InstantiableMixin',
   'WS.Data/Collection/ArrayEnumerator',
   'WS.Data/Functor/Compute',
   'WS.Data/Di',
   'WS.Data/Utils',
   'WS.Data/Shim/Map',
   'WS.Data/Shim/Set',
   'Core/core-merge'
], function (
   Record,
   IInstantiable,
   InstantiableMixin,
   ArrayEnumerator,
   Compute,
   Di,
   Utils,
   Map,
   Set,
   coreMerge
) {
   'use strict';

   /**
    * Абстрактная модель.
    * Модели обеспечивают доступ к данным и поведению объектов предметной области (сущностям).
    * Такими сущностями могут быть, например, товары, пользователи, документы - и другие предметы окружающего мира, которые вы моделируете в своем приложении.
    *
    * В основе абстрактной модели лежит {@link WS.Data/Entity/Record запись}.
    * Основные аспекты модели (дополнительно к аспектам записи):
    * <ul>
    *    <li>определение {@link properties собственных свойств} сущности;</li>
    *    <li>{@link getId уникальный идентификатор сущности} среди ей подобных.</li>
    * </ul>
    *
    * Поведенческие аспекты каждой сущности реализуются ее прикладным модулем в виде публичных методов.
    * Прикладные модели могут внедряться в порождающие их объекты, такие как {@link WS.Data/Source/ISource#model источники данных} или {@link WS.Data/Collection/RecordSet#model рекордсеты}.
    *
    * Для реализации конкретной модели используется наследование от абстрактной либо промежуточной.
    *
    * Для корректной сериализации и клонирования моделей необходимо выносить их в отдельные модули и указывать имя модуля в свойстве _moduleName каждого наследника:
    * <pre>
    * define('My.Awesome.Model', ['WS.Data/Entity/Model'], function (Model) {
    *    'use strict';
    *
    *    var AwesomeModel = Model.extend({
    *      _moduleName: 'My.Awesome.Model'
    *      //...
    *    });
    *
    *    return AwesomeModel;
    * });
    * </pre>
    *
    * Определим модель пользователя:
    * <pre>
    *    define('Application/Model/User', ['WS.Data/Entity/Model', 'Application/Lib/Salt'], function (Model, Salt) {
    *       var User = Model.extend({
    *          _moduleName: 'Application/Model/User'
    *          _$format: [
    *             {name: 'login', type: 'string'},
    *             {name: 'salt', type: 'string'}
    *          ],
    *          _$idProperty: 'login',
    *          authenticate: function(password) {
    *             return Salt.encode(this.get('login') + ':' + password) === this.get('salt');
    *          }
    *       });
    *
    *       return User;
    *    });
    * </pre>
    * Создадим модель пользователя:
    * <pre>
    *    define('Application/Controller/Test/Auth', ['Application/Model/User'], function (User) {
    *       var user = new User();
    *       user.set({
    *          login: 'i.c.wiener',
    *          salt: 'grhS2Nys345fsSW3mL9'
    *       });
    *       var testOk = user.authenticate('its pizza time!');
    *    });
    * </pre>
    *
    * Модели могут объединяться по принципу "матрёшки" - сырыми данными одной модели является другая модель. Для организации такой структуры следует использовать {@link WS.Data/Adapter/RecordSet адаптер рекордсета}:
    * <pre>
    *    var MyEngine, MyTransmission, myCar;
    *
    *    MyEngine = Model.extend({
    *       _$properties: {
    *          fuelType: {
    *             get: function() {
    *                return 'Diesel';
    *             }
    *          }
    *       }
    *    });
    *
    *    MyTransmission = Model.extend({
    *       _$properties: {
    *          transmissionType: {
    *             get: function() {
    *                return 'Manual';
    *             }
    *          }
    *       }
    *    });
    *
    *    myCar = new MyEngine({
    *       rawData: new MyTransmission({
    *          rawData: {
    *             color: 'Red',
    *             fuelType: '',
    *             transmissionType: ''
    *          }
    *       }),
    *       adapter: new RecordSetAdapter()
    *   });
    *
    *   myCar.get('fuelType');//'Diesel'
    *   myCar.get('transmissionType');//'Manual'
    *   myCar.get('color');//'Red'
    * </pre>
    * @class WS.Data/Entity/Model
    * @extends WS.Data/Entity/Record
    * @implements WS.Data/Entity/IInstantiable
    * @mixes WS.Data/Entity/InstantiableMixin
    * @public
    * @ignoreMethods getDefault
    * @author Мальцев Алексей
    */

   var Model = Record.extend([IInstantiable, InstantiableMixin], /** @lends WS.Data/Entity/Model.prototype */{
      '[WS.Data/Entity/Model]': true,

      /**
       * @typedef {Object} Property
       * @property {*|Function} [def] Значение по умолчанию (используется, если свойства нет в сырых данных).
       * @property {Function} [get] Метод, возвращающий значение свойства. Первым аргументом придет значение свойства в сырых данных (если оно там есть).
       * @property {Function} [set] Метод, устанавливающий значение свойства. Если метод вернет значение, отличное от undefined, то будет осуществлена попытка сохранить его в сырых данных.
       */

      _instancePrefix: 'model-',

      /**
       * @cfg {Object.<Property>} Описание собственных свойств модели. Дополняет/уточняет свойства, уже существующие в сырых данных.
       * @name WS.Data/Entity/Model#properties
       * @see Property
       * @see getProperties
       * @example
       * Создадим модель пользователя со свойствами:
       * <ul>
       *    <li>id (чтение/запись, динамическая конвертация, хранится в сырых данных)</li>
       *    <li>group (чтение/запись, хранится в защищенном свойстве)</li>
       *    <li>guid (только чтение, значение по умолчанию генерируется динамически)</li>
       * </ul>
       * <pre>
       *    require(['WS.Data/Entity/Model', 'Core/helpers/random-helpers'], function(Model, randomHelpers) {
       *       var User = Model.extend({
       *          _$properties: {
       *             id: {
       *                get: function(value) {
       *                   return '№' + value;
       *                },
       *                set: function(value) {
       *                   return (value + '')[0] === '№' ? value.substr(1) : value;
       *                }
       *             },
       *             group: {
       *                get: function() {
       *                   return this._group;
       *                },
       *                set: function(value) {
       *                   this._group = value;
       *                }
       *             },
       *             guid: {
       *                def: function() {
       *                   return randomHelpers.createGUID();
       *                },
       *                get: function(value) {
       *                   return value;
       *                }
       *             }
       *          },
       *          _group: null
       *       });
       *
       *       var user = new User({
       *          rawData: {
       *             id: 5,
       *             login: 'Keanu',
       *             firstName: 'Johnny',
       *             lastName: 'Mnemonic',
       *             job: 'Memory stick'
       *          }
       *       });
       *
       *       console.log(user.get('id'));//№5
       *       console.log(user.get('group'));//null
       *       console.log(user.get('guid'));//010a151c-1160-d31d-11b3-18189155cc13
       *       console.log(user.get('job'));//Memory stick
       *       console.log(user.get('uptime'));//undefined
       *
       *       user.set('id', '№6');
       *       console.log(user.getRawData().id);//6
       *
       *       user.set('group', {id: 1, name: 'The One'});
       *       console.log(user.get('group'));//{id: 1, name: 'The One'}
       *
       *       user.set('guid', 'new-one');//ReferenceError 'Model::set(): property "guid" is read only'
       *    });
       * </pre>
       * Создадим модель пользователя со свойством displayName, которое вычисляется с использованием значений других свойств:
       * <pre>
       *    require(['WS.Data/Entity/Model'], function(Model) {
       *       var User = Model.extend({
       *          _$properties: {
       *             displayName: {
       *                get: function() {
       *                   return this.get('firstName') + ' a.k.a "' + this.get('login') + '" ' + this.get('lastName');
       *                }
       *             }
       *          }
       *       });
       *
       *       var user = new User({
       *          rawData: {
       *             login: 'Keanu',
       *             firstName: 'Johnny',
       *             lastName: 'Mnemonic',
       *          }
       *       });
       *       console.log(user.get('displayName'));//Johnny a.k.a "Keanu" Mnemonic
       *    });
       * </pre>
       * Можно явно указать список свойств, от которых зависит другое свойство. В этом случае для свойств-объектов будет сбрасываться кэш, хранящий результат предыдущего вычисления:
       * <pre>
       *    require(['WS.Data/Entity/Model', 'WS.Data/Functor/Compute'], function(Model, Compute) {
       *       var User = Model.extend({
       *          _$properties: {
       *             birthDay: {
       *                get: new Compute(function() {
       *                   return this.get('facebookBirthDay') || this.get('linkedInBirthDay');
       *                }, ['facebookBirthDay', 'linkedInBirthDay'])
       *             }
       *          }
       *       });
       *
       *       var user = new User();
       *       user.set('linkedInBirthDay', new Date(2010, 1, 2));
       *       console.log(user.get('birthDay'));//Tue Feb 02 2010 00:00:00
       *
       *       user.set('facebookBirthDay', new Date(2011, 3, 4));
       *       console.log(user.get('birthDay'));//Mon Apr 04 2011 00:00:00
       *    });
       * </pre>
       */
      _$properties: null,

      /**
       * @cfg {String} Название свойства, содержащего первичный ключ
       * @name WS.Data/Entity/Model#idProperty
       * @see getIdProperty
       * @see setIdProperty
       * @see getId
       * @example
       * Зададим первичным ключом модели свойство с названием id:
       * <pre>
       *    var article = new Model({
       *       idProperty: 'id',
       *       rawData: {
       *          id: 1,
       *          title: 'How to make a Model'
       *       }
       *    });
       *    article.getId();//1
       * </pre>
       */
      _$idProperty: '',

      /**
       * @member {Boolean} Признак, что модель удалена из источника данных
       */
      _isDeleted: false,

      /**
       * @member {Object} Объект, содержащий вычисленные значения свойств по умолчанию
       */
      _defaultPropertiesValues: null,

      /**
       * @member {Map.<String, Set.<Sting>>} Зависимости свойств: название свойства -> названия свойств, которые от него зависят
       */
      _propertiesDependency: null,

      /**
       * @member {String} Имя свойства, для которого сейчас осуществляется сбор зависимостей
       */
      _propertiesDependencyGathering: '',

      /**
       * @member {Set.<String>} Объект, содержащий названия свойств, для которых сейчас выполняется вычисление значения
       */
      _nowCalculatingProperties: null,

      /**
       * @member {Object} Флаг показывающий была ли модель синхронизирована
       */
      _synced: false,

      constructor: function Model(options) {
         Model.superclass.constructor.call(this, options);

         //FIXME: don't allow to inject properties through constructor
         this._propertiesInjected = options && 'properties' in options;

         //FIXME: backward compatibility for _options
         if (this._options) {
            //for _$properties
            if (this._options.properties) {
               var properties = {};
               coreMerge(properties, this._$properties, {clone: true});
               coreMerge(properties, this._options.properties, {clone: true});
               this._$properties = properties;
            }
            //for _$idProperty
            if (this._options.idProperty) {
               this._$idProperty = this._options.idProperty;
            }
         }

         if (!this._$idProperty) {
            this._$idProperty = this._getAdapter().getKeyField(this._getRawData()) || '';
         }
      },

      destroy: function() {
         this._defaultPropertiesValues = null;
         this._propertiesDependency = null;
         this._nowCalculatingProperties = null;

         Model.superclass.destroy.call(this);
      },

      // region WS.Data/Entity/IObject

      get: function (name) {
         this._pushDependency(name);

         if (this._hasInPropertiesCache(name)) {
            return this._getFromPropertiesCache(name);
         }

         var property = this._$properties && this._$properties[name],
            preValue,
            value,
            superValue;

         superValue = Model.superclass.get.call(this, name);
         if (!property) {
            return superValue;
         }

         preValue = superValue;
         if ('def' in property && !this._getRawDataAdapter().has(name)) {
            preValue = this.getDefault(name);
         }

         if (!property.get) {
            return preValue;
         }

         value = this._processCalculatedValue(name, preValue, property, true);

         if (value !== superValue) {
            this._removeChild(superValue);
            this._addChild(value, this._getRelationNameForField(name));
         }

         if (this._isFieldValueCacheable(value)) {
            this._setToPropertiesCache(name, value);
         } else if (this._hasInPropertiesCache(name)) {
            this._unsetFromPropertiesCache(name);
         }

         return value;
      },

      set: function (name, value) {
         if (!this._$properties) {
            Model.superclass.set.call(this, name, value);
            return;
         }

         var map = this._getHashMap(name, value),
            superMap = {},
            key,
            property,
            errors;

         for (key in map) {
            if (!map.hasOwnProperty(key)) {
               continue;
            }

            this._deleteDependencyCache(key);

            value = map[key];

            property = this._$properties && this._$properties[key];

            try {
               //Try to set every property
               if (property) {
                  if (property.set) {
                     if (this._hasInPropertiesCache(key)) {
                        this._removeChild(
                           this._getFromPropertiesCache(key)
                        );
                        this._unsetFromPropertiesCache(key);
                     }

                     value = this._processCalculatedValue(key, value, property, false);
                     if (value === undefined) {
                        continue;
                     }
                  } else if (property.get) {
                     throw new ReferenceError('Property "' + key + '" is read only');
                  }
               }
               superMap[key] = value;
            } catch (e) {
               //Collecting errors for every property
               errors = errors || [];
               errors.push(e);
            }
         }

         Model.superclass.set.call(this, superMap);

         if (errors) {
            if (errors.length === 1) {
               throw errors[0];
            } else {
               throw new ReferenceError(this._moduleName + '::set(): ' + errors.join(', '));
            }
         }
      },

      has: function (name) {
         return (this._$properties && this._$properties.hasOwnProperty(name)) || Model.superclass.has.call(this, name);
      },

      // endregion WS.Data/Entity/IObject

      // region WS.Data/Collection/IEnumerable

      /**
       * Возвращает энумератор для перебора названий свойств модели
       * @return {WS.Data/Collection/ArrayEnumerator}
       * @example
       * Смотри пример {@link WS.Data/Entity/Record#getEnumerator для записи}:
       */
      getEnumerator: function () {
         return new ArrayEnumerator(this._getAllProperties());
      },

      /**
       * Перебирает все свойства модели (включая имеющиеся в "сырых" данных)
       * @param {Function(String, *)} callback Ф-я обратного вызова для каждого свойства. Первым аргументом придет название свойства, вторым - его значение.
       * @param {Object} [context] Контекст вызова callback.
       * @example
       * Смотри пример {@link WS.Data/Entity/Record#each для записи}:
       */
      each: function (callback, context) {
         return Model.superclass.each.call(this, callback, context);
      },

      // endregion WS.Data/Collection/IEnumerable

      // region WS.Data/Entity/SerializableMixin

      _getSerializableState: function() {
         var state = Model.superclass._getSerializableState.call(this);

         //Properties are owned by class, not by instance
         if (!this._propertiesInjected) {
            delete state.$options.properties;
         }

         state._instanceId = this.getInstanceId();
         state._isDeleted = this._isDeleted;
         if (this._defaultPropertiesValues) {
            state._defaultPropertiesValues = this._defaultPropertiesValues;
         }

         return state;
      },

      _setSerializableState: function(state) {
         var fromSuper = Model.superclass._setSerializableState(state);
         return function() {
            fromSuper.call(this);

            this._instanceId = state._instanceId;
            this._isDeleted = state._isDeleted;
            if (state._defaultPropertiesValues) {
               this._defaultPropertiesValues = state._defaultPropertiesValues;
            }
         };
      },

      // endregion WS.Data/Entity/SerializableMixin

      // region Public methods

      /**
       * Возвращает описание свойств модели.
       * @return {Object.<Property>}
       * @see properties
       * @see Property
       * @example
       * Получим описание свойств модели:
       * <pre>
       *    var User = Model.extend({
       *          _$properties: {
       *             id: {
       *                get: function() {
       *                   this._id;
       *                },
       *                set: function(value) {
       *                   this._id = value;
       *                }
       *             },
       *             group: {
       *                get: function() {
       *                   this._group;
       *                }
       *             }
       *          },
       *          _id: 0
       *          _group: null
       *       }),
       *       user = new User();
       *
       *    user.getProperties();//{id: {get: Function, set: Function}, group: {get: Function}}
       * </pre>
       */
      getProperties: function () {
         return this._$properties;
      },

      /**
       * Возвращает значение свойства по умолчанию
       * @param {String} name Название свойства
       * @return {*}
       * @example
       * Получим дефолтное значение свойства id:
       * <pre>
       *    var User = Model.extend({
       *          _$properties: {
       *             id: {
       *                get: function() {
       *                   this._id;
       *                },
       *                def: function(value) {
       *                   return Date.now();
       *                }
       *             }
       *          },
       *          _id: 0
       *       }),
       *       user = new User();
       *
       *    user.getDefault('id');//1466419984715
       *    setTimeout(function(){
       *       user.getDefault('id');//1466419984715
       *    }, 100);
       * </pre>
       */
      getDefault: function (name) {
         var defaultPropertiesValues = this._defaultPropertiesValues;
         if (!defaultPropertiesValues) {
            defaultPropertiesValues = this._defaultPropertiesValues = {};
         }

         if (!defaultPropertiesValues.hasOwnProperty(name)) {
            var property = this._$properties[name];
            if (property && 'def' in property) {
               defaultPropertiesValues[name] = [property.def instanceof Function ? property.def.call(this) : property.def];
            } else {
               defaultPropertiesValues[name] = [];
            }
         }
         return defaultPropertiesValues[name][0];
      },

      /**
       * Объединяет модель с данными другой модели
       * @param {WS.Data/Entity/Model} model Модель, с которой будет произведено объединение
       * @example
       * Объединим модели пользователя и группы пользователей:
       * <pre>
       *    var user = new Model({
       *          rawData: {
       *             id: 1,
       *             login: 'user1',
       *             group_id: 3
       *          }
       *       }),
       *       userGroup = new Model({
       *          rawData: {
       *             group_id: 3,
       *             group_name: 'Domain Users',
       *             group_members: 126
       *          }
       *       });
       *
       *    user.merge(userGroup);
       *    user.get('id');//1
       *    user.get('group_id');//3
       *    user.get('group_name');//'Domain Users'
       * </pre>
       */
      merge: function (model) {
         try {
            var modelData = {};
            model.each(function(key, val) {
               modelData[key] = val;
            });
            this.set(modelData);
         } catch (e) {
            if (e instanceof ReferenceError) {
               Utils.logger.info(this._moduleName + '::merge(): ' + e.toString());
            } else {
               throw e;
            }
         }
      },

      /**
       * Возвращает значение первичного ключа модели
       * @return {*}
       * @see idProperty
       * @see getIdProperty
       * @see setIdProperty
       * @example
       * Получим значение первичного ключа статьи:
       * <pre>
       *    var article = new Model({
       *       idProperty: 'id',
       *       rawData: {
       *          id: 1,
       *          title: 'How to make a Model'
       *       }
       *    });
       *    article.getId();//1
       * </pre>
       */
      getId: function () {
         var idProperty = this.getIdProperty();
         if (!idProperty) {
            Utils.logger.info(this._moduleName + '::getId(): idProperty is not defined');
            return undefined;
         }
         return this.get(idProperty);
      },

      /**
       * Возвращает название свойства, в котором хранится первичный ключ модели
       * @return {String}
       * @see idProperty
       * @see setIdProperty
       * @see getId
       * @example
       * Получим название свойства первичного ключа:
       * <pre>
       *    var article = new Model({
       *       idProperty: 'id',
       *       rawData: {
       *          id: 1,
       *          title: 'How to make a Model'
       *       }
       *    });
       *    article.getIdProperty();//'id'
       * </pre>
       */
      getIdProperty: function () {
         return this._$idProperty;
      },

      /**
       * Устанавливает название свойства, в котором хранится первичный ключ модели
       * @param {String} idProperty Название свойства для первичного ключа модели.
       * @see idProperty
       * @see getIdProperty
       * @see getId
       * @example
       * Зададим название свойства первичного ключа:
       * <pre>
       *    var article = new Model({
       *       rawData: {
       *          id: 1,
       *          title: 'How to make a Model'
       *       }
       *    });
       *    article.setIdProperty('id');
       *    article.getId();//1
       * </pre>
       */
      setIdProperty: function (idProperty) {
         if (idProperty && !this.has(idProperty)) {
            Utils.logger.info(this._moduleName + '::setIdProperty(): property "' + idProperty + '" is not defined');
            return;
         }
         this._$idProperty = idProperty;
      },

      /**
       * Возвращает значения всех свойств в виде объекта ключ-значение
       * @return {Object}
       * @example
       * Получим значения всех свойств в виде объекта:
       * <pre>
       *    var article = new Model({
       *       adapter: 'adapter.xml',
       *       rawData: '<?xml version="1.0"?><response><id>1</id><title>Article 1</title></response>'
       *    });
       *    article.toObject();//{id: 1, title: 'Article 1'}
       * </pre>
       * @deprecated Метод будет удален в 3.7.6, используйте {@link WS.Data/Chain/Abstract#toObject}.
       */
      toObject: function() {
         Utils.logger.stack(this._moduleName + '::toObject(): method is deprecated and will be removed in 3.7.6. Use WS.Data/Chain/Abstract::toObject() instead. See https://wi.sbis.ru/docs/WS/Data/Chain/Abstract/methods/toObject/ for details.', 0, 'error');
         var data = {};
         this.each(function(field, value) {
            data[field] = value;
         });
         return data;
      },

      /**
      * Возвращает значения всех свойств в виде строки формата json
      * @return {String}
       * @example
       * Получим значения всех свойств в виде строки:
       * <pre>
       *    var article = new Model({
       *       rawData: {id: 1, title: 'Article 1'}
       *    });
       *    article.toString();//'{"id": 1, "title": "Article 1"}'
       * </pre>
      */
      toString: function() {
         var result = {};
         this.each(function(key, value) {
            result[key] = value;
         });
         return JSON.stringify(result);
      },


      // endregion Public methods

      //region Protected methods

      /**
       * Возвращает массив названий всех свойств (включая свойства в "сырых" данных)
       * @return {Array.<String>}
       * @protected
       */
      _getAllProperties: function() {
         var fields = this._getRawDataFields(),
            objProps,
            props;

         if (!this._$properties) {
            return fields;
         }

         objProps = this._$properties;
         props = Object.keys(objProps);
         return props.concat(fields.filter(function(field) {
            return !objProps.hasOwnProperty(field);
         }));
      },

      /**
       * Вычисляет/записывает значение свойства
       * @param {String} name Имя свойства
       * @param {*} value Значение свойства
       * @param {Property} property Описание свойства
       * @param {Boolean} isReading Вычисление или запись
       * @return {*}
       * @protected
       */
      _processCalculatedValue: function (name, value, property, isReading) {
         var checkKey = name + '|' + isReading,
            method = isReading ? property.get : property.set,
            isFunctor = isReading && Compute.isFunctor(method),
            doGathering = isReading && !isFunctor,
            nowCalculatingProperties = this._nowCalculatingProperties,
            prevGathering;

         //Check for recursive calculate
         if (!nowCalculatingProperties) {
            nowCalculatingProperties = this._nowCalculatingProperties = new Set();
         }
         if (nowCalculatingProperties.has(checkKey)) {
            throw new Error('Recursive value ' +  (isReading ? 'reading' : 'writing') + ' detected for property ' + name);
         }

         //Automatic dependency gathering
         if (isReading) {
            prevGathering = this._propertiesDependencyGathering;
            this._propertiesDependencyGathering = doGathering ? name : '';
         }
         //Save user defined dependencies
         if (isFunctor) {
            method.properties.forEach(function(dependFor) {
               this._pushDependencyFor(dependFor, name);
            }, this);
         }

         //Get or set property value
         try {
            nowCalculatingProperties.add(checkKey);
            value = method.call(this, value);
         } finally {
            nowCalculatingProperties.delete(checkKey);
            if (isReading) {
               this._propertiesDependencyGathering = prevGathering;
            }
         }

         return value;
      },

      /**
       * Добавляет зависимое свойство для текущего рассчитываемого
       * @param {String} name Название свойства.
       * @protected
       */
      _pushDependency: function (name) {
         if (this._propertiesDependencyGathering && this._propertiesDependencyGathering !== name) {
            this._pushDependencyFor(name, this._propertiesDependencyGathering);
         }
      },

      /**
       * Добавляет зависимое свойство
       * @param {String} name Название свойства.
       * @param {String} dependFor Название свойства, котороое зависит от name
       * @protected
       */
      _pushDependencyFor: function (name, dependFor) {
         var propertiesDependency = this._propertiesDependency,
            data;

         if (!propertiesDependency) {
            propertiesDependency = this._propertiesDependency = new Map();
         }
         if (propertiesDependency.has(name)) {
            data = propertiesDependency.get(name);
         } else {
            data = new Set();
            propertiesDependency.set(name, data);
         }
         if (!data.has(dependFor)) {
            data.add(dependFor);
         }
      },

      /**
      * Удаляет закешированное значение для свойства и всех от него зависимых свойств
      * @param {String} name Название свойства.
      * @protected
      */
      _deleteDependencyCache: function (name) {
         var propertiesDependency = this._propertiesDependency;

         if (propertiesDependency && propertiesDependency.has(name)) {
            propertiesDependency.get(name).forEach(function(related) {
               this._unsetFromPropertiesCache(related);
               this._deleteDependencyCache(related);
            }, this);
         }
      },

      //endregion Protected methods

      //region WS.Data/Mediator/IReceiver

      relationChanged: function (which, route) {
         //Delete cache for properties related of changed one use in-deep route
         var curr = [],
            routeLastIndex = route.length - 1;
         route.forEach(function(name, index) {
            var fieldName = this._getFieldFromRelationName(name);
            curr.push(fieldName);
            if (fieldName) {
               this._deleteDependencyCache(curr.join(ROUTE_SEPEARTOR));

               if (index === routeLastIndex && which.data instanceof Object) {
                  Object.keys(which.data).forEach(function(key) {
                     this._deleteDependencyCache(curr.concat([key]).join(ROUTE_SEPEARTOR));
                  }, this);
               }
            }
         }.bind(this));

         Model.superclass.relationChanged.call(this, which, route);
      },

      //endregion WS.Data/Mediator/IReceiver

      //region Deprecated methods

      setChanged: function (changed) {
         Utils.logger.stack('WS.Data/Entity/Model: method setChanged() is deprecated and will be removed in 3.7.6.', 0, 'error');
         this._isChanged = !!changed;
         if (!this._isChanged) {
            this._clearChanged();
         }
      },

      isChanged: function (name) {
         if (!name && this._isChanged) {
            return true;
         }
         return Model.superclass.isChanged.call(this, name);
      },

      rejectChanges: function(fields) {
         Model.superclass.rejectChanges.call(this, fields);
         if (!(fields instanceof Array)) {
            this._isChanged = false;
         }
      },

      acceptChanges: function(fields) {
         Model.superclass.acceptChanges.call(this, fields);
         if (!(fields instanceof Array)) {
            this._isChanged = false;
         }
      }

      //endregion Deprecated methods
   }),

   ROUTE_SEPEARTOR = '.';

   Model.fromObject = function(data, adapter) {
      var record = Record.fromObject(data, adapter);
      if (!record) {
         return record;
      }
      return new Model({
         rawData: record.getRawData(true),
         adapter: record.getAdapter(),
         format: record._getFormat(true)//"Anakin, I Am Your Son"
      });
   };

   Di.register('entity.$model', Model, {instantiate: false});
   Di.register('entity.model', Model);
   Di.register('model', Model);//deprecated

   return Model;
});
