/* global define */
define('WS.Data/Entity/PropertiesMixin', [
   'Core/helpers/Object/isEmpty'
], function (isEmpty) {
   'use strict';

   /**
    * Примесь, реализующая интерфейсы {@link WS.Data/Entity/IObject} и {@link WS.Data/Entity/IObjectNotify}, оперируя понятием "Свойство".
    * Набор свойств определяется на протоипе объекта через директиву "properties", сигнатура которой аналогична сигнатуре аргумента props метода {@link https://developer.mozilla.org/ru/docs/Web/JavaScript/Reference/Global_Objects/Object/defineProperties Object.defineProperties}.
    * Исходные значения свойств можно внедрять через конструктор экземпляра, передавая ему первым аргументом объект вида "имя свойства" - "значение свойства".
    * @example
    * Создадим модель пользователя со свойствами:
    * <ul>
    *    <li>id, login, firstName, lastName, job (чтение/запись, значение по умолчанию - скалярное значение);</li>
    *    <li>displayName (только чтение, значение вычисляется из трех других свойств - login, firstName, lastName);</li>
    *    <li>group (чтение/запись, хранится в защищенном свойстве);</li>
    *    <li>guid (только чтение, хранится в защищенном свойстве, значение генерируется лениво при первом обращении);</li>
    * </ul>
    * <pre>
    *    //User.js
    *    define('App/Model/User', [
    *       'WS.Data/Entity/Abstract',
    *       'WS.Data/Entity/PropertiesMixin',
    *       'WS.Data/Entity/ObservableMixin',
    *       'Core/helpers/random-helpers'
    *    ], function(Abstract, PropertiesMixin, ObservableMixin, randomHelpers) {
    *       var User = Abstract.extend([PropertiesMixin, ObservableMixin], {
    *          _group: null,
    *          _guid: '',
    *          properties: {
    *             id: 0,
    *             login: '',
    *             firstName: '',
    *             lastName: '',
    *             job: '',
    *             displayName: {
    *                get: function() {
    *                   return this.firstName + ' a.k.a "' + this.login + '" ' + this.lastName;
    *                }
    *             },
    *             group: {
    *                get: function() {
    *                   return this._group;
    *                },
    *                set: function(value) {
    *                   this._group = value;
    *                }
    *             },
    *             guid: {
    *                get: function() {
    *                   return this._guid || (this._guid = randomHelpers.createGUID());
    *                }
    *             },
    *          },
    *          constructor: function(options) {
    *             User.superclass.constructor.call(this, options);
    *             PropertiesMixin.constructor.call(this, options);
    *             ObservableMixin.constructor.call(this, options);
    *             this._publish('onPropertyChange');
    *          },
    *          destroy: function() {
    *             ObservableMixin.destroy.call(this);
    *             User.superclass.destroy.call(this);
    *          },
    *          _notifyPropertyChange: function (changed) {
    *             this._notify('onPropertyChange', changed);
    *          }
    *       });
    *
    *       return User;
    *    });
    *
    *    //getUser.js
    *    require(['App/Model/User'], function(User) {
    *       'use strict';
    *
    *       var user = new User({
    *          id: 5,
    *          login: 'Keanu',
    *          firstName: 'Johnny',
    *          lastName: 'Mnemonic',
    *          job: 'Memory stick'
    *       });
    *
    *       console.log(user.id);//5
    *       console.log(user.group);//null
    *       console.log(user.guid);//010a151c-1160-d31d-11b3-18189155cc13
    *       console.log(user.displayName);//Johnny a.k.a "Keanu" Mnemonic
    *       console.log(user.job);//Memory stick
    *       console.log(user.uptime);//undefined
    *
    *       user.id = 6;
    *       console.log(user.id);//6
    *
    *       user.group = {id: 1, name: 'The One'};
    *       console.log(user.group);//{id: 1, name: 'The One'}
    *
    *       user.guid = 'new-one';//ReferenceError 'Property "guid" is read only'
    *       user.displayName = 'Thomas a.k.a. "Neo" Anderson';//ReferenceError 'Property "displayName" is read only'
    *    });
    * </pre>
    *
    * @class WS.Data/Entity/OptionsMixin
    * @public
    * @author Мальцев Алексей
    */

   var PropertiesMixin = /** @lends WS.Data/Entity/PropertiesMixin.prototype */{
      '[WS.Data/Entity/PropertiesMixin]': true,

      /**
       * @typedef {Object} Property
       * @property {*|Function} [def] Значение по умолчанию (используется, если свойства нет в сырых данных).
       * @property {Function} [get] Метод, возвращающий значение свойства. Первым аргументом придет значение свойства в сырых данных (если оно там есть).
       * @property {Function} [set] Метод, устанавливающий значение свойства. Если метод вернет значение, отличное от undefined, то будет осуществлена попытка сохранить его в сырых данных.
       */

      /**
       * @cfg {Object.<Property>} Описание собственных свойств модели. Дополняет/уточняет свойства, уже существующие в сырых данных.
       * @name WS.Data/Entity/Model#properties
       * @see Property
       * @see getProperties
      /**
       * @member {Object.<Object>} Объявленные свойства
       */
      properties: null,

      /**
       * @member {Object} Хранилище значений свойств при использовании дескриптора данных
       */
      _propertiesValue: null,

      /**
       * @member {Boolean} Свойства были определены
       */
      _propertiesDefined: false,

      /**
       * Конструктор объекта, принимающий набор значений свойств в качестве первого аргумента
       * @param {Object} properties Значения свойств по умолчанию
       */
      constructor: function PropertiesMixin(properties) {
         this.properties = this.properties || {};
         this._propertiesValue = this._propertiesValue || {};

         var proto = Object.getPrototypeOf(this);
         if (!proto._propertiesDefined) {
            PropertiesMixin.defineProperties(proto, this.properties);
            proto._propertiesDefined = true;
         }

         if (properties) {
            for (var name in properties) {
               if (properties.hasOwnProperty(name) && PropertiesMixin.has.call(this, name)) {
                  this.set(name, properties[name]);
               }
            }
         }
      },

      //region WS.Data/Entity/IObject

      get: function (name, prev) {
         if (!PropertiesMixin.has.call(this, name)) {
            return;
         }
         var property = this.properties[name];

         if (typeof property.get === 'function') {
            return property.get.call(this, prev);
         } else if (property.hasOwnProperty('value')) {
            return this._propertiesValue.hasOwnProperty(name) ?
               this._propertiesValue[name] :
               property.value;
         }
      },

      set: function (name, value) {
         var property,
            changed = {},
            errors = [],
            map,
            prev,
            result;
         if (name instanceof Object) {
            map = name;
         } else {
            map = {};
            map[name] = value;
         }

         for (name in map) {
            if (!map.hasOwnProperty(name)) {
               continue;
            }
            value = map[name];
            if (!PropertiesMixin.has.call(this, name)) {
               errors.push(name);
               continue;
            }
            prev = PropertiesMixin.get.call(this, name, value);
            if (prev === value) {
               continue;
            }

            property = this.properties[name];
            if (typeof property.set === 'function') {
               result = property.set.call(this, value);
            } else if (property.hasOwnProperty('value')) {
               this._propertiesValue[name] = value;
            } else {
               throw new Error('Property "' + name + '" is read only.');
            }

            changed[name] = value;
         }

         if (!isEmpty(changed)) {
            this._notifyPropertyChange(changed);
         }

         if (errors.length) {
            throw new ReferenceError(this._moduleName + ': ' +
               (errors.length == 1 ? 'property' : 'properties') + ' ' + errors.join(', ') + ' ' +
               (errors.length == 1 ? 'is' : 'are') + ' not defined.');
         }

         return result;
      },

      has: function (name) {
         return this.properties.hasOwnProperty(name);
      },

      //endregion WS.Data/Entity/IObject

      //region Public methods

      /**
       * Возвращает описание свойств модели.
       * @return {Object.<Property>}
       * @see properties
       * @see Property
       * @example
       * Получим описание свойств модели пользователя:
       * <pre>
       *    var User = Abstract.extend({
       *          properties: {
       *             id: {
       *                get: function() {
       *                   return this._id;
       *                },
       *                set: function(value) {
       *                   this._id = value;
       *                }
       *             },
       *             group: {
       *                get: function() {
       *                   return this._group;
       *                }
       *             }
       *          },
       *          _id: 0
       *          _group: null
       *       }),
       *       user = new User();
       *
       *    user.getProperties();//{id: {get: Function, set: Function}, group: {get: Function}}
       * </pre>
       */
      getProperties: function () {
         return this.properties;
      },

      //endregion Public methods

      //region Protected methods

      _notifyPropertyChange: function (changed) {
         throw new Error('Method must be implemented.');
      }

      //endregion Protected methods
   };

   /**
    * Объявляет набор свойств через defineProperties на объекте, проксируя их вызовы в методы get и set
    * @param {Object} object Объект
    * @param {Object} properties Свойства
    * @static
    */
   PropertiesMixin.defineProperties = function(object, properties) {
      var getter = function(name) {
            return function getProperty() {
               return this.get(name);
            };
         },
         setter = function(name) {
            return function setProperty(value) {
               return this.set(name, value);
            };
         },
         mine = {},
         propType,
         prop,
         name;

      for (name in properties) {
         if (!properties.hasOwnProperty(name)) {
            continue;
         }
         prop = properties[name];
         if (!(prop instanceof Object)) {
            prop = properties[name] = {
               value: prop,
               writable: true
            };
         }

         propType = '';
         if (typeof prop.get === 'function' || typeof prop.set === 'function') {
            propType = 'accessor';
         } else if (prop.hasOwnProperty('value')) {
            propType = 'data';
         }
         if (!propType) {
            throw new TypeError('Wrong property "' + name + '" declaration - data descriptor or accessor descriptor expected. See Object.defineProperty for details.');
         }


         mine[name] = {
            configurable: prop.configurable || false,
            enumerable: prop.enumerable || false,
            get: getter(name)
         };
         if (prop.writable || prop.set) {
            mine[name].set = setter(name);
         }
      }

      Object.defineProperties(object, mine);
   };

   return PropertiesMixin;
});
