/*global define*/
define('WS.Data/Factory', [
   'WS.Data/Di',
   'Core/core-instance',
   'Core/TimeInterval',
   'Core/defaultRenders'
], function (
   Di,
   coreInstance,
   CoreTimeInterval,
   renders
) {
   'use strict';

   /**
    * Фабрика типов - перобразует исходные значения в типизированные и наоборот.
    * @class WS.Data/Factory
    * @public
    * @author Мальцев Алексей
    */

   /**
    * @faq Почему я вижу ошибки от WS.Data/Di::resolve?
    * Для корректной работы с зависимости сначала надо загрузить {@link WS.Data/Entity/Model} и {@link WS.Data/Source/RecordSet}, а уже потом {@link WS.Data/Factory}
    */

   var Factory = {
      '[WS.Data/Factory]': true,

      /**
       * @typedef {String} SimpleType
       * @variant integer Целое число
       * @variant real Действительное число
       * @variant double Действительное число с размером больше real
       * @variant boolean Логический
       * @variant money Деньги
       * @variant link Ссылка
       * @variant date Дата
       * @variant time Время
       * @variant datetime Дата-время
       * @variant timeinterval Временной интервал
       * @variant array Массив со значениями определенного типа
       */

      /**
       * Возвращает типизированное значение из исходного.
       * @param {*} value Исходное значение
       * @param {Function|SimpleType} Type Тип значения
       * @param {Object} [options] Опции
       * @return {*} Типизированное значение
       * @static
       */
      cast: function (value, Type, options) {
         options = options || {};
         if (this._isNullable(Type, value, options)) {
            return value;
         }

         if (typeof Type === 'string') {
            Type = Type.toLowerCase();
            switch (Type) {
               case 'recordset':
                  Type = Di.resolve('collection.$recordset');
                  break;
               case 'record':
                  Type = Di.resolve('entity.$model');
                  break;
               case 'enum':
                  Type = Di.resolve('types.$enum');
                  break;
               case 'flags':
                  Type = Di.resolve('types.$flags');
                  break;
               case 'link':
               case 'integer':
                  return (typeof(value) === 'number') ? value : (isNaN(parseInt(value, 10)) ? null : parseInt(value, 10));
               case 'real':
               case 'double':
                  return (typeof(value) === 'number') ? value : (isNaN(parseFloat(value)) ? null : parseFloat(value));
               case 'boolean':
                  return !!value;
               case 'money':
                  var precision = this._getPrecision(options.format);
                  if (precision > 3) {
                     return renders.real(value, precision, false, true);
                  }
                  return value === undefined ? null : value;
               case 'date':
               case 'time':
               case 'datetime':
                  if (value instanceof Date) {
                     return value;
                  } else if (value === 'infinity') {
                     return Infinity;
                  } else if (value === '-infinity') {
                     return -Infinity;
                  }
                  value = Date.fromSQL('' + value, false);
                  switch (Type) {
                     case 'date':
                        value.setSQLSerializationMode(Date.SQL_SERIALIZE_MODE_DATE);
                        break;
                     case 'time':
                        value.setSQLSerializationMode(Date.SQL_SERIALIZE_MODE_TIME);
                        break;
                     case 'datetime':
                        value.setSQLSerializationMode(Date.SQL_SERIALIZE_MODE_DATETIME);
                        break;
                  }
                  return value;
               case 'timeinterval':
                  if (value instanceof CoreTimeInterval) {
                     return value.toString();
                  }
                  return CoreTimeInterval.toString(value);
               case 'array':
                  var kind = this._getKind(options.format);
                  if (!(value instanceof Array)) {
                     value = [value];
                  }
                  return value.map(function (val) {
                     return this.cast(val, kind, options);
                  }, this);
               default:
                  return value;
            }
         }

         if (typeof Type === 'function') {
            if (value instanceof Type) {
               return value;
            }
            if (Type.prototype['[WS.Data/Entity/IProducible]']) {
               //implements WS.Data/Entity/IProducible
               return Type.prototype.produceInstance.call(
                  Type,
                  value,
                  options
               );
            }
            return new Type(value);
         }

         throw new TypeError('Unknown type ' + Type);
      },

      /**
       * Возвращет исходное значение из типизированного.
       * @param {*} value Типизированное значение
       * @param {Object} [options] Опции
       * @return {*} Исходное значение
       * @static
       */
      serialize: function (value, options) {
         options = options || {};
         var type = this._getTypeName(options.format);

         if (this._isNullable(type, value, options)) {
            return value;
         }

         if (value && typeof value === 'object') {
            if (value['[WS.Data/Entity/FormattableMixin]']) {//mixes WS.Data/Entity/FormattableMixin
               value = value.getRawData(true);
            }  else if (value['[WS.Data/Type/IFlags]']) {
               value = this._serializeFlags(value);
            } else if (value['[WS.Data/Type/IEnum]']) {
               value = value.get();
            } else if (value['[WS.Data/Collection/IList]'] && type === 'recordset') {
               value = this._convertListToRecordSet(value);
            } else if (coreInstance.instanceOfModule(value, 'Deprecated/Record')) {
               throw new TypeError('Deprecated/Record can\'t be used with WS.Data');
            } else if (coreInstance.instanceOfModule(value, 'Deprecated/RecordSet')) {
               throw new TypeError('Deprecated/RecordSet can\'t be used with WS.Data');
            } else if (coreInstance.instanceOfModule(value, 'Deprecated/Enum')) {
               throw new TypeError('Deprecated/Enum can\'t be used with WS.Data');
            }
         }

         switch (type) {
            case 'integer':
               value = this._toScalar(value);
               return typeof(value) === 'number' ?
                  value :
                  (isNaN(value = value - 0) ? null : parseInt(value, 10));
            case 'real':
            case 'double':
               return this._toScalar(value);
            case 'link':
               return parseInt(value, 10);
            case 'money':
               value = this._toScalar(value);
               var precision = this._getPrecision(options.format);
               if (precision > 3) {
                  return renders.real(value, precision, false, true);
               }
               return value;
            case 'date':
            case 'time':
            case 'datetime':
               value = this._toScalar(value);
               var serializeMode;
               switch (type) {
                  case 'datetime':
                     serializeMode = Date.SQL_SERIALIZE_MODE_DATETIME;
                     break;
                  case 'time':
                     serializeMode = Date.SQL_SERIALIZE_MODE_TIME;
                     break;
               }
               if (!value) {
                  value = Date.fromSQL('' + value);
               }
               if (value instanceof Date) {
                  return value.toSQL(serializeMode);
               } else if (value === Infinity) {
                  return 'infinity';
               } else if (value === -Infinity) {
                  return '-infinity';
               }
               return value;
            case 'timeinterval':
               value = this._toScalar(value);
               if (value instanceof CoreTimeInterval) {
                  return value.toString();
               }
               return CoreTimeInterval.toString(value);
            case 'array':
               var kind = this._getKind(options.format);
               if (!(value instanceof Array)) {
                  value = [value];
               }
               return value.map(function (val) {
                  return this.serialize(val, {format: kind});
               }, this);
            default:
               return value;
         }
      },

      _toScalar: function (value) {
         if (Array.isArray(value) && value.length < 2) {
            return value.length ? value[0] : null;
         }
         return value;
      },

      /**
       * Возвращает название типа поля
       * @param {WS.Data/Format/Field|WS.Data/Format/UniversalField|String} format Формат поля
       * @return {String}
       * @protected
       * @static
       */
      _getTypeName: function (format) {
         var type;
         if (typeof format === 'object') {
            type = format.getType ? format.getType() : format.type;
         } else {
            type = format;
         }
         return ('' + type).toLowerCase();
      },

      /**
       * Возвращает точность для поля типа "Вещественное число"
       * @param {WS.Data/Format/Field|WS.Data/Format/UniversalField|String} format Формат поля
       * @return {Number}
       * @protected
       * @static
       */
      _getPrecision: function (format) {
         if (!format) {
            return 0;
         }
         return (format.getPrecision ? format.getPrecision() : format.meta && format.meta.precision) || 0;
      },

      /**
       * Возвращает словарь для поля типа "Словарь"
       * @param {WS.Data/Format/Field|WS.Data/Format/UniversalField|String} format Формат поля
       * @return {Array}
       * @protected
       */
      _getDictionary: function (format) {
         return (format.getDictionary ? format.getDictionary() : format.meta && format.meta.dictionary) || [];
      },

      /**
       * Возвращает тип элементов для поля типа "Массив"
       * @param {WS.Data/Format/Field|WS.Data/Format/UniversalField|String} format Формат поля
       * @return {String}
       * @protected
       * @static
       */
      _getKind: function (format) {
         return (format.getKind ? format.getKind() : format.meta && format.meta.kind) || '';
      },

      /**
       * Сериализует поле флагов
       * @param {WS.Data/Type/Flags} data
       * @return {Array.<Boolean>}
       * @protected
       * @static
       */
      _serializeFlags: function (data) {
         var d = [];
         data.each(function (name) {
            d.push(data.get(name));
         });
         return d;
      },

      /**
       * Возвращает признак, что значение типа может быть null
       * @param {String} type Тип значения.
       * @param {*} value Значение.
       * @param {Object} [options] Опции.
       * @return {Boolean}
       * @protected
       * @static
       */
      _isNullable: function (type, value, options) {
         if (value === undefined || value === null) {
            switch (type) {
               case 'identity':
                  return false;
               case 'enum':
                  return this._isEnumNullable(value, options);
            }

            if (typeof type === 'function') {
               var inst = Object.create(type.prototype);
               if (inst && inst['[WS.Data/Type/IEnum]']) {
                  return this._isEnumNullable(value, options);
               }
            }

            return true;
         }
         return false;
      },

      /**
       * Возвращает признак, что значение типа Enum может быть null
       * @param {*} value Значение.
       * @param {Object} options Опции.
       * @return {Boolean}
       * @protected
       * @static
       */
      _isEnumNullable: function (value, options) {
         var dict = this._getDictionary(options.format);
         if (value === null && !dict.hasOwnProperty(value)) {
            return true;
         }
         if (value === undefined) {
            return true;
         }
         return false;
      },
      /**
       * Конвертирует список записей в рекордсет
       * @param {WS.Data/Collection/List} list Список
       * @return {WS.Data/Collection/RecordSet}
       * @protected
       */
      _convertListToRecordSet: function(list) {
         var adapter = 'adapter.json',
             record,
             count = list.getCount(),
             i;

         for (i = 0; i < count; i++) {
            record = list.at(i);
            //Check for WS.Data/Entity/Record
            if (record && record['[WS.Data/Entity/IObject]'] && record['[WS.Data/Entity/FormattableMixin]']) {
               adapter = record.getAdapter();
               break;
            }
         }

         var rs = Di.resolve('collection.recordset', {
            adapter: adapter
         });
         for (i = 0; i < count; i++) {
            rs.add(list.at(i));
         }

         return rs.getRawData(true);
      }
   };

   Di.register('factory', Factory, {instantiate: false});

   return Factory;
});
