/* global define */
define('WS.Data/Mediator/ManyToMany', [
   'WS.Data/Mediator/IMediator',
   'WS.Data/Entity/Abstract',
   'WS.Data/Shim/Set',
   'WS.Data/Shim/Map',
   'WS.Data/Di'
], function (
   IMediator,
   Abstract,
   Set,
   Map,
   Di
) {
   'use strict';

   /**
    * Посредник, реализующий отношения "многие ко многим".
    * @class WS.Data/Mediator/ManyToMany
    * @extends WS.Data/Entity/Abstract
    * @implements WS.Data/Mediator/IMediator
    * @author Мальцев Алексей
    */

   var ManyToMany = Abstract.extend([IMediator], /** @lends WS.Data/Mediator/ManyToMany.prototype */{
      '[WS.Data/Mediator/ManyToMany]': true,

      /**
       * @member {Map<Object, Set<Object>>} master -> [slave, slave, ...]
       */
      _hasMany: null,

      /**
       * @member {Map<Object, Map<Object, String>>} master -> [name, name, ...]
       */
      _hasManyName: null,

      /**
       * @member {Map<Object, Set<Object>>} slave -> [master, master, ...]
       */
      _belongsTo: null,

      /**
       * @member {Map<Object, Map<Object, String>>} slave -> [name, name, ...]
       */
      _belongsToName: null,

      constructor: function ManyToMany() {
         ManyToMany.superclass.constructor.call(this);
         this._hasMany = new Map();
         this._hasManyName = new Map();
         this._belongsTo = new Map();
         this._belongsToName = new Map();
      },

      destroy: function() {
         this._hasMany = null;
         this._hasManyName = null;
         this._belongsTo = null;
         this._belongsToName = null;
         ManyToMany.superclass.destroy.call(this);
      },

      //region WS.Data/Mediator/IMediator

      //endregion WS.Data/Mediator/IMediator

      //region Public methods

      /**
       * Добавляет отношение между двумя сущностями
       * @param {Object} master Главная сущность
       * @param {Object} slave Зависимая сущность
       * @param {String} [name] Название отношения
       */
      addRelationship : function (master, slave, name) {
         this._addHasMany(master, slave, name);
         this._addBelongsTo(slave, master, name);
      },

      /**
       * Удаляет отношение между двумя сущностями
       * @param {Object} master Главная сущность
       * @param {Object} slave Зависимая сущность
       */
      removeRelationship: function (master, slave) {
         this._removeHasMany(master, slave);
         this._removeBelongsTo(slave, master);
      },

      /**
       * Очищает все отношения указанной сущности
       * @param {Object} entity Сущность
       */
      clear: function (entity) {
         if (this._hasMany.has(entity)) {
            this._hasMany.get(entity).forEach(function(slave) {
               this._removeBelongsTo(slave, entity);
            }.bind(this));
            this._hasMany.delete(entity);
            this._hasManyName.delete(entity);
         }

         if (this._belongsTo.has(entity)) {
            this._belongsTo.get(entity).forEach(function(master) {
               this._removeHasMany(master, entity);
            }.bind(this));
            this._belongsTo.delete(entity);
            this._belongsToName.delete(entity);
         }
      },

      /**
       * Возвращает все зависимые сущности
       * @param {Object} master Главная сущность
       * @param {Function(Object, String)} callback Функция обратного вызова для каждой зависимой сущности
       */
      hasMany: function (master, callback) {
         if (this._hasMany.has(master)) {
            var names = this._hasManyName.get(master);
            this._hasMany.get(master).forEach(function(slave) {
               if (this._isAlive(slave)) {
                  callback.call(
                     this,
                     slave,
                     names.get(slave)
                  );
               }
            }.bind(this));
         }
      },

      /**
       * Возвращает все главные сущности
       * @param {Object} slave Зависимая сущность
       * @param {Function(Object, String)} callback Функция обратного вызова для каждой главной сущности
       */
      belongsTo: function (slave, callback) {
         if (this._belongsTo.has(slave)) {
            var names = this._belongsToName.get(slave);
            this._belongsTo.get(slave).forEach(function(master) {
               if (this._isAlive(master)) {
                  callback.call(
                     this,
                     master,
                     names.get(master)
                  );
               }
            }.bind(this));
         }
      },

      //endregion Public methods

      //region Protected methods

      /**
       * Добавляет отношение вида hasMany
       * @param {Object} master Главная сущность
       * @param {Object} slave Зависимая сущность
       * @param {String} name Название отношения
       * @protected
       */
      _addHasMany: function(master, slave, name) {
         var slaves,
            names;
         if (this._hasMany.has(master)) {
            slaves = this._hasMany.get(master);
            names = this._hasManyName.get(master);
         } else {
            slaves = new Set();
            names = new Map();
            this._hasMany.set(master, slaves);
            this._hasManyName.set(master, names);
         }
         slaves.add(slave);
         names.set(slave, name);
      },

      /**
       * Удаляет отношение вида hasMany
       * @param {Object} master Главная сущность
       * @param {Object} slave Зависимая сущность
       * @protected
       */
      _removeHasMany: function(master, slave) {
         if (this._hasMany.has(master)) {
            var slaves = this._hasMany.get(master);
            slaves.delete(slave);
            this._hasManyName.get(master).delete(slave);

            if (slaves.size === 0) {
               this._hasMany.delete(master);
               this._hasManyName.delete(master);
            }
         }
      },

      /**
       * Добавляет отношение вида belongsTo
       * @param {Object} master Главная сущность
       * @param {Object} slave Зависимая сущность
       * @param {String} name Название отношения
       * @protected
       */
      _addBelongsTo: function(slave, master, name) {
         var masters,
            names;
         if (this._belongsTo.has(slave)) {
            masters = this._belongsTo.get(slave);
            names = this._belongsToName.get(slave);
         } else {
            masters = new Set();
            names = new Map();
            this._belongsTo.set(slave, masters);
            this._belongsToName.set(slave, names);
         }
         masters.add(master);
         names.set(master, name);
      },

      /**
       * Удаляет отношение вида belongsTo
       * @param {Object} master Главная сущность
       * @param {Object} slave Зависимая сущность
       * @protected
       */
      _removeBelongsTo: function(slave, master) {
         if (this._belongsTo.has(slave)) {
            var masters = this._belongsTo.get(slave);
            masters.delete(master);
            this._belongsToName.get(slave).delete(master);

            if (masters.size === 0) {
               this._belongsTo.delete(slave);
               this._belongsToName.delete(slave);
            }
         }
      },

      /**
       * Проверяет, что объект "живой" (не был уничтожен)
       * @param {Object} entity Объект
       * @return {Boolean}
       * @protected
       */
      _isAlive: function(entity) {
         //It's equal to coreInstance.instanceOfModule(entity, 'WS.Data/Entity/Abstract') but faster
         return entity instanceof Object && entity.isDestroyed ? !entity.isDestroyed() : true;
      }

      //endregion Protected methods
   });

   Di.register('mediator.many-to-many', ManyToMany);

   return ManyToMany;
});
