/* global define */
define('WS.Data/Mediator/OneToMany', [
   'WS.Data/Mediator/IMediator',
   'WS.Data/Entity/Abstract',
   'WS.Data/Shim/Set',
   'WS.Data/Shim/Map',
   'WS.Data/Di'
], function (
   IMediator,
   Abstract,
   Set,
   Map,
   Di
) {
   'use strict';

   /**
    * Посредник, реализующий отношения "один ко многим".
    * @class WS.Data/Mediator/OneToMany
    * @extends WS.Data/Entity/Abstract
    * @implements WS.Data/Mediator/IMediator
    * @author Мальцев Алексей
    */

   var OneToMany = Abstract.extend([IMediator], /** @lends WS.Data/Mediator/OneToMany.prototype */{
      '[WS.Data/Mediator/OneToMany]': true,

      /**
       * @member {Map<Object, Set<Object>>} Родитель -> [Ребенок, Ребенок, ...]
       */
      _parentToChild: null,

      /**
       * @member {Map<Object, Object>} Ребенок -> Родитель
       */
      _childToParent: null,

      /**
       * @member {Map<Object, String>} Ребенок -> название отношения
       */
      _childToRelation: null,

      constructor: function OneToMany() {
         OneToMany.superclass.constructor.call(this);
         this._parentToChild = new Map();
         this._childToParent = new Map();
         this._childToRelation = new Map();
      },

      destroy: function() {
         this._parentToChild = null;
         this._childToParent = null;
         this._childToRelation = null;
         OneToMany.superclass.destroy.call(this);
      },

      //region WS.Data/Mediator/IMediator

      //endregion WS.Data/Mediator/IMediator

      //region Public methods

      /**
       * Добавляет отношение "родитель - ребенок"
       * @param {Object} parent Родитель
       * @param {Object} child Ребенок
       * @param {String} [name] Название отношений
       */
      addTo: function (parent, child, name) {
         this._addForParent(parent, child);
         this._addForChild(child, parent, name);
      },

      /**
       * Удаляет отношение "родитель - ребенок"
       * @param {Object} parent Родитель
       * @param {Object} child Ребенок
       */
      removeFrom: function (parent, child) {
         this._removeForParent(parent, child);
         this._removeForChild(child, parent);
      },

      /**
       * Очищает все отношения c детьми у указанного родителя
       * @param {Object} parent Родитель
       */
      clear: function (parent) {
         if (this._parentToChild.has(parent)) {
            this._parentToChild.get(parent).forEach(function(child) {
               this._removeForChild(child, parent);
            }.bind(this));
            this._parentToChild.delete(parent);
         }
      },

      /**
       * Возвращает всех детей для указанного родителя
       * @param {Object} parent Родитель
       * @param {Function(Object, String)} callback Функция обратного вызова для каждого ребенка
       */
      each: function (parent, callback) {
         if (this._parentToChild.has(parent)) {
            this._parentToChild.get(parent).forEach(function(child) {
               if (this._isAlive(child)) {
                  callback.call(
                     this,
                     child,
                     this._childToParent.get(child) === parent ? this._childToRelation.get(child) : undefined
                  );
               }
            }.bind(this));
         }
      },

      /**
       * Возвращает родителя для указанного ребенка
       * @param {Object} child Ребенок
       * @return {Object}
       */
      getParent: function (child) {
         var parent = this._childToParent.get(child);
         return parent !== undefined && this._isAlive(parent) ? parent : undefined;
      },

      //endregion Public methods

      //region Protected methods

      /**
       * Добавляет ребенка в список родителя
       * @param {Object} parent Родитель
       * @param {Object} child Ребенок
       * @protected
       */
      _addForParent: function(parent, child) {
         var children;
         if (this._parentToChild.has(parent)) {
            children = this._parentToChild.get(parent);
         } else {
            children = new Set();
            this._parentToChild.set(parent, children);
         }
         children.add(child);
      },

      /**
       * Удаляет ребенка из списка родителя
       * @param {Object} parent Родитель
       * @param {Object} child Ребенок
       * @protected
       */
      _removeForParent: function(parent, child) {
         if (this._parentToChild.has(parent)) {
            var children = this._parentToChild.get(parent);
            children.delete(child);
            if (children.size === 0) {
               this._parentToChild.delete(parent);
            }
         }
      },

      /**
       * Добавляет связь ребенка с родителем
       * @param {Object} child Ребенок
       * @param {Object} parent Родитель
       * @param {String} name Название отношения
       * @protected
       */
      _addForChild: function(child, parent, name) {
         this._childToParent.set(child, parent);
         this._childToRelation.set(child, name);
      },

      /**
       * Удаляет связь ребенка с родителем
       * @param {Object} child Ребенок
       * @param {Object} parent Родитель
       * @protected
       */
      _removeForChild: function(child, parent) {
         if (this._childToParent.get(child) === parent) {
            this._childToParent.delete(child);
            this._childToRelation.delete(child);
         }
      },

      /**
       * Проверяет, что объект "живой" (не был уничтожен)
       * @param {Object} item Объект
       * @return {Boolean}
       * @protected
       */
      _isAlive: function(item) {
         //It's equal to coreInstance.instanceOfModule(item, 'WS.Data/Entity/Abstract') but faster
         return item instanceof Object && item.isDestroyed ? !item.isDestroyed() : true;
      }

      //endregion Protected methods
   });

   Di.register('mediator.one-to-many', OneToMany);

   return OneToMany;
});
