/* global define */
define('WS.Data/Shim/Set', function () {
   'use strict';


   /**
    * Limited emulation of standard built-in object "Set" if it's not supported.
    * Follow {@link https://developer.mozilla.org/en/docs/Web/JavaScript/Reference/Global_Objects/Set} for details.
    * @class WS.Data/Shim/Set
    * @author Мальцев Алексей
    */

   //Return native implementation if supported
   if (typeof Set !== 'undefined') {
      return Set;
   }
   return (function() {
      var Set = function() {
         this.clear();
      };

      Set.prototype._hash = null;
      Set.prototype._objectPrefix = '{[object]}:';
      Set.prototype._objects = null;

      Object.defineProperty(Set.prototype, 'size', {
         get: function() {
            return Object.keys(this._hash).length;
         },
         enumerable: true,
         configurable: false
      });

      Set.prototype.add = function(value) {
         var key = value;
         if (this._isObject(value)) {
            key = this._addObject(value);
         }
         this._hash[key] = value;
         return this;
      };

      Set.prototype.clear = function() {
         this._hash = {};
         this._objects = [];
      };

      Set.prototype.delete = function(value) {
         var key = value;
         if (this._isObject(value)) {
            key = this._deleteObject(value);
            if (!key) {
               return;
            }
         }
         this._hash[key] = undefined;
      };

      Set.prototype.entries = function() {
         throw new Error('Method is not supported');
      };

      Set.prototype.forEach = function(callbackFn, thisArg) {
         //FIXME: now not in insertion order
         var hash = this._hash;
         for (var key in hash) {
            if (hash.hasOwnProperty(key) && hash[key] !== undefined) {
               callbackFn.call(thisArg, hash[key], hash[key], this);
            }
         }
      };

      Set.prototype.has = function(value) {
         var key = value;
         if (this._isObject(value)) {
            key = this._getObjectKey(value);
            if (!key) {
               return false;
            }
         }
         return this._hash.hasOwnProperty(key) && this._hash[key] !== undefined;
      };

      Set.prototype.keys = function() {
         throw new Error('Method is not supported');
      };

      Set.prototype.values = function() {
         throw new Error('Method is not supported');
      };

      Set.prototype._isObject = function(value) {
         return value && typeof value === 'object';
      };

      Set.prototype._addObject = function(value) {
         var index = this._objects.indexOf(value);
         if (index === -1) {
            index = this._objects.length;
            this._objects.push(value);
         }
         return this._objectPrefix + index;
      };

      Set.prototype._deleteObject = function(value) {
         var index = this._objects.indexOf(value);
         if (index > -1) {
            this._objects[index] = null;
            return this._objectPrefix + index;
         }
         return undefined;
      };

      Set.prototype._getObjectKey = function(value) {
         var index = this._objects.indexOf(value);
         if (index === -1) {
            return undefined;
         }
         return this._objectPrefix + index;
      };

      return Set;
   })();
});
