define('WS.Data/Source/Provider/SbisFile', [
   'WS.Data/Source/Provider/SbisBusinessLogic',
   'Transport/RPCJSON',
   'Transport/XHRTransport',
   'WS.Data/Di',
   'Core/core-extend',
   'Core/constants'
], function (
   SbisBusinessLogic,
   RpcJson,
   XHRTransport,
   Di,
   coreExtend,
   constants
) {
    'use strict';

   /**
    * @param {String} requestField
    * @param {String} meta
    * @param {Object<Lib/File/LocalFile>} files
    * @return {FormData}
    */
    function createFormData (requestField, meta, files) {
        var formData = new FormData();
        formData.append(requestField, meta);
        for (var name in files) {
            formData.append(name, files[name].getData(), files[name].getName());
        }
        return formData;
    }

   /**
    * @typedef {Object} uploadParam
    * @property {String} method Имя метода
    * @property {String} requestField Имя поля, в котором будет лежать тело запроса
    */
   /**
    * @class
    * @public
    * @name WS.Data/Source/Provider/SbisFile
    * @extends WS.Data/Source/Provider/SbisBusinessLogic
    */
   var SbisFile = coreExtend.extend(SbisBusinessLogic, {
      '[WS.Data/Source/Provider/SbisFile]': true,

        /**
         * @cfg {String}
         */
        _$uploadTransport: XHRTransport,
        /**
         * Метод отправки запроса, содержащего файлы.
         * 1) для отправки файлов, они должны быть обёрнуты в FormData
         * 2) JsonRPC - не умеет адекватно отправлять FormData.
         *    Не все браузеры умеют FormData.get() - поэтому просто разрулить приём FormData внутри JsonRPC не выйдет
         * Поэтому пока обходимся решением этих проблем тут. В дальнейшем надо смотреть в сторону  расширение api транспорта
         *
         * @param {Object} cfg
         * @param {uploadParam} cfg.uploadParam
         * @param {Object<Lib/File/LocalFile>} cfg.files
         * @param {*} cfg.meta
         * @return {Core/Deferred}
         * @private
         */
        upload: function (cfg) {
            var endpoint = this.getEndpoint();
            var method = cfg.uploadParam.method;
            if ((method.indexOf('.') === -1) && endpoint.contract) {
                method = endpoint.contract + this._nameSpaceSeparator + method;
            }
            var transport = this._getUploadTransport(endpoint.address);
            var _args = this._prepareRPCArgs(
               method,
               cfg.meta,
               cfg.files,
               cfg.uploadParam.requestField
            );
            var def = transport.execute(_args.body, _args.headers);
            // TODO пока что здесь, выкинуть в Transport
            def.addCallbacks(function (result) {
               var json = result || RpcJson.getEmptyRpcResponse();
                return ('error' in json)?
                   RpcJson.handleRPCError(json.error, method, endpoint.address):
                   json.result;
            }, function (error) {
               return RpcJson.handleHTTPError(error, method)
            });
            return def;
        },
        _prepareRPCArgs: function (method, meta, files, requestField) {
            var rpcArgs = RpcJson.jsonRpcPreparePacket(method, meta);
            new RpcJson()._setHeaders(rpcArgs, 'post');
            return {
                body: createFormData(requestField, rpcArgs.reqBody, files),
                headers: rpcArgs.reqHeaders
            };
        },
        _getUploadTransport: function (url) {
           if (typeof this._$uploadTransport === "function") {
              this._$uploadTransport = new this._$uploadTransport({
                 contentType: false,
                 processData: false,
                 dataType: 'json',
                 method: 'POST',
                 url: url || constants.defaultServiceUrl
              })
           }
           return this._$uploadTransport;
        }
    });

    Di.register('source.provider.sbis-file', SbisFile);

    return SbisFile;
});
