/* global define */
define('WS.Data/Type/Dictionary', [
   'WS.Data/Collection/IEnumerable',
   'WS.Data/Collection/ArrayEnumerator',
   'WS.Data/Collection/ObjectEnumerator',
   'WS.Data/Entity/IEquatable',
   'WS.Data/Entity/Abstract',
   'WS.Data/Entity/OptionsMixin',
   'WS.Data/Entity/ObservableMixin'
], function (
   IEnumerable,
   ArrayEnumerator,
   ObjectEnumerator,
   IEquatable,
   Abstract,
   OptionsMixin,
   ObservableMixin
) {
   'use strict';

   /**
    * Тип данных словарь.
    * Это абстрактный класс, не предусмотрено создание самостоятельных экземпляров.
    * @class WS.Data/Type/Dictionary
    * @implements WS.Data/Collection/IEnumerable
    * @implements WS.Data/Entity/IEquatable
    * @mixes WS.Data/Entity/OptionsMixin
    * @mixes WS.Data/Entity/ObservableMixin
    * @public
    * @author Мальцев Алексей
    */

   var Dictionary = Abstract.extend([IEnumerable, IEquatable, OptionsMixin, ObservableMixin],/** @lends WS.Data/Type/Dictionary.prototype */ {
      '[WS.Data/Type/Dictionary]': true,

      /**
      * @cfg {Array.<String>|Object.<String>} Словарь возможных значений
      * @name WS.Data/Type/Dictionary#dictionary
      */
      _$dictionary: undefined,

      /**
       * @cfg {Array.<String>|Object.<String>} Локализованный словарь возможных значений
       * @name WS.Data/Type/Dictionary#localeDictionary
       */
      _$localeDictionary: undefined,

      /**
       * @member {String} Название типа данных для сериализации в сырой вид
        */
      _type: undefined,

      constructor: function Dictionary(options) {
         Dictionary.superclass.constructor.call(this, options);
         OptionsMixin.constructor.call(this, options);
         ObservableMixin.constructor.call(this, options);
         this._$dictionary = this._$dictionary || [];
      },

      destroy: function() {
         ObservableMixin.destroy.call(this);
         Dictionary.superclass.destroy.call(this);
      },

      //region WS.Data/Collection/IEnumerable

      each: function (callback, context, localize) {
         context = context || this;
         var enumerator = this.getEnumerator(localize);
         while (enumerator.moveNext()) {
            callback.call(
               context,
               enumerator.getCurrent(),
               enumerator.getCurrentIndex()
            );
         }
      },

      getEnumerator: function (localize) {
         var dictionary = localize && this._$localeDictionary ? this._$localeDictionary : this._$dictionary,
            enumerator = dictionary instanceof Array ? new ArrayEnumerator(dictionary) : new ObjectEnumerator(dictionary);

         enumerator.setFilter(function(item, index) {
            return index !== 'null';
         });
         return enumerator;
      },

      //endregion WS.Data/Collection/IEnumerable

      //region WS.Data/Entity/IEquatable

      isEqual: function(to) {
         if (!(to instanceof Dictionary)) {
            return false;
         }

         var enumerator = this.getEnumerator(),
            toEnumerator = to.getEnumerator(),
            item,
            hasItem,
            toItem,
            hasToItem;

         do {
            hasItem = enumerator.moveNext();
            hasToItem = toEnumerator.moveNext();
            item = hasItem ? enumerator.getCurrent() : undefined;
            toItem = hasToItem ? toEnumerator.getCurrent() : undefined;
            if (item !== toItem) {
               return false;
            }
            if (enumerator.getCurrentIndex() !== toEnumerator.getCurrentIndex()) {
               return false;
            }
         } while(hasItem || hasToItem);

         return true;
      },

      //endregion WS.Data/Entity/IEquatable

      //region Protected methods

      /**
       * Возвращает индекс значения в словаре
       * @param {String} name Значение в словаре
       * @param {Boolean} [localize=false] Это локализованное значение
       * @return {Number|String|undefined}
       * @protected
       */
      _getIndex: function (name, localize) {
         var enumerator = this.getEnumerator(localize),
            index = 0;
         while (enumerator.moveNext()) {
            if (enumerator.getCurrent() === name) {
               return index;
            }
            index++;
         }
         return undefined;
      },

      /**
       * Возвращает значение в словаре по индексу
       * @param {Number|String} index Индекс в словаре
       * @param {Boolean} [localize=false] Вернуть локализованное значение
       * @return {String}
       * @protected
       */
      _getValue: function (index, localize) {
         return localize && this._$localeDictionary ? this._$localeDictionary[index] : this._$dictionary[index];
      },

      /**
       * Возвращает словарь из формата
       * @param {WS.Data/Format/Field|WS.Data/Format/UniversalField|String} format Формат поля
       * @return {Array}
       * @protected
       */
      _getDictionaryByFormat: function (format) {
         if (!format) {
            return [];
         }
         return (
            format.getDictionary ? format.getDictionary() : format.meta && format.meta.dictionary
         ) || [];
      },

      /**
       * Возвращает локализованный словарь из формата
       * @param {WS.Data/Format/Field|WS.Data/Format/UniversalField|String} format Формат поля
       * @return {Array|undefined}
       * @protected
       */
      _getLocaleDictionaryByFormat: function (format) {
         if (!format) {
            return;
         }
         return (
            format.getLocaleDictionary ? format.getLocaleDictionary() : format.meta && format.meta.localeDictionary
         ) || undefined;
      }

      //endregion Protected methods
   });

   return Dictionary;
});
