/* global define */
define('WS.Data/Type/Enum', [
   'WS.Data/Type/IEnum',
   'WS.Data/Type/Dictionary',
   'WS.Data/Entity/IProducible',
   'WS.Data/Entity/ManyToManyMixin',
   'WS.Data/Entity/SerializableMixin',
   'WS.Data/Entity/CloneableMixin',
   'WS.Data/Di'
], function (
   IEnum,
   Dictionary,
   IProducible,
   ManyToManyMixin,
   SerializableMixin,
   CloneableMixin,
   Di
) {
   'use strict';

   /**
    * Тип данных перечисляемое.
    * @class WS.Data/Type/Enum
    * @extends WS.Data/Type/Dictionary
    * @implements WS.Data/Type/IEnum
    * @implements WS.Data/Entity/IProducible
    * @mixes WS.Data/Entity/ManyToManyMixin
    * @mixes WS.Data/Entity/SerializableMixin
    * @mixes WS.Data/Entity/CloneableMixin
    * @public
    * @author Мальцев Алексей
    */

   var Enum = Dictionary.extend([IEnum, IProducible, ManyToManyMixin, SerializableMixin, CloneableMixin],/** @lends WS.Data/Type/Enum.prototype */ {
      '[WS.Data/Type/Enum]': true,

      /**
       * @cfg {Number|Null} Текущее значение
       * @name WS.Data/Type/Enum#index
       */
      _$index: null,

      /**
       * @member Название типа данных для сериализации в сырой вид
       */
      _type: 'enum',

      constructor: function Enum() {
         Enum.superclass.constructor.apply(this, arguments);
         this._checkIndex();
      },

      destroy: function() {
         ManyToManyMixin.destroy.call(this);
         Enum.superclass.destroy.call(this);
      },

      //region WS.Data/Type/IEnum

      get: function () {
         return this._$index;
      },

      set: function (index) {
         var value = this._$dictionary[index],
            defined = value !== undefined,
            changed = this._$index !== index;

         if (index === null || defined) {
            this._$index = index;
            this._checkIndex();
         } else {
            throw new ReferenceError(this._moduleName + '::set(): the index "' + index + '" is out of range');
         }

         if (changed) {
            this._notifyChange(this._$index, this.getAsValue());
         }

      },

      getAsValue: function(localize) {
         return this._getValue(this._$index, localize);
      },

      setByValue: function (value, localize) {
         var index = this._getIndex(value, localize),
            changed = index !== this._$index;

         if (value === null) {
            this._$index = value;
         } else if (index === undefined) {
            throw new ReferenceError(this._moduleName + '::setByValue(): the value "' + value + '" doesn\'t found in dictionary');
         } else {
            this._$index = index;
         }

         if (changed) {
            this._notifyChange(index, value);
         }
      },

      //endregion WS.Data/Type/IEnum

      //region WS.Data/Entity/IProducible

      produceInstance: function(data, options) {
         return new this({
            dictionary: this.prototype._getDictionaryByFormat(options && options.format),
            localeDictionary: this.prototype._getLocaleDictionaryByFormat(options && options.format),
            index: data
         });
      },

      //endregion WS.Data/Entity/IProducible

      //region WS.Data/Entity/IEquatable

      isEqual: function(to) {
         if (!(to instanceof Enum)) {
            return false;
         }

         if (!Dictionary.prototype.isEqual.call(this, to)) {
            return false;
         }

         return this.get() === to.get();
      },

      //endregion WS.Data/Entity/IEquatable

      //region Public methods

      valueOf: function() {
         return this.get();
      },

      toString: function() {
         var value = this.getAsValue();
         return value === undefined || value === null ? '' : value;
      },

      //endregion Public methods

      //region Protected methods

      /**
       * Приводит индекс у типу Number
       * @protected
       */
      _checkIndex: function() {
         if (this._$index === null) {
            return;
         }
         this._$index = parseInt(this._$index, 10);
      },

      /**
       * Уведомляет об изменении
       * @param {Number} index Изменившийся индекс
       * @param {String} value Значение в индексе
       * @protected
       */
      _notifyChange: function(index, value) {
         this._childChanged({
            index: index,
            value: value
         });
         this._notify('onChange', index, value);
      }

      //endregion Protected methods
   });

   Di.register('types.$enum', Enum, {instantiate: false});
   Di.register('types.enum', Enum);
   Di.register('data.types.enum', Enum);//deprecated

   return Enum;
});
