/* global define */
define('WS.Data/Type/Flags', [
   'WS.Data/Type/IFlags',
   'WS.Data/Type/Dictionary',
   'WS.Data/Entity/IProducible',
   'WS.Data/Entity/ManyToManyMixin',
   'WS.Data/Entity/SerializableMixin',
   'WS.Data/Entity/CloneableMixin',
   'WS.Data/Di'
], function (
   IFlags,
   Dictionary,
   IProducible,
   ManyToManyMixin,
   SerializableMixin,
   CloneableMixin,
   Di
) {
   'use strict';

   /**
    * Тип данных набор флагов
    * @class WS.Data/Type/Flags
    * @extends WS.Data/Type/Dictionary
    * @implements WS.Data/Type/IFlags
    * @implements WS.Data/Entity/IProducible
    * @mixes WS.Data/Entity/ManyToManyMixin
    * @mixes WS.Data/Entity/SerializableMixin
    * @mixes WS.Data/Entity/CloneableMixin
    * @public
    * @author Мальцев Алексей
    */

   var Flags = Dictionary.extend([IFlags, IProducible, ManyToManyMixin, SerializableMixin, CloneableMixin],/** @lends WS.Data/Type/Flags.prototype */ {
      '[WS.Data/Type/Flags]': true,

      /**
       * @cfg {Array.<Boolean|Null>} Выбранные значения согласно словарю
       * @name WS.Data/Type/Flags#values
       */
      _$values: undefined,

      /**
       * @member Название типа данных для сериализации в сырой вид
       */
      _type: 'flags',

      constructor: function Flags(cfg) {
         Flags.superclass.constructor.call(this, cfg);

         this._$values = this._$values || [];
      },

      destroy: function() {
         ManyToManyMixin.destroy.call(this);
         Flags.superclass.destroy.call(this);
      },

      //region WS.Data/Type/IFlags

      get: function (name, localize) {
         var index = this._getIndex(name, localize);
         if (index !== undefined) {
            return this._prepareValue(this._$values[index]);
         }
         return undefined;
      },

      set: function (name, value, localize) {
         var index = this._getIndex(name, localize);
         if (index === undefined) {
            throw new ReferenceError(this._moduleName + '::set(): the value "' + name + '" doesn\'t found in dictionary');
         }

         value = this._prepareValue(value);
         if (this._$values[index] === value) {
            return;
         }
         this._$values[index] = value;
         this._notifyChange(name, index, value);
      },

      getByIndex: function (index) {
         var name = this._getValue(index),
             valuesIndex = this._getIndex(name);

         return this._$values[valuesIndex];
      },

      setByIndex: function (index, value) {
         var name = this._getValue(index);
         if (name === undefined) {
            throw new ReferenceError(this._moduleName + '::setByIndex(): the index ' + index + ' doesn\'t found in dictionary');
         }

         var valuesIndex = this._getIndex(name);
         value = this._prepareValue(value);
         if (this._$values[valuesIndex] === value) {
            return;
         }

         this._$values[valuesIndex] = value;
         this._notifyChange(name, index, value);
      },

      setFalseAll: function () {
         this._setAll(false);
      },

      setTrueAll: function () {
         this._setAll(true);
      },

      setNullAll: function () {
         this._setAll(null);
      },

      //endregion WS.Data/Type/IFlags

      //region WS.Data/Entity/IProducible

      produceInstance: function(data, options) {
         return new this({
            dictionary: this.prototype._getDictionaryByFormat(options && options.format),
            localeDictionary: this.prototype._getLocaleDictionaryByFormat(options && options.format),
            values: data
         });
      },

      //endregion WS.Data/Entity/IProducible

      //region WS.Data/Entity/IEquatable

      isEqual: function(to) {
         if (!(to instanceof Flags)) {
            return false;
         }

         if (!Dictionary.prototype.isEqual.call(this, to)) {
            return false;
         }

         var enumerator = this.getEnumerator(),
            key;
         while (enumerator.moveNext()) {
            key = enumerator.getCurrent();
            if (this.get(key) !== to.get(key)) {
               return false;
            }
         }

         return true;
      },

      //endregion WS.Data/Entity/IEquatable

      //region Public methods

      toString: function() {
         return '[' + this._$values.map(function(value) {
            return value === null ? 'null' : value;
         }).join(',') + ']';
      },

      //endregion Public methods

      //region Protected methods

      _prepareValue: function (value) {
         return value === null || value === undefined ? null : !!value;
      },

      _setAll: function (value) {
         var dictionary = this._$dictionary,
            values = this._$values,
            enumerator = this.getEnumerator(),
            realIndex = 0;
         while (enumerator.moveNext()) {
            var dictionaryIndex = enumerator.getCurrentIndex();
            if (values[realIndex] !== value) {
               values[realIndex] = value;
               this._notifyChange(dictionary[dictionaryIndex], dictionaryIndex, value);
            }
            realIndex++;
         }
      },

      /**
       * Уведомляет об изменении
       * @param {String} name Имя флага
       * @param {Number} index Изменившийся индекс
       * @param {String} value Значение в индексе
       * @protected
       */
      _notifyChange: function(name, index, value) {
         this._childChanged({
            name: name,
            index: index,
            value: value
         });
         this._notify('onChange', name, index, value);
      }

      //endregion Protected methods
   });

   Di.register('types.$flags', Flags, {instantiate: false});
   Di.register('types.flags', Flags);
   Di.register('data.types.flags', Flags);//deprecated

   return Flags;
});
