/* global define */
define('WS.Data/Type/IEnum', function () {
   'use strict';

   /**
    * Интерфейс перечисляемого типа.
    * Работает на основе словаря, хранящего соотвествие индексов и их значений.
    * @interface WS.Data/Type/IEnum
    * @public
    * @author Мальцев Алексей
    */

   return /** @lends WS.Data/Type/IEnum.prototype */{
      '[WS.Data/Type/IEnum]': true,

      /**
       * @event onChange После изменения текущего индекса.
       * @param {Core/EventObject} event Дескриптор события
       * @param {Number} index Новый индекс
       * @param {String} value Новое значение
       * @example
       * <pre>
       *    requirejs(['WS.Data/Type/Enum'], function(Enum) {
       *       var colors = new Enum({
       *          dictionary: ['Red', 'Green', 'Blue']
       *       });
       *
       *       colors.subscribe('onChange', function(event, index, value) {
       *          console.log('New index: ' + index);
       *          console.log('New value: ' + value);
       *       });
       *
       *       colors.set(0);//'New index: 0', 'New value: Red'
       *       colors.setByValue('Green');//'New index: 1', 'New value: Green'
       *    });
       * </pre>
       */

      /**
       * Возвращает текущий индекс.
       * @return {Number|Null}
       * @example
       * <pre>
       *    requirejs(['WS.Data/Type/Enum'], function(Enum) {
       *       var colors = new Enum({
       *          dictionary: ['Red', 'Green', 'Blue'],
       *          index: 1
       *       });
       *
       *       console.log(colors.get());//1
       *    });
       * </pre>
       */
      get: function() {
         throw new Error('Method must be implemented');
      },

      /**
       * Устанавливает текущий индекс. Если индекс недопустим, кидает исключение.
       * @param {Number|Null} index Текущий индекс
       * @example
       * <pre>
       *    requirejs(['WS.Data/Type/Enum'], function(Enum) {
       *       var colors = new Enum({
       *          dictionary: ['Red', 'Green', 'Blue']
       *       });
       *
       *       colors.set(1);
       *       console.log(colors.get());//1
       *    });
       * </pre>
       */
      set: function(index) {
         throw new Error('Method must be implemented');
      },

      /**
       * Возвращает значение, соответствующее текущему индексу.
       * @param {Boolean} [localize=false] Вернуть локализованное значение
       * @return {String}
       * @example
       * <pre>
       *    requirejs(['WS.Data/Type/Enum'], function(Enum) {
       *       var colors = new Enum({
       *          dictionary: ['Red', 'Green', 'Blue'],
       *          index: 1
       *       });
       *
       *       console.log(colors.getAsValue());//Green
       *    });
       * </pre>
       */
      getAsValue: function(localize) {
         throw new Error('Method must be implemented');
      },

      /**
       * Устанавливает текущий индекс по соответствующему ему значению. Если значение недопустимо, кидает исключение.
       * @param {String} value Текущее значение
       * @param {Boolean} [localize=false] Установить локализованное значение
       * @example
       * <pre>
       *    requirejs(['WS.Data/Type/Enum'], function(Enum) {
       *       var colors = new Enum({
       *          dictionary: ['Red', 'Green', 'Blue'],
       *          index: 1
       *       });
       *
       *       colors.setByValue('Green');
       *       console.log(colors.get());//1
       *    });
       * </pre>
       */
      setByValue: function(value, localize) {
         throw new Error('Method must be implemented');
      }
  };
});
