define('Core/Date', [
   'Core/constants',
   'Core/i18n'
], function(
   constants,
   i18n
) {
   Date.SQL_SERIALIZE_MODE_DATE = undefined;
   Date.SQL_SERIALIZE_MODE_DATETIME = true;
   Date.SQL_SERIALIZE_MODE_TIME = false;
   Date.SQL_SERIALIZE_MODE_AUTO = null;

   var SQL_FORMAT = /([0-9]{2}:[0-9]{2}:[0-9]{2}(?:\.[0-9]{1,9})?)([+-])([0-9]{2})[:-]*([0-9]{2})*$/;

   /**
    * @class
    * @name Core/Date
    * @public
    * @author Бегунов А.В.
    */
   /**
    * Приводит объект Date() к виду, необходимому для передачи в SQL.
    * @function
    * @name Core/Date#toSQL
    * @param {Boolean|Object} [mode] Признак: необходимость выводить время.
    * <ul>
    *    <li>undefined - сериализуется как Дата.</li>
    *    <li>true - сериализуется как Дата и время.</li>
    *    <li>false - сериализуется как Время.</li>
    *    <li>null - выбрать тип автоматически (см. setSQLSerializationMode).</li>
    * <ul>
    * @returns {String}
    */
   Date.prototype.toSQL = function(mode) {

      if (mode === Date.SQL_SERIALIZE_MODE_AUTO) {
         mode = this._serializeMode;
      }

      var
         year = this.getFullYear(),
         month = this.getMonth() + 1,
         day = this.getDate(),
         hours = this.getHours(),
         minutes = this.getMinutes(),
         seconds = this.getSeconds(),
         milliseconds = this.getMilliseconds(),
         offsetNum = this.getTimezoneOffset(),
         offset = ['+', 0],
         someDig = function(num, dig) { // функция для форматирования чисел с нужным количеством цифр/ведущих нулей
            if (dig === undefined || dig < 2) {
               dig = 2;
            }
            var
               dec = num % 10;
            num -= dec;
            num /= 10;
            return (dig === 2 ? '' + num : someDig(num, dig - 1)) + dec;
         },
         data = '';
      if (mode !== Date.SQL_SERIALIZE_MODE_TIME) {
         data = year + '-' + someDig(month) + '-' + someDig(day);
      }
      if (mode !== Date.SQL_SERIALIZE_MODE_DATE) {
         if (mode === Date.SQL_SERIALIZE_MODE_DATETIME) {
            data += ' ';
         }
         data += someDig(hours) + ':' + someDig(minutes) + ':' + someDig(seconds);
         if (milliseconds) // выводим милисекунды, если они заданы
         {
            data += '.' + someDig(milliseconds, 3);
         }
         if (offsetNum > 0) // добавляем указание часового пояса локали
         {
            offset[0] = '-';
         } else {
            offsetNum = -offsetNum;
         }
         //offset[3] = offsetNum % 60;
         offsetNum -= offsetNum % 60;
         offset[1] = offsetNum / 60;
         offset[1] = someDig(offset[1]);
         //offset[3] = someDig(offset[3]);
         data += offset.join('');
      }
      return data;
   };

   /**
    * Производит сравнение текущей даты с другой.
    * @function
    * @name Core/Date#equals
    * @param {Date} d Другая дата.
    * @return {Boolean} Если даты равны, возвращает true, иначе - false.
    */
   Date.prototype.equals = function(d) {
      var res = false;
      if (d instanceof Date) {
         res = this.getTime() === d.getTime();
      }
      return res;
   };

   /**
    * Производит установку режима сериализации даты в SQL-формат.
    * @function
    * @name Core/Date#setSQLSerializationMode
    * @param {Boolean} mode Режим сериализации текущего инстанса даты в SQL-формат по умолчанию.
    * <ul>
    *    <li>undefined - сериализуется как Дата.</li>
    *    <li>true - сериализуется как Дата и время.</li>
    *    <li>false - сериализуется как Время.</li>
    * <ul>
    */
   Date.prototype.setSQLSerializationMode = function(mode) {
      this._serializeMode = mode;
   };
   /**
    * Вовзращает значение режима, который установлен для сериализации даты в SQL-формат.
    * @function
    * @name Core/Date#getSQLSerializationMode
    * @returns {Boolean|*}
    * <ul>
    *    <li>undefined - сериализуется как Дата.</li>
    *    <li>true - сериализуется как Дата и время.</li>
    *    <li>false - сериализуется как Время.</li>
    * <ul>
    */
   Date.prototype.getSQLSerializationMode = function() {
      return this._serializeMode;
   };

   /**
    * Производит откат даты.
    * @function
    * @name Core/Date#rollback
    * @param {Number} timeValue Число, равное количеству миллисекунд, прошедших с полуночи 1 января 1970 года по универсальному времени.
    */
   Date.prototype.rollback = function(timeValue) {
      this.setTime(timeValue);
   };

   /**
    * Производит форматирование даты.
    * @remark
    * Во многом похож на аналог из пхп.
    * <a href="http://php.net/manual/ru/function.strftime.php">http://php.net/manual/ru/function.strftime.php</a>
    *
    * Отображение года:
    * <ul>
    *    <li>'%y': Двухзначный порядковый номер года;</li>
    *    <li>'%Y': Четырехзначный номер года;</li>
    *    <li>'%C': Двухзначный порядковый номер столетия (год, делённый на 100, усечённый до целого);</li>
    *    <li>'%g': <b>См. пример 2!</b> Двухзначный номер года в соответствии со стандартом ISO-8601:1988 (см. %V);</li>
    *    <li>'%G': <b>См. пример 2!</b> Полная четырёхзначная версия %g.</li>
    * </ul>
    * Отображение месяца:
    * <ul>
    *    <li>'%b': Аббревиатура названия месяца (Янв);</li>
    *    <li>'%v': Аббревиатура названия месяца со строчной буквы (янв);</li>
    *    <li>'%K': Аббревиатура названия месяца со строчной буквы в родительном падеже(мая, дек);</li>
    *    <li>'%B': Полное название месяца (Январь);</li>
    *    <li>'%f': Полное название месяца со строчной буквы (январь);</li>
    *    <li>'%q': Имя месяца с маленькой буквы в родительном падеже (января);</li>
    *    <li>'%m': Двухзначный порядковый номер месяца (01, 02, ...).</li>
    * </ul>
    * Отображение дня:
    * <ul>
    *    <li>'%d': Двухзначное представление дня месяца (с ведущими нулями) (01, 02, ...);</li>
    *    <li>'%e': День месяца, с ведущим пробелом, если он состоит из одной цифры. ( 1,  2, ...);</li>
    *    <li>'%j': Порядковый номер в году, 3 цифры с ведущими нулями;</li>
    *    <li>'%u': Порядковый номер дня недели согласно стандарту ISO-8601.</li>
    * </ul>
    * Отображение недели и дня недели:
    * <ul>
    *    <li>'%a': Сокращённое название дня недели (Пн, Вт, ...);</li>
    *    <li>'%A': Полное название дня недели (Понедельник, ...);</li>
    *    <li>'%w': Порядковый номер дня недели (0 - воскресенье, 1 - понедельник, ...);</li>
    *    <li>'%U': Порядковый номер недели в указанном году, начиная с первого воскресенья в качестве первой недели;</li>
    *    <li>'%W': Порядковый номер недели в указанном году, начиная с первого понедельника в качестве первой недели;</li>
    *    <li>'%V': <b>См. пример 2!</b> Порядковый номер недели в указанном году в соответствии со стандартом ISO-8601:1988.</li>
    * </ul>
    * Отображение времени:
    * <ul>
    *    <li>'%R': Аналогично '%H:%M' (21:17);</li>
    *    <li>'%T': Аналогично '%H:%M:%S' (21:17:56);</li>
    *    <li>'%H': Двухзначный номер часа в 24-часовом формате;</li>
    *    <li>'%I': Двухзначный номер часа в 12-часовом формате;</li>
    *    <li>'%l': Час в 12-часовом формате, с пробелом перед одиночной цифрой;</li>
    *    <li>'%p': 'AM' или 'PM' в верхнем регистре, в зависимости от указанного времени;</li>
    *    <li>'%P': 'am' или 'pm' в зависимости от указанного времени;</li>
    *    <li>'%M': Двухзначный номер минуты;</li>
    *    <li>'%S': Двухзначный номер секунды;</li>
    *    <li>'%z': Смещение временной зоны относительно UTC либо аббревиатура;</li>
    *    <li>'%Z': Смещение временной зоны/аббревиатура, НЕ выдаваемая опцией %z.</li>
    * </ul>
    * Прочее:
    * <ul>
    *    <li>'%Q': квартал (I, II, ...);</li>
    *    <li>'%E': полугодие (I, II);</li>
    *    <li>'%s': Метка времени Эпохи Unix;</li>
    *    <li>'%%': Символ процента ("%").</li>
    * </ul>
    *
    * @example
    * 1. Пример вывода даты.
    * <pre>
    *    var date = new Date();
    *
    *    date.strftime('Сегодня %e %q %Y года.');
    *    // > "Сегодня 16 апреля 2014 года."
    *
    *    date.setMonth(date.getMonth()+1);
    *    date.setDate(0);
    *    date.strftime('Последний день текущего месяца будет %e %q %Y года.');
    *    // > "Последний день текущего месяца будет 30 апреля 2014 года."
    * </pre>
    * 2. Про %V, %g и %G.
    * По стандарту ISO-8601:1988 счет недель начинается с той, которая содержит минимум 4 дня текущего года.
    * Неделя начинается с понедельника, даже если он выпал на предыдущий год.
    * <pre>
    *    var date = new Date(2013,11,30);
    *
    *    date.toString();
    *    // > "Mon Dec 30 2013 00:00:00 GMT+0400 (Московское время (зима))"
    *
    *    date.strftime('Дата %d %q %Y года по ISO-8601:1988 выпадает на %V неделю %G года (%G-%V).');
    *    // > "Дата 30 декабря 2013 года по ISO-8601:1988 выпадает на 01 неделю 2014 года (2014-01)."
    * </pre>
    * @param {String} format Формат вывода
    * @returns {String} Возвращает дату в выбранном формате.
    * @function
    * @name Core/Date#strftime
    */
   Date.prototype.strftime = function() {
      function _xPad(x, pad, r) {
         if (typeof r === 'undefined') {
            r = 10;
         }
         for (; parseInt(x, 10) < r && r > 1; r /= 10) {
            x = ('' + pad) + x;
         }
         return '' + x;
      }

      var leadingZero = '000';
      var _formats = {
         a: function(d) {
            return constants.Date.days[d.getDay()];
         },
         A: function(d) {
            return constants.Date.longDays[d.getDay()];
         },
         b: function(d) {
            return constants.Date.months[d.getMonth()];
         },
         B: function(d) {
            return constants.Date.longMonths[d.getMonth()];
         },
         q: function(d) {
            return constants.Date.monthsWithDays[d.getMonth()];
         },
         K: function(d) {
            return constants.Date.monthsSmallWithDays[d.getMonth()];
         },
         f: function(d) {
            return constants.Date.longMonthsSmall[d.getMonth()];
         },
         C: function(d) {
            return _xPad(parseInt(d.getFullYear() / 100, 10), 0);
         },
         d: ['getDate', '0'],
         e: ['getDate', ' '],
         g: function(d) {
            return _xPad(parseInt(_formats.G(d) % 100, 10), 0);
         },
         G: function(d) {
            var y = d.getFullYear();
            var V = parseInt(_formats.V(d), 10);
            var W = parseInt(_formats.W(d), 10);

            if (W > V) {
               y++;
            } else if (W === 0 && V >= 52) {
               y--;
            }

            return y;
         },
         H: ['getHours', '0'],
         i: 'getDate',
         I: function(d) {
            var I = d.getHours() % 12;
            return _xPad(I === 0 ? 12 : I, 0);
         },
         j: function(d) {
            var ms = d - new Date('' + d.getFullYear() + '/1/1 GMT');
            ms += d.getTimezoneOffset() * 60000; // Line differs from Yahoo implementation which would be equivalent to replacing it here with:
            // ms = new Date('' + d.getFullYear() + '/' + (d.getMonth()+1) + '/' + d.getDate() + ' GMT') - ms;
            var doy = parseInt(ms / 60000 / 60 / 24, 10) + 1;
            return _xPad(doy, 0, 100);
         },
         k: ['getHours', '0'],
         // not in PHP, but implemented here (as in Yahoo)
         l: function(d) {
            var l = d.getHours() % 12;
            return _xPad(l === 0 ? 12 : l, ' ');
         },
         m: function(d) {
            return _xPad(d.getMonth() + 1, 0);
         },
         M: ['getMinutes', '0'],
         p: function(d) {
            return ['AM', 'PM'][d.getHours() >= 12 ? 1 : 0];
         },
         P: function(d) {
            return ['am', 'pm'][d.getHours() >= 12 ? 1 : 0];
         },
         Q: function(d) {
            return ['I', 'II', 'III', 'IV'][parseInt(d.getMonth() / 3, 10)];
         },
         E: function (d) {
            return ['I', 'II'][parseInt(d.getMonth() / 6, 10)];
         },
         s: function(d) { // Yahoo uses return parseInt(d.getTime()/1000, 10);
            return Date.parse(d) / 1000;
         },
         S: ['getSeconds', '0'],
         u: function(d) {
            var dow = d.getDay();
            return ((dow === 0) ? 7 : dow);
         },
         U: function(d) {
            var doy = parseInt(_formats.j(d), 10);
            var rdow = 6 - d.getDay();
            var woy = parseInt((doy + rdow) / 7, 10);
            return _xPad(woy, 0);
         },
         v: function(d) {
            var month = d.getMonth();
            return constants.Date.monthsSmall[month];
         },
         V: function(d) {
            var woy = parseInt(_formats.W(d), 10);
            var dow1_1 = (new Date('' + d.getFullYear() + '/1/1')).getDay();
            // First week is 01 and not 00 as in the case of %U and %W,
            // so we add 1 to the final result except if day 1 of the year
            // is a Monday (then %W returns 01).
            // We also need to subtract 1 if the day 1 of the year is
            // Friday-Sunday, so the resulting equation becomes:
            var idow = woy + (dow1_1 > 4 || dow1_1 <= 1 ? 0 : 1);
            if (idow === 53 && (new Date('' + d.getFullYear() + '/12/31')).getDay() < 4) {
               idow = 1;
            } else if (idow === 0) {
               idow = _formats.V(new Date('' + (d.getFullYear() - 1) + '/12/31'));
            }
            return _xPad(idow, 0);
         },
         w: 'getDay',
         W: function(d) {
            var doy = parseInt(_formats.j(d), 10);
            var rdow = 7 - _formats.u(d);
            var woy = parseInt((doy + rdow) / 7, 10);
            return _xPad(woy, 0, 10);
         },
         y: function(d) {
            return _xPad(d.getFullYear() % 100, 0);
         },
         Y: 'getFullYear',
         z: function(d) {
            var o = d.getTimezoneOffset();
            var H = _xPad(parseInt(Math.abs(o / 60), 10), 0);
            var M = _xPad(o % 60, 0);
            return (o > 0 ? '-' : '+') + H + M;
         },
         Z: function(d) {
            var Z = d.toString().match(/^[^(]+\((.+)\)$/);
            return Z ? Z[1] : '';
         },
         L: 'getMilliseconds',
         J: function (d) {
            var
               ms = d.getMilliseconds().toString();
            return leadingZero.substr(0, 3 - ms.length) + ms;
         },
         '%': function(d) {
            return '%';
         }
      };

      var _aggregates = {
         c: '%a %d %b %Y %r %Z',
         D: '%d/%m/%y',
         F: '%y-%m-%d',
         h: '%b',
         n: '\n',
         r: '%I:%M:%S %p',
         R: '%H:%M',
         t: '\t',
         T: '%H:%M:%S',
         x: '%d/%m/%Y',
         X: '%r'
      };

      return function(format) {
         var self = this;

         // First replace aggregates (run in a loop because an agg may be made up of other aggs)
         while (format.match(/%[cDFhnrRtTxX]/)) {
            format = format.replace(/%([cDFhnrRtTxX])/g, function(m0, m1) {
               return _aggregates[m1];
            });
         }

         // Now replace formats - we need a closure so that the date object gets passed through
         return format.replace(/%([aAbBCdeEfgGHiIjkKlmMpPsqQSuUvVwWyYzZLJ%])/g, function(m0, m1) {
            var f = _formats[m1];
            if (typeof f === 'string') {
               return self[f]();
            } else if (typeof f === 'function') {
               return f(self);
            } else if (typeof f === 'object' && typeof(f[0]) === 'string') {
               return _xPad(self[f[0]](), f[1]);
            } else { // Shouldn't reach here
               return m1;
            }
         });
      }
   }();

   /**
    * Разбирает дату из БД в объект Date. Если передать дату со временем или только время, то оно будет приведено к местному.
    * @param {String} dateTime Дата и/или время в формате БД
    * @param {Boolean} [useDefaultTimeZone=true] Использовать временную зону по умолчанию (Москва), если в dateTime указано время без временной зоны. Если передать false, то при отсутствии временной зоны коррекция к местному времени производиться не будет.
    * @returns {Date}
    * @function
    * @name Core/Date#fromSQL
    */
   Date.fromSQL = function(dateTime, useDefaultTimeZone) {
      switch (dateTime) {
         case 'infinity':
            throw new TypeError('Type Infinity is not supported');
         case '-infinity':
            throw new TypeError('Type -Infinity is not supported');
      }

      if (useDefaultTimeZone === undefined) {
         useDefaultTimeZone = true;
      }

      var
         dateSep = dateTime.indexOf('-'),
         timeSep = dateTime.indexOf(':'),
         millisecSep = dateTime.indexOf('.'),
         tz = dateTime.match(SQL_FORMAT),
         tzSep = tz && tz.index + tz[1].length || -1,
         retval = new Date(),
         ms = null,
         msStr,
         timeOffset,
         y,
         m,
         d;

      retval.setHours(0, 0, 0, 0);

      if (timeSep === -1 && dateSep === -1) {
         return retval;
      }

      //Если передали только время с отрицательным часовым поясом
      if (timeSep > -1 && dateSep > timeSep) {
         dateSep = -1;
      }

      if (millisecSep !== -1) {
         msStr = dateTime.substr(millisecSep + 1, (tzSep === -1 ? dateTime.length : tzSep) - (millisecSep + 1));
         if (msStr.length > 3) {
            msStr = msStr.substr(0, 3);
         }
         ms = parseInt(msStr, 10);
         if (msStr.length < 3) {
            ms *= (msStr.length === 2 ? 10 : 100);
         }
      }

      if (dateSep !== -1) {
         y = parseInt(dateTime.substr(dateSep - 4, 4), 10);
         m = parseInt(dateTime.substr(dateSep + 1, 2), 10) - 1;
         d = parseInt(dateTime.substr(dateSep + 4, 2), 10);

         if (constants.compatibility.dateBug && m === 0 && d === 1) {
            retval.setHours(1);
         }

         retval.setFullYear(y, m, d);
      }

      if (timeSep !== -1) {
         retval.setHours(
            parseInt(dateTime.substr(timeSep - 2, 2), 10),
            parseInt(dateTime.substr(timeSep + 1, 2), 10),
            parseInt(dateTime.substr(timeSep + 4, 2), 10),
            ms);

         //Указана временная зона или используется временная зона по умолчанию
         if (tz || useDefaultTimeZone) {
            //Приводим время к местному из серверного, если дата передана с сервера со временем
            timeOffset = tz ? ((tz[2] + '1') * (tz[3] * 60 + (tz[4] * 1 || 0))) : constants.moscowTimeOffset;
            retval.setMinutes(retval.getMinutes() - timeOffset - retval.getTimezoneOffset());
         }
      }

      return retval;
   };

   /**
    * Отменяет перевод времени, производимый в функции fromSQL.
    * @returns {Date} Возвращает текущую уже изменённую дату.
    * @function
    * @name Core/Date#toServerTime
    */
   Date.prototype.toServerTime = function() {
      // Приводим время к серверному (московскому) из местного
      this.setMinutes(this.getMinutes() + this.getTimezoneOffset() + constants.moscowTimeOffset);
      return this;
   };

   /**
    * Устанавливает дату на последний день месяца.
    * @param {Number} [month] Номер месяца 0 - 11, последний день которого нужен.
    * Если не указан берется из даты.
    * @returns {Date}
    * @function
    * @name Core/Date#setLastMonthDay
    */
   Date.prototype.setLastMonthDay = function(month) {
      month = month === undefined ? this.getMonth() : parseInt(month, 10);
      this.setDate(1);
      this.setMonth(month + 1);
      this.setDate(0);
      return this;
   };

   //Dirty hack for Node.js: use client cookies to detect it's time zone (if possible)
   if (constants.isNodePlatform) {
      //Save system time zone
      var systemTimeZone = new Date().getTimezoneOffset();
      //Overwrite method on Date prototype
      Date.prototype._getTimezoneOffset = Date.prototype.getTimezoneOffset;
      Date.prototype.getTimezoneOffset = function() {
         //Return time zone from cookies if possible. Use typecasting.
         var clientTimeZone = process && process.domain && process.domain.req && process.domain.req.cookies && process.domain.req.cookies.tz;
         return parseInt(clientTimeZone || systemTimeZone, 10);
      };
   }

   var locales = {
         'ru-RU': {
            days: ['Вс', 'Пн', 'Вт', 'Ср', 'Чт', 'Пт', 'Сб'],
            daysSmall: ['вс', 'пн', 'вт', 'ср', 'чт', 'пт', 'сб'],
            longDays: ['Воскресенье', 'Понедельник', 'Вторник', 'Среда', 'Четверг', 'Пятница', 'Суббота'],
            months: ['Янв', 'Фев', 'Мар', 'Апр', 'Май', 'Июн', 'Июл', 'Авг', 'Сен', 'Окт', 'Ноя', 'Дек'],
            monthsSmall: ['янв', 'фев', 'мар', 'апр', 'май', 'июн', 'июл', 'авг', 'сен', 'окт', 'ноя', 'дек'],
            monthsBig: ['ЯНВ', 'ФЕВ', 'МАР', 'АПР', 'МАЙ', 'ИЮН', 'ИЮЛ', 'АВГ', 'СЕН', 'ОКТ', 'НОЯ', 'ДЕК'],
            longMonths: [
               'Январь', 'Февраль', 'Март', 'Апрель', 'Май', 'Июнь', 'Июль', 'Август', 'Сентябрь', 'Октябрь', 'Ноябрь',
               'Декабрь'
            ],
            longMonthsSmall: [
               'январь', 'февраль', 'март', 'апрель', 'май', 'июнь', 'июль', 'август', 'сентябрь', 'октябрь', 'ноябрь',
               'декабрь'
            ],
            monthsWithDays: [
               'января', 'февраля', 'марта', 'апреля', 'мая', 'июня', 'июля', 'августа', 'сентября', 'октября', 'ноября',
               'декабря'
            ],
            monthsSmallWithDays: [
               'янв', 'фев', 'мар', 'апр', 'мая', 'июн', 'июл', 'авг', 'сен', 'окт', 'ноя', 'дек'
            ],
            quarters: ['I', 'II', 'III', 'IV'],
            longQuarters: ['I квартал', 'II квартал', 'III квартал', 'IV квартал'],
            masks: {
               date: '%d.%m.%y',
               min: '%H:%M',
               sec: '%H:%M:%S',
               msec: '%H:%M:%S.%J'
            }
         },
         'en-US': {
            days: ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
            daysSmall: ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
            longDays: ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'],
            months: ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
            monthsSmall: ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
            monthsBig: ['JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN', 'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC'],
            longMonths: [
               'January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October',
               'November', 'December'
            ],
            longMonthsSmall: [
               'January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October',
               'November', 'December'
            ],
            monthsWithDays: [
               'January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October',
               'November', 'December'
            ],
            monthsSmallWithDays: ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],
            quarters: ['I', 'II', 'III', 'IV'],
            longQuarters: ['I quarter', 'II quarter', 'III quarter', 'IV quarter'],
            masks: {
               date: '%m/%d/%y',
               min: '%I:%M %p',
               sec: '%I:%M:%S %p',
               msec: '%I:%M:%S.%J %p'
            }
         }
      },
      defaultLang = 'ru-RU';

   //FIXME: isEnabled() result can be changed during runtime
   if (i18n.isEnabled()) {
      var global = (function(){ return this || (0,eval)('this'); }());
      if (global.Intl) {
         Date.prototype._toLocaleString = Date.prototype.toLocaleString;
         Date.prototype._toLocaleDateString = Date.prototype.toLocaleDateString;
         Date.prototype._toLocaleTimeString = Date.prototype.toLocaleTimeString;

         Date.prototype.toLocaleString = function(locales, options) {
            return this._toLocaleString(locales || i18n.getLang() || [], options);
         };
         Date.prototype.toLocaleDateString = function(locales, options) {
            return this._toLocaleDateString(locales || i18n.getLang() || [], options);
         };
         Date.prototype.toLocaleTimeString = function(locales, options) {
            return this._toLocaleTimeString(locales || i18n.getLang() || [], options);
         };
      }
   }

   Object.defineProperty(constants, 'Locales', {
      configurable: true,
      get: function() {
         return Object.keys(locales);
      }
   });

   Object.defineProperty(constants, 'Date', {
      configurable: true,
      get: function() {
         var lang = i18n.isEnabled() ? i18n.getLang() : defaultLang;
         return locales[lang] || locales[defaultLang];
      }
   });

   return Date;
});