define('Core/LocalStorageNative', [
    'Core/UserInfo',
    'Core/IoC'
], function (UserInfo, IoC) {
    "use strict";
    ///region const
    var GLOBAL = (function () { return this || (0, eval)('this'); })();
    var SID_KEY = "__sid";
    var S3SU_KEY = "__s3su";
    var WS_STORAGE_KEYS = "ws-local-keys";
    // Ключ по которому хранится список полей, которые необходимо сохранить в хранилище при смене пользователя
    var WS_STORE_ALWAYS_KEYS = "ws-store-always-keys";
    // Нативный localStorage, либо фейковый объект, если нативный не доступен
    var STORAGE = getLocalStorage();
    // Максимальный допустимый для записи размер строки
    var MAXIMUM_ITEM_SIZE = 1024 * 512;
    ///endregion const
    ///region utils
    /**
     * Логирование сообщения
     * @param {String} message
     * @param {"info" | "error" | "log"} [type] тип сообщения лог/инфо/ошибка
     */
    function log(message, type) {
        type = type || "log";
        IoC.resolve('ILogger')[type]('Core/LocalStorageNative', message);
    }
    /**
     * Метод получения нативного localStorage,
     * либо реализацию интерфейса нативного localStorage вокруг объекта-пустышки, в случае возникновении ошибки при работе с первым
     *
     * Возможные ошибки с получением нативной реализации
     * 1) не нашли localStorage в текущем контексте исполнения или в глобальном
     * 2) нету доступа на запись в хранилище.
     * 3) возвращаемое значение не равно записываемому - так же проблема, что и выше, но setItem не выбросил ошибку
     * @private
     * @return {Storage}
     */
    function getLocalStorage() {
        try {
            var storage = this && this.localStorage || GLOBAL && GLOBAL.localStorage;
            var testItem = Math.random().toString();
            var tetItemName = "PingLocalStorage" + testItem;
            storage.setItem(tetItemName, testItem);
            if (testItem !== storage.getItem(tetItemName)) {
                throw new Error("localStorage not supported, using fake storage");
            }
            storage.removeItem(tetItemName);
            return storage;
        }
        catch (err) {
            return new FakeStorage();
        }
    }
    /**
     * Фейковая реализация интерфейса хранилища
     * @private
     * @constructor
     */
    function FakeStorage() {
        var _storage = {};
        this.setItem = function (key, data) {
            _storage[key] = data && data.toString() || ("" + data);
            this.length++;
        };
        this.removeItem = function (key) {
            delete _storage[key];
            this.length--;
        };
        this.getItem = function (key) {
            return _storage[key] || null;
        };
        this.clear = function () {
            _storage = {};
            this.length = 0;
        };
        this.length = 0;
    }
    /**
     * Возвращает размер строки
     * @private
     * @param {String} str
     * @return {Number}
     */
    function getItemSize(str) {
        var match = encodeURIComponent(str).match(/%[89ABab]/g);
        return str.length + (match ? match.length : 0);
    }
    ///endregion utils
    // try/catch to call native methods
    var tryToSet = function (key, data) {
        try {
            STORAGE.setItem(key, data);
            return true;
        }
        catch (err) {
            log(err);
            return false;
        }
    };
    var tryToGet = function (key) {
        try {
            return STORAGE.getItem(key);
        }
        catch (err) {
            log(err);
            return null;
        }
    };
    var tryToRemove = function (key) {
        try {
            STORAGE.removeItem(key);
        }
        catch (err) {
            log(err);
        }
    };
    /**
     * Очищает хранилище, если сменился пользователь
     * @void
     */
    var clearIfInvalidUser = function () {
        var fromStorage = tryToGet(S3SU_KEY);

        if (!UserInfo.isValid(fromStorage)) {
            //сбросим значения всех параметров, которые мы сохраняли, ибо сессия поменялась
            clearStorage();
            var fromCookie = UserInfo.getCurrent();
            // null, undefined, false, 0 не могут быть сидом, а записывая их как строки мы создаём себе проблем при валидации
            if (fromCookie) {
                tryToSet(S3SU_KEY, fromCookie);
            } else {
                tryToRemove(S3SU_KEY);
            }
        }
    };
    /**
     * Метод очистки значений хранилища, которые были установлены через наш модуль
     *
     * Необходимо перевести платформу на работу с хранилищем через модуль
     * Тогда мы можем без угрезений совести очищать всё хранилище через .clear,
     * сохранив при этом и потом записав обратно storage-always поля
     */
    var clearStorage = function () {
        var localKeys = getStorageKeys();
        var alwaysStorageKeys;
        try {
            // [].concat для того чтобы не упало, если кто-то вместо массива в поле запишет одно значение
            alwaysStorageKeys = [].concat(JSON.parse(tryToGet(WS_STORE_ALWAYS_KEYS)));
        } catch (e) {
            alwaysStorageKeys = [];
        }
        var newKeys = [];
        localKeys.forEach(function (key) {
            if (alwaysStorageKeys.indexOf(key) === -1) {
                return tryToRemove(key);
            }
            newKeys.push(key);
        });
        tryToSet(WS_STORAGE_KEYS, JSON.stringify(newKeys));
    };
    /**
     * Возвращает массив из ключей, записанных через наш модуль
     * @return {Array<string>}
     */
    var getStorageKeys = function () {
        try {
            return JSON.parse(tryToGet(WS_STORAGE_KEYS)) || [];
        }
        catch (e) {
            return [];
        }
    };
    // Зачистим sid за собой, удалить после выхода 3.17.200
    tryToRemove(SID_KEY);
    /**
     * Обёртка для безопасной работы с локальным хранилищем
     * @class Core/LocalStorageNative
     * @public
     */
    var LocalStorageNative = /** @lends Core/LocalStorageNative.prototype*/{
        /**
         * Возвращает значение из localStorage по ключу.
         * @param {String} key - ключ
         * @returns {String || null} Значение localStorage по key
         */
        getItem: function (key) {
            clearIfInvalidUser();
            return tryToGet(key);
        },
        /**
         * Устанавливает значение в localStorage по ключу.
         * @param {String} key - ключ
         * @param {String} data - Значение
         * @return {Boolean} Флаг успешности записи
         */
        setItem: function (key, data) {
            // Не даём перезаписать пользователя
            if (key === S3SU_KEY) {
                return;
            }
            var value = "" + data;
            var size = getItemSize(value);
            if (size > MAXIMUM_ITEM_SIZE) {
                log("Maximum size of item for write local storage = " + MAXIMUM_ITEM_SIZE / 1024 + "kb, item size = " + (size / 1024).toFixed(2) + "kb", "error");
                return false;
            }
            clearIfInvalidUser();
            var res = tryToSet(key, value);
            // Запоминаем ключ
            var localKeys = getStorageKeys();
            /** 
             * если мы еще не запомнили, что мы используем этот ключ, то запомним его в список
             * IE бросает событие о изменении хранилища, даже если данные фактически не поменялись
             * и бросает их на всех вкладках, включая вкладку-инициатор,
             * + стреляются события сессионного хранилища
             * при активном использовании IE10 не стправляется с нагрузкой и пропускает некоторые события, либо доставляет их с задержкой в несколько минут
             * поэтому тут и ниже пишем свой хелперы только если они изменили значение
             */
            if ((localKeys.indexOf(key) === -1)  && (key !== WS_STORE_ALWAYS_KEYS)) {
                localKeys.push(key);
                tryToSet(WS_STORAGE_KEYS, JSON.stringify(localKeys));
            }
            return res;
        },
        /**
         * Удаляет значение в localStorage по ключу.
         * @param {String} key - ключ
         * @void
         */
        removeItem: function (key) {
            if (key === S3SU_KEY) {
                return;
            }
            tryToRemove(key);
            var localKeys = getStorageKeys();
            localKeys = localKeys.filter(function(_key) {
                return key !== _key;
            });
            tryToSet(WS_STORAGE_KEYS, JSON.stringify(localKeys));
            var storeAlwaysKeys;
            try {
                storeAlwaysKeys = JSON.parse(tryToGet(WS_STORE_ALWAYS_KEYS)) || [];
            } catch (e) {
                storeAlwaysKeys = [];
            }
            if (storeAlwaysKeys.indexOf(key) !== -1) {
                storeAlwaysKeys = storeAlwaysKeys.filter(function(_key) {
                    return key !== _key;
                });
                tryToSet(WS_STORE_ALWAYS_KEYS, JSON.stringify(storeAlwaysKeys));
            }
        },
        /**
         * Удаляет все значения в localStorage, записанные через данный модуль
         * @void
         */
        clear: function () {
            clearStorage();
            clearIfInvalidUser();
        },
        /**
         * Возвращает количество записей хранилища
         * @return {number}
         */
        getLength: function () {
            return STORAGE.length;
        },
        /**
         * Возвращает массив ключей, записанных через данный модуль
         * @return {Array<String>}
         */
        getKeys: function () {
            return getStorageKeys();
        }
    };
    return LocalStorageNative;
});
