define('Core/Sanitize', [
   'Core/markup/ParserUtilities',
   'Core/helpers/Object/find',
   'Core/core-merge',
   'Core/core-clone',
   'Core/helpers/String/escapeHtml'
], function(Parser, objectFind, coreMerge, coreClone, escapeHtml) {
   'use strict';
   /**
    * Очищает контент от вредоносных скриптов
    * @class Core/Sanitize
    * @author Авраменко А.С.
    * @public
    */

   var
      validNodes = { // Допустимые типы нод
         html: true, head: true, body: true, code: true,// Основные элементы
         p: true, div: true, span: true, img: true, br: true, a: true, pre: true, label: true, // Основные элементы
         b: true, strong: true, i: true, em: true, u: true, s: true, strike: true, q: true, blockquote: true, font: true, // Стили
         h1: true, h2: true, h3: true, h4: true, h5: true, h6: true, // Заголовки
         dd: true, dir: true, dl: true, dt: true, li: true, menu: true, ol: true, ul: true, // Списки
         table: true, thead: true, tbody: true, caption: true, col: true, colgroup: true, td: true, tfoot: true, th: true, tr: true, // Таблицы
         input: true, textarea: true, // инпуты
         iframe: {
            src: function(content, attributeName) {
               var
                  whiteList =  [
                     '//www.youtube.com/embed/'
                  ],
                  srcInWhiteList = objectFind(whiteList, function(url) {
                     return content.attributes[attributeName].value.indexOf(url) >= 0;
                  });

               if (!srcInWhiteList) {
                  delete content.attributes[attributeName];
               }
            },
            allowfullscreen: true
         }
      },

      validAttributes = { // Допустимые атрибуты
         style: true,
         src: true,
         height: true,
         width: true,
         colspan: true,
         rowspan: true,
         'class': true,
         id: true,
         tabindex: true,
         title: true,
         href: true,
         target: true,
         rel: true,// Для ссылок важно иметь атрибут rel="noopener noreferrer"
         alt: true,
         hasmarkup: true,
         config: true,
         name: true
      },

      fullEscapeNodes = { // Типы нод, которые нужно экранировать полностью, вместе с содержимым.
         script: true,    // Если ноды нет в этом списке, в текст превращаются только открывающие и закрывающие теги,
         style: true      // а содержимое остается тегами и также проходит проверку на валидность
      },

      defaultOptions = {
         validNodes: validNodes,
         validAttributes: validAttributes,
         fullEscapeNodes: fullEscapeNodes,
         checkDataAttribute: true,
         escapeInvalidTags: true // не вырезать неразрешенные теги, а экранировать и превращать в текст
      },
   
      // Для того чтобы data-атрибуты не вырезались
      checkDataAttribute = function checkDataAttribute(name) {
         return /^data-([\w-])*$/gi.test(name);
      },

      jsRegExp = /javascript:/gim;

   function validateAttributes(content, settings) {
      var
         attributeSettings;
      for (var attributeName in content.attributes) {
         if (content.attributes.hasOwnProperty(attributeName)) {
            var
               attribute = content.attributes[attributeName];
            attributeSettings = settings.validAttributes[attributeName] || (settings.checkDataAttribute ? checkDataAttribute(attributeName): false);
            if (typeof attributeSettings !== 'function') {
               if (attributeSettings) {
                  while (jsRegExp.test(attribute.value)) {
                     attribute.value = attribute.value.replace(jsRegExp, '');
                  }
               } else {
                  delete content.attributes[attributeName];
               }
            } else {
               // если attributeSettings - функция, то она должна сама вырезать или оставить атрибут
               attributeSettings(content, attributeName);
            }
         }
      }
   }

   function validateContent(content, settings) {
      if (content.childNodes) {
         var idx = content.childNodes.length;

         while (idx--) {
            var
               node = content.childNodes[idx],
               nodeType = node.nodeType,
               nodeName = node.nodeName,
               nodeSettings = settings.validNodes[(nodeName || '').toLowerCase()];

            if (typeof nodeSettings !== 'object') {
               if (nodeType !== 1 || nodeSettings) {
                  // детей component и его атрибуты не надо валидировать
                  if (nodeName !== 'component') {
                     validateContent(node, settings);
                  }
               } else {
                  // обрабатываем неразрешенный тег в зависимости от опции
                  if (settings.escapeInvalidTags) {
                     // экранируем запрещенный тег
                     if (settings.fullEscapeNodes[nodeName]) {
                        // создаем новую текстовую ноду, с содержимым в виде текста неразрешенного тега
                        var escapedTag = escapeHtml(node.outerHTML());
                        content.childNodes[idx] = new Parser.Node({nodeType: 3, text: escapedTag});
                     } else {
                        // считаем за текст открывающий и закрывающий теги, внутренние ноды пока оставляем без изменений
                        var innerNodes = node.childNodes; // берем внутренние ноды экранируемого тега
                        if (node.startTag) {
                           // экранируем открывающий тег и добавляем в начало innerNodes
                           innerNodes.unshift(new Parser.Node({ nodeType: 3, text: escapeHtml(node.startTag) }));
                        }
                        if (node.closeTag) {
                           // экранируем закрывающий тег и добавляем в конец innerNodes
                           innerNodes.push(new Parser.Node({ nodeType: 3, text: escapeHtml(node.closeTag) }));
                        }

                        // заменяем запрещенный тег на [экранированный открывающий + вложенные теги + экранированный закрывающий
                        content.childNodes.splice.apply(content.childNodes, [ idx, 1 ].concat(innerNodes));

                        // сдвигаем указатель проверки, так как нужно проверить ноды, которые раньше лежали
                        // внутри замененного тега
                        idx += innerNodes.length;
                     }
                  } else {
                     // вырезаем запрещенный тег
                     content.childNodes.splice(idx, 1);
                  }
               }
            } else {
               // если nodeSettings - object, значит это набор атрибутов с их валидаторами
               var tempSettings = coreMerge(coreClone(settings), {validAttributes: nodeSettings});
               validateContent(content.childNodes[idx], tempSettings);
            }
         }
      }
      validateAttributes(content, settings);
   }

   return function(content, options) {
      var parsed, settings;

      if (typeof(content) !== 'string') {
         return content;
      }
      
      // нет <[A-z] значит точно нет тегов и нечего санитайзить
      // мы должны удостовериться, что это именно что-то похожее на тег < + какой-то alphaсимвол
      if (content.search(/<[A-zA-я]/g) === -1) {
         return content;
      }

      parsed = Parser.parse(content);
      settings = options ? coreMerge(coreClone(defaultOptions), options) : defaultOptions;
      validateContent(parsed, settings);
      return parsed.innerHTML();
   };

});
