define('Core/cookie', [], function() {
   'use strict';

   var MS_IN_DAY = 24 * 60 * 60 * 1000,
      getRequest = function() {
         return process && process.domain && process.domain.req || {};
      },
      cookie;

   /**
    * Модуль для работы с cookie.
    * @class Core/cookie
    * @public
    * @author Мальцев А.А.
    */
   cookie =/** @lends Core/cookie.prototype */{
      /**
       * @typedef {Object} CookieOptions
       * @property {Number|Date} [expires] Дата окончания действия cookie. Если не указана, то cookie исчезнет после завершении сессии браузера.
       * @property {String} [path] Путь на сервере, для которого действует cookie. По умолчанию - текущий путь.
       * @property {String} [domain] Домен, на который следует установить cookie. По умолчанию - текущий домен.
       * @property {Boolean} [secure=false] Передавать cookie только через безопасный протокол
       */

      /**
       * Возвращает значение cookie с указанным именем
       * @param {String} name Имя cookie
       * @returns {*} Значение cookie
       */
      get: function (name) {
         if (cookie.SERVER_SCRIPT) {
            var request = getRequest();
            return request.cookies ? request.cookies[name] : null;
         }

         var value = null;
         if (document.cookie) {
            var cookies = document.cookie.split(';'),
               item;
            for (var i = 0; i < cookies.length; i++) {
               item = String(cookies[i]).trim();
               if (item.substring(0, name.length + 1) === (name + '=')) {
                  value = decodeURIComponent(item.substring(name.length + 1));
                  break;
               }
            }
         }
         return value;
      },

      /**
       * Устанавливает cookie с указанным именем
       * @param {String} name Имя cookie
       * @param {String} value Значение cookie
       * @param {CookieOptions} [options] Дополнительные свойства для установки cookie
       */
      set: function (name, value, options) {
         var expires = '',
            path,
            domain,
            secure;

         options = options || {};

         if (cookie.SERVER_SCRIPT) {
            throw new Error('Set cookie on server is not supported');
         }

         if (value === null) {
            value = '';
            options.expires = -1;
         }

         if (options.expires) {
            var date;
            if (typeof options.expires === 'number') {
               date = new Date();
               date.setTime(date.getTime() + (options.expires * MS_IN_DAY));
            } else if (options.expires.toUTCString) {
               date = options.expires;
            } else {
               throw new TypeError('Option "expires" should be a Number or Date instance')
            }
            expires = '; expires=' + date.toUTCString();
         }

         path = options.path ? '; path=' + options.path : '';
         domain = options.domain ? '; domain=' + options.domain : '';
         secure = options.secure ? '; secure' : '';

         document.cookie = [name, '=', encodeURIComponent(value), expires, path, domain, secure].join('');
      },

      SERVER_SCRIPT: typeof window === 'undefined'
   };

   return cookie;
});