define('Core/core-functions', [
   'require',
   'Core/core-merge',
   'Core/core-clone',
   'Core/core-hash',
   'Core/property-merge',
   'Core/core-classicExtend',
   'Core/constants',
   'Core/IoC',
   'Core/Deferred',
   'Core/helpers/Function/shallowClone',
   'Core/helpers/Object/isEmpty'
], function(
   require,
   coreMerge,
   coreClone,
   coreHash,
   propertyMerge,
   classicExtend,
   _const,
   IoC,
   Deferred,
   shallowClone
) {
   var coreFunctions = /** @lends $ws.core.prototype */{
      /**
       * Объединение двух hash.
       * @param {Object} hash Исходный хэш.
       * @param {Object} hashExtender хэш-расширение.
       * @param {Object} [cfg] Параметры работы.
       * @param {Boolean} [cfg.preferSource=false] Сохранять или нет исходное значение. По умолчанию исходное значение не сохраняется.
       * @param {Boolean} [cfg.rec=true] Рекурсивное объединение (по умолчанию - рекурсивное объединение включено).
       * @param {Boolean} [cfg.clone=false] Клонировать элементы или передавать по ссылке (по умолчанию - по ссылке).
       * @param {Boolean} [cfg.create=true] Создавать элементы, отсутствующие в исходном объекте. По умолчанию включено.
       * @param {Boolean} [cfg.noOverrideByNull=false] Запретить заменять исходные значения на null. По умолчанию выключено.
       */
      merge: coreMerge,
      propertyMerge: propertyMerge,

      clone: coreClone,
      /**
       * Функция, делающая поверхностное (без клонирования вложенных объектов и массивов) копирование объекта или массива. Сделана на основе <strong>$ws.core.merge</strong>.
       * @param hash Исходный объект или массив.
       * @return {Object} Скопированный объект или массив.
       */
      shallowClone: shallowClone,

      hash: coreHash,

      /**
       * Функция разбивает путь на класс и компонент
       * @param path
       * @returns {{component: *, className: *, moduleName: *, isModule: *}}
       */
      extractComponentPath: function(path) {
         var className, modName;

         var ClassMapper = require.defined("Deprecated/ClassMapper") ? require("Deprecated/ClassMapper") : null;
         if (ClassMapper === null && IoC.has('ILogger')) {
            IoC.resolve('ILogger').error('extractComponentPath', 'Для работы функции extractComponentPath требуется Deprecated/ClassMapper. Подключите этот модуль в свои зависимости.')
         }
         path = ClassMapper ? ClassMapper.getClassMapping(path) : path;

         if (path.indexOf(':') == -1) {
            path = path.split('/');
            path.push(className = path.pop());
            if (className.indexOf('@') != -1) {
               className = className.split('@').pop();
            }
         } else {
            path = path.split(':');
            className = path.pop();
            path = path.pop().split('/');
         }

         modName = className in _const.jsCoreModules ? className : ((className in _const.jsModules) ? className : 'SBIS3.CORE.' + className);

         return {
            component: path.join('/'),
            className: className,
            moduleName: modName,
            isModule: modName in _const.jsCoreModules || modName in _const.jsModules
         };
      },
      /**
       * Проверяет, имеется ли для файла альтернативный URL
       *
       * @param {String} url исходный URL файла.
       * @returns {String} альтернативный URL для файла (с учетом пакета).
       */
      checkPackages: function(url) {
         var shortUrl = url.replace(_const.wsRoot, '/');
         var packages = _const.jsPackages;
         return packages[shortUrl] ? (_const.wsRoot + packages[shortUrl]) : url;
      },
      getNamedContents: function(name) {
         var isPath = name.lastIndexOf('/');
         if (isPath != -1) {
            name = name.substring(isPath + 1);
         }
         return _const.hdlBindings[name] || [];
      },
      /**
       * Устанавливает курсор загрузки или стандартный
       * @param {Boolean} def -- ставить ли стандартный курсор.
       */
      setCursor: function(def) {
         $('body').toggleClass('ws-progress', !def);
      },
      /**
       * Возвращает URL сервиса по имени
       * @param {String} serviceName имя сервиса (как указано на 4 шаге мастера выгрузки).
       * @returns {String} URL сервиса.
       */
      getServiceByName: function(serviceName) {
         return _const.services[serviceName] || _const.defaultServiceUrl;
      },

      classicExtend: classicExtend,

      /**
       * Возвращает случайный хост из списка хостов
       *
       * @param {String} url
       * @returns {String} адрес хоста
       */
      urlWithHost: function(url) {
         var l = _const.hosts.length,
            host = '';
         if (l > 0) {
            host = _const.hosts[Math.floor(Math.random() * l)];
            if (url.substring(0, 1) == '.') {
               var curPath = window.location.pathname;
               curPath = curPath.substring(0, curPath.lastIndexOf('/') + 1);
               url = host + curPath + url.substring(2);
            } else {
               url = host + (url.substring(0, 1) == '/' ? '' : '/') + url;
            }
         }
         return url;
      },
      /**
       * Вызывает диалог-сообщение
       * @param {String} message - текст сообщения.
       * @param {String} [detail] - текст-расшифровка.
       * @param {String} [type] - тип сообщения (info || error || success), по умолчанию "info".
       */
      alert: function(message, detail, type) {
         if (typeof window != 'undefined') {
            IoC.resolve('ILogger').info('Deprecated', 'Core/core-functions::alert помечен как deprecated и будет удален в 3.18.');
         }

         var
            defOnClose = new Deferred(),
            types = {"error": 0, "success": 0, "info": 0},
            windowManager = require('Core/WindowManager');

         if (arguments.length == 2 && detail in types) {
            type = detail;
            detail = "";
         }
         require(['Deprecated/Controls/DialogAlert/DialogAlert'], function(dAlert){
            var myAd = new dAlert({
               message: message,
               opener: windowManager && windowManager.getActiveWindow(),
               detail: detail,
               type: type,
               handlers: {
                  onAfterClose: function() {
                     defOnClose.callback();
                  }
               }
            })
         }, function(err){
            IoC.resolve('ILogger').log('alert', err.message);

             require (['Deprecated/helpers/string-helpers'], function (strHelpers) {

                 message = strHelpers.htmlToText(message || '');
                 detail = strHelpers.htmlToText(detail || '');

                 alert(message + '\n\n' + detail);//Если не удалось создать диалог, показываем простой браузерный алёрт
             });
         });
         return defOnClose;
      }
   };
   return coreFunctions;
});
