define('Core/helpers/date-helpers', [], function(){
    /**
     * @class Core/helpers/date-helpers
     * @deprecated
     * @public
     * @author Мальцев А.А.
     */
   return /** @lends  Core/helpers/date-helpers.prototype */{
      periodTypes: {
         years: 'years',
         year: 'year',
         halfyears_different_years: 'halfyears_different_years',
         halfyear: 'halfyear',
         quarters: 'quarters',
         quarters_different_years: 'quarters_different_years',
         quarter: 'quarter',
         months: 'months',
         months_different_years: 'months_different_years',
         month: 'month',
         day: 'day'
      },
      /**
       * Проверяет строковое значение даты на соответствие формату ISO 8601.
       * @param {String} value Строковое значение даты.
       * @returns {Boolean} true, если строковое значение даты соответствует формату ISO.
       * @see dateFromISO
       * @see compareDates
       * @see getFormattedDateRange
       */
      isISODate: function (value) {
         return /^(\d{4}\-\d\d\-\d\d([tT][\d:\.]*)?)([zZ]|([+\-])(\d\d):?(\d\d))?$/g.test(value);
      },
      /**
       * Преобразует строковое значение даты в формате ISO 8601 в Date.
       * @param {String} value строковое значение даты в формате ISO 8601.
       * @returns {Date|NaN} Date - в случае успешного преобразования, NaN - в противном случае.
       * @see isISODate
       * @see compareDates
       * @see getFormattedDateRange
       */
      dateFromISO: function(value) {
         var day, tz,
            rx= /^(\d{4}\-\d\d\-\d\d([tT][\d:\.]*)?)([zZ]|([+\-])(\d\d):?(\d\d))?$/,
            p= rx.exec(value) || [];
         if(p[1]){
            day = p[1].split(/\D/).map(function(itm){
               return parseInt(itm, 10) || 0;
            });
            day[1] -= 1;
            var secToStr = (day[6] !== undefined ? day[6] + '' : '');
            if(secToStr.length > 3) {
               day[6] = parseInt(secToStr.substr(0,3), 10);
            }
            day= new Date(Date.UTC.apply(Date, day));
            if(!day.getDate()) return NaN;
            if(p[5]){
               tz= parseInt(p[5], 10)*60;
               if(p[6]) tz += parseInt(p[6], 10);
               if(p[4]== "+") tz*= -1;
               if(tz) day.setUTCMinutes(day.getUTCMinutes()+ tz);
            }
            return day;
         }
         return NaN;
      },
      /**
       * Производит сравнение двух дат без учета времени.
       * @remark
       * В качестве примера, "12.12.2012 13:23:12" и "12.12.2012 14:43:52" равны.
       * @param {Date} date1 Первая дата.
       * @param {Date} date2 Вторая дата.
       * @param {String} sign Тип сравнения дат. Например, " >= " - первая дата больше или равна второй.
       * Доступны значения: "<", "<=", "=", "==", ">=", ">" и "!=".
       * @returns {boolean}
       * @see isISODate
       * @see dateFromISO
       * @see getFormattedDateRange
       */
      compareDates: function (date1, sign, date2) {
         if (!date1 || !date2)
            return false;

         var equal = date1.toSQL() == date2.toSQL();

         switch (sign) {
            case '<':
               return !equal && date1 < date2;
            case '<=':
               return equal || date1 < date2;
            case '=':
               return equal;
            case '==':
               return equal;
            case '>=':
               return equal || date1 > date2;
            case '>':
               return !equal && date1 > date2;
            case '!=':
               return !equal;
            default:
               return false;
         }
      },

      getPeriodLength: function (start, end) {
         var oneDay = 24*60*60*1000;
         return Math.round(Math.abs((start.getTime() - end.getTime())/(oneDay))) + 1;
      },

      /**
       * Возвращает массив содержащий начальную и конечную даты текущего периода включительно
       * @param {String} periodType тип возвращаемого периода:(month, quarter, halfyear, year)
       * <ul>
       *    <li>month: текущий месяц;</li>
       *    <li>quarter: текущий квартал;</li>
       *    <li>halfyear: текущее полугодие;</li>
       *    <li>year: текущий год;</li>
       * </ul>
       * @returns {Array} Массив содержащий даты начала и конца теккущего периода
       */
      getCurrentPeriod: function (periodType) {
         var now = new Date();
         if (periodType === this.periodTypes.month) {
            return [
               new Date(now.getFullYear(), now.getMonth(), 1),
               new Date(now.getFullYear(), now.getMonth() + 1, 0)
            ]
         } else if (periodType === 'quarter') {
            return [
               new Date(now.getFullYear(), (now.getMonth()/3|0)*3, 1),
               new Date(now.getFullYear(), (now.getMonth()/3|0)*3 + 3, 0)
            ]
         } else if (periodType === 'halfyear') {
            return [
               new Date(now.getFullYear(), (now.getMonth()/6|0)*6, 1),
               new Date(now.getFullYear(), (now.getMonth()/6|0)*6 + 6, 0)
            ]
         } else if (periodType === 'year') {
            return [
               new Date(now.getFullYear(), 0, 1),
               new Date(now.getFullYear(), 11, 31)
            ]
         }
      },

      /**
       * Возвращает тип периода по переданным датам его начала и конца.
       *
       * @param {Date} startDate - первая дата
       * @param {Date} endDate - вторая дата
       * @returns {String} Строковое представление типа периода:
       *  <ul>
       *    <li>years: несколько лет;</li>
       *    <li>year: один год;</li>
       *    <li>halfyears_different_years: несколько полугодий включающие разные года;</li>
       *    <li>halfyear: одно полугодие;</li>
       *    <li>quarters: несколько кварталов в рамках 1 года;</li>
       *    <li>quarters_different_years: несколько кварталов в рамках нескольких лет;</li>
       *    <li>quarter: один квартал;</li>
       *    <li>months: несколько месяцев в рамках 1 года;</li>
       *    <li>months_different_years: несколько месяцев в рамках несколькиз лет;</li>
       *    <li>month: одно месяц;</li>
       *    <li>day: один день;</li>
       * </ul>
       */
      getPeriodType: function (startDate, endDate) {
         var
             lastDay,
             isSameYear,
             isFullMonth,
             isFullQuarter,
             isFullHalfYear,
             isFullYear,
             isOneDay,
             startDateParts,
             endDateParts;

         lastDay = new Date(endDate).setLastMonthDay().getDate();
         startDateParts = [startDate.getDate(), startDate.getMonth(), startDate.getFullYear(),
            Math.floor(startDate.getMonth()/6), Math.floor(startDate.getMonth()/3)];
         endDateParts = [endDate.getDate(), endDate.getMonth(), endDate.getFullYear(),
            Math.floor(endDate.getMonth()/6), Math.floor(endDate.getMonth()/3)];
         isSameYear = startDateParts[2] === endDateParts[2];
         isFullMonth = startDateParts[0] == 1 && endDateParts[0] == lastDay;
         isFullQuarter = isFullMonth && startDateParts[1]%3 === 0 && (endDateParts[1] + 1)%3 === 0;
         isFullHalfYear = isFullMonth && startDateParts[1]%6 === 0 && (endDateParts[1] + 1)%6 === 0;
         isFullYear = isFullMonth && startDateParts[1] === 0 && endDateParts[1] === 11;
         isOneDay = startDateParts[0] == endDateParts[0] && startDateParts[1] == endDateParts[1] && startDateParts[2] == endDateParts[2];

         // TODO: Вынести куда нибудь возвращаемые константы
         if (isFullYear) {
            if (isSameYear) {
               return this.periodTypes.year;
            } else {
               return this.periodTypes.years;
            }
         } else if (isFullHalfYear) {
            if (isSameYear) {
               if (startDateParts[3] === endDateParts[3]) {
                  return this.periodTypes.halfyear;
               } else {
                  return this.periodTypes.halfyears;
               }
            } else {
               return this.periodTypes.halfyears_different_years;
            }
         } else if (isFullQuarter) {
            if (isSameYear) {
               if (startDateParts[4] === endDateParts[4]) {
                  return this.periodTypes.quarter;
               } else {
                  return this.periodTypes.quarters;
               }
            } else {
               return this.periodTypes.quarters_different_years;
            }
         } else if (isFullMonth) {
            if (isSameYear) {
               if (startDateParts[1] == endDateParts[1]) {
                  return this.periodTypes.month;
               } else {
                  return this.periodTypes.months;
               }
            } else {
               return this.periodTypes.months_different_years;
            }
         } else if (isOneDay) {
            return this.periodTypes.day;
         }
      },

      /**
       * Возвращает длинну периода в месяцах по его текстовому представлению
       * @param periodType
       * @returns {number}
       * @see getPeriodType
       */
      getPeriodLengthInMonthByType: function (periodType) {
         switch (periodType) {
            case this.periodTypes.month:
               return 1;
            case this.periodTypes.quarter:
               return 3;
            case this.periodTypes.halfyear:
               return 6;
            case this.periodTypes.year:
               return 12;
         }
      },

      /**
       * Форматирует значения диапазона дат в строку.
       * @param {Date} startDate Первая дата.
       * @param {Date} endDate Вторая дата.
       * @returns {String} Строковое представление диапазона дат.
       * @see isISODate
       * @see dateFromISO
       * @see compareDates
       */
      getFormattedDateRange: function(startDate, endDate, options) {
         options = options || {};
         options.fullNameOfMonth = typeof options.fullNameOfMonth !== 'undefined' ? options.fullNameOfMonth : false;
         options.contractToMonth = typeof options.contractToMonth !== 'undefined' ? options.contractToMonth : true;
         options.contractToQuarter = typeof options.contractToQuarter !== 'undefined' ? options.contractToQuarter : false;
         options.contractToHalfYear = typeof options.contractToHalfYear !== 'undefined' ? options.contractToHalfYear : false;
         options.contractToYear = typeof options.contractToYear !== 'undefined' ? options.contractToYear : true;
         options.shortYear = typeof options.shortYear !== 'undefined' ? options.shortYear : true;

         var
             fullNameOfMonth = typeof options.fullNameOfMonth !== 'undefined' ? options.fullNameOfMonth : false,
             contractToMonth = typeof options.contractToMonth !== 'undefined' ? options.contractToMonth : false,
             contractToQuarter = typeof options.contractToQuarter !== 'undefined' ? options.contractToQuarter : false,
             contractToHalfYear = typeof options.contractToHalfYear !== 'undefined' ? options.contractToHalfYear : false,
             contractToYear = typeof options.contractToYear !== 'undefined' ? options.contractToYear : true,
             shortYear = typeof options.shortYear !== 'undefined' ? options.shortYear : true,
             emptyPeriodTitle = typeof options.emptyPeriodTitle !== 'undefined' ? options.emptyPeriodTitle : '...',
             lastDay,
             result = "",
             yearMask = shortYear? "'%y": " %Y",
             isSameYear,
             isFullMonth,
             isFullQuarter,
             isFullHalfYear,
             isFullYear,
             isOneDay,
             startDateParts,
             endDateParts;
         // TODO: использовать getPeriodType
         if (startDate && endDate) {
            lastDay = new Date(endDate).setLastMonthDay().getDate();
            startDateParts = [startDate.getDate(), startDate.getMonth(), startDate.getFullYear(),
               Math.floor(startDate.getMonth()/6), Math.floor(startDate.getMonth()/3)];
            endDateParts = [endDate.getDate(), endDate.getMonth(), endDate.getFullYear(),
               Math.floor(endDate.getMonth()/6), Math.floor(endDate.getMonth()/3)];
            isSameYear = startDateParts[2] === endDateParts[2];
            isFullMonth = startDateParts[0] == 1 && endDateParts[0] == lastDay;
            isFullQuarter = isFullMonth && isSameYear && startDateParts[1]%3 === 0 && (endDateParts[1] + 1)%3 === 0;
            isFullHalfYear = isFullMonth && isSameYear && startDateParts[3] === endDateParts[3] && startDateParts[1]%6 === 0 && (endDateParts[1] + 1)%6 === 0;
            isFullYear = isFullMonth && startDateParts[1] === 0 && endDateParts[1] === 11;
            isOneDay = startDateParts[0] == endDateParts[0] && startDateParts[1] == endDateParts[1] && startDateParts[2] == endDateParts[2];

            if (isFullYear) {
               if (isSameYear) {
                  result = startDate.strftime("%Y");
               } else if (contractToYear) {
                  result = startDate.strftime("%Y") + '-' + endDate.strftime("%Y");
               } else {
                  result = startDate.strftime("%d.%m.%y") + ' - ' + endDate.strftime("%d.%m.%y");
               }
            } else if (contractToHalfYear && isFullHalfYear) {
               result = startDate.strftime("%E полугодие" + yearMask);
            } else if (contractToQuarter && isFullQuarter) {
               if (startDateParts[4] === endDateParts[4]) {
                  result = startDate.strftime("%Q квартал" + yearMask);
               } else {
                  result = startDate.strftime("%Q") + '-' + endDate.strftime("%Q квартал" + yearMask);
               }
            } else if (isFullMonth) {
               if (isSameYear) {
                  if (startDateParts[1] == endDateParts[1]) {
                     result = startDate.strftime("%B" + yearMask);
                  } else {
                     if (contractToMonth) {
                        result = startDate.strftime('%f') + ' - ' + endDate.strftime('%f' + yearMask);
                     } else {
                        result = startDate.strftime('%d.%m.%y') + ' - ' + endDate.strftime('%d.%m.%y');
                     }
                  }
               } else {
                  if (contractToMonth) {
                     result = startDate.strftime('%f' + yearMask) + ' - ' + endDate.strftime('%f' + yearMask);
                  } else {
                     result = startDate.strftime('%d.%m.%y') + ' - ' + endDate.strftime('%d.%m.%y');
                  }
               }
            } else if (isOneDay) {
               if (fullNameOfMonth) {
                  result = startDate.strftime("%i %q'%y");
               } else {
                  result = startDate.strftime("%d.%m.%y");
               }
            } else {
               result = startDate.strftime("%d.%m.%y") + ' - ' + endDate.strftime("%d.%m.%y");
            }
         } else if (startDate) {
            result = startDate.strftime("%d.%m.%y") + ' - ...';
         } else if (endDate) {
            result = '... - ' + endDate.strftime("%d.%m.%y");
         } else {
            result = emptyPeriodTitle;
         }
         return result;
      }
   }
});