define('Core/helpers/event-helpers', [
   'Core/constants'
], function(constants) {
   //MOVE_TO КРАЙНОВ

    /**
     * @class Core/helpers/event-helpers
     * @public
     * @author Мальцев А.А.
     */
   var eventHelpers = /** @lends  Core/helpers/event-helpers.prototype */ {
      /**
       * Устанавливает обработчик на событие keydown.
       * @remark
       * Для одинаковой обработки событий в разных браузерах.
       * Актуально для старых версий, например, Opera до 12.10.
       * @param {jQuery} object Объект, на который вешается обработчик события нажатия клавиши.
       * @param {Function} callback Функция, которая сработает при нажатии клавиши.
       * @see checkCapsLock
       * @see wheel
       * @see getTouchEventCoords
       * @see getTouchEventClientCoords
       */
      keyDown: function (object, callback) {
         object[constants.compatibility.correctKeyEvents ? 'keydown' : 'keypress'](function (e) {
            if (!constants.compatibility.correctKeyEvents) {
               if (e.which === e.keyCode) {
                  if (e.which >= 97 && e.which <= 122) {
                     e.which = e.keyCode - 32;
                  }
                  else if (e.which in constants.operaKeys) {
                     e.which = constants.key[constants.operaKeys[e.which]];
                  }
               }
               if (e.which === 0) {
                  e.which = e.keyCode;
               }
               e.keyCode = e.which;
            }
            callback(e);
         });
      },
      /**
       * Проверяет нажата ли клавиша CapsLock.
       * @param {Event} keypressEvent jQuery-событие keypress, для которого проверяется нажат капс-лок или нет.
       * @returns {Boolean} Зажат ли CapsLock. undefined возвращается, когда определить невозможно.
       * @see keyDown
       * @see wheel
       * @see getTouchEventCoords
       * @see getTouchEventClientCoords
       */
      checkCapsLock: function (keypressEvent) {
         if (!keypressEvent || !keypressEvent.which)
            return undefined;
         var asciiCode = keypressEvent.which;
         var letter = String.fromCharCode(asciiCode);
         var upper = letter.toUpperCase();
         var lower = letter.toLowerCase();
         var shiftKey = keypressEvent.shiftKey;

         // Если верхние и нижние регистры равны, то нет возможности определить нажат ли капс-лок
         if (upper !== lower) {

            // Если вводится верхний символ при ненажатом шифте, значит капс-лок включен
            if (letter === upper && !shiftKey) {
               return true;
               // Если нижний, то выключен.
            } else if (letter === lower && !shiftKey) {
               return false;
               // Если нижний при нажажатом шифте, то включен
            } else if (letter === lower && shiftKey) {
               return true;
            } else if (letter === upper && shiftKey) {
               if (navigator.platform.toLowerCase().indexOf("win") !== -1) {
                  // Если на Windows, то выключен
                  return false;
               } else {
                  if (navigator.platform.toLowerCase().indexOf("mac") !== -1) {
                     // Если на Mac, то выключен
                     return false;
                  } else {
                     return undefined;
                  }
               }
            } else {
               return undefined;
            }
         } else {
            return undefined;
         }
      },
      /**
       * Устанавливает обработчик на вращение колеса мыши.
       * @param {jQuery} element Элемент, над которым нужно вращать колесо мыши. В некоторых браузерах также необходимо, чтобы страница имела фокус.
       * @param {Function} callback Функция, которая будет вызвана. Получит один аргумент - event, объект jQuery-события. У него будет задано свойство wheelDelta. При вращении колеса вниз значение будет отрицательным, вверх - положительным. Значение примерно равно количеству пикселей, на которое будет проскроллен блок, но не гарантируется, что в разных браузерах это значение будет одинаковым.
       * @return {jQuery}
       * @see keyDown
       * @see checkCapsLock
       * @see getTouchEventCoords
       * @see getTouchEventClientCoords
       */
      wheel: function (element, callback) {
         var support = constants.compatibility.wheel;
         return element.bind(support, function (event) {
            var originalEvent = event.originalEvent;
            if (support === 'wheel') {
               event.wheelDelta = -originalEvent.deltaY;
               if (originalEvent.deltaMode === 1) {
                  event.wheelDelta *= 40;
               }
            }
            else if (support === 'DOMMouseScroll') {
               event.wheelDelta = -originalEvent.detail * 40;
            } else if (support === 'mousewheel') {
               event.wheelDelta = originalEvent.wheelDelta;
            }
            callback.call(this, event);
         });
      },
      /**
       * Возвращает координаты курсора мыши из jQuery-события.
       * @remark
       * Координаты возвращаются относительно левого верхнего угла <strong>документа</strong>, как в функции jQuery.offset().
       * @param {Object} event jQuery-событие.
       * @returns {{x: Number, y: Number}}
       * @see keyDown
       * @see checkCapsLock
       * @see wheel
       * @see getTouchEventClientCoords
       */
      getTouchEventCoords: function (event) {
         var x = event.pageX,
            y = event.pageY;
         if (!x) {
            var touch = event.originalEvent.touch || (event.originalEvent.touches && event.originalEvent.touches[0]) ||
                (event.originalEvent.changedTouches && event.originalEvent.changedTouches[0]);
            if (touch) {
               x = touch.pageX;
               y = touch.pageY;
            }
         }
         return {'x': x, 'y': y};
      },
      /**
       * Возвращает координаты курсора мыши из jQuery-события.
       * @remark
       * Координаты возвращаются относительно левого верхнего угла <strong>Window</strong>, как в функции element.getBoundingClientRect.
       * @param {Object} event jQuery-событие.
       * @returns {{x: Number, y: Number}}
       * @see keyDown
       * @see checkCapsLock
       * @see wheel
       * @see getTouchEventCoords
       */
      getTouchEventClientCoords: function(event) {
         var x = event.clientX,
            y = event.clientY;
         if (x === undefined) {
            var touch = event.originalEvent.touch || event.originalEvent.touches && event.originalEvent.touches[0] ||
               event.originalEvent.changedTouches && event.originalEvent.changedTouches[0];
            if (touch) {
               x = touch.clientX;
               y = touch.clientY;
            }
         }
         return {'x': x, 'y': y};
      }
   };

   return eventHelpers;
});