define('Core/vdom/Synchronizer/Synchronizer',
   [  'require',
      'Core/vdom/Utils/Monad',
      'Core/vdom/Synchronizer/resources/DirtyChecking',
      'Core/vdom/Synchronizer/resources/DOMEnvironment',
      'Core/helpers/getModuleNameWithPlugin',
      'Core/helpers/Function/runDelayed',
      'Core/helpers/Object/isEmpty',
      'Core/helpers/Array/uniq',
      'Core/Serializer',
      'Core/vdom/vdomutils',
      'View/Runner/common',
      'js'],
   function(require,
            Monad,
            DirtyChecking,
            DOMEnvironment,
            getModuleNameWithPlugin,
            runDelayed,
            isEmptyObject,
            uniqueArray,
            Serializer,
            vdomutils,
            utils
   ) {

   'use strict';

   var Debug;

   var
      DirtyKind = DirtyChecking.DirtyKind,
      rebuildNode = DirtyChecking.rebuildNode,
      createNode  = DirtyChecking.createNode,
      RebuildResultWriter = DirtyChecking.RebuildResultWriter,
      MAX_REBUILD = 5,
      isDebug = false;

   if (isDebug) {
      require(['Core/vdom/Synchronizer/resources/Debug'], function(_Debug){
         Debug = _Debug;
      });
   }
   function forEachNodeParents(node, fn) {
      var parent = node.parent;
      while (parent) {
         fn(parent);
         parent = parent.parent;
      }
   }

   function checkControlNodes(controlNodes, controlNode) {
      if (controlNodes && controlNodes.length > 0) {
         for (var i = 0; i < controlNodes.length; i++) {
            if (controlNode.id === controlNodes[i].id) {
               return true;
            }
         }
      }
      return false;
   }

   function findNodeForControlNode(rootNode, controlNode) {
      var children, node;
      if (checkControlNodes(rootNode.controlNodes, controlNode)) {
         return rootNode;
      }
      if (rootNode.children && rootNode.children.length > 0) {
         children = rootNode.children;
         for (var i = 0; i < children.length; i++) {
            if (children[i].controlNodes) {
               if (checkControlNodes(children[i].controlNodes, controlNode)) {
                  node = children[i];
                  return node;
               }
               node = findNodeForControlNode(children[i], controlNode);
            }
         }
      }
      return node;
   }

   function loadControlClassIfString(controlClass) {
      if (Object.prototype.toString.call(controlClass) === '[object String]') {
         return require(getModuleNameWithPlugin(controlClass));
      } 
      return controlClass;
   }

   /**
    * Функция для сбора массива созданных/измененных контрол-нод
    * @param arr
    */
   function mapIds(arr) {
      return arr.map(function reduceIdsReduce(controlNode) { return controlNode.id });
   }

   var
      _rootNodes = [],
      _environments = [],
      _controlNodes = {},
      _dirties = {};

   var VDomSynchronizer = {
      _rebuild: function() {
         function doRebuild(timeMeasureFn) {
            var dirties = _dirties;

            if (Object.keys(dirties).length === 0) {
               /*Ключей нет, значит мы пытаемся обновлять вторую корневую ноду, которую уже обновили с первой*/
               return {
                  dirties: dirties,
                  oldRoots: _rootNodes,
                  newRoots: _rootNodes,
                  rebuildChanges: null,
                  applyResults: null
               };
            }


            _dirties = {};

            var
               oldRoots = _rootNodes,
               rootsRebuild = Monad.mapM(oldRoots, function rebuildOneRootNode(node, i) {
                  node.requestDirtyCheck = requestDirtyCheck.bind(node);
                  return rebuildNode(_environments[i], dirties, node, undefined, true);
               }, RebuildResultWriter, this);


            if (rootsRebuild.callback) {
               rootsRebuild.addCallback(function (rootRebuildVal) {
                  var
                     newRoots = rootRebuildVal.value,
                     rebuildChanges = rootRebuildVal.memo,
                     rebuildChangesIds = mapIds(newRoots),
                     /**
                      * Типы нод, для которых нужно запустить _afterUpdate
                      * @type {{createdNodes: boolean, updatedChangedNodes: boolean, selfDirtyNodes: boolean}}
                      */
                     rebuildChangesFields = {
                        'createdNodes': true,
                        'updatedChangedNodes': true,
                        'selfDirtyNodes': true
                     },
                     applyResults;



                  //Сохраняем id созданных/обновленных контрол нод, чтобы вызвать afterUpdate или afterMound только у них
                  for (var field in rebuildChangesFields) {
                     if (rebuildChanges[field].length > 0) {
                        rebuildChangesIds = rebuildChangesIds.concat(mapIds(rebuildChanges[field]));
                     }
                  }

                  rebuildChangesIds = uniqueArray(rebuildChangesIds);

                  var applyNewVNode = function(env, i) {
                     return env.applyNewVNode(newRoots[i].fullMarkup, rebuildChangesIds);
                  };
                  if (timeMeasureFn) {
                     applyNewVNode = timeMeasureFn(applyNewVNode);
                  }

                  rebuildChanges.createdNodes.forEach(function(node) {
                     node.requestDirtyCheck = requestDirtyCheck.bind(node);
                     _controlNodes[node.id] = node;
                  }, this);

                  rebuildChanges.destroyedNodes.forEach(function(node) {
                     delete _controlNodes[node.id];
                  }, this);

                  applyResults = _environments.map(applyNewVNode);

                  _rootNodes = newRoots;

                  return {
                     dirties: dirties,
                     oldRoots: oldRoots,
                     newRoots: newRoots,
                     rebuildChanges: rebuildChanges,
                     applyResults: applyResults
                  };
               });
            } else {
               var
                  newRoots = rootsRebuild.value,
                  rebuildChanges = rootsRebuild.memo,
                  rebuildChangesIds = mapIds(newRoots).map(function (current) {
                     /*Запускаем afterUpdate только для тех, кто реально Dirty
                      * Остальные - это ChildDirty*/
                     return (dirties[current] & DirtyKind.DIRTY) ? current : undefined;
                  }),
                  /**
                   * Типы нод, для которых нужно запустить _afterUpdate
                   * @type {{createdNodes: boolean, updatedChangedNodes: boolean, selfDirtyNodes: boolean}}
                   */
                  rebuildChangesFields = {
                     'createdNodes': true,
                     'updatedChangedNodes': true,
                     'selfDirtyNodes': true
                  },
                  applyResults;


               //Сохраняем id созданных/обновленных контрол нод, чтобы вызвать afterUpdate или afterMound только у них
               for (var field in rebuildChangesFields) {
                  if (rebuildChanges[field].length > 0) {
                     rebuildChangesIds = rebuildChangesIds.concat(mapIds(rebuildChanges[field]));
                  }
               }

               rebuildChangesIds = uniqueArray(rebuildChangesIds);

               var applyNewVNode = function(env, i) {
                  return (newRoots[i]&&newRoots[i].fullMarkup)?env.applyNewVNode(newRoots[i].fullMarkup, rebuildChangesIds):null;
               };
               if (timeMeasureFn) {
                  applyNewVNode = timeMeasureFn(applyNewVNode);
               }

               rebuildChanges.createdNodes.forEach(function(node) {
                  node.requestDirtyCheck = requestDirtyCheck.bind(node);
                  _controlNodes[node.id] = node;
               }, this);

               rebuildChanges.destroyedNodes.forEach(function(node) {
                  delete _controlNodes[node.id];
               }, this);

               applyResults = _environments.map(applyNewVNode);

               _rootNodes = newRoots;

               return {
                  dirties: dirties,
                  oldRoots: oldRoots,
                  newRoots: newRoots,
                  rebuildChanges: rebuildChanges,
                  applyResults: applyResults
               };
            }

         }

         var rebuildImpl = isDebug ? Debug.withLogRebuild(doRebuild, {silent: true}) : doRebuild, i;

         for (i = 0; i !== MAX_REBUILD; i++) {
            rebuildImpl.call(this);
            if (isEmptyObject(_dirties)) {
               break;
            }
         }
         if (i === MAX_REBUILD) {
            throw new Error('SBIS3.CORE.VDOM.Synchronizer: i === MAX_REBUILD');
         }
      },
      controlStateChangedCallback: function(controlId) {
         VDomSynchronizer.requestRebuild(controlId);
      },
      mountControlToDOM: function(controlClass, options, domElement) {
         domElement = domElement[0] ? domElement[0] : domElement;
         var
            hasMountedComponent,
            rootAttrs = {
               'data-component': domElement.getAttribute('data-component')
            },
            serializedId = domElement.getAttribute('data-component-root-id'),
            jsModules = serializedId && window.vdomJsModules[serializedId],
            controlClass = loadControlClassIfString(controlClass);
         
            var doMount = function doMount() {
               var
                  controlNode, environment, state,
                  Slr = new Serializer();


               environment = new DOMEnvironment(domElement, this.controlStateChangedCallback, rootAttrs, serializedId);

               controlNode = createNode(controlClass, { user: options }, undefined, environment, null, state);
               controlNode.requestDirtyCheck = requestDirtyCheck.bind(controlNode);

               _controlNodes[controlNode.id] = controlNode;

               _environments.push(environment);
               _rootNodes.push(controlNode);
               
               // храним в рутовом виртуальном компоненте ссылку на environment, 
               // иначе до него не докопаться, только после ребилда, а это слишком поздно для системы фокусов
               controlNode.control.environment = environment;

               var carrier = vdomutils.getReceivedState(controlNode, {controlProperties: options}, Slr);
               controlNode['element'] = domElement;
               /**
                * Сделать final проверку
                */
               if (controlNode.control.saveOptions) {
                  controlNode.control.saveOptions(controlNode.options, controlNode);
               } else {
                  /**
                   * Поддержка для совместимости версий контролов
                   */
                  controlNode.control._options = controlNode.options;
                  controlNode.control._container = $(controlNode.element);
               }
               /**
                * Обработка асинхронного построения для рутовой ноды
                * @type {*}
                */
               if (carrier) {
                  carrier.addCallback(function (receivedState) {
                     controlNode.receivedState = receivedState;
                     controlNode.control = utils.plainMerge(controlNode.control, receivedState);
                     this.requestRebuild(controlNode.id);
                     return receivedState;
                  }.bind(this));
               } else {
                  this.requestRebuild(controlNode.id);
               }

            }.bind(this);

         hasMountedComponent = _environments.some(function (env) {
            return (env instanceof DOMEnvironment) && env.getDOMNode() === domElement;
         });

         if (hasMountedComponent) {
            throw new Error('На этом DOM-элементе уже есть смонтированный корневой компонент');
         }
         if (jsModules && jsModules.length > 0) {
            require(jsModules, doMount);
         } else {
            doMount();
         }
      },

      unMountControlFromDOM: function unMountControlFromDOM(control, node) {

         var i, j,
            domElement = node[0] ? node[0] : node;
         for (i=0;i<_rootNodes.length;i++) {
            if (domElement === _rootNodes[i].environment._rootDOMNode) {
               delete _controlNodes[_rootNodes[i].id];
               var env = _environments.splice(i,1);
               _rootNodes.splice(i,1);
               if (!control.isDestroyed()) {
                  control.destroy();
               }
               env[0].destroy();
               break;
            }
         }
      },

      requestRebuild: function(controlId) {
         //контрол здесь точно должен найтись, или быть корневым - 2 варианта попадания сюда:
         //    из конструктора компонента (тогда его нет, но тогда синхронизация активна) - тогда не нужно с ним ничего делать
         //    из внутреннего события компонента, меняющего его состояние, и вызывающего requestRebuild
         var
            controlNode = _controlNodes[controlId],
            dirties = _dirties,
            self = this;

         if (controlNode) {
            dirties[controlId] |= DirtyKind.DIRTY;

            forEachNodeParents(controlNode, function (parent) {
               dirties[parent.id] |= DirtyKind.CHILD_DIRTY;
            });

            if (!controlNode.environment._haveRebuildRequest) {
               controlNode.environment._haveRebuildRequest = true;
               runDelayed(function requestRebuildDelayed() {
                  var needSetFocusControl;
                  // ищем в каком компоненте находится фокус
                  for (var nodeId in _controlNodes) {
                     if (_controlNodes.hasOwnProperty(nodeId)) {
                        var control = _controlNodes[nodeId];
                        if (control && control.environment && !control.environment._rootDOMNode){
                           delete _controlNodes[nodeId];
                           continue;
                        }
                        if (control.environment && control.environment._rootDOMNode.contains(document.activeElement)) {
                           needSetFocusControl = control;
                        }
                     }
                  }
                  
                  self._rebuild();

                  controlNode.environment._haveRebuildRequest = false;

                  controlNode.environment.addTabListener();
                  // если фокус слетел, и есть куда его восстановить, восстановим его
                  if (needSetFocusControl && document.activeElement === document.body || controlNode.control.__$focusing) {
                     controlNode.environment.autofocusChanges.call(this, controlNode);
                  }
               }.bind(this));
            }
         }
      },

      setDebugMode: function (val) {
         isDebug = val;
      }
   };

   function requestDirtyCheck() {
      VDomSynchronizer.requestRebuild(this.id);
   }

   return VDomSynchronizer;
});
