define('Core/vdom/Synchronizer/resources/Hooks', [
   'Core/helpers/Function/shallowClone',
   'Core/vdom/Utils/Functional',
   'View/Runner/expressions/attr',
   'View/Runner/expressions/event'

], function (shallowClone, functionalHelpers, utils, eventExpressions) {
   'use strict';

   function getEventHandlerBase(eventDescr) {
      function eventUtil(event) {
         var evArgs = event.nativeEvent && event.nativeEvent.args || [],
            args = evArgs.concat(eventDescr.args);
         eventDescr.fn.apply(eventDescr.controlNode.control, [event].concat(args));
         var res = event.nativeEvent.result;
         if(res && res.addCallback) {
            res.addCallback(function(){
               eventDescr.controlNode.requestDirtyCheck();
            });
         } else {
            eventDescr.controlNode.requestDirtyCheck();
         }
         return [eventDescr.controlNode.control, event];
      }

      if (eventDescr.name === 'event') {
         return eventUtil;
      } else {
         throw new Error('getEventHandlerBase: wrong event type');
      }
   }

   var ControlNodeHook = function (controlNode) {
      this.controlNode = controlNode;
   };

   function pluck(propName) {
      return function (obj) {
         return obj[propName];
      };
   }

   // Хук именованных элементов внутри контрола
   var NamedNodeHook = function NamedNodeHookC(name, control) {
      this.nameForNode = name;
      this.control = control;
   };

   NamedNodeHook.hookName = 'named-hook';
   NamedNodeHook.attributeName = 'name';
   NamedNodeHook.restrictedTags = {
      'component': true,
      'option'   : true
   };

   NamedNodeHook.prototype.hook = function namedNodeHook(node) {
      if (this.control && this.control._children) {
         this.control._children[this.nameForNode] = node;
      }
   };



   function updateControlNodes(element, controlNode, prevNextControlNode, fn) {
      var
         controlNodes = element['controlNodes'] || [],
         controlNodeIdx = controlNodes.indexOf(controlNode),
         prevNextId = prevNextControlNode && prevNextControlNode.id,
         ids, modNames;

      fn(controlNodes, controlNodeIdx);

      element['controlNodes'] = controlNodes;
      for(var i=0; i<controlNodes.length; i++) {
         controlNodes[i]['element'] = element;
         controlNodes[i].control.saveOptions(controlNodes[i].control._options, controlNodes[i]);
      }

      if (prevNextId !== controlNode.id) {
         ids = controlNodes.map(pluck('id')).join(',');
         //modNames = controlNodes.map(function (node) { return node.control.get('moduleClassName'); }).join(',');
         if (element.cids !== ids) {
            element.cids = ids;
            element.setAttribute('data-cids', ids);
         }

         //element.setAttribute('data-cmods', modNames);
      }
   }

   ControlNodeHook.prototype.hook = function (element, propertyName, prev) {
      if ((prev !== this) && (!prev || prev.controlNode !== this.controlNode)) {
         updateControlNodes(element, this.controlNode, prev && prev.controlNode, function (controlNodes, controlNodeIdx) {
            var haveNode = controlNodeIdx !== -1;

            if (haveNode) {
               throw new Error('Хук для контрола ' + this.controlNode.id + ' на узле уже есть');
            }

            if (prev && prev.type !== this.type) {
               throw new Error('Хук для контрола: previousValue.type !== this.type');
            }

            controlNodes.push(this.controlNode);
         }.bind(this));
      }
   };

   ControlNodeHook.prototype.unhook = function (element, propertyName, next) {
      if ((next !== this) && (!next || next.controlNode !== this.controlNode)) {
         updateControlNodes(element, this.controlNode, next && next.controlNode, function (controlNodes, controlNodeIdx) {
            var haveNode = controlNodeIdx !== -1;

            if (!haveNode) {
               throw new Error('Хук для контрола ' + this.controlNode.id + ' на узле должен быть');
            }

            if (next && next.type !== this.type) {
               throw new Error('Хук для контрола: nextValue.type !== this.type');
            }

            controlNodes.splice(controlNodeIdx, 1);
         }.bind(this));
      }
   };

   ControlNodeHook.prototype.type = 'ControlNodeHook';

   function setControlNodeHook(controlNode) {
      return function (tagName, props, children, key) {
         var
            newProps = shallowClone(props),
            id = controlNode.id,
            hookName = 'control-' + id;

         if (newProps[hookName]) {
            throw new Error('На узле уже есть хук для контрола ' + id);
         }

         newProps[hookName] = new ControlNodeHook(controlNode);

         return [tagName, newProps, children, key];
      };
   }

   ///////////////////////////////////////////////////////////////////////////

   function eventDescrsEqual(d1, d2) {
      var result = d1.name === d2.name && d1.value === d2.value && !!d1.args === !!d2.args, i, ln;
      if (result && d1.args) {
         result = d1.args.length === d2.args.length;
         if (result) {
            for (i = 0, ln = d1.args.length; result && i !== ln; i++) {
               result = d1.args[i] === d2.args[i];
            }
         }
      }
      return result;
   }

   function eventDescrsArrEqual(da1, da2) {
      if (isFunction(da1) && isFunction(da2)) {
         result = !!da1 === !!da2;
      } else {
         var result = !!da1 === !!da2, i, ln;
         if (result && da1) {
            result = da1.length === da2.length;
            if (result) {
               for (i = 0, ln = da1.length; i !== ln && result; i++) {
                  result = eventDescrsEqual(da1[i], da2[i]);
               }
            }
         }
      }
      return result;
   }

   function isFunction(entity) {
      return Object.prototype.toString.call(entity) === '[object Function]';
   }

   function reverseArray(arr) {
      arr = arr.arguments || arr;
      var ln = arr.length, result = new Array(ln), j = ln - 1, i;

      if (typeof ln !== 'number') {
         throw new Error('reverseArray - wrong arg');
      }

      for (i = 0; i !== ln; i++) {
         result[j] = arr[i];
         j--;
      }

      return result;
   }

   function getEventHandler(eventDescrsArr) {
      var
         descrFns = reverseArray(eventDescrsArr).map(getEventHandlerBase),
         handler = functionalHelpers.composeWithResultApply(descrFns);

      return function(event) {
         return handler(event);
      };
   }

   ///////////////////////////////////////////////////////////////////////////

   function BubblingPhaseEventHook(eventName, eventDescrArr) {
      this.eventName = eventName;
      this.eventDescrArr = eventDescrArr;
      this.handlerFn = null;
   }

   BubblingPhaseEventHook.prototype.type = 'BubblingPhaseEventHook';

   BubblingPhaseEventHook.prototype.hook = function (element, propertyName, previousValue) {
      var prev = previousValue;
      if ((prev !== this) && (!prev || !eventDescrsArrEqual(prev.eventDescrArr, this.eventDescrArr))) {
         this.handlerFn = getEventHandler(this.eventDescrArr);
         element.addEventListener(this.eventName, this.handlerFn);
      }
   };

   BubblingPhaseEventHook.prototype.unhook = function (element, propertyName, nextValue) {
      var next = nextValue;
      if ((next !== this) && (!next || !eventDescrsArrEqual(this.eventDescrArr, next.eventDescrArr))) {
         element.removeEventListener(this.eventName, this.handlerFn);
      }
   };

   ///////////////////////////////////////////////////////////////////////////

   function CapturePhaseEventHook(eventName, eventDescrArr, environment, controlNode) {
      this.eventName = eventName;
      this.environment = environment;
      this.eventDescrArr = eventDescrArr.map(function (val) {
         val.controlNode = controlNode;
         return val;
      });
   }



   CapturePhaseEventHook.prototype.type = 'CapturePhaseEventHook';

   CapturePhaseEventHook.prototype.hook = function (element, propertyName, previousValue) {
      var prev = previousValue,
         handler;
      if ((prev !== this) && (!prev || !eventDescrsArrEqual(prev.eventDescrArr, this.eventDescrArr))) {
         if (!element.eventProperties) {
            element.eventProperties = {};
            element.eventPropertiesCnt = 0;
         }

         element.eventProperties[propertyName] = this.eventDescrArr;
         element.eventPropertiesCnt++;
         this.environment.addCaptureEventHandler(this.eventName, element);
      }
   };

   CapturePhaseEventHook.prototype.unhook = function (element, propertyName, nextValue) {
      var next = nextValue;
      if ((next !== this) && (!next || !eventDescrsArrEqual(this.eventDescrArr, next.eventDescrArr))) {
         this.environment.removeCaptureEventHandler(this.eventName);

         if (element.eventProperties) {
            delete element.eventProperties[this.eventName];
            element.eventPropertiesCnt--;
            if (element.eventPropertiesCnt === 0) {
               delete element.eventPropertiesCnt;
               delete element.eventProperties;
            }
         }
      }
   };

   ///////////////////////////////////////////////////////////////////////////

   function setEventHooks(domEnvironment) {
      return function(tagName, props, children, key, controlNode) {
         var
            haveEvents = Object.keys(props).some(eventExpressions.isEvent),
            newProps = {};

         if (haveEvents) {
            for (var propName in props) {
               if (props.hasOwnProperty(propName)) {
                  var
                     propVal = props[propName],
                     newProp;
                  if (eventExpressions.isEvent(propName)) {
                     if (domEnvironment.haveCapturePhase()) {
                        newProp = new CapturePhaseEventHook(eventExpressions.getEventName(propName), propVal, domEnvironment, controlNode);
                     } else {
                        newProp = new BubblingPhaseEventHook(eventExpressions.getEventName(propName), propVal, controlNode);
                     }
                  } else {
                     newProp = propVal;
                  }
                  newProps[propName] = newProp;
               }
            }
         } else {
            newProps = props;
         }

         return [tagName, newProps, children, key];
      };
   }

   ///////////////////////////////////////////////////////////////////////////

   return {
      setControlNodeHook: setControlNodeHook,
      setEventHooks: setEventHooks,
      namedNodeHook: NamedNodeHook
   };
});