
define('Lib/Control/FileStorageLoader/FileStorageLoader', [
   "Core/helpers/random-helpers",
   "Core/EventBus",
   "Lib/Control/FileLoader/FileLoader",
   "WS.Data/Di",
   "Core/Deferred",
   "i18n!Lib/Control/FileStorageLoader/FileStorageLoader"
], function(randomHelpers, cEventBus, FileLoader, Di, Deferred) {

   'use strict';

   var
      uploaderInstance,
      uploadedFiles = {};

   function uploadErrorHandler(event, data) {
      var fileName = data.UploadName,
         message = data.Message;

      if (uploadedFiles[fileName]) {
         this._logError(message, 'FileStorageLoader');
         uploadedFiles[fileName].callback({code:500});
         delete uploadedFiles[fileName];
      }
   }
   function uploadFileStorageErrorHandler(event, data) {
      var fileName = data.UploadName,
         message = data.Message,
         guid = data.UploadGuid;

      if (uploadedFiles[fileName]) {
         this._getFileLoaderPlugin().addCallback(function(plugin) {
            plugin.call('сancelUploadToBigFileStorage', {Guid:guid});
         });
         this._logError(message, 'FileStorageLoader');
         uploadedFiles[fileName].callback({code:500});
         delete uploadedFiles[fileName];
      }
   }
   function uploadFileStorage(event, data) {
      var fileName = data.UploadName,
         url = data.UploadId;

      if (uploadedFiles[fileName]) {
         uploadedFiles[fileName].callback({code:201, url: url});
         delete uploadedFiles[fileName];
      }
   }

   /**
    * Модуль 'Компонент загрузки файла'.
    * Умеет загружать файлы  в файловое хранилище
    * Использует стандартный input для выбора файлов
    * В случае вызова диалога выбора из кода методом selectFile,
    * Используется для Chrome плагин. Для IE8 всегда используется ActiveX
    * **С версии 3.17.310 использовать не рекомендуется.**
    * @class Lib/Control/FileStorageLoader/FileStorageLoader
    * @extends Lib/Control/FileLoader/FileLoader
    * @author Крайнов Д.О.
    * @category Loaders
    * @control
    * @public
    * @nonVisual
    */
   var FileStorageLoader = FileLoader.extend(/** @lends Lib/Control/FileStorageLoader/FileStorageLoader.prototype */{
      _loadDeferred: undefined,
      $constructor: function() {
         var channel = cEventBus.channel('SbisPluginEvent');
         channel.subscribe('uploadToBigFileError', uploadErrorHandler, this);
         channel.subscribe('uploadToBigFileStorageError', uploadFileStorageErrorHandler, this);
         channel.subscribe('uploadToBigFileStorageEvent', uploadFileStorage, this);
      },
      /**
       * Непосредственно загружает файл на файловое хранилище
       * @param status
       * @param filePath
       * @private
       */
      _postFileStorageFinished: function(status, filePath) {
         var result = '';
         //Проверяем, как закончился запрос
         if (status !== 201) { //Если полученный статус не 201, то картинка на сервере не сохранилась
            switch (status) { //Ругаемся в зависимости от статуса запроса
               case 413:
                  result = rk('Размер выбранного изображения превышает максимально допустимый.');
                  break;
               default:
                  result = rk('Не удалось загрузить изображение. Повторите попытку.');
            }
         } else {
            result = rk('Изображение успешно загружено на сервер');
         }
         this._postFinished(JSON.stringify({
            jsonprc: '2.0',
            result: {
               code: status,
               message: result,
               filePath: filePath
            },
            id: 1,
            protocol: 3
         }));
      },

      /**
       * Загрузка файла на файловое хранилище
       * Не поддерживает множественный выбор файлов
       * @private
       */
      _postInput: function(files) {
         var self = this, file,
             xhr = new XMLHttpRequest(),
             filePath = '/cstorage/img?id=' + randomHelpers.createGUID(); //Генерируем псевдо-UID (подробности в описании функции "createGUID")

         // Берем первый файл
         for (var i in files) {
            if (files.hasOwnProperty(i) && typeof(i) !== 'function') {
               file = files[i];
               break;
            }
         }

         this._upload(function() {
            self._loading = true;
            //Подготавливаем XHR
            xhr.open('PUT', filePath, true);
            xhr.setRequestHeader('Content-Disposition', "inline; filename*=utf-8''" + encodeURI(file.name));
            xhr.setRequestHeader('Content-Type', file.type);
            xhr.onreadystatechange = function() { //Выполняется по завершению запроса
               if (xhr.readyState == 4) {
                  self._postFileStorageFinished(xhr.status, filePath);
               }
            };
            xhr.send(file);
         });
      },

      destroy: function() {
         var channel = cEventBus.channel('SbisPluginEvent');
         channel.unsubscribe('uploadToBigFileError', uploadErrorHandler, this);
         channel.unsubscribe('uploadToBigFileStorageError', uploadFileStorageErrorHandler, this);
         channel.unsubscribe('uploadToBigFileStorageEvent', uploadFileStorage, this);
         FileStorageLoader.superclass.destroy.apply(this, arguments);
      },

      //////////////////////////////////////////
      /// Вся работа с плагином тут
      //////////////////////////////////////////

      /**
       * Вызывается перед загрузкой файлов
       * @param {String} uploadedFile Выбранный файл
       */
      _postPlugin: function(uploadedFile) {
         if (!uploadedFile || uploadedFile == 'undefined') {
            return;
         }

         uploadedFile = uploadedFile[0];
         this._notify('onChange', this._curval = decodeURIComponent(uploadedFile));

         var self = this;
         this._upload(function() {
            if (!self._isDestroyed) {
               self._loading = true;
               if (self._options.showIndicator) {
                  self._showIndicator();
               }

               var requestParams = self._getRequestParameters();
               requestParams.fileName = uploadedFile;
               var ret = self._uploadFile(requestParams);
               if (ret instanceof  Deferred) {
                  ret.addErrback(function(error) {
                     if (self._options.showIndicator) {
                        self._hideIndicator();
                     }
                     self._loading = false;
                     self._logError(error);
                  });
               }
            }
         });
      },

      /**
       * Вызвает метод загрузки файлов только для fileStorage
       * @param requestParams
       * @private
       */
      _uploadFile: function(requestParams) {
         var self = this;
         return this._getFileLoaderPlugin().addCallback(function(plugin) {
            var host= [window.location.protocol,window.location.hostname].join('//');
            if (window.location.port && window.location.protocol == 'http:') {
               host += ':' + window.location.port;
            }
            requestParams.targetURL = host + '/cstorage/img?id=' + randomHelpers.createGUID();
            var def = new  Deferred();

            def.addErrback(function (e) {
               return e;
            });

            uploadedFiles[requestParams.fileName] = def;

            def.addCallback(function(result) {
               self._postFileStorageFinished(result.code, result.url ? result.url.replace(host, '') : '');
            });

            plugin.call('uploadToBigFileStorage', {Params:JSON.stringify(requestParams)}).addErrback(function(err) {
               def.errback(err);
            });

            return def;
         });
      },

      startFileLoad: function(target) {
         var
            loadDeferred = new Deferred(),
            resulObj = {},
            onLoadedHandler =function(event, json) {
               if (!loadDeferred.isReady()) {
                  if (json.result && json.result.code === 201) {
                     resulObj = json.result;
                     loadDeferred.callback();
                  } else {
                     loadDeferred.errback();
                  }
               }
               this.unsubscribe('onLoaded', onLoadedHandlerInstance);
            },
            onLoadedHandlerInstance = onLoadedHandler.bind(this);

         this.subscribe('onLoaded', onLoadedHandlerInstance);

         this.selectFile();

         if (this._loadDeferred && !this._loadDeferred.isReady()) {
            this._loadDeferred.errback();
         }
         this._loadDeferred = loadDeferred;
         loadDeferred.addCallback(function(){
            return resulObj;
         });
         return loadDeferred;
      }

      //////////////////////////////////////////
      /// Конец
      //////////////////////////////////////////
   });

   if (!Di.isRegistered('ImageUploader')) {
      Di.register('ImageUploader', {
         getFileLoader: function(){
               uploaderInstance = uploaderInstance ? uploaderInstance : new FileStorageLoader({
                  fileStorage: true,
                  extensions: ['image'],
                  element: $('<div></div>')
               });
               return uploaderInstance;
            },
            canMultiSelect: false
         }, {
            instantiate: false
      });
   }

   return FileStorageLoader;

});