define("Lib/File/Attach/BaseAttach", ["require", "exports", "Core/Abstract", "Core/core-simpleExtend", "Core/Deferred", "Lib/File/Attach/Container/GetterContainer", "Lib/File/Attach/Container/SourceContainer", "Core/moduleStubs"], function (require, exports, Abstract, CoreExtend, Deferred, GetterContainer, SourceContainer, moduleStubs) {
    "use strict";
    var UPLOADER_LINK = "Lib/File/Attach/Uploader";
    /**
     * Класс, реализующий выбор и загрузку файлов через разные источники данных
     * <br/>
     * Выбор и загрузка ресурсов:
     * <pre>
     *   var attach = new BaseAttach({
     *      multiSelect: false,
     *      fileProperty: "Файл"
     *   });
     *
     *   var scanner = new ScannerGetter();
     *   var fs = new FSGetter();
     *   attach.registerGetter(scanner);
     *   attach.registerGetter(fs);
     *
     *   var sourceOption = {
     *      endpoint: {
     *          address: "/service/",
     *          contract: "simple"
     *      },
     *      binding: {
     *          create: "ЗагрузитьВНикуда"
     *      },
     *      idProperty: "link"
     *   };
     *   attach.registerSource(new BL_SbisPluginSource(sourceOption));
     *   attach.registerSource(new SbisFileSource(sourceOption));
     *
     *   self.getChildControlByName("fsBtn").subscribe("onActivated", function(){
     *      attach.choose(fs.getType());
     *   });
     *   self.getChildControlByName("scanBtn").subscribe("onActivated", function(){
     *      attach.choose(scanner.getType());
     *   });
     *   self.getChildControlByName("uploadBtn").subscribe("onActivated", function(){
     *      attach.upload({
     *          // Дополнительные мета-данные для отправки
     *      }).addCallback(function(results){
     *          // вывод результатов загрузки
     *      });
     *   });
     * </pre>
     * @public
     * @class
     * @name Lib/File/Attach/BaseAttach
     * @extends Core/Abstract
     * @author Заляев А.В.
     */
    var BaseAttach = CoreExtend.extend(Abstract, /** @lends Lib/File/Attach/BaseAttach.prototype*/ {
        /**
         * {Array<LocalFile|LocalFileLink|HttpFileLink>}
         * @private
         */
        _selectedResources: [],
        /**
         * {Lib/File/Attach/Container/GetterContainer}
         * @private
         */
        _getterContainer: null,
        /**
         * {Lib/File/Attach/Container/SourceContainer}
         * @private
         */
        _sourceContainer: null,
        _uploadHandlers: null,
        _$options: {
            /**
             * @cfg {Boolean} Множетвенный выбор
             */
            multiSelect: true,
            /**
             * @cfg {String} Имя параметра, содержащее файл при отправке на сервер
             */
            fileProperty: "File"
        },
        constructor: function (opt) {
            BaseAttach.superclass.constructor.apply(this, arguments);
            this._$options = Object.assign({}, this._$options, opt);
            this._getterContainer = new GetterContainer();
            this._sourceContainer = new SourceContainer();
            this._uploadHandlers = {};
            this._publish('onProgress', 'onWarning', 'onLoadedFolder', 'onLoaded', 'onLoadError', 'onLoadResourceError', 'onLoadedResource');
        },
        ///region IAttacher
        /**
         * Регестрация IResourceGetter, для получения файлов
         * @param {IResourceGetter} ResourceGetter
         * @example
         * Регестрация сканера:
         * <pre>
         *    require([
         *    'Lib/File/Attach/BaseAttach'
         *    'js!SBIS3.Plugin/components/Extensions/Integration/FileGetter/ScannerGetter'
         *    ], function (BaseAttach, ScannerGetter) {
         *       var attach = new BaseAttach();
         *       var getter = new ScannerGetter();
         *       attach.registerGetter(getter);
         *       // теперь доступно прикрепление сканов
         *       attach.choose(getter.getType()).addCallback(function(links){ // links: Array<Lib/File/LocalFileLink>
         *          ...
         *       });
         *    });
         * </pre>
         * @method
         * @name Lib/File/Attach/BaseAttach#registerGetter
         * @see Lib/File/ResourceGetter/IResourceGetterBase
         */
        registerGetter: function (ResourceGetter) {
            this._getterContainer.push(ResourceGetter);
        },
        /**
         * Регистрация источников данных для загрузки определённого типа файла
         * @param {Function} type конструктор обёртки над файлом
         * @param {ISource} source источник данных
         * @example
         * Регестрация источника, умеющего загружать LocalFileLink:
         * <pre>
         *    require([
         *      'Lib/File/Attach/BaseAttach',
         *      'js!SBIS3.Plugin/components/Extensions/Integration/Source/BL_SbisPluginSource'
         *    ], function (BaseAttach, BL_SbisPluginSource) {
         *       var attach = new BaseAttach();
         *       attach.registerSource(new BL_SbisPluginSource({
         *          endpoint: {
         *              address: "/service/",
         *              contract: "simple"
         *          },
         *          binding: {
         *              create: "ЗагрузитьВНикуда"
         *          },
         *          idProperty: "link"
         *       }));
         *       ...
         *    });
         * </pre>
         * @method
         * @name Lib/File/Attach/BaseAttach#registerSource
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        registerSource: function (type, source) {
            return this._sourceContainer.register(type, source);
        },
        ///endregion
        ///region IDirectInsertFile
        /**
         * Устанавливает ресурсы в список выбранных
         * @param {Array<IFileData> | IFileData} files файл или набор устанавливаемых файлов
         * @example
         * Привязка файлов, полученных путём Drag&Drop к Attach для последующей загрузки
         * <pre>
         *    myArea.subscribe("onDrop", function(event, data) {
         *      if (data.file instanceof Blob) {
         *          attach.setSelectedResource(new LocalFile(data.file))
         *      }
         *    });
         * </pre>
         * @method
         * @name Lib/File/Attach/BaseAttach#setSelectedResource
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        setSelectedResource: function (files) {
            this._selectedResources = files instanceof Array ? files : [files];
        },
        /**
         * Очищает набор выбраных ресурсов
         * @void
         * @method
         * @name Lib/File/Attach/BaseAttach#clearSelectedResource
         */
        clearSelectedResource: function () {
            this._selectedResources = [];
        },
        /**
         * Возвращает набор выбраных ресурсов
         * @return {Array<IFileData>}
         * @method
         * @name Lib/File/Attach/BaseAttach#getSelectedResource
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        getSelectedResource: function () {
            return this._selectedResources || [];
        },
        /**
         * Добавляет ресурсы к списку выбранных
         * @return {Array<IFileData>}
         * @method
         * @name Lib/File/Attach/BaseAttach#addSelectedResource
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        addSelectedResource: function (files) {
            var fileArray = [];
            if (this._$options.multiSelect) {
                fileArray = this.getSelectedResource();
            }
            this.setSelectedResource(fileArray.concat(files));
        },
        ///endregion
        ///region IUpload
        /**
         * Загрузка выбранных ресурсов
         * @param {*} [meta] Дополнительные мета-данные для отправки
         * @return {Core/Deferred<Array<WS.Data/Entity/Model | Error>>}
         * @example
         * Загрузка выбранных сканов:
         * <pre>
         *    require([
         *      'Lib/File/Attach/BaseAttach',
         *      'js!SBIS3.Plugin/components/Extensions/Integration/FileGetter/ScannerGetter',
         *      'js!SBIS3.Plugin/components/Extensions/Integration/Source/BL_SbisPluginSource'
         *    ], function (BaseAttach, ScannerGetter, BL_SbisPluginSource) {
         *       var attach = new BaseAttach();
         *       attach.registerGetter(new ScannerGetter());
         *       attach.registerSource(new BL_SbisPluginSource({
         *          endpoint: {
         *              address: "/service/",
         *              contract: "simple"
         *          },
         *          binding: {
         *              create: "ЗагрузитьВНикуда"
         *          },
         *          idProperty: "link"
         *       }));
         *       attach.choose(getter.getType()).addCallback(function(links){ // links: Array<Lib/File/LocalFileLink>
         *          attach.upload({
         *              "ИдО": 12345
         *          }).addCallback(function(results){
         *              // вывод результатов загрузки
         *          });;
         *       });
         *    });
         * </pre>
         * @method
         * @name Lib/File/Attach/BaseAttach#upload
         */
        upload: function (meta) {
            /*
             * забираем выбранные файлы себе, очищая набор,
             * чтобы файлы, выбранные после начала upload, не попали в текущую пачку загрузки
             */
            var files = this.getSelectedResource();
            if (!files.length) {
                return Deferred.fail("Not selected files");
            }
            this.clearSelectedResource();
            return this._getUploader().addCallback(function (loader) {
                return loader.upload(files, meta);
            });
        },
        /**
         * Асинхронное получение сущности загрузчика ресурсов
         * @return {Core/Deferred<Lib/File/Attach/Uploader>}
         * @private
         * @method
         * @name Lib/File/Attach/BaseAttach#_getUploader
         */
        _getUploader: function () {
            var _this = this;
            if (this._loader) {
                return Deferred.success(this._loader);
            }
            return moduleStubs.require([UPLOADER_LINK]).addCallback(function (modules) {
                var Uploader = modules[0];
                _this._loader = new Uploader(_this._sourceContainer, _this._$options.fileProperty, function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    _this._notify.apply(_this, args);
                });
                return _this._loader;
            });
        },
        ///endregion IUpload,
        /**
         * Метод вызова выбора файла через контейнер, по средствам имени модуля IResourceGetter
         * @param {String} getterName
         * @return {Core/Deferred<Array<IFileData | Error>>}
         * @example
         * Выбор и загрузка ресурсов:
         * <pre>
         *   var attach = new BaseAttach();
         *
         *   var scanner = new ScannerGetter();
         *   var fs = new FSGetter();
         *   attach.registerGetter(scanner);
         *   attach.registerGetter(fs);
         *
         *   var sourceOption = {
         *      endpoint: {
         *          address: "/service/",
         *          contract: "simple"
         *      },
         *      binding: {
         *          create: "ЗагрузитьВНикуда"
         *      },
         *      idProperty: "link"
         *   };
         *   attach.registerSource(new BL_SbisPluginSource(sourceOption));
         *   attach.registerSource(new SbisFileSource(sourceOption));
         *
         *   self.getChildControlByName("fsBtn").subscribe("onActivated", function(){});
         *      attach.choose(fs.getType());
         *   }
         *   self.getChildControlByName("scanBtn").subscribe("onActivated", function(){});
         *      attach.choose(scanner.getType());
         *   }
         *   self.getChildControlByName("uploadBtn").subscribe("onActivated", function(){});
         *      attach.upload({
         *          // Дополнительные мета-данные для отправки
         *      }).addCallback(function(results){
         *          // вывод результатов загрузки
         *      });;
         *   }
         * </pre>
         * @method
         * @name Lib/File/Attach/BaseAttach#choose
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        choose: function (getterName) {
            var _this = this;
            return this._getterContainer.get(getterName).addCallback(function (getter) {
                return getter.getFiles();
            }).addCallback(function (files) {
                /* добавляем к ранее сохранёным ресурсам отфильтрованные от ошибок при выборе
                 * они не нужны нам во внутренем состоянии, а пользователь о них будет уведомлен,
                 * т.к. возвращаем что было выбрано в текущей операции
                 */
                _this.addSelectedResource(files.filter(function (f) { return !(f instanceof Error); }));
                return files;
            });
        },
        /**
         * Возвращает список зарегестрированый конструкторов над файлами
         * @return {Array<Function>}
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        getRegisteredResource: function () {
            return this._sourceContainer.getRegisteredResource();
        }
    });
    return BaseAttach;
});
/**
 * @event onProgress
 * @name Lib/File/Attach/BaseAttach#onProgress
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onWarning
 * @name Lib/File/Attach/BaseAttach#onWarning
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onLoadedFolder
 * @name Lib/File/Attach/BaseAttach#onLoadedFolder
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onLoaded
 * Событые окончания загрузки ресурсов
 *
 * @name Lib/File/Attach/BaseAttach#onLoaded
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Array<Error | WS.Data/Entity/Model>} results Массив, содержащий результаты загрузки выбранных ресурсов
 *
 * @see WS.Data/Entity/Model
 */
/**
 * @event onLoadError
 * Событые ошибки начала загрузки.
 *
 * @name Lib/File/Attach/BaseAttach#onLoadError
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Error} error
 */
/**
 * @event onLoadResourceError
 * Событые ошибки загрузки отдельного ресурса
 *
 * @name Lib/File/Attach/BaseAttach#onLoadResourceError
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {IFileData} resource загружаемый ресурс
 * @param {Error} error
 *
 * @see Lib/File/LocalFile
 * @see Lib/File/LocalFileLink
 * @see Lib/File/HttpFileLink
 */
/**
 * @event onLoadedResource
 * Событые загрузки отдельного ресурса
 *
 * @name Lib/File/Attach/BaseAttach#onLoadedResource
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {IFileData} resource загружаемый ресурс
 * @param {Model} model Результат загрузки
 *
 * @see Lib/File/LocalFile
 * @see Lib/File/LocalFileLink
 * @see Lib/File/HttpFileLink
 * @see WS.Data/Entity/Model
 */ 
