/// <amd-module name="Lib/File/Attach/BaseAttach" />
// dependency for types
import Uploader = require("Lib/File/Attach/Uploader");
import Model = require("WS.Data/Entity/Model");
import ISource = require("WS.Data/Source/ISource");
import IResourceGetter = require("Lib/File/ResourceGetter/IResourceGetter");
import {IFileData, IFileDataConstructor} from "Lib/File/Interface";
// real dependency
import Abstract = require("Core/Abstract");
import CoreExtend = require('Core/core-simpleExtend');
import Deferred = require("Core/Deferred");
import GetterContainer = require("Lib/File/Attach/Container/GetterContainer");
import SourceContainer = require("Lib/File/Attach/Container/SourceContainer");
import moduleStubs = require("Core/moduleStubs");

const UPLOADER_LINK = "Lib/File/Attach/Uploader";
/**
 * Класс, реализующий выбор и загрузку файлов через разные источники данных
 * <br/>
 * Выбор и загрузка ресурсов:
 * <pre>
 *   var attach = new BaseAttach({
 *      multiSelect: false,
 *      fileProperty: "Файл"
 *   });
 *
 *   var scanner = new ScannerGetter();
 *   var fs = new FSGetter();
 *   attach.registerGetter(scanner);
 *   attach.registerGetter(fs);
 *
 *   var sourceOption = {
 *      endpoint: {
 *          address: "/service/",
 *          contract: "simple"
 *      },
 *      binding: {
 *          create: "ЗагрузитьВНикуда"
 *      },
 *      idProperty: "link"
 *   };
 *   attach.registerSource(new BL_SbisPluginSource(sourceOption));
 *   attach.registerSource(new SbisFileSource(sourceOption));
 *
 *   self.getChildControlByName("fsBtn").subscribe("onActivated", function(){
 *      attach.choose(fs.getType());
 *   });
 *   self.getChildControlByName("scanBtn").subscribe("onActivated", function(){
 *      attach.choose(scanner.getType());
 *   });
 *   self.getChildControlByName("uploadBtn").subscribe("onActivated", function(){
 *      attach.upload({
 *          // Дополнительные мета-данные для отправки
 *      }).addCallback(function(results){
 *          // вывод результатов загрузки
 *      });
 *   });
 * </pre>
 * @public
 * @class
 * @name Lib/File/Attach/BaseAttach
 * @extends Core/Abstract
 * @author Заляев А.В.
 */
let BaseAttach = CoreExtend.extend(Abstract, /** @lends Lib/File/Attach/BaseAttach.prototype*/ {
    /**
     * {Array<LocalFile|LocalFileLink|HttpFileLink>}
     * @private
     */
    _selectedResources: [],
    /**
     * {Lib/File/Attach/Container/GetterContainer}
     * @private
     */
    _getterContainer: null,
    /**
     * {Lib/File/Attach/Container/SourceContainer}
     * @private
     */
    _sourceContainer: null,
    _uploadHandlers: null,
    _$options: {
        /**
         * @cfg {Boolean} Множетвенный выбор
         */
        multiSelect: true,
        /**
         * @cfg {String} Имя параметра, содержащее файл при отправке на сервер
         */
        fileProperty: "File"
    },
    constructor(opt) {
        BaseAttach.superclass.constructor.apply(this, arguments);
        this._$options = Object.assign({}, this._$options, opt);
        this._getterContainer = new GetterContainer();
        this._sourceContainer = new SourceContainer();
        this._uploadHandlers = {};
        this._publish('onProgress', 'onWarning', 'onLoadedFolder',
            'onLoaded', 'onLoadError', 'onLoadResourceError', 'onLoadedResource'
        );
    },
    
    ///region IAttacher
    /**
     * Регестрация IResourceGetter, для получения файлов
     * @param {IResourceGetter} ResourceGetter
     * @example
     * Регестрация сканера:
     * <pre>
     *    require([
     *    'Lib/File/Attach/BaseAttach'
     *    'js!SBIS3.Plugin/components/Extensions/Integration/FileGetter/ScannerGetter'
     *    ], function (BaseAttach, ScannerGetter) {
     *       var attach = new BaseAttach();
     *       var getter = new ScannerGetter();
     *       attach.registerGetter(getter);
     *       // теперь доступно прикрепление сканов
     *       attach.choose(getter.getType()).addCallback(function(links){ // links: Array<Lib/File/LocalFileLink>
     *          ...
     *       });
     *    });
     * </pre>
     * @method
     * @name Lib/File/Attach/BaseAttach#registerGetter
     * @see Lib/File/ResourceGetter/IResourceGetterBase
     */
    registerGetter(ResourceGetter: IResourceGetter): void {
        this._getterContainer.push(ResourceGetter);
    },
    /**
     * Регистрация источников данных для загрузки определённого типа файла
     * @param {Function} type конструктор обёртки над файлом
     * @param {ISource} source источник данных
     * @example
     * Регестрация источника, умеющего загружать LocalFileLink:
     * <pre>
     *    require([
     *      'Lib/File/Attach/BaseAttach',
     *      'js!SBIS3.Plugin/components/Extensions/Integration/Source/BL_SbisPluginSource'
     *    ], function (BaseAttach, BL_SbisPluginSource) {
     *       var attach = new BaseAttach();
     *       attach.registerSource(new BL_SbisPluginSource({
     *          endpoint: {
     *              address: "/service/",
     *              contract: "simple"
     *          },
     *          binding: {
     *              create: "ЗагрузитьВНикуда"
     *          },
     *          idProperty: "link"
     *       }));
     *       ...
     *    });
     * </pre>
     * @method
     * @name Lib/File/Attach/BaseAttach#registerSource
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    registerSource(type: IFileDataConstructor, source: ISource): void {
        return this._sourceContainer.register(type, source);
    },
    ///endregion

    ///region IDirectInsertFile
    /**
     * Устанавливает ресурсы в список выбранных
     * @param {Array<IFileData> | IFileData} files файл или набор устанавливаемых файлов
     * @example
     * Привязка файлов, полученных путём Drag&Drop к Attach для последующей загрузки
     * <pre>
     *    myArea.subscribe("onDrop", function(event, data) {
     *      if (data.file instanceof Blob) {
     *          attach.setSelectedResource(new LocalFile(data.file))
     *      }
     *    });
     * </pre>
     * @method
     * @name Lib/File/Attach/BaseAttach#setSelectedResource
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    setSelectedResource(files: IFileData | IFileData[]): void {
        this._selectedResources = files instanceof Array ? files : [files];
    },
    /**
     * Очищает набор выбраных ресурсов
     * @void
     * @method
     * @name Lib/File/Attach/BaseAttach#clearSelectedResource
     */
    clearSelectedResource(): void {
        this._selectedResources = [];
    },
    /**
     * Возвращает набор выбраных ресурсов
     * @return {Array<IFileData>}
     * @method
     * @name Lib/File/Attach/BaseAttach#getSelectedResource
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    getSelectedResource(): IFileData[] {
        return this._selectedResources || [];
    },
    /**
     * Добавляет ресурсы к списку выбранных
     * @return {Array<IFileData>}
     * @method
     * @name Lib/File/Attach/BaseAttach#addSelectedResource
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    addSelectedResource(files: IFileData | IFileData[]): void {
        let fileArray = [];
        if (this._$options.multiSelect) {
            fileArray = this.getSelectedResource();
        }
        this.setSelectedResource(fileArray.concat(files));
    },
    ///endregion
    
    ///region IUpload
    /**
     * Загрузка выбранных ресурсов
     * @param {*} [meta] Дополнительные мета-данные для отправки
     * @return {Core/Deferred<Array<WS.Data/Entity/Model | Error>>}
     * @example
     * Загрузка выбранных сканов:
     * <pre>
     *    require([
     *      'Lib/File/Attach/BaseAttach',
     *      'js!SBIS3.Plugin/components/Extensions/Integration/FileGetter/ScannerGetter',
     *      'js!SBIS3.Plugin/components/Extensions/Integration/Source/BL_SbisPluginSource'
     *    ], function (BaseAttach, ScannerGetter, BL_SbisPluginSource) {
     *       var attach = new BaseAttach();
     *       attach.registerGetter(new ScannerGetter());
     *       attach.registerSource(new BL_SbisPluginSource({
     *          endpoint: {
     *              address: "/service/",
     *              contract: "simple"
     *          },
     *          binding: {
     *              create: "ЗагрузитьВНикуда"
     *          },
     *          idProperty: "link"
     *       }));
     *       attach.choose(getter.getType()).addCallback(function(links){ // links: Array<Lib/File/LocalFileLink>
     *          attach.upload({
     *              "ИдО": 12345
     *          }).addCallback(function(results){
     *              // вывод результатов загрузки
     *          });;
     *       });
     *    });
     * </pre>
     * @method
     * @name Lib/File/Attach/BaseAttach#upload
     */
    upload(meta?: Object): Deferred<Array<Model | Error>> {
        /*
         * забираем выбранные файлы себе, очищая набор,
         * чтобы файлы, выбранные после начала upload, не попали в текущую пачку загрузки
         */
        let files = this.getSelectedResource();
        if (!files.length) {
            return Deferred.fail("Not selected files");
        }
        this.clearSelectedResource();
        return this._getUploader().addCallback((loader: Uploader) => {
            return loader.upload(files, meta);
        });
    },
    /**
     * Асинхронное получение сущности загрузчика ресурсов
     * @return {Core/Deferred<Lib/File/Attach/Uploader>}
     * @private
     * @method
     * @name Lib/File/Attach/BaseAttach#_getUploader
     */
    _getUploader(): Deferred<Uploader> {
        if (this._loader) {
            return Deferred.success(this._loader);
        }
        return moduleStubs.require([UPLOADER_LINK]).addCallback((modules) => {
            let Uploader = modules[0];
            this._loader = new Uploader(this._sourceContainer, this._$options.fileProperty, (...args) => {
                this._notify(...args);
            });
            return this._loader;
        });
    },
    ///endregion IUpload,
    
    /**
     * Метод вызова выбора файла через контейнер, по средствам имени модуля IResourceGetter
     * @param {String} getterName
     * @return {Core/Deferred<Array<IFileData | Error>>}
     * @example
     * Выбор и загрузка ресурсов:
     * <pre>
     *   var attach = new BaseAttach();
     *
     *   var scanner = new ScannerGetter();
     *   var fs = new FSGetter();
     *   attach.registerGetter(scanner);
     *   attach.registerGetter(fs);
     *
     *   var sourceOption = {
     *      endpoint: {
     *          address: "/service/",
     *          contract: "simple"
     *      },
     *      binding: {
     *          create: "ЗагрузитьВНикуда"
     *      },
     *      idProperty: "link"
     *   };
     *   attach.registerSource(new BL_SbisPluginSource(sourceOption));
     *   attach.registerSource(new SbisFileSource(sourceOption));
     *
     *   self.getChildControlByName("fsBtn").subscribe("onActivated", function(){});
     *      attach.choose(fs.getType());
     *   }
     *   self.getChildControlByName("scanBtn").subscribe("onActivated", function(){});
     *      attach.choose(scanner.getType());
     *   }
     *   self.getChildControlByName("uploadBtn").subscribe("onActivated", function(){});
     *      attach.upload({
     *          // Дополнительные мета-данные для отправки
     *      }).addCallback(function(results){
     *          // вывод результатов загрузки
     *      });;
     *   }
     * </pre>
     * @method
     * @name Lib/File/Attach/BaseAttach#choose
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    choose(getterName: string): Deferred<IFileData[]> {
        return this._getterContainer.get(getterName).addCallback((getter: IResourceGetter) => {
            return getter.getFiles();
        }).addCallback((files: IFileData[]) => {
            /* добавляем к ранее сохранёным ресурсам отфильтрованные от ошибок при выборе
             * они не нужны нам во внутренем состоянии, а пользователь о них будет уведомлен,
             * т.к. возвращаем что было выбрано в текущей операции
             */
            this.addSelectedResource(files.filter((f) => {return !(f instanceof Error)}));
            return files;
        })
    },
    /**
     * Возвращает список зарегестрированый конструкторов над файлами
     * @return {Array<Function>}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    getRegisteredResource(): IFileDataConstructor[] {
        return this._sourceContainer.getRegisteredResource();
    }
});
export = BaseAttach;

/**
 * @event onProgress
 * @name Lib/File/Attach/BaseAttach#onProgress
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onWarning
 * @name Lib/File/Attach/BaseAttach#onWarning
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onLoadedFolder
 * @name Lib/File/Attach/BaseAttach#onLoadedFolder
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onLoaded
 * Событые окончания загрузки ресурсов
 *
 * @name Lib/File/Attach/BaseAttach#onLoaded
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Array<Error | WS.Data/Entity/Model>} results Массив, содержащий результаты загрузки выбранных ресурсов
 *
 * @see WS.Data/Entity/Model
 */
/**
 * @event onLoadError
 * Событые ошибки начала загрузки.
 *
 * @name Lib/File/Attach/BaseAttach#onLoadError
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Error} error
 */
/**
 * @event onLoadResourceError
 * Событые ошибки загрузки отдельного ресурса
 *
 * @name Lib/File/Attach/BaseAttach#onLoadResourceError
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {IFileData} resource загружаемый ресурс
 * @param {Error} error
 *
 * @see Lib/File/LocalFile
 * @see Lib/File/LocalFileLink
 * @see Lib/File/HttpFileLink
 */
/**
 * @event onLoadedResource
 * Событые загрузки отдельного ресурса
 *
 * @name Lib/File/Attach/BaseAttach#onLoadedResource
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {IFileData} resource загружаемый ресурс
 * @param {Model} model Результат загрузки
 *
 * @see Lib/File/LocalFile
 * @see Lib/File/LocalFileLink
 * @see Lib/File/HttpFileLink
 * @see WS.Data/Entity/Model
 */