/// <amd-module name="Lib/File/Attach/Container/GetterContainerLazy" />
// dependency for types
import {HashMap, IResourceGetterConstructor} from "Lib/File/Interface";
import IResourceGetter = require("Lib/File/ResourceGetter/IResourceGetter");
// real dependency
import Deferred = require("Core/Deferred");
import moduleStubs = require("Core/moduleStubs");
import GetterContainer = require("Lib/File/Attach/Container/GetterContainer");

/**
 * Контейнер для работы с различными реализациями IResourceGetter, позволяющий работать посредством ленивой подгрузки
 * @class Lib/File/Attach/Container/GetterContainerLazy
 * @private
 * @see Lib/File/ResourceGetter/IResourceGetterBase
 * @author Заляев А.В.
 */
class GetterContainerLazy extends GetterContainer /** @lends Lib/File/Attach/Container/GetterContainerLazy.prototype*/{
    /**
     * ссылки на экземпляры IResourceGetter для ленивой загрузки
     * @private
     */
    private _links: HashMap<string>;
    private _options: HashMap<any>;
    
    constructor() {
        super();
        this._links = Object.create(null);
        this._options = Object.create(null);
    }
    /**
     * Метод асинхронного получения экземпляра IResourceGetter
     * @param name
     * @return {Deferred<IResourceGetter>}
     * @see Lib/File/ResourceGetter/IResourceGetterBase
     */
    get(name: string): Deferred<IResourceGetter> {
        return super.get(name).addErrback((error) => {
            if (!this._links[name]) {
                return error;
            }
            /*
             * Загружаем Модуль через optional! т.к. не все проекты включают в себя модули из репозитория СБИС Плагина
             */
            return moduleStubs.require(
                "optional!" + this._links[name]
            ).addCallback((modules: [IResourceGetterConstructor]) => {
                let ResourceGetter = modules[0];
                delete this._links[name];
                if (!ResourceGetter) {
                    return Deferred.fail(`ResourceGetter \"${name}\" is not supported in this project`);
                }
                let getter =  new ResourceGetter(this._options[name]);
                this.push(getter);
                return getter;
            });
        });
    }
    /**
     * Зарегестрирован ли источник ресурсов в контейнере
     * @param {String} name Имя источника
     * @return {Boolean}
     * @see Lib/File/ResourceGetter/IResourceGetterBase
     */
    has(name: string): boolean {
        return super.has(name) || !!this._links[name];
    }
    /**
     * Регистрация ссылки для последующей ленивой загрузки
     * @param {String} name Имя модуля
     * @param {String} link Ссылка на модуль
     * @param {} options Параметры для конструирования
     * @see Lib/File/ResourceGetter/IResourceGetterBase
     */
    register(name: string, link: string, options?): void {
        this._links[name] = link;
        this._options[name] = options;
    }
}

export = GetterContainerLazy;