define("Lib/File/Attach/Container/SourceContainer", ["require", "exports", "Core/Deferred"], function (require, exports, Deferred) {
    "use strict";
    var filterForType = function (wrappers, FileData) {
        return wrappers.filter(function (sourceObj) {
            return FileData == sourceObj.fileType;
        });
    };
    /**
     * Класс для хранения экземпляров ISource по типам отправляеммых файлов
     * @class Lib/File/Attach/Container/SourceContainer
     * @author Заляев А.В.
     * @private
     */
    var SourceContainer = /** @class */ (function () {
        function SourceContainer() {
            this._sourceWrappers = [];
        }
        /**
         * Регистрация источников данных для загрузки определённого типа ресурса
         * @param {Deferred<ISource>} source источник данных
         * @param {IFileDataConstructor} FileData конструктор обёртки над ресурсом
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        SourceContainer.prototype.push = function (source, FileData) {
            if (!(FileData instanceof Function)) {
                throw new Error("Invalid type of the first argument, expected constructor");
            }
            // Если для данного типа файла ISource уже зарегестрирован - поменяем его. нет - добавим
            var registeredSource = filterForType(this._sourceWrappers, FileData)[0];
            if (registeredSource) {
                registeredSource.source = source || registeredSource.source;
                return;
            }
            this._sourceWrappers.push({
                source: source,
                fileType: FileData
            });
        };
        /**
         * Зарегестрирован ли для текущего ресурса источник данных
         * @param {IFileData} file
         * @return {boolean}
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        SourceContainer.prototype.has = function (file) {
            return !!this._get(file);
        };
        /**
         * Возвращает источник данных для ресурса
         * @param {IFileData} file
         * @return {Deferred<ISource>}
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        SourceContainer.prototype.get = function (file) {
            var source = this._get(file);
            return Deferred.success(source ||
                new Error("Not registered source for this file type"));
        };
        /**
         * Возвращает список зарегестрированый обёрток
         * @return {Array<IFileDataConstructor>}
         * @see Lib/File/LocalFile
         * @see Lib/File/LocalFileLink
         * @see Lib/File/HttpFileLink
         */
        SourceContainer.prototype.getRegisteredResource = function () {
            return this._sourceWrappers.map(function (wrapper) {
                return wrapper.fileType;
            });
        };
        SourceContainer.prototype._get = function (file) {
            var wrapper = this._sourceWrappers.filter(function (wrapper) {
                return file instanceof wrapper.fileType;
            })[0];
            return wrapper && wrapper.source;
        };
        return SourceContainer;
    }());
    return SourceContainer;
});
