/// <amd-module name="Lib/File/Attach/Container/SourceContainer" />
import {IFileData, IFileDataConstructor} from "lib/File/Interface";
import IContainer = require("Lib/File/Attach/Container/IContainer");
import ISource = require("WS.Data/Source/ISource");
import Deferred = require("Core/Deferred");

type SourceWrapper = {
    fileType: IFileDataConstructor;
    source: ISource;
}

let filterForType = (wrappers: SourceWrapper[], FileData: IFileDataConstructor) => {
    return wrappers.filter((sourceObj) => {
        return FileData == sourceObj.fileType;
    });
};
/**
 * Класс для хранения экземпляров ISource по типам отправляеммых файлов
 * @class Lib/File/Attach/Container/SourceContainer
 * @author Заляев А.В.
 * @private
 */
class SourceContainer implements IContainer<ISource> /** @lends Lib/File/Attach/Container/SourceContainer.prototype*/{
    private _sourceWrappers: SourceWrapper[];
    
    constructor() {
        this._sourceWrappers = [];
    }
    
    /**
     * Регистрация источников данных для загрузки определённого типа ресурса
     * @param {Deferred<ISource>} source источник данных
     * @param {IFileDataConstructor} FileData конструктор обёртки над ресурсом
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    push(source: ISource, FileData: IFileDataConstructor) {
        if (!(FileData instanceof Function)) {
            throw new Error("Invalid type of the first argument, expected constructor");
        }
        // Если для данного типа файла ISource уже зарегестрирован - поменяем его. нет - добавим
        let registeredSource = filterForType(this._sourceWrappers, FileData)[0];
        if (registeredSource) {
            registeredSource.source = source || registeredSource.source;
            return;
        }
        this._sourceWrappers.push({
            source: source,
            fileType: FileData
        });
    }
    /**
     * Зарегестрирован ли для текущего ресурса источник данных
     * @param {IFileData} file
     * @return {boolean}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    has(file: IFileData): boolean {
        return !!this._get(file);
    }
    /**
     * Возвращает источник данных для ресурса
     * @param {IFileData} file
     * @return {Deferred<ISource>}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    get(file: IFileData): Deferred<ISource> {
        let source = this._get(file);
        return Deferred.success(
            source ||
            new Error("Not registered source for this file type")
        );
    }
    /**
     * Возвращает список зарегестрированый обёрток
     * @return {Array<IFileDataConstructor>}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    getRegisteredResource(): IFileDataConstructor[] {
        return this._sourceWrappers.map((wrapper) => {
            return wrapper.fileType;
        })
    }
    private _get(file: IFileData): ISource {
        let wrapper = this._sourceWrappers.filter((wrapper) => {
            return file instanceof wrapper.fileType;
        })[0];
        return wrapper && wrapper.source;
    }
}

export = SourceContainer