/// <amd-module name="Lib/File/Attach/Container/SourceContainerLazy" />
import {IFileData, IFileDataConstructor} from "lib/File/Interface";
import SourceContainer = require("Lib/File/Attach/Container/SourceContainer");
import ISource = require("WS.Data/Source/ISource");
import Deferred = require("Core/Deferred");
import moduleStubs = require("Core/moduleStubs");

type Wrapper = {
    fileType: IFileDataConstructor;
}
type CreateWrappers = Wrapper & {
    source: Deferred<ISource>;
}
type LazyWrapper = Wrapper & {
    link: string;
    options: any;
}

/**
 * Фильтрует массив обёрток над ISource по типу файла
 * @param {Array<Wrapper>} array
 * @param {IFileData} file
 * @return {Wrapper}
 * @private
 * @author Заляев А.В.
 */
let instanceFilter = <T> (array: Wrapper[], file) => {
    return array.filter((wrapper) => {
        return file instanceof wrapper.fileType;
    })[0];
};

/**
 * Класс для хранения экземпляров ISource по типам отправляеммых ресурсов
 * @class Lib/File/Attach/Container/SourceContainer
 * @author Заляев А.В.
 * @private
 */
class SourceContainerLazy extends SourceContainer /** @lends Lib/File/Attach/Container/SourceContainer.prototype*/{
    private _lazyWrappers: LazyWrapper[];
    private _createWrappers: CreateWrappers[];
    
    constructor() {
        super();
        this._lazyWrappers = [];
        this._createWrappers = [];
    }
    
    /**
     * Ленивая егистрация источников данных для загрузки определённого типа ресурса
     * @param {IFileDataConstructor} fileType конструктор обёртки над ресурсом
     * @param {String} link Ссылка на источник данных
     * @param {*} options Параметры вызова конструктора обёртки
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    register(fileType: IFileDataConstructor, link: string, options: any): void {
        this._lazyWrappers.push({
            fileType,
            link,
            options
        })
    }
    /**
     * Зарегестрирован ли для текущего ресурса источник данных
     * @param {IFileData} file
     * @return {boolean}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    has(file: IFileData): boolean {
        return super.has(file) || !!this._getLazyWrapper(file);
    }
    /**
     * Возвращает источник данных для ресурса
     * @param {IFileData} file
     * @return {Deferred<ISource>}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    get(file: IFileData): Deferred<ISource> {
        return super.get(file).addErrback((error) => {
            let lazyWrapper = this._getLazyWrapper(file);
            if (!lazyWrapper) {
                return error;
            }
            return this._createSource(lazyWrapper);
        });
    }
    /**
     * Возвращает список зарегестрированый обёрток
     * @return {Array<IFileDataConstructor>}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    getRegisteredResource(): IFileDataConstructor[] {
        return super.getRegisteredResource().concat(this._lazyWrappers.map((wrapper) => {
            return wrapper.fileType;
        }));
    }
    private _getLazyWrapper(file: IFileData): LazyWrapper {
        return <LazyWrapper>instanceFilter(this._lazyWrappers, file);
    }
    private _createSource(wrapper: LazyWrapper): Deferred<ISource> {
        let createWrapper = <CreateWrappers>instanceFilter(this._createWrappers, wrapper.fileType);
        if (createWrapper) {
            return createWrapper.source.createDependent()
        }
        // Подгружаем и сохраняем инстанс ISource
        return moduleStubs.require(
            "optional!" + wrapper.link
        ).addCallback((modules) => {
            if (!modules[0]) {
                return Deferred.fail(`FileSource \"${wrapper.link}\" is not supported in this project`);
            }
            let source = new modules[0](wrapper.options);
            this.push(source, wrapper.fileType);
            return source;
        }).addBoth((result) => {
            // убираем найденную обёрку из списка ленивых
            this._lazyWrappers = this._lazyWrappers.filter((w) => {
                return w !== wrapper;
            });
            // так же чистим список ожидающх
            this._createWrappers = this._createWrappers.filter((w) => {
                return w !== createWrapper;
            });
            return result;
        });
    }
}

export = SourceContainerLazy