/// <amd-module name="Lib/File/Attach/SbisAttach" />
// dependency for types
import { IFileData, IFileDataConstructor } from "Lib/File/Interface";
import Deferred = require("Core/Deferred");
import EventObject = require("Core/EventObject");
import Model = require("WS.Data/Entity/Model");
import SourceOption = require("Lib/File/Attach/Option/Source");
import ResourceGetterOption = require("Lib/File/Attach/Option/ResourceGetter");
// real dependency
import LazyAttach = require("Lib/File/Attach/LazyAttach");
import CoreExtend = require('Core/core-simpleExtend');
import Abstract = require("Core/Abstract");

type Options = {
    sourceOptions: SourceOption[];
    getterOptions: ResourceGetterOption[];
    attachOptions: {
        fileProperty: string;
        multiselect: boolean;
    }
    [propName: string]: any;
}
/**
 * Класс, агрегирующий работу с BaseAttach
 * <pre>
 *   var sourceOptions = {
 *      endpoint: {
 *          address: "/service/",
 *          contract: "simple"
 *      },
 *      binding: {
 *          create: "ЗагрузитьВНикуда"
 *      },
 *      idProperty: "link"
 *   }
 *   var attach = new SbisAttach({
 *       sourceOptions: [ // загрузка на бизнеслогику
 *          // модуль опций: Lib/File/Attach/Option/Sources/BLXHR для загрузки через ajax
 *          new BLXHR(sourceOptions),
 *          // модуль опций: Lib/File/Attach/Option/Sources/BLPlugin для загрузки через СБИС Плагин
 *          new BLPlugin(sourceOptions)
 *       ],
 *       //  Или фабрика опций для ISource: Lib/File/Attach/Option/Sources/Fabric
 *       // sourceOptions: Fabric.getBLSourceOptions(sourceOptions),
 *       getterOptions: [
 *          // модуль опций: Lib/File/Attach/Option/Getters/FS для нативного окна выбора файлов
 *          new FS({multiselect: true, extensions: ["image"]}),
 *          // модуль опций: Lib/File/Attach/Option/Getters/Rich для "богатого окна выбора"
 *          new Rich({multiselect: true, extensions: ["image"]}),
 *          // модуль опций: Lib/File/Attach/Option/Getters/Scanner
 *          new Scanner({
 *              // Объект конфигурации сканера
 *          }),
 *          // модуль опций: Lib/File/Attach/Option/Getters/PhotoCam
 *          new PhotoCam({
 *              // Объект конфигурации веб-камеры
 *          }),
 *          // модуль опций: Lib/File/Attach/Option/Getters/Clipboard
 *          new Clipboard({
 *              // Объект конфигурации выбора через буфер обмена
 *          }),
 *          // модуль опций: Lib/File/Attach/Option/Getters/Dialogs
 *          new Dialogs({
 *              // Объект конфигурации окна выбора файлов через СБИС Плагин
 *          })
 *      ],
 *      attachOptions: {
 *          fileProperty: "Файл",
 *          multiselect: true
 *      }
 *   });
 *
 *   self.getChildControlByName("fsBtn").subscribe("onActivated", function(){
 *      attach.choose("FSGetter");
 *   });
 *   self.getChildControlByName("clipboardBtn").subscribe("onActivated", function(){
 *      attach.choose("ClipboardGetter");
 *   });
 *   self.getChildControlByName("scanBtn").subscribe("onActivated", function(){
 *      attach.choose("ScannerGetter");
 *   });
 *   self.getChildControlByName("uploadBtn").subscribe("onActivated", function(){
 *      attach.upload({
 *          // Дополнительные мета-данные для отправки
 *      });
 *   });
 * </pre>
 * @public
 * @class
 * @name Lib/File/Attach/SbisAttach
 * @extends Core/Abstract
 * @author Заляев А.В.
 */
let SbisAttach  = CoreExtend.extend(Abstract,/** @lends Lib/File/Attach/SbisAttach.prototype*/ {
    _$options: {
        /**
         * @cfg {Object} Объект конфигурирования BaseAttach
         * @example
         * <pre>
         *   var attach = new SbisAttach({
         *      attachOptions: {
         *          fileProperty: "Файл",
         *          multiselect: true
         *      }
         *      ...
         *   });
         * </pre>
         */
        attachOptions: {
            fileProperty: "Файл"
        },
        /**
         * @cfg {Array<Lib/File/Attach/Option/Source>} Набор параметров для регестрации ISource
         * @example
         * <pre>
         *   var attach = new SbisAttach({
         *       // фабрика опций для ISource: Lib/File/Attach/Option/Sources/Fabric
         *       sourceOptions: Fabric.getBLSourceOptions({ // загрузка на бизнеслогику
         *          endpoint: {
         *              address: "/service/",
         *              contract: "simple"
         *          },
         *          binding: {
         *              create: "ЗагрузитьВНикуда"
         *          },
         *          idProperty: "link"
         *       }),
         *       ...
         *   });
         * </pre>
         */
        sourceOptions: [],
        /**
         * @cfg {Array<Lib/File/Attach/Option/ResourceGetter>} Набор параметров для регестрации ResourceGetter
         * @example
         * <pre>
         *   var attach = new SbisAttach({
         *       getterOptions: [
         *          // модуль опций: Lib/File/Attach/Option/Getters/FS для нативного окна выбора файлов
         *          new FS({multiSelect: true, extensions: ["image"]}),
         *          // модуль опций: Lib/File/Attach/Option/Getters/Rich для "богатого окна выбора"
         *          new Rich({multiSelect: true, extensions: ["image"]}),
         *          // модуль опций: Lib/File/Attach/Option/Getters/Scanner
         *          new Scanner({
         *              // Объект конфигурации сканера
         *          }),
         *          // модуль опций: Lib/File/Attach/Option/Getters/PhotoCam
         *          new PhotoCam({
         *              // Объект конфигурации веб-камеры
         *          }),
         *          // модуль опций: Lib/File/Attach/Option/Getters/Clipboard
         *          new Clipboard({
         *              // Объект конфигурации выбора через буфер обмена
         *          }),
         *          // модуль опций: Lib/File/Attach/Option/Getters/Dialogs
         *          new Dialogs({
         *              // Объект конфигурации окна выбора файлов через СБИС Плагин
         *          })
         *      ],
         *      ...
         *   });
         * </pre>
         */
        getterOptions: []
    },
    constructor(opt: Options) {
        SbisAttach.superclass.constructor.apply(this, arguments);
        this._$options = Object.assign({}, this._$options, opt);
        this._attacher = new LazyAttach(this._$options.attachOptions);
        let events = [
            "onProgress", "onWarning", "onLoadedFolder", "onLoaded",
            "onLoadError", "onLoadResourceError", "onLoadedResource"
        ];
        this._publish.apply(this, events);
        // пробрасываем события загрузки файлов
        events.forEach((eventName: string) => {
            this._attacher.subscribe(eventName, (event: EventObject, ...args) => {
                this._notify(eventName, ...args);
            });
        });
        // регестрируем список IResourceGetter из наследников
        opt.getterOptions.forEach((getterOpt) => {
            let getter = getterOpt.getGetter();
            if (typeof getter === "string") {
                return this._attacher.registerLazyGetter(
                    getterOpt.getName(),
                    getter,
                    getterOpt.getOptions()
                );
            }
            this._attacher.registerGetter(getter);
        });
        opt.sourceOptions.forEach((srcOpt) => {
            this._attacher.registerLazySource(
                srcOpt.getResourceType(),
                srcOpt.getSourceName(),
                srcOpt.getOptions()
            )
        });
    },

    ///region IDirectInsertFile
    /**
     * Устанавливает ресурсы в список выбранных
     * @param {Array<IFileData> | IFileData} files файл или набор устанавливаемых файлов
     * @example
     * Привязка файлов, полученных путём Drag&Drop к Attach для последующей загрузки
     * <pre>
     *    myArea.subscribe("onDrop", function(event, data) {
     *      if (data.file instanceof Blob) {
     *          attach.setSelectedResource(new LocalFile(data.file))
     *      }
     *    });
     * </pre>
     * @method
     * @name Lib/File/Attach/SbisAttach#setSelectedResource
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    setSelectedResource(files: IFileData | IFileData[]): void {
        return this._attacher.setSelectedResource(files);
    },
    /**
     * Очищает набор выбраных ресурсов
     * @void
     * @method
     * @name Lib/File/Attach/SbisAttach#clearSelectedResource
     */
    clearSelectedResource(): void {
        return this._attacher.clearSelectedResource();
    },
    /**
     * Возвращает набор выбраных ресурсов
     * @return {Array<IFileData>}
     * @method
     * @name Lib/File/Attach/SbisAttach#getSelectedResource
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    getSelectedResource(): IFileData[] {
        return this._attacher.getSelectedResource();
    },
    ///endregion
    
    ///region IUpload
    /**
     * Загрузка выбраных ресурсов
     * @param {*} [meta] Дополнительные мета-данные для отправки
     * @return {Deferred<Array<WS.Data/Entity/Model | Error>>}
     * @example
     * Выбор и загрузка файла:
     * <pre>
     *   var attach = new SbisAttach({...});
     *
     *   self.getChildControlByName("fsBtn").subscribe("onActivated", function(){});
     *      attach.choose("FSGetter").addCallback(function(files) {
     *          attach.upload({
     *              // Дополнительные мета-данные для отправки
     *          }).addCallback(function(results){
     *              // вывод результатов загрузки
     *          });;
     *      });
     *   }
     * </pre>
     * @method
     * @name Lib/File/Attach/SbisAttach#upload
     */
    upload(meta?: Object): Deferred<Array<Model | Error>> {
        return this._attacher.upload(meta);
    },
    ///endregion IUpload
    
    /**
     * Метод вызова выбора ресурсов, по средствам имени модуля IResourceGetter
     * @param {String} getterName
     * @return {Core/Deferred<Array<IFileData>>}
     * @example
     * Выбор и загрузка файла:
     * <pre>
     *   var attach = new SbisAttach({...});
     *
     *   self.getChildControlByName("fsBtn").subscribe("onActivated", function(){});
     *      attach.choose("FSGetter").addCallback(function(files) {
     *          attach.upload({
     *              // Дополнительные мета-данные для отправки
     *          }).addCallback(function(results){
     *              // вывод результатов загрузки
     *          });;
     *      });
     *   }
     * </pre>
     * @method
     * @name Lib/File/Attach/SbisAttach#choose
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    choose(getterName: string): Deferred<IFileData[]> {
        return this._attacher.choose(getterName);
    },
    /**
     * Возвращает список зарегестрированый конструкторов над файлами
     * @return {Array<Function>}
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     * @see Lib/File/HttpFileLink
     */
    getRegisteredResource(): IFileDataConstructor[] {
        return this._attacher.getRegisteredResource();
    }
});

export = SbisAttach;

/**
 * @event onProgress
 * @name Lib/File/Attach/SbisAttach#onProgress
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onWarning
 * @name Lib/File/Attach/SbisAttach#onWarning
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onLoadedFolder
 * @name Lib/File/Attach/SbisAttach#onLoadedFolder
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Object} data
 * @param {IFileData} file
 */
/**
 * @event onLoaded
 * Событые окончания загрузки ресурсов
 *
 * @name Lib/File/Attach/SbisAttach#onLoaded
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Array<Error | WS.Data/Entity/Model>} results Массив, содержащий результаты загрузки выбранных ресурсов
 *
 * @see WS.Data/Entity/Model
 */
/**
 * @event onLoadError
 * Событые ошибки начала загрузки.
 *
 * @name Lib/File/Attach/SbisAttach#onLoadError
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {Error} error
 */
/**
 * @event onLoadResourceError
 * Событые ошибки загрузки отдельного ресурса
 *
 * @name Lib/File/Attach/SbisAttach#onLoadResourceError
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {IFileData} resource загружаемый ресурс
 * @param {Error} error
 *
 * @see Lib/File/LocalFile
 * @see Lib/File/LocalFileLink
 * @see Lib/File/HttpFileLink
 */
/**
 * @event onLoadedResource
 * Событые загрузки отдельного ресурса
 *
 * @name Lib/File/Attach/SbisAttach#onLoadedResource
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {IFileData} resource загружаемый ресурс
 * @param {Model} model Результат загрузки
 *
 * @see Lib/File/LocalFile
 * @see Lib/File/LocalFileLink
 * @see Lib/File/HttpFileLink
 * @see WS.Data/Entity/Model
 */