/// <amd-module name="Lib/File/ResourceGetter/FSGetter" />
/// <amd-dependency path="i18n!Lib/File/ResourceGetter/IResourceGetterBase" />
define("Lib/File/ResourceGetter/FSGetter", ["require", "exports", "tslib", "Lib/File/ResourceGetter/IResourceGetterBase", "Core/Deferred", "tmpl!Lib/File/ResourceGetter/FSGetter", "json!Lib/File/ResourceGetter/resources/MimeTypes", "Lib/File/LocalFile", "Core/core-merge", "Core/detection", "Lib/File/utils/ExtensionsError", "i18n!Lib/File/ResourceGetter/IResourceGetterBase"], function (require, exports, tslib_1, IResourceGetterBase, Deferred, dotTpl, MimeTypes, LocalFile, merge, detection, ExtensionsError) {
    "use strict";
    var AFTER_CLOSE_DELAY = 2000;
    /**
     * Детектим отмену выбора файлов пользователем
     * @param {Function} handler
     */
    var onClose = function (handler) {
        /*
         * Различные модели телефонов могут стрелять onChange с большой задержкой при выборе нескольких файлов
         * и для них опасно стрелять Deferred.cancel по какому-то таймауту
         * поэтому, если пользователь отменил выбор файлов на мобилке, оставим подвисший Deferred
         * то тех пор, пока не будет позван метод получения файлов ещё раз, либо не вызовется destroy метод
         * как доставерные показатели, что предыдущее окно было закрыто
         */
        if (detection.isMobilePlatform) {
            return;
        }
        /*
         * Ловим получение фокуса страницой, как признак того, что окно выбора закрылось
         * Для этого ожидаем событие focusin - если браузер стреляет событие при закрытии окна (Chromium, Edge, IE)
         *
         * Firefox и Safari не стреляют focusin при закрытии окна, поэтому у них Deferred.cancel вызовится после нажатии
         * в область DOM, которая не была в фокусе до открытия окна выбора.
         * Это можно было покрыть событием mouseover, но неокоторые браузеры (включая Safari) стреляют им при открытом окне выбора файлов
         */
        var focus = function (event) {
            // Фокус поймаем раньше, чем файлы появятся в input и тот стрельнёт onChange
            // поэтому с небольшой задержкой стреляем отмену Deferred'а, если нету файлов и он ещё не стрельнувший
            setTimeout(handler, AFTER_CLOSE_DELAY);
            document.removeEventListener("focusin", focus);
        };
        /*
         * навешивание обработчиков фокуса вынесем в асинхронную функцию:
         * Lib/Control/Control при клике на кнопку может отдать фокус другому контролу
         * в зависимости от опции activableByClick
         * но сделает это синхронно
         *
         * Если же не выносить через setTimeout то можем получить ситуацию, когда окошко выбора файлов открывается,
         * а фокус в этот момет придёт другому контролу и обработчик сработает, а мы соответственно будем бумать, что окно закрылось
         */
        setTimeout(function () {
            document.addEventListener("focusin", focus);
        }, 0);
    };
    /**
     * Формирует строку mime-types из extensions
     * @returns {string}
     * @private
     * @author Заляев А.В.
     */
    function createExtensionMimeTypes(extensions) {
        return extensions.map(function (ext) {
            if (ext == 'audio' || ext == 'video' || ext == 'image') {
                return ext + '/*';
            }
            if (ext in MimeTypes) {
                return MimeTypes[ext];
            }
            return ext;
        }).join(',');
    }
    /**
     * Проверяет файл на валидность расширения
     * @param {File} file
     * @param {Array<String>} extensions
     * @returns {Boolean}
     * @private
     */
    function verifyFileExtension(file, extensions) {
        var fileExt = ((file.name.match(/^\S[\S\s]*\.([\S]+)$/) || [])[1] || "").toLowerCase();
        if (extensions.indexOf(fileExt) > -1) {
            return true;
        }
        var type = file.type || '';
        var key;
        if (type) {
            for (key in MimeTypes) {
                if (MimeTypes.hasOwnProperty(key) &&
                    MimeTypes[key] === type &&
                    extensions.indexOf(key) != -1) {
                    return true;
                }
            }
        }
        return false;
    }
    /**
     * Класс, реализующий интерфейс получения файлов IResourceGetter через нативное окошко
     *
     * В связи с политиками безопасности браузеров для выбора файла необходимо:
     * <ul>
     *      <li> Экземпляр класса должен быть подготовлен заранее, т.к. input-элемент должен находиться в DOM до начала операции.</li>
     *      <li> Наличие пользовательского события</li>
     *      <li> Между пользовательским событием и вызовом метода .getFiles() не должно быть ленивых подгрузок модулей, а также вызовов БЛ</li>
     * </ul>
     *
     * @class
     * @name Lib/File/ResourceGetter/FSGetter
     * @extends Lib/File/ResourceGetter/IResourceGetterBase
     * @public
     * @author Заляев А.В.
     */
    var FSGetter = /** @class */ (function (_super) {
        tslib_1.__extends(FSGetter, _super); /** @lends Lib/File/ResourceGetter/FSGetter.prototype*/
        function FSGetter(cfg) {
            var _this = _super.call(this) || this;
            _this.name = "FSGetter";
            _this._options = {
                /**
                 * @cfg {Boolean} Выбрать несколько файлов
                 * <wiTag group="Управление">
                 * Позволяет выбрать несколько файлов
                 */
                multiSelect: false,
                /**
                 * @cfg {Array<String>} Список расширений выбираемых файлов
                 * <wiTag group="Управление">
                 * Помимо перечисления массива конкретных расширений файлов, можно также передать в массив значения:
                 * <ul>
                 *      <li> "image" - доступен выбор всех типов изображений</li>
                 *      <li> "audio" - доступен выбор всех типов аудио файлов</li>
                 *      <li> "video" - доступен выбор всех типов видео файлов</li>
                 * </ul>
                 * @example
                 * <pre>
                 *    extensions: ["image"]
                 *    // extensions: ["jpe","jpg","jpeg","gif","png","bmp","ico","svg","svgz","tif","tiff","pct","psd"]
                 * </pre>
                 */
                extensions: null,
                /**
                 * @cfg {HTMLElement} DOM элемент - контейнер, в котором будет построен невидивый input[type=file] посредством которого открывается окошко выбора файлов
                 * <wiTag group="Управление">
                 * По умолчанию: document.body
                 */
                element: null
            };
            _this._options = merge(_this._options, typeof cfg === "object" ? cfg : {});
            var container;
            container = _this._options.element;
            if (!(container instanceof HTMLElement)) {
                container = document.body;
            }
            var element = document.createElement("div");
            element.innerHTML = dotTpl(_this._options);
            container.appendChild(element);
            _this._form = element.querySelector("form");
            _this._inputBtn = _this._form.querySelector("input[type=file]");
            _this._inputBtn.onchange = function () {
                _this._onChangeInput(_this._inputBtn.files);
            };
            if (_this._options.multiSelect) {
                _this._inputBtn.setAttribute("multiple", "true");
            }
            _this._setMIMEType(_this._inputBtn);
            return _this;
        }
        /**
         * Осуществляет выбор файлов через нативное окно
         * @description Deferred.cancel стреляет с задержкой после закрытия окна выбора.
         * Это связано необходимым временем между получением фокуса и отработкой onChange у input-элемента на маломощных эвм
         * Необходимо это учитывать при обработке errback у результата метода
         * @return {Core/Deferred<Array<Lib/File/LocalFile | Error>>}
         * @method
         * @name Lib/File/ResourceGetter/FSGetter#getFiles
         * @see Lib/File/LocalFile
         */
        FSGetter.prototype.getFiles = function () {
            var _this = this;
            if (this.isDestroy()) {
                return Deferred.fail("Resource getter is destroyed");
            }
            if (this._selectDef) {
                /*
                 * сюда попадём, если не успели поймать фокус и обработать закрытие окна
                 * например: клик по кнопке выбора -> esc -> клик по кнопке выбора
                 */
                this._selectDef.cancel();
            }
            var def;
            def = this._selectDef = new Deferred();
            // Перестает стрелять change'ом если выбрать тот же файл, поэтому зануляем форму перед кликом
            this._form.reset();
            // Детектим отмену выбора файлов пользователем
            onClose(function () {
                if (!def.isReady() && !_this._inputBtn.files.length) {
                    def.cancel();
                    _this._selectDef = null;
                }
            });
            this._inputBtn.click();
            return def;
        };
        /**
         * Возможен ли выбор файлов
         * @return {Core/Deferred<Boolean>}
         * @method
         * @name Lib/File/ResourceGetter/FSGetter#canExec
         */
        FSGetter.prototype.canExec = function () {
            return Deferred.success(!this.isDestroy());
        };
        /**
         * Устанавливает фильтр MIME-типов на input-жлемент
         * @param {HTMLInputElement} input
         * @private
         * @method
         * @name Lib/File/ResourceGetter/FSGetter#_setMIMEType
         * @void
         */
        FSGetter.prototype._setMIMEType = function (input) {
            var _this = this;
            this._extensions = [];
            if (this._options.extensions && this._options.extensions.length) {
                input.setAttribute("accept", createExtensionMimeTypes(this._options.extensions));
                this._options.extensions.forEach(function (ext) {
                    if ((ext == "audio" || ext == "video" || ext == "image")) {
                        _this._extensions.concat(MimeTypes[ext]);
                    }
                    else {
                        _this._extensions.push(ext);
                    }
                });
            }
        };
        /**
         * Обработчик события выбора файла в input
         * @param {FileList} selectedFiles
         * @private
         * @method
         * @name Lib/File/ResourceGetter/FSGetter#_onChangeInput
         * @void
         */
        FSGetter.prototype._onChangeInput = function (selectedFiles) {
            if (this.isDestroy()) {
                return;
            }
            var files = [];
            /*
             * Нельзя обходить FileList через "for in" + ".hasOwnProperty"
             * https://w3c.github.io/FileAPI/#filelist-section
             * Обход надо делать только по числовому индексу и получать через FileList.item({Number}) или FileList[{Number}]
             */
            for (var i = 0; i < selectedFiles.length; i++) {
                var error = void 0;
                var file = selectedFiles.item && selectedFiles.item(i) || selectedFiles[i];
                if (this._extensions.length && !verifyFileExtension(file, this._extensions)) {
                    error = new ExtensionsError(file.name);
                }
                files.push(error || new LocalFile(file));
            }
            this._selectDef.callback(files);
            this._selectDef = null;
        };
        FSGetter.prototype.destroy = function () {
            this._form.remove && this._form.remove();
            this._form = null;
            this._inputBtn = null;
            if (this._selectDef && !this._selectDef.isReady()) {
                this._selectDef.cancel();
            }
            _super.prototype.destroy.call(this);
        };
        return FSGetter;
    }(IResourceGetterBase /** @lends Lib/File/ResourceGetter/FSGetter.prototype*/));
    return FSGetter;
});
