/// <amd-module name="Lib/File/ResourceGetter/FSGetter" />
/// <amd-dependency path="i18n!Lib/File/ResourceGetter/IResourceGetterBase" />

import IResourceGetterBase = require("Lib/File/ResourceGetter/IResourceGetterBase");
import Deferred = require("Core/Deferred");
import dotTpl = require("tmpl!Lib/File/ResourceGetter/FSGetter");
import MimeTypes = require("json!Lib/File/ResourceGetter/resources/MimeTypes");
import LocalFile = require("Lib/File/LocalFile");
import merge = require("Core/core-merge");
import detection = require("Core/detection");
import ExtensionsError = require("Lib/File/utils/ExtensionsError");

const AFTER_CLOSE_DELAY = 2000;

/**
 * Детектим отмену выбора файлов пользователем
 * @param {Function} handler
 */
let onClose = (handler) => {
    /*
     * Различные модели телефонов могут стрелять onChange с большой задержкой при выборе нескольких файлов
     * и для них опасно стрелять Deferred.cancel по какому-то таймауту
     * поэтому, если пользователь отменил выбор файлов на мобилке, оставим подвисший Deferred
     * то тех пор, пока не будет позван метод получения файлов ещё раз, либо не вызовется destroy метод
     * как доставерные показатели, что предыдущее окно было закрыто
     */
    if (detection.isMobilePlatform) {
        return;
    }
    /*
     * Ловим получение фокуса страницой, как признак того, что окно выбора закрылось
     * Для этого ожидаем событие focusin - если браузер стреляет событие при закрытии окна (Chromium, Edge, IE)
     *
     * Firefox и Safari не стреляют focusin при закрытии окна, поэтому у них Deferred.cancel вызовится после нажатии
     * в область DOM, которая не была в фокусе до открытия окна выбора.
     * Это можно было покрыть событием mouseover, но неокоторые браузеры (включая Safari) стреляют им при открытом окне выбора файлов
     */
    
    let focus = (event) => {
        // Фокус поймаем раньше, чем файлы появятся в input и тот стрельнёт onChange
        // поэтому с небольшой задержкой стреляем отмену Deferred'а, если нету файлов и он ещё не стрельнувший
        setTimeout(handler, AFTER_CLOSE_DELAY);
        document.removeEventListener("focusin", focus);
    };
    /*
     * навешивание обработчиков фокуса вынесем в асинхронную функцию:
     * Lib/Control/Control при клике на кнопку может отдать фокус другому контролу
     * в зависимости от опции activableByClick
     * но сделает это синхронно
     *
     * Если же не выносить через setTimeout то можем получить ситуацию, когда окошко выбора файлов открывается,
     * а фокус в этот момет придёт другому контролу и обработчик сработает, а мы соответственно будем бумать, что окно закрылось
     */
    setTimeout(() => {
        document.addEventListener("focusin", focus);
    }, 0);
};
/**
 * Формирует строку mime-types из extensions
 * @returns {string}
 * @private
 * @author Заляев А.В.
 */
function createExtensionMimeTypes(extensions: string[]): string {
	return extensions.map((ext) => {
		if (ext == 'audio' || ext == 'video' || ext == 'image') {
			return ext + '/*';
		}
		if (ext in MimeTypes) {
			return MimeTypes[ext];
		}
		return ext;
	}).join(',');
}

/**
 * Проверяет файл на валидность расширения
 * @param {File} file
 * @param {Array<String>} extensions
 * @returns {Boolean}
 * @private
 */
function verifyFileExtension(file: File, extensions: string[]): boolean{
	let fileExt = ((file.name.match(/^\S[\S\s]*\.([\S]+)$/) || [])[1] || "").toLowerCase();
	
	if (extensions.indexOf(fileExt) > -1) {
		return true;
	}
	let type = file.type || '';
	let key;
	if (type) {
		for (key in MimeTypes) {
			if (MimeTypes.hasOwnProperty(key) &&
				MimeTypes[key] === type &&
				extensions.indexOf(key) != -1
			) {
				return true;
			}
		}
	}
	return false;
}
/**
 * Класс, реализующий интерфейс получения файлов IResourceGetter через нативное окошко
 *
 * В связи с политиками безопасности браузеров для выбора файла необходимо:
 * <ul>
 *      <li> Экземпляр класса должен быть подготовлен заранее, т.к. input-элемент должен находиться в DOM до начала операции.</li>
 *      <li> Наличие пользовательского события</li>
 *      <li> Между пользовательским событием и вызовом метода .getFiles() не должно быть ленивых подгрузок модулей, а также вызовов БЛ</li>
 * </ul>
 *
 * @class
 * @name Lib/File/ResourceGetter/FSGetter
 * @extends Lib/File/ResourceGetter/IResourceGetterBase
 * @public
 * @author Заляев А.В.
 */
class FSGetter extends IResourceGetterBase /** @lends Lib/File/ResourceGetter/FSGetter.prototype*/{
	protected name = "FSGetter";
	private _extensions: string[];
	private _form: HTMLFormElement;
	private _inputBtn: HTMLInputElement;
	private _selectDef: Deferred<LocalFile[]>;
	_options = {
		/**
		 * @cfg {Boolean} Выбрать несколько файлов
		 * <wiTag group="Управление">
		 * Позволяет выбрать несколько файлов
		 */
		multiSelect: false,
		/**
		 * @cfg {Array<String>} Список расширений выбираемых файлов
         * <wiTag group="Управление">
		 * Помимо перечисления массива конкретных расширений файлов, можно также передать в массив значения:
		 * <ul>
		 *      <li> "image" - доступен выбор всех типов изображений</li>
		 *      <li> "audio" - доступен выбор всех типов аудио файлов</li>
		 *      <li> "video" - доступен выбор всех типов видео файлов</li>
		 * </ul>
		 * @example
		 * <pre>
		 *    extensions: ["image"]
		 *    // extensions: ["jpe","jpg","jpeg","gif","png","bmp","ico","svg","svgz","tif","tiff","pct","psd"]
		 * </pre>
		 */
		extensions: null,
		/**
		 * @cfg {HTMLElement} DOM элемент - контейнер, в котором будет построен невидивый input[type=file] посредством которого открывается окошко выбора файлов
         * <wiTag group="Управление">
		 * По умолчанию: document.body
		 */
		element: null
	};
	constructor(cfg) {
		super();
		this._options = merge(this._options, typeof cfg === "object"? cfg: {});
		let container;
		container = this._options.element;
		if (!(container instanceof HTMLElement)){
			container = document.body;
		}
		let element = document.createElement("div");
		element.innerHTML = dotTpl(this._options);
		container.appendChild(element);
		this._form = element.querySelector("form");
		this._inputBtn = <HTMLInputElement> this._form.querySelector("input[type=file]");
		this._inputBtn.onchange = () => {
			this._onChangeInput(this._inputBtn.files);
		};
		if (this._options.multiSelect){
			this._inputBtn.setAttribute("multiple", "true");
		}
		this._setMIMEType(this._inputBtn);
	}
    /**
     * Осуществляет выбор файлов через нативное окно
     * @description Deferred.cancel стреляет с задержкой после закрытия окна выбора.
     * Это связано необходимым временем между получением фокуса и отработкой onChange у input-элемента на маломощных эвм
     * Необходимо это учитывать при обработке errback у результата метода
     * @return {Core/Deferred<Array<Lib/File/LocalFile | Error>>}
     * @method
     * @name Lib/File/ResourceGetter/FSGetter#getFiles
     * @see Lib/File/LocalFile
     */
	getFiles(): Deferred<Array<LocalFile | Error>> {
		if (this.isDestroy()) {
			return Deferred.fail("Resource getter is destroyed");
		}
		if (this._selectDef) {
			/*
			 * сюда попадём, если не успели поймать фокус и обработать закрытие окна
			 * например: клик по кнопке выбора -> esc -> клик по кнопке выбора
			 */
            this._selectDef.cancel()
		}
		let def: Deferred<Array<LocalFile | Error>>;
		def = this._selectDef = new Deferred();
        // Перестает стрелять change'ом если выбрать тот же файл, поэтому зануляем форму перед кликом
		this._form.reset();
		
        // Детектим отмену выбора файлов пользователем
        onClose(() => {
            if (!def.isReady() && !this._inputBtn.files.length) {
                def.cancel();
                this._selectDef = null;
            }
		});
		
        this._inputBtn.click();
		return def;
	}
    /**
     * Возможен ли выбор файлов
     * @return {Core/Deferred<Boolean>}
     * @method
     * @name Lib/File/ResourceGetter/FSGetter#canExec
     */
    canExec(): Deferred<boolean> {
        return Deferred.success(!this.isDestroy());
    }
    /**
     * Устанавливает фильтр MIME-типов на input-жлемент
     * @param {HTMLInputElement} input
     * @private
     * @method
     * @name Lib/File/ResourceGetter/FSGetter#_setMIMEType
     * @void
     */
    private _setMIMEType(input: HTMLInputElement) {
		this._extensions = [];
		if (this._options.extensions && this._options.extensions.length){
			input.setAttribute("accept", createExtensionMimeTypes(this._options.extensions));
			this._options.extensions.forEach((ext: string) => {
				if ((ext == "audio" || ext == "video" || ext == "image")){
					this._extensions.concat(MimeTypes[ext])
				}else{
					this._extensions.push(ext);
				}
			});
		}
	}
    
    /**
     * Обработчик события выбора файла в input
     * @param {FileList} selectedFiles
     * @private
     * @method
     * @name Lib/File/ResourceGetter/FSGetter#_onChangeInput
     * @void
     */
	private _onChangeInput(selectedFiles: FileList) {
        if (this.isDestroy()) {
            return;
        }
		let files = [];
    
        /*
         * Нельзя обходить FileList через "for in" + ".hasOwnProperty"
         * https://w3c.github.io/FileAPI/#filelist-section
         * Обход надо делать только по числовому индексу и получать через FileList.item({Number}) или FileList[{Number}]
         */
		for (let i = 0; i < selectedFiles.length; i++) {
            let error: Error;
            let file = selectedFiles.item && selectedFiles.item(i) || selectedFiles[i];
            if (this._extensions.length && !verifyFileExtension(file, this._extensions)) {
                error = new ExtensionsError(file.name);
            }
            files.push(error || new LocalFile(file));
		}
		this._selectDef.callback(files);
		this._selectDef = null;
	}
	destroy() {
        this._form.remove && this._form.remove();
        this._form = null;
        this._inputBtn = null;
        if (this._selectDef && !this._selectDef.isReady()) {
            this._selectDef.cancel();
        }
        super.destroy();
	}
}

export = FSGetter;