/// <amd-module name="Lib/File/ResourceGetter/IResourceGetterBase" />
import {IFileData} from "Lib/File/Interface";
import IResourceGetter = require("Lib/File/ResourceGetter/IResourceGetter");
import Deferred = require("Core/Deferred");

/**
 * Абстрактный класс, реализующий интерфейс получения файлов IResourceGetter
 * <br/>
 * Получение скана:
 * <pre>
 *    require(['js!SBIS3.Plugin/components/Extensions/Integration/FileGetter/ScannerGetter'], function (ScannerGetter) {
 *       var getter = new ScannerGetter();
 *       getter.getFiles().addCallbacks(function(links){ // links: Array<Lib/File/LocalFileLink | Error>
 *          // действия с отсканированным документом
 *       }, function(err) {
 *          log(err);
 *          alert("Сканирование не доступно");
 *       });
 *    });
 * </pre>
 * @public
 * @class
 * @name Lib/File/ResourceGetter/IResourceGetterBase
 * @implements IResourceGetter
 * @author Заляев А.В.
 * @abstract
 */
/**
 * Осуществляет выбор ресурсов
 * @method
 * @name Lib/File/ResourceGetter/IResourceGetterBase#getFiles
 * @abstract
 * @return {Core/Deferred<Array<IFileData | Error>>}
 * @example
 * Получение скана:
 * <pre>
 *    require(['js!SBIS3.Plugin/components/Extensions/Integration/FileGetter/ScannerGetter'], function (ScannerGetter) {
 *       var getter = new ScannerGetter();
 *       getter.getFiles().addCallbacks(function(links){ // links: Array<Lib/File/LocalFileLink | Error>
 *          ...
 *       }, function(err) {
 *          // Ошибка недоступности ResourceGetter
 *          log(err);
 *          alert("Сканирование не доступно");
 *       });
 *    });
 * </pre>
 * Выбор из файловой системы:
 * <pre>
 *    require(['Lib/File/ResourceGetter/FSGetter'], function (FSGetter) {
 *       var getter = new FSGetter({
 *          extensions: ["png", "jpg"]
 *       });
 *       getter.getFiles().addCallbacks(function(files){ // files: Array<Lib/File/LocalFile | Error>
 *          files.forEach(function(f){
 *              if (f instanceOf Error) {
 *                  // Ошибка выбора файла (прим: выбран файл неверного расширения)
 *                  alert("Ошибка выбора файла: " + f.message);
 *              }
 *          });
 *       }, function(err) {
 *          // Ошибка недоступности ResourceGetter
 *          log(err);
 *          alert("Выбор файлов не доступен");
 *       });
 *    });
 * </pre>
 * @see Lib/File/LocalFile
 * @see Lib/File/LocalFileLink
 * @see Lib/File/HttpFileLink
 */
/**
 * Возможен ли выбор ресурса
 * @abstract
 * @method
 * @name Lib/File/ResourceGetter/IResourceGetterBase#canExec
 * @return {Core/Deferred<boolean>}
 * @example
 * Доступны ли сканеры:
 * <pre>
 *    require(['js!SBIS3.Plugin/components/Extensions/Integration/FileGetter/ScannerGetter'], function (ScannerGetter) {
 *       var getter = new ScannerGetter();
 *       getter.canExec().addCallback(function(isCan){
 *          if (!isCan) {
 *              alert("Сканирование не доступно")
 *          }
 *       });
 *    });
 * </pre>
 */

abstract class IResourceGetterBase implements IResourceGetter /** @lends Lib/File/ResourceGetter/IResourceGetterBase.prototype*/{
    protected name: string;
    private _isDestroyed: boolean = false;
    abstract getFiles(): Deferred<Array<IFileData | Error>>;
    abstract canExec(): Deferred<boolean>;
    destroy(): void {
        this._isDestroyed = true;
    };
    isDestroy(): boolean {
        return this._isDestroyed;
    }
    /**
     * Возвращает строку - тип ресурса
     * @method
     * @name Lib/File/ResourceGetter/IResourceGetterBase#getType
     * @return {String}
     */
    getType(): string {
        return this.name || "";
    }
}

export = IResourceGetterBase;
