/// <amd-module name="Lib/File/ResourceGetter/PhotoCam" />

import IResourceGetterBase = require("Lib/File/ResourceGetter/IResourceGetterBase");
import Deferred = require("Core/Deferred");
import LocalFile = require("Lib/File/LocalFile");
import LocalFileLink = require("Lib/File/LocalFileLink");
import detection = require("Core/detection");
import merge = require("Core/core-merge");
import OpenDialog = require("SBIS3.CONTROLS/Action/OpenDialog");

const DIALOG = "Lib/File/ResourceGetter/PhotoCam/Dialog";
const DIALOG_PLUGIN = "Lib/File/ResourceGetter/PhotoCam/DialogPlugin";

type OpenDialogOptions = {
    mode: string;
    template: string;
    dialogOptions: any;
    componentOptions: any;
}
type Option = {
    /**
     * @name Lib/File/ResourceGetter/PhotoCam#openDialogOptions
     * @cfg {Object} dialogOptions Объект параметров, пробрасываемый в OpenDialog.
     * @see SBIS3.CONTROLS/Action/OpenDialog
     */
    openDialog: Partial<OpenDialogOptions>;
}
const DEFAULT: Option = {
    openDialog: {
        mode: "dialog",
        dialogOptions: {
            autoHeight: true,
            resizeable: false,
            autoWidth: true,
            /**
             * Проверить доступность камеры можно только попробовав к ней подключиться
             * В случаях, если
             * а) пльзователь уже запретил доступ к камере
             * б) обращение к медиа-девайсам запрещено политиками безопасности
             * в) камера отсуствует
             * то получим эффект "моргания окошка" которое отрисовывает текст инициализации и тут же дестроится
             *
             * поэтому скрываем диалоговое окно по умолчанию
             */
            visible: false
        },
        template: detection.isIE10 || detection.isIE11? // В IE нет поддержки userMedia в Edge всё норм
            DIALOG_PLUGIN :
            DIALOG
    }
};
/**
 * Класс, для получения фотографии с камеры, реализующий интерфейс IResourceGetter
 * @class
 * @name Lib/File/ResourceGetter/PhotoCam
 * @extends Lib/File/ResourceGetter/IResourceGetterBase
 * @public
 * @author Заляев А.В.
 */
class PhotoCam extends IResourceGetterBase /** @lends Lib/File/ResourceGetter/PhotoCam.prototype*/{
    protected name = "PhotoCam";
    protected _$options : Option;
    private _openDialog: OpenDialog;
    private _chooseDef: Deferred;
    constructor (opt: Partial<Option>) {
        super();
        this._$options = merge(merge({}, DEFAULT), opt || {});
        
        this._openDialog = new OpenDialog(this._$options.openDialog);
    }
    /**
     * Осуществляет получение изображения с веб-камеры устройства
     * @return {Core/Deferred<Array<Lib/File/LocalFile | Lib/File/LocalFileLink>>}
     * @method
     * @name Lib/File/ResourceGetter/PhotoCam#getFiles
     * @see Lib/File/LocalFile
     * @see Lib/File/LocalFileLink
     */
    getFiles(): Deferred<Array<LocalFileLink | LocalFile>> {
        if (this.isDestroy()) {
            return Deferred.fail("Resource getter is destroyed");
        }
        this._chooseDef = new Deferred().addBoth((result) => {
            this._chooseDef = null;
            return result;
        });
        this._openDialog.execute({
            
            componentOptions: {
                resultDeferred: this._chooseDef
            }
        }).addErrback((err) => {
            this._chooseDef.errback(err);
        });
        return this._chooseDef;
    };
    /**
     * Возможен ли выбор файлов
     * @return {Core/Deferred<boolean>}
     * @method
     * @name Lib/File/ResourceGetter/PhotoCam#canExec
     */
    canExec(): Deferred<boolean> {
        return Deferred.success(!this._chooseDef);
    }
}

export = PhotoCam;