/// <amd-module name="Lib/ServerEvent/class/RabbitEnv" />
import {SEB} from "../interfaces";
import Deferred = require('Core/Deferred');
import Xhr = require('Transport/XHRTransport');
import CONST = require('Lib/ServerEvent/class/Constants');
import ConnectOptions = require('Lib/ServerEvent/class/ConnectOptions');

const REQUEST_LIMIT = 2 * 1000;

class RabbitEnv {
    constructor(private watcher: SEB.IWatchDogSystem) {
    }

    /**
     * Пытаемся установить соединение с RabbitMQ.
     * При люббой ошибке связи кроме как от прямого ответа о неработоспособности кролика
     *  пытаемся переподключиться каждые 5 секунд.
     * Повторять процесс соединения нужно полностью, т.к. hash может устареть,
     *  при попытке создания обменника
     * @param {Lib/ServerEvent/class/ConnectOptions} connectOptions
     * @return {Deferred<string>}
     */
    up(connectOptions: ConnectOptions): Deferred<string> {
        let def = new Deferred<string>();
        let timer = Date.now();
        this.tryCreateChannel(connectOptions).addCallback((hash) => {
            if (Date.now() - timer > REQUEST_LIMIT) {
                console.warn(`[STOMP][timeout] /!info request to long: ${Date.now() - timer}ms`);
            }
            def.callback(hash);
        }).addErrback((err: Error) => {
            if (err.message == CONST.ERR_MSG_RABBIT_OFF) {
                def.errback(err);
                return;
            }
            setTimeout(() => {
                this.up(connectOptions).addCallback((hash) => {
                    def.callback(hash);
                })
            }, 5000);
        });

        return def;
    }

    /**
     * Запрос создания обменника на RabbitMQ
     * @param {Lib/ServerEvent/class/ConnectOptions} connectOptions
     * @return {Deferred}
     */
    private tryCreateChannel(connectOptions: ConnectOptions) {
        let url = connectOptions.getUrl();
        return new Xhr({
            url: url,
            method: 'GET',
            dataType: 'json'
        }).execute().addErrback((err) => {
            this.watcher.logConnect({message: 'error get info', url, err});
            return err;
        }).addCallback((response) => {
            if (!response.websocket) {
                return new Error(CONST.ERR_MSG_RABBIT_OFF);
            }
            return connectOptions.hash;
        });
    }
}

export = RabbitEnv;