/// <amd-module name="Lib/ServerEvent/class/creator/CreatorLocalWebSocket" />
import {SEB} from "../../interfaces";
import Deferred = require('Core/Deferred');
import detection = require('Core/detection');
import UserInfo = require('Core/UserInfo');
import moduleStubs = require('Core/moduleStubs');
import LocalStorage = require('Lib/Storage/LocalStorage');
import ConnectOptions = require('Lib/ServerEvent/class/ConnectOptions');
import CONST = require('Lib/ServerEvent/class/transport/Constants');

/**
 * @class Lib/ServerEvent/class/creator/CreatorLocalWebSocket
 * @memberOf module:ServerEvent.class.creator
 */
abstract class CreatorLocalWebSocket implements SEB.ITransportCreator {
    static STOMP_HEARTBEAT_OUT_TIME = 60000;
    static STOMP_HEARTBEAT_IN_TIME = 0;
    static LOGIN = 'stomp_user';
    static PASSWORD = 'stomp_user';
    private isGuestAccess: boolean;
    private ls: LocalStorage = new LocalStorage('SEB');

    constructor(private connectOptions: ConnectOptions) {
        let userName = UserInfo.get('ЛогинПользователя') || '';
        this.isGuestAccess = userName.indexOf('__сбис__гость__') > -1;
    }

    protected abstract getWebSocket(url): Deferred<WebSocket>;

    protected abstract getResources(): Deferred<any[]>;

    isAvailableInEnv(isExclusive: boolean) {
        if (isExclusive) {
            return false;
        }
        let isMobile = detection.isMobileAndroid || detection.isMobileIOS;
        return !isMobile && !detection.safari11 && !this.connectOptions.isDesktop();
    }

    build(hash) {
        return this.chooseTransportConstructor()
            .addCallback<SEB.ITrackedTransport>((ctor) => {
                return this.createTransport(ctor, hash);
            });
    }

    private createTransport(ctor: SEB.ITransportConstructor, hash): Deferred<SEB.ITrackedTransport> {
        let def = new Deferred<SEB.ITrackedTransport>();
        this.getResources().addCallback(() => {
            if (ctor.getLocalName() === 'LocalSlaveTransport') {
                def.callback(new ctor());
                return;
            }
            this.connect().addCallbacks((connect: Stomp.Client) => {
                if (ctor.getLocalName() != 'LocalPageTransport' &&
                    ctor.getLocalName() != 'LocalMainTransport') {
                    return;
                }
                def.callback(new ctor(
                    connect, hash,
                    !this.isGuestAccess,
                    this.connectOptions.exchange
                ));
            }, (err) => {
                def.errback(err);
            });
        });
        return def;
    }

    private chooseTransportConstructor(): Deferred<SEB.ITransportConstructor> {
        let timestamp = this.ls.getItem(CONST.KEY_ACTUAL_DATE);
        if (!timestamp) {
            return moduleStubs.require('Lib/ServerEvent/class/transport/LocalMainTransport')
                .createDependent().addCallback(([transportClass]) => {
                    return transportClass;
                });
        }

        let last = new Date(timestamp + CONST.CHECK_MIN_INTERVAL * 2);
        if (last < new Date() || isNaN(last.valueOf())) {
            return moduleStubs.require('Lib/ServerEvent/class/transport/LocalMainTransport')
                .createDependent().addCallback(([transportClass]) => {
                    return transportClass;
                });
        }

        return moduleStubs.require('Lib/ServerEvent/class/transport/LocalSlaveTransport')
                .createDependent().addCallback(([transportClass]) => {
                    return transportClass;
                });
    }

    private connect(): Deferred<Stomp.Client> {
        let def = new Deferred<Stomp.Client>();
        try {
            this.getWebSocket(this.connectOptions.getWebSocketUrl()).addCallback((webSocket) => {
                let stomp = Stomp.over(webSocket);
                stomp.heartbeat.outgoing = CreatorLocalWebSocket.STOMP_HEARTBEAT_OUT_TIME;
                stomp.heartbeat.incoming = CreatorLocalWebSocket.STOMP_HEARTBEAT_IN_TIME;
                stomp.debug = function () {};
                stomp.connect(
                    CreatorLocalWebSocket.LOGIN,
                    CreatorLocalWebSocket.PASSWORD,
                    () => {
                        def.callback(stomp);
                    },
                    (data) => {
                        if (def.isReady()) {
                            return console.warn(data);
                        }
                        def.errback(data);
                    }
                );
            }).addErrback((e) => {
                def.errback(e);
            });
        } catch (e) {
            def.errback(e);
        }

        return def;
    }
}

export = CreatorLocalWebSocket;