/// <amd-module name="Lib/ServerEvent/class/transport/LocalMainTransport" />
import {SEB} from "../../interfaces";
import CONST = require('Lib/ServerEvent/class/transport/Constants');
import LocalPageTransport = require('Lib/ServerEvent/class/transport/LocalPageTransport');
import TabMessage = require('Lib/TabMessage/TabMessage');
import Subscribe = require('Lib/ServerEvent/class/Subscribe');
import LocalStorage = require('Lib/Storage/LocalStorage');
import LocalStorageNative = require('Core/LocalStorageNative');

class LocalMainTransport implements SEB.ITrackedTransport {
    private transport: LocalPageTransport;
    private ls: LocalStorage = new LocalStorage(CONST.LS_PREFIX);
    private taber: TabMessage = new TabMessage();
    private mainChecker: number;

    private subFromTaber: (event, data: SEB.ISubscribeData) => void;
    private unsubFromTaber: (event, data: SEB.ISubscribeData) => void;
    private fnQuit: () => void;

    /**
     * Ключ, котоырй указывает, что он не единственный мастер
     * Стейт появляется лишь при закрытии, что бы оно пошло по другому.
     * @type {boolean}
     */
    private isAloneMain = true;
    /**
     * Необходима, что бы при обаружении смены мастера, вызвать пересоздание
     */
    private disconnectCallback: Function;

    constructor(private stomp: Stomp.Client, private hash: string,
                private persistent: boolean = true, private exchangeName: string) {
        /**
         * Нужно всем сообщить, что мастер готов.
         */
        this.fixMain();

        //region Разрешение гонок создания
        /**
         * Кто последний стал мастером, тот и главный.
         * Все асинхронное, поэтому проверяем по дате создания
         * @type {number}
         */
        let now = Date.now();
        this.taber.notify(CONST.KEY_MAINREADY, now);
        this.fnQuit = (event, other) => {
            if (other < now) {
                return;
            }
            this.taber.unsubscribe(CONST.KEY_MAINREADY, this.fnQuit);
            this.isAloneMain = false;
            this.close("Detect later main.");
        };
        this.taber.subscribe(CONST.KEY_MAINREADY, this.fnQuit);
        //endregion

        this.ls.removeItem(CONST.KEY_TRY_CREATE_MAIN);
        this.mainChecker = setInterval(() => {
            this.fixMain();
        }, CONST.CHECK_MIN_INTERVAL);

        let isAck = false;
        if ('true' == LocalStorageNative.getItem('shared-bus-ack')) {
            isAck = true;
        }
        this.transport = new LocalPageTransport(
            stomp, hash, persistent, exchangeName, isAck
        );

        this.subFromTaber = (event, data: SEB.ISubscribeData) => {
            this.subscribe(Subscribe.create(data.channelName, data.channeled, false, data.target));
        };
        this.unsubFromTaber = (event, data: SEB.ISubscribeData) => {
            this.unsubscribe(Subscribe.create(data.channelName, data.channeled, false, data.target));
        };
        this.taber.subscribe(CONST.KEY_SUBSCRIBE, this.subFromTaber);
        this.taber.subscribe(CONST.KEY_UNSUBSCRIBE, this.unsubFromTaber);

        window && window.addEventListener('unload', () => {
            this.close();
        });
    }

    private fixMain() {
        this.ls.setItem(CONST.KEY_ACTUAL_DATE, Date.now());
    }

    static getLocalName() {
        return 'LocalMainTransport';
    }

    subscribe(subscribe) {
        this.transport.subscribe(subscribe);
    }

    unsubscribe(subscribe: SEB.ISubscribe) {
        this.transport.unsubscribe(subscribe);
    }

    destructor(msg?) {
        /*
         Может прилететь второй вызов при остановке транспорта черзе хедлер
         */
        if (this.taber) {
            this.taber.unsubscribe(CONST.KEY_SUBSCRIBE, this.subFromTaber);
            this.taber.unsubscribe(CONST.KEY_UNSUBSCRIBE, this.unsubFromTaber);
            this.taber.unsubscribe(CONST.KEY_MAINREADY, this.fnQuit);
            this.taber = undefined;
        }

        if (this.isAloneMain) {
            this.ls.removeItem(CONST.KEY_ACTUAL_DATE);
        }
        clearInterval(this.mainChecker);
        let callback = this.disconnectCallback;
        this.disconnectCallback = () => {
        };
        callback(msg);
    }

    /**
     * Обработчик вызывается после деструктора
     * @param {(event) => any} fn
     */
    setDisconnectHandler(fn: (event) => any) {
        this.disconnectCallback = fn;
        this.transport.setDisconnectHandler((evt) => {
            this.destructor();
        });
    }

    /**
     * Закрыте транспорта.
     * При закрытии транспорта не вызывается DisconnectHandler
     */
    close(msg?) {
        this.transport.close();
        this.destructor(msg);
    }

    setDelivery(delivery: SEB.IEventDeliver) {
        this.transport.setDelivery(delivery);
    }

    setWatchDog(watcher: SEB.IWatchDog) {
        this.transport.setWatchDog(watcher);
    }
}

export = LocalMainTransport;
