define("Lib/ServerEvent/class/transport/TransportChooser", ["require", "exports", "Core/Deferred", "Lib/ServerEvent/class/logger/WatchDogAggregator", "Lib/ServerEvent/class/RabbitEnv", "Lib/ServerEvent/class/creator/CreatorSharedWorker", "Lib/ServerEvent/class/creator/CreatorWebSocket", "Lib/ServerEvent/class/creator/CreatorSockJs", "Lib/ServerEvent/class/creator/CreatorExclusive", "Lib/ServerEvent/class/creator/CreatorExclusiveSockJs", "Lib/ServerEvent/class/deliver/DeliveryChooser"], function (require, exports, Deferred, Watcher, RabbitEnv, CreatorSharedWorker, CreatorWebSocket, CreatorSockJs, CreatorExclusive, CreatorExclusiveSockJs, DeliveryChooser) {
    "use strict";
    /**
     * @class Lib/ServerEvent/class/transport/TransportChooser
     * @memberOf module:ServerEvent.class.transport
     */
    var TransportChooser = /** @class */ (function () {
        /**
         * @param {Lib/ServerEvent/class/ConnectOptions} connectOptions
         * @param {Function} onclose коллбэк закрытия соединения. Пробрасывается в транспорт
         * @param {boolean} isExclusive создавать ли выделенное соединение для вкладки
         * @param {SEB.IWatchDogSystem} watcher класс логирования
         */
        function TransportChooser(connectOptions, onclose, isExclusive, watcher) {
            if (onclose === void 0) { onclose = function (v) {
                return v;
            }; }
            if (isExclusive === void 0) { isExclusive = false; }
            if (watcher === void 0) { watcher = new Watcher(); }
            this.connectOptions = connectOptions;
            this.onclose = onclose;
            this.isExclusive = isExclusive;
            this.watcher = watcher;
            this.builders = [];
            this.deliveryChooser = new DeliveryChooser(watcher);
            var strategies = [
                CreatorSharedWorker,
                CreatorWebSocket,
                CreatorSockJs,
                CreatorExclusive,
                CreatorExclusiveSockJs
            ];
            for (var _i = 0, strategies_1 = strategies; _i < strategies_1.length; _i++) {
                var strategy = strategies_1[_i];
                var builder = new strategy(this.connectOptions);
                if (!builder.isAvailableInEnv(this.isExclusive)) {
                    continue;
                }
                this.builders.push(builder);
            }
        }
        TransportChooser.prototype.destructor = function () {
            var _this = this;
            this.defStrategy && this.defStrategy.addCallback(function (transport) {
                _this.watcher && _this.watcher.logDisconnect('Manual disconnect');
                transport.close();
            });
        };
        /**
         * Выбор стратегии транспорта сообщений.
         * @return {Deferred<ITransport>}
         */
        TransportChooser.prototype.choose = function () {
            var _this = this;
            if (this.defStrategy) {
                return this.defStrategy.createDependent();
            }
            this.defStrategy = this.build();
            this.defStrategy.addCallback(function (transport) {
                transport.setDisconnectHandler(function (msg) {
                    _this.defStrategy = undefined;
                    _this.onclose(msg);
                });
                transport.setWatchDog(_this.watcher);
                _this.deliveryChooser.choose(transport)
                    .addCallback(function (deliver) {
                    transport.setDelivery(deliver);
                });
                return transport;
            });
            return this.defStrategy.createDependent();
        };
        /**
         * Метод строит транспорт и возвращает deferred
         * @return {Deferred}
         */
        TransportChooser.prototype.build = function () {
            var _this = this;
            var rbt = new RabbitEnv(this.watcher);
            return rbt.up(this.connectOptions).addCallback(function (hash) {
                return _this.createTransport(hash);
            });
        };
        TransportChooser.prototype.createTransport = function (hash) {
            var def = new Deferred();
            var _loop_1 = function (builder) {
                def.addErrback(function () {
                    return builder.build(hash);
                });
            };
            for (var _i = 0, _a = this.builders; _i < _a.length; _i++) {
                var builder = _a[_i];
                _loop_1(builder);
            }
            def.addErrback(function (err) {
                console.error("No one server event bus transport choice.");
                return err;
            });
            /**
             * Запускаем цепочку выбора
             */
            def.errback();
            return def;
        };
        return TransportChooser;
    }());
    return TransportChooser;
});
