/// <amd-module name="Lib/Storage/LocalStorage" />

import Abstract = require("Core/Abstract");
import EventBus = require("Core/EventBus");
import Serializer = require("Core/Serializer");
import simpleExtend = require("Core/core-simpleExtend");
import isPlainObject = require("Core/helpers/Object/isPlainObject");
import localStorage = require("Lib/Storage/ie/LocalStorageNative");

///region Const
const global = (function () { return this || (0, eval)('this'); })();
// Ключ по которому хранится список полей, которые необходимо сохранить в хранилище при смене sid
const WS_STORE_ALWAYS_KEYS = "ws-store-always-keys";
// Список всех событий хранилища
const EVENT_LIST = ["onChange", "onRemove", "onClear"];
// Канал для передачи всех событий хранилища
const channel = () => EventBus.channel("local_storage");
// Экземпляр сериализатора
let serializer = new Serializer();
///endregion const
///region EventBus
let isAddedGlobalStorageHandler = false;
function addStorageEventListener() {
    if (isAddedGlobalStorageHandler || !global) {
        return;
    }
    isAddedGlobalStorageHandler = true;
    if (global.addEventListener) {
        return global.addEventListener("storage", storageGlobalHandler, false);
    }
}
// storage handlers
/**
 * Глобальный обработчик события onstorage
 * @param {StorageEvent} event
 */
function storageGlobalHandler(event) {
    if (!event && global.event) {
        event = global.event;
    }
    if (!event) {
        return;
    }
    let {key, newValue, oldValue} = event;
    // Если в событии отсуствует ключ -> было вызвана очистка хранилища
    if (!key) {
        channel().notify("onClear");
        return;
    }
    if (!newValue) {
        return channel().notify("onRemove", key);
    }
    return channel().notify("onChange", key, newValue);
}
/**
 * Обработчик события, который вызовится у каждого инстанса LocalStorage
 * @param {Core/EventObject} event
 * @param {String} key
 * @param {*} value
 * @return {*}
 */
function storageHandler(event, key, value) {
    // Если нету ключа => onClear, иначе смотрим на то чтобы ключ начинался на наш префикс
    if (!key || startsWith(key, this.prefix)) {
        return this._notify(event.name, removePrefix(key, this.prefix), deserializeItem(value));
    }
}
///endregion EventBus
///region LocalStorageNative
/**
 * Возвращение десериализованной записи хранилища по ключу
 * @param {String} key
 * @return {*}
 */
function getItem(key) {
    return deserializeItem(localStorage.getItem(key));
}
/**
 * Возвращает массив всех ключей, содержащих тот же префикс
 * при этом сам префикс не обрезаются
 * @param prefix префикс по которому необходимо отфильтровать ключи
 * @return {Array<String>}
 */
function getKeys(prefix: string): string[] {
    return localStorage.getKeys().filter((key: string) => {
        return startsWith(key, prefix);
    })
}
///endregion LocalStorageNative
///region utils
/**
 * Проверка начинается ли строка с конструкциюи вида prefix/, если передан префикс.
 * в случае отсуствия префикса вернёт true
 * @param {String} str
 * @param {String} prefix
 * @return {Boolean}
 */
function startsWith(str, prefix) {
    return !prefix || str.indexOf(prefix + "/") === 0;
}
/**
 * Добавляет префикс к строке
 * @param {String} str
 * @param {String} prefix
 * @return {String}
 */
function addPrefix(str, prefix) {
    return prefix && (prefix + "/" + str) || str;
}
/**
 * Возвращает строку без профекса
 * @param {String} str
 * @param {String} prefix
 * @return {String}
 */
function removePrefix(str, prefix) {
    return prefix && startsWith(str, prefix) ? str.substr(prefix.length + 1) : str;
}
/**
 * Сериализация даных перед записью в хранилище
 * @param {*} item
 * @return {String}
 * @private
 */
function serializeItem(item) {
    let json = item;
    /*
     * У объектов WS.Data.Entity/* и WS.Data.Collection/* перед сереализацией необходимо вызвать toJSON
     * Для корректрой работы серилизации и обратной десерилизации в Record|Model|RecordSet|...
     */
    if (item && typeof item.toJSON === "function") {
        json = item.toJSON();
    }
    try {
        /**
         * Core/Serializer не умеет нормально сериализовывать сложные объекты,
         * инстансы каких-либо классов (не прямой потомок Object)
         */
        if (!json || isPlainObject(json)) {
            return JSON.stringify(json, serializer.serialize)
        }
        return JSON.stringify(json);
    }
    catch (err) {
        return "" + json;
    }
}
/**
 * Десериализация данных
 * @param {String} item
 * @return {*}
 * @private
 */
function deserializeItem(item) {
    if (!item) {
        return null;
    }
    let result;
    try {
        result = JSON.parse(item, serializer.deserialize);
    }
    catch (err) {
        return item;
    }
    return result;
}
///endregion utils
/**
 * Класс для работы с локальным хранилищем и его событиями
 * Позволяет хранить и получать сложные типы данных, включая WS.Data/Entity/* и WS.Data/Collections/*
 * @class Lib/Storage/LocalStorage
 * @param {String} prefix Префикс хранимых в хранилище ключей
 * @param {Boolean} [storeAlways] Необходимо ли сохранять значение при смене пользователя. Используется для хранения конфигурации, не привязаной к пользовательским данным.
 * @public
 * @author Заляев А.В.
 */
let LocalStorage = simpleExtend.extend(Abstract, /** @lends Lib/Storage/LocalStorage.prototype*/ {
    /**
     * @cfg {string} [prefix] Префикс хранилища
     * @example
     * Изменить подпись флага в зависимости от его состояния.
     * <pre>
     *  require(["Lib/Storage/LocalStorage"], function(LocalStorage){
     *      storage = new LocalStorage("news");
     *      storage.setItem("lastFilter", {...});
     *      let resetFilter = function(filterValue){
     *          // Обработчик события изменения значения в хранилище
     *          // Обработчик не будет вызван на вкладке. производящей изменения
     *          ...
     *      };
     *      storage.subscribe("onChange", resetFilter);
     *      storage.subscribe("onRemove", resetFilter);
     *  });
     * </pre>
     */
    prefix: "",
    /**
     * @cfg {Boolean} [storeAlways] Необходимо ли сохранять значение при смене пользователя. Используется для хранения конфигурации, не привязаной к пользовательским данным.
     */
    storeAlways: false,
    constructor(prefix, storeAlways) {
        Abstract.apply(this, arguments);
        this.prefix = prefix || "";
        this.storeAlways = storeAlways || false;
        this._onStorage = storageHandler.bind(this);
        EVENT_LIST.forEach((event) => {
            channel().subscribe(event, this._onStorage);
        });
        this._publish.apply(this, EVENT_LIST);
        addStorageEventListener();
    },
    destroy() {
        EVENT_LIST.forEach((event) => {
            channel().unsubscribe(event, this._onStorage);
        });
        return LocalStorage.superclass.destroy.call(this);
    },
    /**
     * Устанавливает значение в хранилище.
     * @param {String} key Ключ записываемого поля хранилища.
     * @param {*} item Значение.
     * @return {Boolean}
     */
    setItem(key, item): boolean {
        if (typeof item === "undefined") {
            return this.removeItem(key);
        }
        let fullKey = addPrefix(key, this.prefix);
        let resultItem = serializeItem(item);
        let res = localStorage.setItem(fullKey, resultItem);
        if (!res || !this.storeAlways) { return res }
        let keepKeys = getItem(WS_STORE_ALWAYS_KEYS) || [];
        if (keepKeys.indexOf(fullKey) === -1) {
            keepKeys.push(fullKey);
            localStorage.setItem(WS_STORE_ALWAYS_KEYS, serializeItem(keepKeys));
        }
        return res;
    },
    /**
     * Возвращает значение хранилища.
     * @param {String} key Ключ возвращаемого поля хранилища.
     */
    getItem(key) {
        return getItem(addPrefix(key, this.prefix));
    },
    /**
     * Устанавливает значение в хранилище.
     * @param {String} key Ключ удаляемого поля хранилища.
     * @void
     */
    removeItem(key) {
        return localStorage.removeItem(addPrefix(key, this.prefix));
    },
    /**
     * Очищает хранилище от значений имеющих тот же префикс, что и текущий экземпляр класса.
     * @void
     */
    clear() {
        getKeys(this.prefix).forEach((key: string) => {
            localStorage.removeItem(key);
        });
    },
    /**
     * Возвращает массив всех ключей, содержащих тот же префикс
     * @return {Array<String>}
     */
    getKeys(): string[] {
        return getKeys(this.prefix).map((key: string) => {
            return removePrefix(key, this.prefix);
        });
    },
    /**
     * Возвращает объект, содержащий все значения хранилища, соответствующие текущему префиксу
     * @return {Object}
     */
    toObject(): {
        [propName: string]: any
    } {
        let result = {};
        getKeys(this.prefix).forEach((key: string) => {
            result[key] = getItem(key);
        });
        return result;
    }
});
/**
 * Очищает всё хранилище вне зависимости от используемых префиксов.
 * @static
 * @void
 */
LocalStorage.clearAll = () => {
    localStorage.clear();
};

export =  LocalStorage;

/**
 * @event onChange
 * Происходит при изменении/добавлении данных в хранилище.
 * @name Lib/Storage/LocalStorage#onChange
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {String} key Ключ добавленного/изменённого в хранилище поля.
 * @param {*} value Значение поля в хранилище.
 */
/**
 * @event onRemove
 * Происходит при удалении данных из хранилища.
 * @name Lib/Storage/LocalStorage#onRemove
 * @param {Core/EventObject} eventObject Дескриптор события.
 * @param {String} key Ключ удалённого из хранилища поля.
 */
/**
 * @event onClear
 * Происходит при очистке хранилища.
 * @name Lib/Storage/LocalStorage#onClear
 * @param {Core/EventObject} eventObject Дескриптор события.
 */
