define("Lib/Storage/ie/LocalStorageNative", ["require", "exports", "Core/detection", "Core/EventBus", "Core/LocalStorageNative"], function (require, exports, detection, EventBus, localStorageNative) {
    "use strict";
    /**
     * Обёртка над Core/LocalStorageNative для выполнения операций событийными хаками
     * 1) Заглушка событий, вознихших на той же вкладке, где были произведены изменения в хранилище
     * 2) Эмуляция не возникщего события, при записи в хранилище полей длинее ~2kb
     */
    var GLOBAL = (function () { return this || (0, eval)('this'); })();
    // Директория системных данных модуля
    var SYSTEM_STORAGE_NAME = "__storage";
    // Канал для передачи всех событий хранилища
    var channel = function () { return EventBus.channel("local_storage"); };
    // Ключ по которому хранится список полей, которые необходимо сохранить в хранилище при смене пользователя
    var WS_STORE_ALWAYS_KEYS = "ws-store-always-keys";
    var WS_STORAGE_KEYS = "ws-local-keys";
    // Имя записи, содержащее последнее изменённое поле хранилища для IE.
    var FIX_EVENT_KEY = addPrefix("ie-fix-event", SYSTEM_STORAGE_NAME);
    // Хэш-данные по последнему пойманному события хранилища для IE
    var fixEventsData = [];
    /**
     *
     * @param {String} value
     * @return {Boolean}
     */
    function isNeedFixEvent(value) {
        return value.length >= 2143;
    }
    // заглушка событий на той же вкладке
    var IE_ITEM_PREFIX_REGEXP = /\$ie-fix\{(\d+)\}/i;
    var WINDOW_ID = Math.random().toString().substr(2);
    ///region EventBus
    var lastEvent;
    /**
     * Глобальный обработчик события onstorage
     * @param {StorageEvent} event
     */
    function storageGlobalHandler(event) {
        event = event || GLOBAL.event;
        if (!event) {
            return;
        }
        var key = event.key, newValue = event.newValue, oldValue = event.oldValue;
        if (newValue == oldValue) {
            // отсекаем обработку событий, если данные в хранилище не поменялись
            event.stopPropagation();
            event.stopImmediatePropagation();
            return;
        }
        /**
         * В редких версиях IE есть баг, когда в iframe вызывается два события storage вместо одного
         * смотрим когда у нас данные события совпадают с последним событием, и не старше 2 секунд
         * https://connect.microsoft.com/IE/feedback/details/811546/ie11-localstorage-events-fire-twice-or-not-at-all-in-iframes
         */
        var date = Date.now();
        if (lastEvent && (lastEvent.key === key) && (lastEvent.newValue === newValue) &&
            (lastEvent.oldValue === oldValue) && (date - lastEvent.date <= 2 * 1000)) {
            return;
        }
        lastEvent = { key: key, newValue: newValue, oldValue: oldValue, date: date };
        var id, needNewEvent;
        if (newValue) {
            var match = newValue.match(IE_ITEM_PREFIX_REGEXP);
            if (match) {
                newValue = newValue.replace(match[0], "");
                id = match[1];
                needNewEvent = true;
                event.stopPropagation();
                event.stopImmediatePropagation();
            }
        }
        // Если так же вкладка, то завершаем обработку
        if (id === WINDOW_ID) {
            return;
        }
        if (startsWith(key, SYSTEM_STORAGE_NAME)) {
            return storageSystemHandler(key, newValue, oldValue);
        }
        if (isNeedFixEvent(newValue)) {
            // если пришло событие, которое прошло условие для вызова фикса,
            // то запомним его, чтобы в событии фикса не вызвать его повторно
            fixEventsData.push({
                key: key,
                length: newValue.length
            });
        }
        /*
         * Если нашли оверхед, но вкладка не инициатор, то надо заглушить текущее событие и вызвать новое, без оверхеда,
         * чтобы дальнейшие обработчики не знали ничего о нём
         */
        if (needNewEvent) {
            channel().notify('onChange', key, newValue);
        }
    }
    /**
     * Обработчик событий, вызванных самим модулем
     * @param {String} key Ключ изменённого поля хранилища
     * @param {String} value Новое значение
     * @param {String} oldValue Предыдущее значение
     */
    function storageSystemHandler(key, value, oldValue) {
        var eventKey = removePrefix(key, SYSTEM_STORAGE_NAME);
        var systemItem = value;
        if (eventKey === FIX_EVENT_KEY) {
            /*
             * Посылаем асинхронно т.к. поулчении данных для восстановления события
             * размером ~~5kb и больше мы получем null
             * т.к. фактическая запись в хранилище происходит одновреммено с вызовом событий
             */
            setTimeout(function () {
                var value = localStorage.getItem(systemItem);
                var ieFixData = fixEventsData.filter(function (data) {
                    return (data.key === systemItem) && (data.length === value.length);
                });
                // Если нету в списке значений, которые прошли условие фикс события,
                // то значит событие по этому ключу не пришло
                if (!ieFixData) {
                    channel().notify('onChange', systemItem, value);
                }
                // убираем из списка обработанных событий
                fixEventsData = fixEventsData.filter(function (data) {
                    return data !== ieFixData;
                });
            }, 10);
        }
    }
    ///endregion EventBus
    ///region utils
    function removeWindowId(item) {
        return item && item.replace(IE_ITEM_PREFIX_REGEXP, "") || item;
    }
    /**
     * Проверка начинается ли строка с конструкциюи вида prefix/, если передан префикс.
     * в случае отсуствия префикса вернёт true
     * @param {String} str
     * @param {String} prefix
     * @return {Boolean}
     */
    function startsWith(str, prefix) {
        return !prefix || str.indexOf(prefix + "/") === 0;
    }
    /**
     * Добавляет префикс к строке
     * @param {String} str
     * @param {String} prefix
     * @return {String}
     */
    function addPrefix(str, prefix) {
        return prefix && (prefix + "/" + str) || str;
    }
    /**
     * Возвращает строку без профекса
     * @param {String} str
     * @param {String} prefix
     * @return {String}
     */
    function removePrefix(str, prefix) {
        return prefix && startsWith(str, prefix) ? str.substr(prefix.length + 1) : str;
    }
    ///endregion utils
    ///region localStorage
    var localStorage;
    if (detection.isIE) {
        localStorage = {
            /**
             * Возвращение десериализованной записи хранилища по ключу
             * @param {String} key
             * @return {*}
             */
            getItem: function (key) {
                return removeWindowId(localStorageNative.getItem(key));
            },
            /**
             * Устанавливает значение в localStorage по ключу.
             * @param {String} key - ключ
             * @param {String} item - Значение
             * @return {Boolean} Флаг успешности записи
             */
            setItem: function (key, item) {
                // поля для внутренних механизмов оставим как есть
                if (key === WS_STORE_ALWAYS_KEYS || key === WS_STORAGE_KEYS) {
                    return localStorageNative.setItem(key, item);
                }
                // Докидываем оверхед в виде строки с id вкладки, чтобы в дальнейшем по нему заглушить события
                item = "$ie-fix{" + WINDOW_ID + "}" + item;
                var res = localStorageNative.setItem(key, item);
                // Фикс для IE, т.к событие onstorage не стрельнет, если длина сообщения больше 2143 символа
                // f3f77e596dd3d19889490fff390b3ef5e4b44d9e
                if (res && isNeedFixEvent(item)) {
                    localStorageNative.setItem(FIX_EVENT_KEY, key);
                    localStorageNative.removeItem(FIX_EVENT_KEY);
                }
                return res;
            }
        };
        // Добавим методы, которые не нуждаются в обёртке
        ["getLength", "getKeys", "removeItem", "clear"].forEach(function (f) {
            localStorage[f] = localStorageNative[f].bind(localStorage);
        });
        if (GLOBAL.addEventListener) {
            GLOBAL.addEventListener("storage", storageGlobalHandler, false);
        }
    }
    else {
        localStorage = localStorageNative;
    }
    return localStorage;
});
