/// <amd-module name="Lib/TabMessage/TabMessage" />

import LocalStorage = require("Lib/Storage/LocalStorage");
import EventBus = require("Core/EventBus");
import detection = require("Core/detection");

const NAME = "TabMessage";
let removeTimers = {};

/**
 * @param {Storage} storage
 * @param {String} key
 */
function remove(storage, key: string) {
    if (!detection.isIE) {
        return storage.removeItem(key);
    }
    /**
     * Необходимо почистить хранилище за собой, однако в IE событие onStorage на больших строках не вызывается
     * поэтому внутри он записывает вспомогательное значение. чтобы вызвать фейкоое событие на всех вкладках
     * тут получается проблема, что есть делать set + remove на одной вкладке, вспомогательное событие на другой
     * может отработать позже remove и когда полезет за данными через get их уже не будет там,
     * что в итоге приведёт к двум событиям onremove на вкладках получателях.
     */
    if (removeTimers[key]) {
        clearTimeout(removeTimers[key]);
    }
    removeTimers[key] = setTimeout(() => {
        storage.removeItem(key);
    }, 500);
}

type Handler = (event, data: any) => void;

/**
 * Класс, позволяющий пересылать сообщения между вкладками
 *
 * <pre>
 *     // ...
 *     tabMessage = new TabMessage();
 *     // подписываемся на изменение статуса загрузки
 *     tabMessage.subscribe("upload-status", function (event, data){
 *          switch (data.get('status')) {
 *              case "start": {
 *                  log("start upload", data.get('file'));
 *                  break;
 *              }
 *              case "finish": {
 *                  log("start finish", data.get('file'), data.get('result'));
 *                  break;
 *              }
 *              case "progress": {
 *                  log("start progress", data.get('file'), );
 *                  break;
 *              }
 *          }
 *     });
 *     // ...
 *     // при старте загрузки с одной вкладки отправляем статус на другие вкладки
 *     self.getChildControlByName("uploadBtn").subscribe("onactivated", function (event){
 *          self.upload();
 *          var record = self.getUploadInfo();
 *          tabMessage.notify("upload-status", record);
 *     });
 * </pre>
 * Уведомление об открытом документе
 * <pre>
 *     // Пример кода, который выполняется при открытии документа
 *     require(["Lib/TabMessage/TabMessage"], function (TabMessage) {
 *         var tabMessage = new TabMessage();
 *         var documentId = 245;
 *         var eventName = "document.open:%id".replace("%id", documentId);
 *         tabMessage.subscribe(eventName, function (event, data) {
 *             if (data.readOnly) {
 *                 return;
 *             }
 *             // Нашалогика. Решаем кто и как будет реагировать. Например, покажем пользователю открытый документ*
 *             console.warn("Документ уже открыт на другой вкладке.");
 *             alert("Вот открытый вами документ.");
 *         });
 *         tabMessage.notify(eventName, {readOnly: true})
 *     ;});
 * </pre>
 *
 * @class Lib/TabMessage/TabMessage
 * @public
 * @author Заляев А.В.
 */
class TabMessage {
    _storage;
    _channel: EventBus;
    constructor() {
        this._storage = new LocalStorage(NAME);
        this._channel = EventBus.channel(NAME + "-" + Math.random().toFixed(6).toString().substr(2), {});
        this._storage.subscribe('onChange', (event, messageName, message) => {
            this._channel.notifyWithTarget(messageName, this, message);
        });
    }
    /**
     * Разрушить экземпляр класса.
     * @method
     * @name Lib/TabMessage/TabMessage#destroy
     */
    destroy() {
        this._storage.destroy();
        if (this._channel) {
            this._channel.unsubscribeAll();
            this._channel.destroy();
        }
    }
    /**
     * Добавить обработчик на меж-оконное сообщение.
     * @method
     * @name Lib/TabMessage/TabMessage#subscribe
     * @param {String} messageName Имя сообщения, на которое следует подписать обработчик.
     * @param {Function} handler Функция-делегат, обработчик сообщения.
     * @return {Lib/TabMessage/TabMessage} Экземпляр класса.
     * @example
     * <pre>
     *     ...
     *     tabMessage.subscribe("best-button-click", function (e, messageData){
     *          log("my favorite button is click!");
     *     });
     * </pre>
     */
     subscribe(messageName: string, handler: Handler) {
        this._channel.subscribe(messageName, handler, this);
        return this;
     }
    /**
     * Выполнить обработчик меж-оконного сообщения единожды.
     * @method
     * @name Lib/TabMessage/TabMessage#once
     * @param {String} messageName Имя сообщения, на которое следует подписать обработчик.
     * @param {Function} handler Функция-делегат, обработчик сообщения.
     * @return {Lib/TabMessage/TabMessage} Экземпляр класса.
     */
     once(messageName: string, handler: Handler) {
        this._channel.once(messageName, handler, this);
        return this;
     }
    /**
     * Снять обработчик с указанного меж-оконного сообщения.
     * @method
     * @name Lib/TabMessage/TabMessage#unsubscribe
     * @param {String} messageName Имя сообщения, на которое следует подписать обработчик.
     * @param {Function} handler Функция-делегат, обработчик сообщения.
     * @return {Lib/TabMessage/TabMessage} Экземпляр класса.
     */
    unsubscribe(messageName: string, handler: Handler) {
        this._channel.unsubscribe(messageName, handler);
        return this;
    }
    // dispatchEvent
    /**
     * Отправить меж-оконное сообщение.
     * @method
     * @name Lib/TabMessage/TabMessage#notify
     * @param {String} messageName Имя сообщения, на которое следует подписать обработчик.
     * @param {*} data Данные сообщения.
     * @example
     * <pre>
     *     ...
     *     bestButtonHeandler: function (event){
     *     ...
     *          tabMessage.notify("best-button-click", {
     *              date:       Date.now(),
     *              property:   "text"
     *          });
     *     }
     * </pre>
     */
    notify(messageName: string, data: any) {
        if (data == undefined) {
            data = "" + data;
        }
        this._storage.setItem(messageName, data);
        remove(this._storage, messageName);
    }
}

export = TabMessage;

