define('Transport/RPCJSON', [
   'Core/core-extend',
   'Core/Deferred',
   'Core/constants',
   'Core/UserInfo',
   'Core/IoC',
   'Core/EventBus',
   'Core/helpers/Object/isEmpty',
   'Core/helpers/random-helpers',
   'Core/helpers/string-helpers-min',
   'Transport/HTTPError',
   'Transport/TransportError',
   'Core/base64',
   'Core/i18n',
   'Transport/XHRTransport'
], function(coreExtend, Deferred, constants, userInfo, IoC, EventBus, isEmptyObject, randomHelpers, strHelpersMin, HTTPError, TransportError, base64, i18n) {

   var FALLBACK_EVENT_NAME = 'system.bl_async_replies';

   /**
    * Подготавливает пакет для отправки запроса JSON RPC.
    * @param {String} method Название метода.
    * @param {Object} params Параметры метода.
    * @param {String|Number} [id=1] Идентификатор.
    * @returns {Object} Объект с тремя полями:
    * <ul>
    *    <li>reqBody — тело запроса в виде строки,</li>
    *    <li>reqHeaders — объект с необходимыми заголовками.</li>
    *    <li>reqUrl — строка с параметрами для GET запроса.</li>
    * </ul>
    * @see prepareSorting
    * @see prepareGetRPCInvocationURL
    * @see fetchFile
    * @see newRecordSet
    */
   function jsonRpcPreparePacket(method, params, id) {
      var
         body = {
            jsonrpc: '2.0',
            protocol: constants.JSONRPC_PROOTOCOL_VERSION,
            method: method,
            params: params,
            id: id !== undefined ? id : 1
         },
         headers = {
            'X-CalledMethod'        : strHelpersMin.transliterate("" + method),
            'X-OriginalMethodName'  : base64.encode("" + method),
            'Accept-Language'       : (i18n.getLang() || 'ru-RU') + ';q=0.8,en-US;q=0.5,en;q=0.3'
         },
         url = '?protocol=' + constants.JSONRPC_PROOTOCOL_VERSION +
            '&method=' + encodeURI(method) +
            '&params=' + encodeURIComponent(strHelpersMin.serializeURLData(params)) +
            '&id=' + (id !== undefined ? id : 1);

      return {
         reqBody: JSON.stringify(body),
         reqUrl: url,
         reqHeaders: headers
      };
   }

   function jsonRpcPrepareBodyHeaders(method, params) {
      var body = {
            jsonrpc: '2.0',
            protocol: constants.JSONRPC_PROOTOCOL_VERSION,
            method: method,
            params: params,
            id: 1
         };
      var headers = {
         'X-CalledMethod': strHelpersMin.transliterate("" + method),
         'X-OriginalMethodName': base64.encode("" + method),
         'Accept-Language': (i18n.getLang() || 'ru-RU') + ';q=0.8,en-US;q=0.5,en;q=0.3'
      };

      return {
         reqBody: JSON.stringify(body),
         reqHeaders: headers
      };
   }

   function isGetMethod(method) {
      var cachedMethods = (typeof window !== 'undefined') && !isEmptyObject(window.cachedMethods || []) ?
            window.cachedMethods :
            [];
      return  cachedMethods.indexOf(method) > -1;
   }

   /**
    * Пытаемся поправить req. Если это невозможно то возвращаем false
    * @param req {Object} in. Объект запроса. Изменяется функцией.
    * @param method
    * @param params
    * @return {Boolean} false - если запрос не подходит для метода GET
    *
    */
   function prepareGetReq(req, method, params) {
      var url = '?protocol=' + constants.JSONRPC_PROOTOCOL_VERSION +
         '&method=' + encodeURI(method) +
         '&params=' + encodeURIComponent(strHelpersMin.serializeURLData(params)) +
         '&id=1';
      req.reqUrl = url;
      if (url.length >= 2 * 1024) {
         return false;
      }
      req.reqBody = '';
      req.reqUrl = url;
      return true;
   }

   var RpcJsonHeaders = {
      init: function() {
         if (this._channel || typeof window == 'undefined') return;

         var self = this;
         this._channel = EventBus.channel('Transport');
         this._channel.subscribe('onResponseSuccess', function(event, xhr, def, url) {
            self.__cache[url] = self.__cache[url] || {};
            self.__cache[url]['X-LastModification'] = xhr.getResponseHeader('X-LastModification') || self.__cache[url]['X-LastModification'] || undefined;
         });
      },
      getHeader: function(url, header) {
         return this.__cache[url] ? this.__cache[url][header] : '';
      },
      __cache: {}
   };

   RpcJsonHeaders.init();
   function deprecateLog(method) {
      // Пока не удалим ненужные приватные методы, но пометим как устаревшие
      IoC.resolve('ILogger').log('Transport/RPCJSON', 'Private method "' + method + '" is deprecated and will be deleted in 3.17.220');
   }
   /**
    * Класс для вызова методов бизнес-логики, в том числе и с других сервисов.
    * @example
    * <pre>
    *    var rpcjson = new RPCJSON({
    *          serviceUrl: '/telephony/service/', // Обращение в сервис "Телефония"
    *          fallback: true, // Асинхронный вызов метода без гарантии доставки
    *          timeout: 120000 // Время ожидания ответа - 2 минуты
    *       }),
    *       res = rpcjson.callMethod('Phonebook.ReadNumber', { ... }); // Производим вызов метода БЛ
    *    res.addCallbacks( ... ); // Обработка результата
    * </pre>
    * @class Transport/RPCJSON
    * @author Бегунов А.В.
    * @public
    * @TODO переписать полностью.
    */
   var RPCJSON = coreExtend({}, /** @lends Transport/RPCJSON.prototype */ {
      $protected: {
         _transportPOST: null,
         _transportGET: null,
         _currentTransport: null,
         _useGet: false,
         _options: {
            /**
             * @cfg {String} Устанавливает виртуальный адрес сервиса БЛ, на который будет произведён запрос.
             * @remark
             * Точка входа в другой сервис (указывается в свойствах приложения), и все методы будут вызваны с указанного сервиса.
             * Все точки входа, заданные в свойствах приложения, попадают в интерфейсном движке WS в массив — $ws._const.services.
             * Если значение не установлено, то будет произведено вызов с сервиса по умолчанию - "/service/".
             */
            serviceUrl: '',
            /**
             * @cfg {Boolean} Устанавливает признак асинхронного вызова метода без гарантии доставки.
             * @remark
             * Подробнее о назначении этого параметра вы можете прочитать в разделе <a href="http://wi.sbis.ru/doc/platform/developmentapl/cooperationservice/subscription-to-events-in-the-cloud/#fallback">Событийные операции с контрактом и fallback</a>.
             * Поведение вызова изменяется в зависимости от значения параметра: если fallback=false, то {@link timeout} не учитывается, и вызов метода {@link callMethod} будет подобен вызову класса {@link Deprecated/BLObject}.
             */
            fallback: false,
            /**
             * @cfg {Number} Устанавливает время ожидания ответа.
             * @remark
             * Значение устанавливается в мс. При превышении этого времени в качестве результата возвращается errback с сообщение "Timeout".
             * Опция актуальна только при использовании совместно с опцией {@link fallback} в значении true.
             */
            timeout: 60000,
            async: false
         }
      },
      $constructor: function() {
         this._transportPOST = this._createTransport("POST");
         this._useGet = typeof window !== 'undefined' && !isEmptyObject(window.cachedMethods || []);
      },
       /**
        * @param {Core/Deferred} dResult
        * @param {String} method
        * @param {*} args
        * @param {*} response
        * @void
        * @private
        * @deprecated
        */
      _handleRPCError: function(dResult, method, args, response) {
           deprecateLog("_handleRPCError");
         dResult.errback(RPCJSON.handleRPCError(method, response.error));
      },
       /**
        * @param {Core/Deferred} dResult
        * @param {String} method
        * @param {*} args
        * @param {Error} error
        * @void
        * @private
        * @deprecated
        */
      _handleHTTPError: function(dResult, method, args, error) {
           deprecateLog("_handleHTTPError");
         dResult.errback(RPCJSON.handleHTTPError(method, error));
      },
      _fallbackResponse: function(dResult, guid, event, data) {
         if (data && typeof data === 'string') {
            data = JSON.parse(data);
         }
         if (data && data.guid === guid) {
            clearTimeout(this.timeout);
            dResult.callback(data.data);
         }
      },
       /**
        * @param method
        * @return {boolean}
        * @private
        * @deprecated
        */
      _useGetMethod: function(method) {
         deprecateLog("_useGetMethod");
         return (window.cachedMethods || []).indexOf(method) > -1;
      },
       /**
        * Можно ли вызывать метод через GET запрос
        * @param {String} method Имя вызываемого метода БЛ. Значение устанавливается в формате "ИмяОбъекта.ИмяМетода".
        * @param {String} url
        * @return {boolean}
        * @private
        */
      _canUseGetMethod: function(method, url) {
         var cachedMethods = (typeof window !== 'undefined') && !isEmptyObject(window.cachedMethods || []) ?
            window.cachedMethods :
            [];
         return (
            (cachedMethods.indexOf(method) > -1) &&
            (url.length < 2 * 1024)
         );
      },
      _setHeaders: function(req, method, recent) {
         if (method === 'get') {
            var sid = userInfo.getCurrent();
            if (sid) {
               var tmp = sid.split('-'),
                   client = tmp[0],
                   user = tmp[1];

               req.reqHeaders['X-SbisSessionsID'] = sid;
               req.reqHeaders['X-CID'] = client;
               req.reqHeaders['X-UID'] = user;
            }
         } else {
            var url = this._options.serviceUrl || constants.defaultServiceUrl,
               lastModHeader = (RpcJsonHeaders.getHeader(url, 'X-LastModification'));
            if (lastModHeader) {
               req.reqHeaders['X-LastModification'] = lastModHeader;
            }
         }

         if (this._options.fallback) {
            req.reqHeaders['X-ASYNCINVOKE'] = true;
            req.reqHeaders['X-PublishAsyncResponse'] = true;
            req.reqHeaders['X-UNIQ-ID'] = randomHelpers.createGUID();
         }

         if (this._options.async) {
            req.reqHeaders['X-ASYNCINVOKE'] = true;
         }

         if (recent) {
            req.reqHeaders['X-RequireRecent'] = true;
         }
      },
      /**
       * Производит вызов метода БЛ.
       * @remark
       * Вызов может производиться в двух режимах, которые устанавливают опцией {@link fallback}.
       * @param {String} method Имя вызываемого метода БЛ. Значение устанавливается в формате "ИмяОбъекта.ИмяМетода".
       * @param {Array|Object} args Аргументы вызова метода. Аргументы платформенных типов методов описаны в разделе <a href="https://wi.sbis.ru/doc/platform/developmentapl/workdata/logicworkapl/objects/blmethods/">Типы методов бизнес-логики</a>.
       * Передаются без изменений со следующими исключениями:
       * <ul>
       *    <li>Deprecated/Record, Deprecated/RecordSet - сериализуется по формату обмена;</li>
       *    <li>Date - сериализуется в строку через Date.toSQL(null).</li>
       * </ul>
       * @param {Boolean} recent Признак, по которому чтение данных будет произведено из master-базы.
       * @returns {Core/Deferred} При использовании опции {@link timeout} в ответ может прийти errback с сообщением "Timeout", если время ожидания ответа превысило допустимое значение.
       * @see abort
       */
      callMethod: function(method, args, recent) {
         var
            self = this,
            dResult = new Deferred({
               cancelCallback: this.abort.bind(this)
            }),
            dExecute;

         var req = jsonRpcPrepareBodyHeaders(method, args);

         var isGet = isGetMethod(method);
         if (isGet) {
            isGet = prepareGetReq(req, method, args);
            if (isGet) {
               this._setHeaders(req, 'get', recent);
               if (!this._transportGET) {
                  this._transportGET = this._createTransport("GET");
               }
               this._transportGET.setUrl((this._options.serviceUrl || constants.defaultServiceUrl) + req.reqUrl);
               this._currentTransport = this._transportGET;
            }
         }

         if (!isGet) {
            this._setHeaders(req, 'post', recent);
            this._currentTransport = this._transportPOST;
         }

         if (this._options.fallback) {
            require(['Lib/ServerEvent/Bus'], function(ServerEventBus) {
               ServerEventBus.serverChannel(FALLBACK_EVENT_NAME).once('onReady', function() {
                  var ctx = {},
                      fallbackResponse = self._fallbackResponse.bind(ctx, dResult, req.reqHeaders['X-UNIQ-ID']);
                  ctx.timeout = setTimeout(function() {
                     ServerEventBus.serverChannel(FALLBACK_EVENT_NAME).unsubscribe('onMessage', fallbackResponse, ctx);
                     dResult.errback('Timeout');
                  }, self._options.timeout);
                  ServerEventBus.serverChannel(FALLBACK_EVENT_NAME).subscribe('onMessage', fallbackResponse, ctx);
                  dExecute = self._currentTransport.execute(req.reqBody, req.reqHeaders);
               });
            }, function(err) {
               dResult.errback(err);
            });
         } else {
            dExecute = this._currentTransport.execute(req.reqBody, req.reqHeaders);
            dExecute.addCallbacks(
               function(r) {
                  r = r || RPCJSON.getEmptyRpcResponse();
                  // 200, но пустой ответ или вложенная ошибка
                  var result = ('error' in r)?
                      RPCJSON.handleRPCError(r.error, method, self._options.serviceUrl):
                      r.result;
                  dResult.callback(result);
                  return r;
               },
               function(e) {
                  // НЕ 200 ОК, какая-то ошибка, возможно не просто HTTP.
                   dResult.callback(RPCJSON.handleHTTPError(e, method));
                  return e;
               }
            );
         }
         dResult.addErrback(function (error) {
            var details, code = error.httpError || 0;
            details = code > 500? rk("Попробуйте заглянуть сюда через 15 минут"): "";
            error.details = error.details || details;
            return error;
         });
         return dResult;
      },
      /**
       * Прерывает последний запрос.
       * @see callMethod
       * @void
       */
      abort: function() {
         this._currentTransport && this._currentTransport.abort();
      },
       /**
        * Создаёт экземпляр ITransport
        * @param {"GET" | "POST"} method
        * @return {ITransport}
        * @private
        */
      _createTransport: function(method) {
         return IoC.resolve('ITransport', {
             url: this._options.serviceUrl || constants.defaultServiceUrl,
             method: method,
             dataType: 'json',
             contentType: 'application/json; charset=utf-8'
         });
      }
   });
    /**
     *
     * @name Transport/RPCJSON#jsonRpcPreparePacket
     * @function
     */
   RPCJSON.jsonRpcPreparePacket = jsonRpcPreparePacket;
    /**
     * @return {{error: {message: string, code: string, details: string}}}
     */
   RPCJSON.getEmptyRpcResponse = function () {
       return {
           error: {
               message: "Получен пустой ответ от сервиса",
               code: "",
               details: ""
           }
       }
   };
    /**
     * @param {Error} error
     * @param {String} method Имя вызываемого метода БЛ.
     * @return {Transport/TransportError | Transport/HTTPError}
     */
   RPCJSON.handleHTTPError = function (error, method) {
       if (!(error instanceof HTTPError)) {
          return error;
       }
       var errInst, payload;
       var message = error.message, details = '', code = 0, classid = '', errType = '', addinfo = '', error_code = '';
       try {
           payload = JSON.parse(error.payload);
           if (payload.error) {
               message = payload.error.message || message;
               details = payload.error.details;
               code = payload.error.code;
               if (payload.error.data) {
                   classid = payload.error.data.classid || classid;
                   addinfo = payload.error.data.addinfo || addinfo;
                   error_code = payload.error.data.error_code || error_code;
               }
               errType = payload.error.type;
           }
       } catch (e) {}
       errInst = new TransportError(message, error.httpError, code, method, details, error.url, classid, errType || 'error', addinfo || '', error_code);
       EventBus.channel('errors').notify('onRPCError', errInst);

       if(errInst.processed){
           error.processed = true;
       }
       return errInst;
   };
    /**
     * @param {Error | Object} error
     * @param {String} method Имя вызываемого метода БЛ.
     * @param {String} [url]
     * @return {Transport/TransportError}
     */
   RPCJSON.handleRPCError = function (error, method, url) {
       var transportError = new TransportError(
           error.message,
           '',
           error.code,
           method,
           error.details || '',
           url || constants.defaultServiceUrl,
           error.data && error.data.classid || '',
           error.type || 'error',
           error.data && error.data.addinfo || '',
           error.data && error.data.error_code || ''
        );
       EventBus.channel('errors').notify('onRPCError', transportError);
       return transportError;
   };
   return RPCJSON;
});