define('Transport/XHRTransport', [
   'Transport/ITransport',
   'Core/Deferred',
   'Core/EventBus',
   'Core/UserInfo',
   'Core/constants',
   'Transport/HTTPError',
   'Core/IoC',
   'Core/detection',
   'Transport/ajax-emulator'
], function(ITransport, Deferred, EventBus, UserInfo, _const, HTTPError, IoC, detection, ajax) {
   var XHRTransport,
      TRANSPORT_CHANNEL = EventBus.channel('Transport'),
      ERROR_CHANNEL = EventBus.channel('errors');

   function handleOfflineError(error, data) {
      if (XHRTransport.isOfflineError(error)) {
         var methodInfo;
          /* TODO
          * нехорошо, когда транспорт общего назначения знает о внутреней структуре отправляеммых пакетов
          * обернул пока в try/catch, дальше надо исправить
           */
         try{
             methodInfo = JSON.parse(data).method.split('.');
             methodInfo.push(error);
         } catch (err) {
             methodInfo = [data, error];
         }
         return !!XHRTransport.notifyOfflineError(methodInfo);
      }
   }

   /**
    * @cfg {Boolean} Указывает на перезагрузку страницы
    * По умолчанию false.
    */
   var isUnloadProcess = false;
   if (typeof window !== 'undefined') {
      if (detection.firefox) {
         window.addEventListener('beforeunload', function () {
            isUnloadProcess = true;
            setTimeout(function () {
               isUnloadProcess = false;
            }, 600);
         });
      } else {
         window.addEventListener('unload', function () {
            isUnloadProcess = true;
         });
      }
   }
   /**
    * Асинхронный XHR Transport через Deferred
    * @class Transport/XHRTransport
    * @extends Transport/ITransport
    * @public
    * @author Бегунов А.В.
    */
   XHRTransport = ITransport.extend(/** @lends Transport/XHRTransport.prototype */{
      $protected: {
         _options: {
            /**
             * @cfg {String} Метод запроса. POST или GET.
             * По умолчанию = GET.
             */
            method: 'GET',
            /**
             * @cfg {String} Тип данных, который Вы ожидаете от сервера.
             * По умолчанию text.
             */
            dataType: 'text',
            /**
             * @cfg {String} contentType Тип данных при отсылке данных на сервер.
             * По умолчанию application/x-www-form-urlencoded.
             */
            contentType: 'application/x-www-form-urlencoded',
            /**
             * @cfg {String} url URL, по которому отправляется запрос.
             */
            url: '',
            /**
             * @cfg {Boolean} processData преобразование отправляемых данных в строку запроса,
             * подходящую для типа "application/x-www-form-urlencoded".
             * Если Вы хотите отправить DOM документ, или другие необрабатываемые данные,
             * установите значение данного параметр в false.
             * По умолчанию true.
             */
            processData: true
         },
         _xhr: undefined      //Последний запрос
      },
      $constructor: function() {
         // Если тексты ошибок не локализованы и присутствует функция локализации (не ПП)
         if (!isTranslated && typeof rk !== 'undefined'){
            isTranslated = true;
            for(var err in XHRTransport.ERRORS_TEXT){
               if (XHRTransport.ERRORS_TEXT.hasOwnProperty(err)){
                  XHRTransport.ERRORS_TEXT[err] = rk(XHRTransport.ERRORS_TEXT[err]);
               }
            }
         }
         if (this._options.url === '') {
            throw new Error("Request with no URL is ambiguous");
         }
      },
      /**
       * @param {String} data Данные для отправки
       * @param {Object} [headers] Заголовки
       * @return {Core/Deferred}
       */
      execute: function(data, headers) {
         var dResult = new Deferred({
               cancelCallback: this.abort.bind(this)
            }),
            self = this;

         /**
          * Проверка куки, если изменилась - кидаем ошибку авторизации
          */
         if (_const.checkSessionCookie && !UserInfo.isValid()) {
            dResult.errback(XHRTransport.ERRORS_TEXT.sessionError);
            ERROR_CHANNEL.notify('onAuthError');
            return dResult;
         }
         try {
            this._xhr = ajax({
               type: this._options.method,
               dataType: this._options.dataType,
               contentType: this._options.contentType,
               url: this._options.url,
               processData: this._options.processData,
               headers: headers || {},
               data: data,
                /**
                 * TODO: Вынести обработчики запроса из метода.
                 * 1) Не читабельно
                 * 2) Лишняя работа на оптимизатор браузера
                 * 3) Услажнён дебаг с правками из браузера из-за оптимизатора
                 */
               beforeSend: function(xhr, settings) {
                  return TRANSPORT_CHANNEL.notify('onBeforeSend', xhr, settings);
               },

               success: function(result) {
                  if (handleOfflineError(result.error, data)) {
                     dResult.errback(HTTPError.offlineModeError());
                     return;
                  }

                  var resultDfr = Deferred.success(result);

                  TRANSPORT_CHANNEL.notify('onResponseSuccess', self._xhr, resultDfr, self._options.url);
                  dResult.dependOn(resultDfr);
                  return result;
               },
               error: function(xhr, textStatus) {
                  var humanReadableErrors, textError, resultDfr, error, status;
                  if (xhr.responseJSON && handleOfflineError(xhr.responseJSON.error, data)) {
                     dResult.errback(HTTPError.offlineModeError());
                     return;
                  }
                  status = xhr.status;
                  humanReadableErrors = XHRTransport.ERRORS_TEXT;
                  textError = humanReadableErrors[status] ||
                      humanReadableErrors[textStatus] ||
                      humanReadableErrors.unknown || "";

                  if ((status === 0) && (xhr.getAllResponseHeaders() === "") && (textStatus !== 'abort')) {
                     var site = (self._options.url.charAt(0) === '/') ?
                         window.location.hostname :
                         self._options.url;
                     textError = humanReadableErrors.lossOfConnection + " " + site;
                  }

                  error = new HTTPError(textError, status, self._options.url, xhr.responseText);

                  // Извещаем о HTTP-ошибке
                  ERROR_CHANNEL.notify('onHTTPError', error);

                  //обрабатываем ситуацию истекшей сессии
                  if (status == "401") {
                     // Старый способ, он пока важнее. Надо выпилить в 3.8
                     //Не нашёл где инизилизируется, поэтому закоментил, если ни чего не сломает, можно удалить
                     /*if (typeof $ws.core._authError == "function") {
                        $ws.core._authError();
                        return;
                     }*/

                     // Новый вариант рассылки ошибки о проблеме аутентификации
                     if (ERROR_CHANNEL.notify('onAuthError') === true) {
                        return;
                     }
                  }

                  //Здесь ошибку точно надо отдавать в dResult
                  resultDfr = new Deferred();
                  TRANSPORT_CHANNEL.notify('onResponseError', self._xhr, resultDfr);

                  resultDfr.addCallback(function(result) {
                     dResult.callback(result);
                  }).addErrback(function(error) {
                     dResult.errback(error);
                  });

                  /**
                   *  На слабых компьютерах события обрыва запроса вызывает
                   *  негативные эффекты (отображение окошек потери связи), при переходе на новую ссылку или обновления страницы.
                   *  Поэтому подавляем ошибки.
                   */
                  if(isUnloadProcess){
                     resultDfr.cancel();
                  } else {
                     resultDfr.errback(error);
                  }
               }
            });
         } catch (e) {
            dResult.errback("JavaScript exception while trying to execute request: " + e.message);
         }

         return dResult;
      },

      /**
       * Заменяет url
       * @param {String} url
       */
      setUrl: function(url) {
         if (typeof url == 'string') {
            this._options.url = url;
         }
      },

      /**
       * Прерывает загрузку
       */
      abort: function() {
         if (this._xhr) {
            this._xhr.abort();
         }
      }
   });
   // на ПП мы не имеем rk, поэтому тексты ошибок держим для рускоязычного сегмента, с последующим вызовом rk через map
   var isTranslated;
    /**
     * ERRORS_TEXT
     * @static
     */
   XHRTransport.ERRORS_TEXT = {
       timeout: "Таймаут запроса",
       unknown: "Неизвестная ошибка",
       parsererror: "Ошибка разбора документа",
       abort: "Запрос был прерван",
       lossOfConnection: "Потеряна связь с сайтом",
       sessionError: "Ошибка авторизации. Текущая сессия принадлежит другому пользователю/клиенту",
       403: "У вас недостаточно прав для выполнения данного действия",
       404: "Документ не найден",
       413: "Превышен допустимый размер загружаемого файла",
       423: "Действие заблокировано лицензией",
       500: "Внутренняя ошибка сервера",
       502: "Сервис находится на техническом обслуживании",
       503: "Сервис находится на техническом обслуживании",
       504: "Сервис находится на техническом обслуживании"
   };
    /**
     * isOfflineError
     * @static
     * @param {Error | HTTPError} error
     * @return {Boolean}
     */
   XHRTransport.isOfflineError = function (error) {
      var code;
      code = error && (
          error.data && error.data.error_code ||
          error.httpError ||
          error.code
       );
       return (code == 300 || code == 405);
   };
    /**
     * notifyOfflineError
     * @static
     * @param {Array<*>} args
     * @return {*}
     */
   XHRTransport.notifyOfflineError = function (args) {
       var globalCanel, notifyArgs, result;
       globalCanel = EventBus.globalChannel();
       notifyArgs = ["onOfflineModeError"].concat(args);
       result = globalCanel.notify.apply(globalCanel, notifyArgs);
       if (result) {
           IoC.resolve('ILogger').info('XHRTransport', 'Работа в offline-режиме');
       }
       return result;
   };
    /**
     * Возвращает URL сервиса по имени
     * @param {String} serviceName имя сервиса (как указано на 4 шаге мастера выгрузки).
     * @returns {String} URL сервиса.
     */
   XHRTransport.getServiceByName =  function(serviceName) {
       return _const.services[serviceName] || _const.defaultServiceUrl;
   };


   if (!IoC.has('ITransport')) {
      IoC.bind('ITransport', XHRTransport);
   }

   return XHRTransport;
});
